
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class ConsoleHostRawUserInterfaceStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal ConsoleHostRawUserInterfaceStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.PowerShell.ConsoleHost.resources.ConsoleHostRawUserInterfaceStrings", typeof(ConsoleHostRawUserInterfaceStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the operation because the supplied coordinate is not valid. Specify a coordinate within the buffer area of {0}.
    ///  
    /// </summary>
    internal static string CoordinateOutOfBufferErrorTemplate {
        get {
            return ResourceManager.GetString("CoordinateOutOfBufferErrorTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot set the buffer size because the size specified is too large or too small.
    ///  
    /// </summary>
    internal static string InvalidBufferSizeError {
        get {
            return ResourceManager.GetString("InvalidBufferSizeError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot set the console color because the value specified is not valid. Specify a valid color as defined by the System.ConsoleColor type.
    ///  
    /// </summary>
    internal static string InvalidConsoleColorError {
        get {
            return ResourceManager.GetString("InvalidConsoleColorError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process CursorSize because the cursor size specified is not valid.
    ///  
    /// </summary>
    internal static string InvalidCursorSizeError {
        get {
            return ResourceManager.GetString("InvalidCursorSizeError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot read key options. To read options, set one or both of the following: IncludeKeyDown, IncludeKeyUp.
    ///  
    /// </summary>
    internal static string InvalidReadKeyOptionsError {
        get {
            return ResourceManager.GetString("InvalidReadKeyOptionsError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0} should be greater than or equal to {1}.
    ///  
    /// </summary>
    internal static string InvalidRegionErrorTemplate {
        get {
            return ResourceManager.GetString("InvalidRegionErrorTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot use the specified Window X (column) position because it extends past the width of the screen buffer. Specify another X position, starting with 0 as the left most column of the buffer.
    ///  
    /// </summary>
    internal static string InvalidXWindowPositionError {
        get {
            return ResourceManager.GetString("InvalidXWindowPositionError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot use the specified Window Y (row) position because it extends past the height of the screen buffer. Specify another Y position, starting with 0 as the top-most row of the buffer.
    ///  
    /// </summary>
    internal static string InvalidYWindowPositionError {
        get {
            return ResourceManager.GetString("InvalidYWindowPositionError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Window width cannot be less than 1.
    ///  
    /// </summary>
    internal static string WindowWidthTooSmallError {
        get {
            return ResourceManager.GetString("WindowWidthTooSmallError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Window height must be at least 1.
    ///  
    /// </summary>
    internal static string WindowHeightTooSmallError {
        get {
            return ResourceManager.GetString("WindowHeightTooSmallError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Window cannot be wider than the screen buffer.
    ///  
    /// </summary>
    internal static string WindowWidthLargerThanBufferError {
        get {
            return ResourceManager.GetString("WindowWidthLargerThanBufferError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Window cannot be taller than the screen buffer.
    ///  
    /// </summary>
    internal static string WindowHeightLargerThanBufferError {
        get {
            return ResourceManager.GetString("WindowHeightLargerThanBufferError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Window cannot be wider than {0}.
    ///  
    /// </summary>
    internal static string WindowWidthTooLargeErrorTemplate {
        get {
            return ResourceManager.GetString("WindowWidthTooLargeErrorTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Window cannot be taller than {0}.
    ///  
    /// </summary>
    internal static string WindowHeightTooLargeErrorTemplate {
        get {
            return ResourceManager.GetString("WindowHeightTooLargeErrorTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Window size is too narrow.
    ///  
    /// </summary>
    internal static string WindowTooNarrowError {
        get {
            return ResourceManager.GetString("WindowTooNarrowError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Window size is too short.
    ///  
    /// </summary>
    internal static string WindowTooShortError {
        get {
            return ResourceManager.GetString("WindowTooShortError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Window title cannot be empty.
    ///  
    /// </summary>
    internal static string WindowTitleTooShortError {
        get {
            return ResourceManager.GetString("WindowTitleTooShortError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Window title cannot be longer than {0} characters.
    ///  
    /// </summary>
    internal static string WindowTitleTooLongErrorTemplate {
        get {
            return ResourceManager.GetString("WindowTitleTooLongErrorTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Administrator: 
    ///  
    /// </summary>
    internal static string WindowTitleElevatedPrefix {
        get {
            return ResourceManager.GetString("WindowTitleElevatedPrefix", resourceCulture);
        }
    }

}

