
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


namespace Microsoft.PowerShell.LocalAccounts {

using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class Strings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal Strings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.PowerShell.LocalAccounts.resources.Microsoft.PowerShell.LocalAccounts.Strings", typeof(Strings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Access denied.
    ///  
    /// </summary>
    internal static string AccessDenied {
        get {
            return ResourceManager.GetString("AccessDenied", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Account {0} was not found.
    ///  
    /// </summary>
    internal static string AccountNotFound {
        get {
            return ResourceManager.GetString("AccountNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Add member {0}
    ///  
    /// </summary>
    internal static string ActionAddGroupMember {
        get {
            return ResourceManager.GetString("ActionAddGroupMember", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Disable local user
    ///  
    /// </summary>
    internal static string ActionDisableUser {
        get {
            return ResourceManager.GetString("ActionDisableUser", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Enable local user
    ///  
    /// </summary>
    internal static string ActionEnableUser {
        get {
            return ResourceManager.GetString("ActionEnableUser", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Create new local group
    ///  
    /// </summary>
    internal static string ActionNewGroup {
        get {
            return ResourceManager.GetString("ActionNewGroup", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Create new local user
    ///  
    /// </summary>
    internal static string ActionNewUser {
        get {
            return ResourceManager.GetString("ActionNewUser", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Remove local group
    ///  
    /// </summary>
    internal static string ActionRemoveGroup {
        get {
            return ResourceManager.GetString("ActionRemoveGroup", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Remove member {0}
    ///  
    /// </summary>
    internal static string ActionRemoveGroupMember {
        get {
            return ResourceManager.GetString("ActionRemoveGroupMember", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Remove local user
    ///  
    /// </summary>
    internal static string ActionRemoveUser {
        get {
            return ResourceManager.GetString("ActionRemoveUser", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Rename local group to {0}
    ///  
    /// </summary>
    internal static string ActionRenameGroup {
        get {
            return ResourceManager.GetString("ActionRenameGroup", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Rename local user to {0}
    ///  
    /// </summary>
    internal static string ActionRenameUser {
        get {
            return ResourceManager.GetString("ActionRenameUser", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Modify local group
    ///  
    /// </summary>
    internal static string ActionSetGroup {
        get {
            return ResourceManager.GetString("ActionSetGroup", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Modify local user
    ///  
    /// </summary>
    internal static string ActionSetUser {
        get {
            return ResourceManager.GetString("ActionSetUser", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Group {0} already exists.
    ///  
    /// </summary>
    internal static string GroupExists {
        get {
            return ResourceManager.GetString("GroupExists", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The group {0} still has members.
    ///  
    /// </summary>
    internal static string GroupHasMembers {
        get {
            return ResourceManager.GetString("GroupHasMembers", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Group {0} was not found.
    ///  
    /// </summary>
    internal static string GroupNotFound {
        get {
            return ResourceManager.GetString("GroupNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The operation is not allowed for group {0}.
    ///  
    /// </summary>
    internal static string InvalidForGroup {
        get {
            return ResourceManager.GetString("InvalidForGroup", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The operation is not allowed for user {0}.
    ///  
    /// </summary>
    internal static string InvalidForUser {
        get {
            return ResourceManager.GetString("InvalidForUser", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The name {0} is invalid.
    ///  
    /// </summary>
    internal static string InvalidName {
        get {
            return ResourceManager.GetString("InvalidName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Parameter {0} and parameter {1} may not be used together.
    ///  
    /// </summary>
    internal static string InvalidParameterPair {
        get {
            return ResourceManager.GetString("InvalidParameterPair", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Invalid password.
    ///  
    /// </summary>
    internal static string InvalidPassword {
        get {
            return ResourceManager.GetString("InvalidPassword", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot remove the last Administrator
    ///  
    /// </summary>
    internal static string LastAdmin {
        get {
            return ResourceManager.GetString("LastAdmin", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0} is already a member of group {1}.
    ///  
    /// </summary>
    internal static string MemberExists {
        get {
            return ResourceManager.GetString("MemberExists", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Member {0} was not found in group {1}.
    ///  
    /// </summary>
    internal static string MemberNotFound {
        get {
            return ResourceManager.GetString("MemberNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    User {0} may not be removed from its primary group.
    ///  
    /// </summary>
    internal static string MembersPrimaryGroup {
        get {
            return ResourceManager.GetString("MembersPrimaryGroup", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The name {0} is already in use.
    ///  
    /// </summary>
    internal static string NameInUse {
        get {
            return ResourceManager.GetString("NameInUse", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Group
    ///  
    /// </summary>
    internal static string ObjectClassGroup {
        get {
            return ResourceManager.GetString("ObjectClassGroup", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Other
    ///  
    /// </summary>
    internal static string ObjectClassOther {
        get {
            return ResourceManager.GetString("ObjectClassOther", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    User
    ///  
    /// </summary>
    internal static string ObjectClassUser {
        get {
            return ResourceManager.GetString("ObjectClassUser", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The password cannot be set because a password restriction is in place.
    ///  
    /// </summary>
    internal static string PasswordRestriction {
        get {
            return ResourceManager.GetString("PasswordRestriction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Principal {0} was not found.
    ///  
    /// </summary>
    internal static string PrincipalNotFound {
        get {
            return ResourceManager.GetString("PrincipalNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    RID {0} was not found.
    ///  
    /// </summary>
    internal static string RidToSidFailed {
        get {
            return ResourceManager.GetString("RidToSidFailed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    An unspecified error occurred.
    ///  
    /// </summary>
    internal static string UnspecifiedError {
        get {
            return ResourceManager.GetString("UnspecifiedError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    An unspecified error occurred: status = {0}
    ///  
    /// </summary>
    internal static string UnspecifiedErrorNtStatus {
        get {
            return ResourceManager.GetString("UnspecifiedErrorNtStatus", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    An unspecified error occurred: error code = {0}
    ///  
    /// </summary>
    internal static string UnspecifiedErrorWin32Error {
        get {
            return ResourceManager.GetString("UnspecifiedErrorWin32Error", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    User {0} already exists.
    ///  
    /// </summary>
    internal static string UserExists {
        get {
            return ResourceManager.GetString("UserExists", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    User {0} was not found.
    ///  
    /// </summary>
    internal static string UserNotFound {
        get {
            return ResourceManager.GetString("UserNotFound", resourceCulture);
        }
    }

}

}

