
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class FormatAndOutXmlLoadingStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal FormatAndOutXmlLoadingStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.FormatAndOutXmlLoadingStrings", typeof(FormatAndOutXmlLoadingStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: The XML element {2} does not allow attributes.
    ///  
    /// </summary>
    internal static string AttributesNotAllowed {
        get {
            return ResourceManager.GetString("AttributesNotAllowed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: Node {2} cannot have child objects.
    ///  
    /// </summary>
    internal static string NoChildrenAllowed {
        get {
            return ResourceManager.GetString("NoChildrenAllowed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: {2} is not valid.
    ///  
    /// </summary>
    internal static string InvalidNode {
        get {
            return ResourceManager.GetString("InvalidNode", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: There must be at least one default {2}.
    ///  
    /// </summary>
    internal static string NoDefaultShapeEntry {
        get {
            return ResourceManager.GetString("NoDefaultShapeEntry", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: There cannot be more than one default {2}.
    ///  
    /// </summary>
    internal static string TooManyDefaultShapeEntry {
        get {
            return ResourceManager.GetString("TooManyDefaultShapeEntry", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: The control name cannot be null or empty.
    ///  
    /// </summary>
    internal static string NullControlName {
        get {
            return ResourceManager.GetString("NullControlName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: Out Of Band views can only have CustomControl or ListControl.
    ///  
    /// </summary>
    internal static string InvalidControlForOutOfBandView {
        get {
            return ResourceManager.GetString("InvalidControlForOutOfBandView", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: An Out Of Band view cannot have GroupBy.
    ///  
    /// </summary>
    internal static string OutOfBandGroupByConflict {
        get {
            return ResourceManager.GetString("OutOfBandGroupByConflict", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: The view cannot be loaded.
    ///  
    /// </summary>
    internal static string ViewNotLoaded {
        get {
            return ResourceManager.GetString("ViewNotLoaded", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: "{2}" is not a valid alignment value.
    ///  
    /// </summary>
    internal static string InvalidAlignmentValue {
        get {
            return ResourceManager.GetString("InvalidAlignmentValue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: A positive integer is expected.
    ///  
    /// </summary>
    internal static string ExpectPositiveInteger {
        get {
            return ResourceManager.GetString("ExpectPositiveInteger", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: The column header definition is not valid; all headers are discarded.
    ///  
    /// </summary>
    internal static string InvalidColumnHeader {
        get {
            return ResourceManager.GetString("InvalidColumnHeader", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: The row item count = {2} on alternative set #{3} does not match default row item count = {4}.
    ///  
    /// </summary>
    internal static string IncorrectRowItemCount {
        get {
            return ResourceManager.GetString("IncorrectRowItemCount", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: The header item count = {2} does not match default row item count = {3}.
    ///  
    /// </summary>
    internal static string IncorrectHeaderItemCount {
        get {
            return ResourceManager.GetString("IncorrectHeaderItemCount", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: At least one list view item must be specified.
    ///  
    /// </summary>
    internal static string NoListViewItem {
        get {
            return ResourceManager.GetString("NoListViewItem", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: The property entry is not valid.
    ///  
    /// </summary>
    internal static string InvalidPropertyEntry {
        get {
            return ResourceManager.GetString("InvalidPropertyEntry", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: The definition list is missing.
    ///  
    /// </summary>
    internal static string NoDefinitionList {
        get {
            return ResourceManager.GetString("NoDefinitionList", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: A Boolean value is expected.
    ///  
    /// </summary>
    internal static string ExpectBoolean {
        get {
            return ResourceManager.GetString("ExpectBoolean", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: A non-negative integer is expected.
    ///  
    /// </summary>
    internal static string ExpectNaturalNumber {
        get {
            return ResourceManager.GetString("ExpectNaturalNumber", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: An integer is expected.
    ///  
    /// </summary>
    internal static string ExpectInteger {
        get {
            return ResourceManager.GetString("ExpectInteger", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: The inner text value is missing.
    ///  
    /// </summary>
    internal static string MissingInnerText {
        get {
            return ResourceManager.GetString("MissingInnerText", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: The custom control token list cannot be empty.
    ///  
    /// </summary>
    internal static string EmptyCustomControlList {
        get {
            return ResourceManager.GetString("EmptyCustomControlList", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: {2} failed to load.
    ///  
    /// </summary>
    internal static string LoadTagFailed {
        get {
            return ResourceManager.GetString("LoadTagFailed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: {2} cannot be specified without an expression.
    ///  
    /// </summary>
    internal static string NodeWithoutExpression {
        get {
            return ResourceManager.GetString("NodeWithoutExpression", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: {2} cannot be specified with an expression.
    ///  
    /// </summary>
    internal static string NodeWithExpression {
        get {
            return ResourceManager.GetString("NodeWithExpression", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: A format string is missing.
    ///  
    /// </summary>
    internal static string NoFormatString {
        get {
            return ResourceManager.GetString("NoFormatString", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: Script block text is missing.
    ///  
    /// </summary>
    internal static string NoScriptBlockText {
        get {
            return ResourceManager.GetString("NoScriptBlockText", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: A property is missing.
    ///  
    /// </summary>
    internal static string NoProperty {
        get {
            return ResourceManager.GetString("NoProperty", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: Script block "{2}" is not valid.
    ///  
    /// </summary>
    internal static string InvalidScriptBlock {
        get {
            return ResourceManager.GetString("InvalidScriptBlock", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: The string {2} from resource {3} in assembly {4} is not found.
    ///  
    /// </summary>
    internal static string StringResourceNotFound {
        get {
            return ResourceManager.GetString("StringResourceNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: The resource {2} in assembly {3} is not found.
    ///  
    /// </summary>
    internal static string ResourceNotFound {
        get {
            return ResourceManager.GetString("ResourceNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: The assembly {2} is not found.
    ///  
    /// </summary>
    internal static string AssemblyNotFound {
        get {
            return ResourceManager.GetString("AssemblyNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: The node must be an XmlElement.
    ///  
    /// </summary>
    internal static string NonXmlElementNode {
        get {
            return ResourceManager.GetString("NonXmlElementNode", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: An expression is expected.
    ///  
    /// </summary>
    internal static string ExpectExpression {
        get {
            return ResourceManager.GetString("ExpectExpression", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: Cannot have control or Label without an expression.
    ///  
    /// </summary>
    internal static string ControlLabelWithoutExpression {
        get {
            return ResourceManager.GetString("ControlLabelWithoutExpression", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: Cannot have control and Label at the same time.
    ///  
    /// </summary>
    internal static string ControlAndLabel {
        get {
            return ResourceManager.GetString("ControlAndLabel", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: Cannot have SelectionSetName and TypeName at the same time.
    ///  
    /// </summary>
    internal static string SelectionSetNameAndTypeName {
        get {
            return ResourceManager.GetString("SelectionSetNameAndTypeName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: No type or condition is specified for applying the view.
    ///  
    /// </summary>
    internal static string EmptyAppliesTo {
        get {
            return ResourceManager.GetString("EmptyAppliesTo", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: The {2} value is not valid.
    ///  
    /// </summary>
    internal static string InvalidNodeValue {
        get {
            return ResourceManager.GetString("InvalidNodeValue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: A duplicate node exists.
    ///  
    /// </summary>
    internal static string DuplicatedNode {
        get {
            return ResourceManager.GetString("DuplicatedNode", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: {2} and {3} are mutually exclusive.
    ///  
    /// </summary>
    internal static string MutuallyExclusiveNode {
        get {
            return ResourceManager.GetString("MutuallyExclusiveNode", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: {2}, {3} and {4} are mutually exclusive.
    ///  
    /// </summary>
    internal static string ThreeMutuallyExclusiveNode {
        get {
            return ResourceManager.GetString("ThreeMutuallyExclusiveNode", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: {2} is an unknown node.
    ///  
    /// </summary>
    internal static string UnknownNode {
        get {
            return ResourceManager.GetString("UnknownNode", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: {2} is an unknown attribute.
    ///  
    /// </summary>
    internal static string UnknownAttribute {
        get {
            return ResourceManager.GetString("UnknownAttribute", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: {2} is a missing attribute.
    ///  
    /// </summary>
    internal static string MissingAttribute {
        get {
            return ResourceManager.GetString("MissingAttribute", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: Node {2} is missing.
    ///  
    /// </summary>
    internal static string MissingNode {
        get {
            return ResourceManager.GetString("MissingNode", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: A node is missing from {2}.
    ///  
    /// </summary>
    internal static string MissingNodeFromList {
        get {
            return ResourceManager.GetString("MissingNodeFromList", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: {2} is an empty node.
    ///  
    /// </summary>
    internal static string EmptyNode {
        get {
            return ResourceManager.GetString("EmptyNode", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error at XPath {0} in file {1}: {2} is an empty attribute.
    ///  
    /// </summary>
    internal static string EmptyAttribute {
        get {
            return ResourceManager.GetString("EmptyAttribute", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error in file {0}: {1}
    ///  
    /// </summary>
    internal static string ErrorInFile {
        get {
            return ResourceManager.GetString("ErrorInFile", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Too many errors in file {0}.
    ///  
    /// </summary>
    internal static string TooManyErrors {
        get {
            return ResourceManager.GetString("TooManyErrors", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Errors occurred while loading the format data file: {0}
    ///  
    /// </summary>
    internal static string FormatLoadingErrors {
        get {
            return ResourceManager.GetString("FormatLoadingErrors", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    (Global Assembly Cache) {0}
    ///  
    /// </summary>
    internal static string AssemblyInGAC {
        get {
            return ResourceManager.GetString("AssemblyInGAC", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0}, {1}
    ///  
    /// </summary>
    internal static string MshSnapinQualifiedError {
        get {
            return ResourceManager.GetString("MshSnapinQualifiedError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The path {0} is not fully qualified. Specify a fully qualified format file path.
    ///  
    /// </summary>
    internal static string FormatFileNotRooted {
        get {
            return ResourceManager.GetString("FormatFileNotRooted", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The FormatTable cannot be updated because the FormatTable might have been created outside of the runspace.
    ///  
    /// </summary>
    internal static string SharedFormatTableCannotBeUpdated {
        get {
            return ResourceManager.GetString("SharedFormatTableCannotBeUpdated", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Errors occurred while loading the FormatTable. View the contents of the Errors property to get detailed error messages.
    ///  
    /// </summary>
    internal static string FormatTableLoadErrors {
        get {
            return ResourceManager.GetString("FormatTableLoadErrors", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error in formatting data "{0}": {1}
    ///  
    /// </summary>
    internal static string ErrorInFormattingData {
        get {
            return ResourceManager.GetString("ErrorInFormattingData", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error in view data with type name {0} at index {1}: The header item count = {2} does not match the default row item count = {3}.
    ///  
    /// </summary>
    internal static string IncorrectHeaderItemCountInFormattingData {
        get {
            return ResourceManager.GetString("IncorrectHeaderItemCountInFormattingData", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error in view data with type name {0} at index {1}: Formatting data "{2}" is not valid.
    ///  
    /// </summary>
    internal static string InvalidFormattingData {
        get {
            return ResourceManager.GetString("InvalidFormattingData", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error in view data with type name {0} at index {1}: Script block "{2}" is not valid.
    ///  
    /// </summary>
    internal static string InvalidScriptBlockInFormattingData {
        get {
            return ResourceManager.GetString("InvalidScriptBlockInFormattingData", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error in view data with type name {0} at index {1}: {2} failed to load.
    ///  
    /// </summary>
    internal static string LoadTagFailedInFormattingData {
        get {
            return ResourceManager.GetString("LoadTagFailedInFormattingData", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error in view data with type name {0} at index {1}: A TableControl should contain only one {2}.
    ///  
    /// </summary>
    internal static string MultipleRowEntriesFoundInFormattingData {
        get {
            return ResourceManager.GetString("MultipleRowEntriesFoundInFormattingData", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error in view data with type name {0} at index {1}: There must be at least one default {2}.
    ///  
    /// </summary>
    internal static string NoDefaultShapeEntryInFormattingData {
        get {
            return ResourceManager.GetString("NoDefaultShapeEntryInFormattingData", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error in view data with type name {0} at index {1}: At least one list view item must be specified.
    ///  
    /// </summary>
    internal static string NoListViewItemInFormattingData {
        get {
            return ResourceManager.GetString("NoListViewItemInFormattingData", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error in view data with type name {0} at index {1}: There cannot be more than one default {2}.
    ///  
    /// </summary>
    internal static string TooManyDefaultShapeEntryInFormattingData {
        get {
            return ResourceManager.GetString("TooManyDefaultShapeEntryInFormattingData", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Too many errors in the formatting data for type "{0}".
    ///  
    /// </summary>
    internal static string TooManyErrorsInFormattingData {
        get {
            return ResourceManager.GetString("TooManyErrorsInFormattingData", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A shared format table cannot be updated with more than one entry.
    ///  
    /// </summary>
    internal static string FormatTableCannotCoExist {
        get {
            return ResourceManager.GetString("FormatTableCannotCoExist", resourceCulture);
        }
    }

}

