
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class CatalogStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal CatalogStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.CatalogStrings", typeof(CatalogStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unable to generate catalog definition file.
    ///  
    /// </summary>
    internal static string CatalogDefinitionFileNotGenerated {
        get {
            return ResourceManager.GetString("CatalogDefinitionFileNotGenerated", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Adding file '{0}' to the catalog. Relative path of file in catalog is '{1}'.
    ///  
    /// </summary>
    internal static string AddFileToCatalog {
        get {
            return ResourceManager.GetString("AddFileToCatalog", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Skipping validating file {0} from catalog. 
    ///  
    /// </summary>
    internal static string SkipValidationOfCatalogFile {
        get {
            return ResourceManager.GetString("SkipValidationOfCatalogFile", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Found file {0} in the catalog with hash of {1}. 
    ///  
    /// </summary>
    internal static string FoundFileHashInCatalogItem {
        get {
            return ResourceManager.GetString("FoundFileHashInCatalogItem", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Paths for catalog contains multiple files with same relative path {0}. 
    ///  
    /// </summary>
    internal static string FoundDuplicateFilesRelativePath {
        get {
            return ResourceManager.GetString("FoundDuplicateFilesRelativePath", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Found file {0} on disk with hash of {1}.
    ///  
    /// </summary>
    internal static string FoundFileInPath {
        get {
            return ResourceManager.GetString("FoundFileInPath", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Skipping validating file {0} from path.
    ///  
    /// </summary>
    internal static string SkipValidationOfPathFile {
        get {
            return ResourceManager.GetString("SkipValidationOfPathFile", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unable to acquire a handle to a catalog administrator context for a given hash algorithm {0}.
    ///  
    /// </summary>
    internal static string UnableToAcquireHashAlgorithmContext {
        get {
            return ResourceManager.GetString("UnableToAcquireHashAlgorithmContext", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unable to create the hash for file {0}. 
    ///  
    /// </summary>
    internal static string UnableToCreateFileHash {
        get {
            return ResourceManager.GetString("UnableToCreateFileHash", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unable to open catalog file {0}.
    ///  
    /// </summary>
    internal static string UnableToOpenCatalogFile {
        get {
            return ResourceManager.GetString("UnableToOpenCatalogFile", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Catalog version is not valid. We only support catalog version {0} and version {1}.
    ///  
    /// </summary>
    internal static string UnKnownCatalogVersion {
        get {
            return ResourceManager.GetString("UnKnownCatalogVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unable to open catalog definition file.
    ///  
    /// </summary>
    internal static string UnableToOpenCatalogDefinitionFile {
        get {
            return ResourceManager.GetString("UnableToOpenCatalogDefinitionFile", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Found multiple entries of file member {0} in the catalog.
    ///  
    /// </summary>
    internal static string FoundDuplicateFileMemberInCatalog {
        get {
            return ResourceManager.GetString("FoundDuplicateFileMemberInCatalog", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unable to find file name or path for catalog member  {0}.
    ///  
    /// </summary>
    internal static string UnableToFindFileNameOrPathForCatalogMember {
        get {
            return ResourceManager.GetString("UnableToFindFileNameOrPathForCatalogMember", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unable to find file {0} to hash.
    ///  
    /// </summary>
    internal static string UnableToFindFileToHash {
        get {
            return ResourceManager.GetString("UnableToFindFileToHash", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unable to read file {0} to calculate its hash.
    ///  
    /// </summary>
    internal static string UnableToReadFileToHash {
        get {
            return ResourceManager.GetString("UnableToReadFileToHash", resourceCulture);
        }
    }

}

