
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class CimCmdletStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal CimCmdletStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.Management.Infrastructure.CimCmdlets.resources.CimCmdletStrings", typeof(CimCmdletStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Operation '{0}' complete.
    ///    {0} is a placeholder for operation name. (i.e, GetCimInstance)
    ///  
    /// </summary>
    internal static string CimOperationCompleted {
        get {
            return ResourceManager.GetString("CimOperationCompleted", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Create CimInstance
    ///  
    /// </summary>
    internal static string CimOperationNameCreateInstance {
        get {
            return ResourceManager.GetString("CimOperationNameCreateInstance", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Delete CimInstance
    ///  
    /// </summary>
    internal static string CimOperationNameDeleteInstance {
        get {
            return ResourceManager.GetString("CimOperationNameDeleteInstance", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Enumerate Associated CimInstances
    ///  
    /// </summary>
    internal static string CimOperationNameEnumerateAssociatedInstances {
        get {
            return ResourceManager.GetString("CimOperationNameEnumerateAssociatedInstances", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Enumerate CimClasses
    ///  
    /// </summary>
    internal static string CimOperationNameEnumerateClasses {
        get {
            return ResourceManager.GetString("CimOperationNameEnumerateClasses", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Enumerate CimInstances
    ///  
    /// </summary>
    internal static string CimOperationNameEnumerateInstances {
        get {
            return ResourceManager.GetString("CimOperationNameEnumerateInstances", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Get CimClass
    ///  
    /// </summary>
    internal static string CimOperationNameGetClass {
        get {
            return ResourceManager.GetString("CimOperationNameGetClass", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Get CimInstance
    ///  
    /// </summary>
    internal static string CimOperationNameGetInstance {
        get {
            return ResourceManager.GetString("CimOperationNameGetInstance", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Invoke CimMethod
    ///  
    /// </summary>
    internal static string CimOperationNameInvokeMethod {
        get {
            return ResourceManager.GetString("CimOperationNameInvokeMethod", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Modify CimInstance
    ///  
    /// </summary>
    internal static string CimOperationNameModifyInstance {
        get {
            return ResourceManager.GetString("CimOperationNameModifyInstance", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Query CimInstances
    ///  
    /// </summary>
    internal static string CimOperationNameQueryInstances {
        get {
            return ResourceManager.GetString("CimOperationNameQueryInstances", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Subscribe CimIndication
    ///  
    /// </summary>
    internal static string CimOperationNameSubscribeIndication {
        get {
            return ResourceManager.GetString("CimOperationNameSubscribeIndication", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Perform operation '{0}' with following parameters, '{1}'.
    ///    {0} is a placeholder for operation name; {1} is a placeholder for parameters value
    ///  
    /// </summary>
    internal static string CimOperationStart {
        get {
            return ResourceManager.GetString("CimOperationStart", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Parameter '{0}' cannot be used with the parameter '{1}'.
    ///    {0} is a placeholder for parameter name; {1} is a placeholder for another parameter name;
    ///  
    /// </summary>
    internal static string ConflictParameterWasSet {
        get {
            return ResourceManager.GetString("ConflictParameterWasSet", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Could not find CimSession with the given {0} = {1}
    ///    {0} is a placeholder for property name; {1} is a placeholder for property value.
    ///  
    /// </summary>
    internal static string CouldNotFindCimsessionObject {
        get {
            return ResourceManager.GetString("CouldNotFindCimsessionObject", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Could not find the following properties from the given class {0}: {1}.
    ///    {0} is a placeholder for class name; {1} is a placeholder for list of property names.
    ///  
    /// </summary>
    internal static string CouldNotFindPropertyFromGivenClass {
        get {
            return ResourceManager.GetString("CouldNotFindPropertyFromGivenClass", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Could not modify readonly property '{0}' of object '{1}'.
    ///    {0} is a placeholder for propertyname; {1} is a placeholder for object string.
    ///  
    /// </summary>
    internal static string CouldNotModifyReadonlyProperty {
        get {
            return ResourceManager.GetString("CouldNotModifyReadonlyProperty", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Default status description.
    ///    N/A
    ///  
    /// </summary>
    internal static string DefaultStatusDescription {
        get {
            return ResourceManager.GetString("DefaultStatusDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot perform operation because the wildcard path {0} did not resolve to a file.
    ///    {0} is a placeholder for a path
    ///  
    /// </summary>
    internal static string DidNotResolveFile {
        get {
            return ResourceManager.GetString("DidNotResolveFile", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Authentication type '{0}' is invalid without credential. Only following authentication type are allowed without credential, '{1}', '{2}', '{3}', or '{4}'.
    ///    {0} is a placeholder for authentication type. {1}-{4} are placeholders for authentication types.
    ///  
    /// </summary>
    internal static string InvalidAuthenticationTypeWithNullCredential {
        get {
            return ResourceManager.GetString("InvalidAuthenticationTypeWithNullCredential", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Can not find method '{0}' in class '{1}'.
    ///    {0} is a placeholder for method name. {1} is a placeholders for class name.
    ///  
    /// </summary>
    internal static string InvalidMethod {
        get {
            return ResourceManager.GetString("InvalidMethod", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Can not find Parameter '{0}' in method '{1}' of class '{2}'.
    ///    {0} is a placeholder for parameter name; {1} is a placeholder for method name; {2} is a placeholder for class name.
    ///  
    /// </summary>
    internal static string InvalidMethodParameter {
        get {
            return ResourceManager.GetString("InvalidMethodParameter", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Invalid operation. Current cmdlet already have operation created.
    ///    N/A
    ///  
    /// </summary>
    internal static string InvalidOperation {
        get {
            return ResourceManager.GetString("InvalidOperation", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Argument '{0}' contains characters that are not allowed in parameter '{1}'. Supply an argument that is valid and then try the command again.
    ///    {0} stand for argument value, {1} stand for parameter name.
    ///  
    /// </summary>
    internal static string InvalidParameterValue {
        get {
            return ResourceManager.GetString("InvalidParameterValue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot perform operation because the path resolved to more than one file. This command cannot operate on multiple files.
    ///  
    /// </summary>
    internal static string MultipleFilesNotSupported {
        get {
            return ResourceManager.GetString("MultipleFilesNotSupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Argument '{0}' can not be null.
    ///    N/A
    ///  
    /// </summary>
    internal static string NullArgument {
        get {
            return ResourceManager.GetString("NullArgument", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    CimSession proxy object already have operation in progress.
    ///    N/A
    ///  
    /// </summary>
    internal static string OperationInProgress {
        get {
            return ResourceManager.GetString("OperationInProgress", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot open file because the current provider ({0}) cannot open a file.
    ///    {0} is a placeholder for PowerShell filesystem-like provider name (i.e. registry provider)
    ///  
    /// </summary>
    internal static string ReadWriteFileNotFileSystemProvider {
        get {
            return ResourceManager.GetString("ReadWriteFileNotFileSystemProvider", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unable to add property '{0}' to input object '{1}'. The class schema does not contain the property.
    ///    {0} stand for property name, {1} stand for cim instance path.
    ///  
    /// </summary>
    internal static string UnableToAddPropertyToInstance {
        get {
            return ResourceManager.GetString("UnableToAddPropertyToInstance", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unable to resolve the parameter set name.
    ///    N/A
    ///  
    /// </summary>
    internal static string UnableToResolveParameterSetName {
        get {
            return ResourceManager.GetString("UnableToResolveParameterSetName", resourceCulture);
        }
    }

}

