
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class TypesXmlStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal TypesXmlStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.TypesXmlStrings", typeof(TypesXmlStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0}, {1}({2}) : Error: {3}
    ///  
    /// </summary>
    internal static string FileLineError {
        get {
            return ResourceManager.GetString("FileLineError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0}, {1}({2}) : Error in type "{3}": {4}
    ///  
    /// </summary>
    internal static string FileLineTypeError {
        get {
            return ResourceManager.GetString("FileLineTypeError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Node "{0}" must occur only once under "{1}". The parent node, "{1}", will be ignored.
    ///  
    /// </summary>
    internal static string NotMoreThanOnceOne {
        get {
            return ResourceManager.GetString("NotMoreThanOnceOne", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The node {0} is not allowed. The following nodes are allowed: {1}.
    ///  
    /// </summary>
    internal static string UnknownNode {
        get {
            return ResourceManager.GetString("UnknownNode", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Node "{0}" should not have an inner text.
    ///  
    /// </summary>
    internal static string NodeShouldNotHaveInnerText {
        get {
            return ResourceManager.GetString("NodeShouldNotHaveInnerText", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Node "{0}" should have an inner text.
    ///  
    /// </summary>
    internal static string NodeShouldHaveInnerText {
        get {
            return ResourceManager.GetString("NodeShouldHaveInnerText", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Node "{0}" was not found. It should occur only once under "{1}". The parent node, "{1}", will be ignored.
    ///  
    /// </summary>
    internal static string NodeNotFoundOnce {
        get {
            return ResourceManager.GetString("NodeNotFoundOnce", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The "Type" node must have "Members", "TypeConverters", or "TypeAdapters".
    ///  
    /// </summary>
    internal static string TypeNodeShouldHaveMembersOrTypeConverters {
        get {
            return ResourceManager.GetString("TypeNodeShouldHaveMembersOrTypeConverters", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot create an instance of the type converter for type {0} due to exception: {1}.
    ///  
    /// </summary>
    internal static string UnableToInstantiateTypeConverter {
        get {
            return ResourceManager.GetString("UnableToInstantiateTypeConverter", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    PowerShell cannot create an instance of the type adapter for the type {0} because of the following exception: {1}.
    ///  
    /// </summary>
    internal static string UnableToInstantiateTypeAdapter {
        get {
            return ResourceManager.GetString("UnableToInstantiateTypeAdapter", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The adapted type "{0}" is not valid.
    ///  
    /// </summary>
    internal static string InvalidAdaptedType {
        get {
            return ResourceManager.GetString("InvalidAdaptedType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The TypeConverter was ignored because it already occurs.
    ///  
    /// </summary>
    internal static string TypeConverterAlreadyPresent {
        get {
            return ResourceManager.GetString("TypeConverterAlreadyPresent", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The TypeAdapter was ignored because it already occurs.
    ///  
    /// </summary>
    internal static string TypeAdapterAlreadyPresent {
        get {
            return ResourceManager.GetString("TypeAdapterAlreadyPresent", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Type "{0}" should be either TypeConverter or PSTypeConverter.
    ///  
    /// </summary>
    internal static string TypeIsNotTypeConverter {
        get {
            return ResourceManager.GetString("TypeIsNotTypeConverter", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Type "{0}" should be a PSPropertyAdapter.
    ///  
    /// </summary>
    internal static string TypeIsNotTypeAdapter {
        get {
            return ResourceManager.GetString("TypeIsNotTypeAdapter", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The member {0} is already present.
    ///  
    /// </summary>
    internal static string DuplicateMember {
        get {
            return ResourceManager.GetString("DuplicateMember", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The following member name is reserved: {0}
    ///  
    /// </summary>
    internal static string ReservedNameMember {
        get {
            return ResourceManager.GetString("ReservedNameMember", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Exception: {0}
    ///  
    /// </summary>
    internal static string Exception {
        get {
            return ResourceManager.GetString("Exception", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The ScriptProperty should have a getter or setter.
    ///  
    /// </summary>
    internal static string ScriptPropertyShouldHaveGetterOrSetter {
        get {
            return ResourceManager.GetString("ScriptPropertyShouldHaveGetterOrSetter", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The CodeProperty should have a getter or setter.
    ///  
    /// </summary>
    internal static string CodePropertyShouldHaveGetterOrSetter {
        get {
            return ResourceManager.GetString("CodePropertyShouldHaveGetterOrSetter", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0}, {1} : {2}
    ///  
    /// </summary>
    internal static string FileError {
        get {
            return ResourceManager.GetString("FileError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The value should be either TRUE or FALSE instead of {0}.
    ///  
    /// </summary>
    internal static string ValueShouldBeTrueOrFalse {
        get {
            return ResourceManager.GetString("ValueShouldBeTrueOrFalse", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Node "{0}" should not have "{1}" attribute.
    ///  
    /// </summary>
    internal static string IsHiddenNotSupported {
        get {
            return ResourceManager.GetString("IsHiddenNotSupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0}, {1}: The file was not found.
    ///  
    /// </summary>
    internal static string FileNotFound {
        get {
            return ResourceManager.GetString("FileNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0}, {1}: The file was skipped because it was already loaded by {2}.
    ///  
    /// </summary>
    internal static string DuplicateFile {
        get {
            return ResourceManager.GetString("DuplicateFile", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot find the registry key: {0}{1}. Using {2} to load the configuration files.
    ///  
    /// </summary>
    internal static string CannotFindRegistryKey {
        get {
            return ResourceManager.GetString("CannotFindRegistryKey", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot find the path {0} specified in the registry key: {1}{2}. Using {3} to load the configuration files.
    ///  
    /// </summary>
    internal static string CannotFindRegistryKeyPath {
        get {
            return ResourceManager.GetString("CannotFindRegistryKeyPath", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0}, {1}: The file was skipped because it does not have the ps1xml file name extension.
    ///  
    /// </summary>
    internal static string EntryShouldBeMshXml {
        get {
            return ResourceManager.GetString("EntryShouldBeMshXml", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0}, {1}: The file was skipped because of the following validation exception: {2}.
    ///  
    /// </summary>
    internal static string ValidationException {
        get {
            return ResourceManager.GetString("ValidationException", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Member "{0}" must be a note.
    ///  
    /// </summary>
    internal static string MemberShouldBeNote {
        get {
            return ResourceManager.GetString("MemberShouldBeNote", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot convert note "{0}":"{1}".
    ///  
    /// </summary>
    internal static string ErrorConvertingNote {
        get {
            return ResourceManager.GetString("ErrorConvertingNote", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Do not use the member "{0}" here.
    ///  
    /// </summary>
    internal static string MemberShouldNotBePresent {
        get {
            return ResourceManager.GetString("MemberShouldNotBePresent", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Member "{0}" must have type "{1}".
    ///  
    /// </summary>
    internal static string MemberShouldHaveType {
        get {
            return ResourceManager.GetString("MemberShouldHaveType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    "{0}" must be present when the "{1}" is "{2}" and "{3}" is "{4}".
    ///  
    /// </summary>
    internal static string MemberMustBePresent {
        get {
            return ResourceManager.GetString("MemberMustBePresent", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A previous error caused all serialization settings to be ignored.
    ///  
    /// </summary>
    internal static string SerializationSettingsIgnored {
        get {
            return ResourceManager.GetString("SerializationSettingsIgnored", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    "{0}" is not a standard member and will be ignored.
    ///  
    /// </summary>
    internal static string NotAStandardMember {
        get {
            return ResourceManager.GetString("NotAStandardMember", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The {0} path is not fully qualified. Specify a fully qualified type file path.
    ///  
    /// </summary>
    internal static string TypeFileNotRooted {
        get {
            return ResourceManager.GetString("TypeFileNotRooted", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The TypeTable cannot be updated because the TypeTable might have been created outside of the runspace.
    ///  
    /// </summary>
    internal static string SharedTypeTableCannotBeUpdated {
        get {
            return ResourceManager.GetString("SharedTypeTableCannotBeUpdated", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    There were errors loading TypeTable. Look in the Errors property to get detailed error messages.
    ///  
    /// </summary>
    internal static string TypeTableLoadErrors {
        get {
            return ResourceManager.GetString("TypeTableLoadErrors", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error in TypeData "{0}": {1}
    ///  
    /// </summary>
    internal static string TypeDataTypeError {
        get {
            return ResourceManager.GetString("TypeDataTypeError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    "{0}" should have a value for its property "{1}".
    ///  
    /// </summary>
    internal static string TypeDataShouldHaveValue {
        get {
            return ResourceManager.GetString("TypeDataShouldHaveValue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    "{0}" should not have null or an empty string in its property "{1}".
    ///  
    /// </summary>
    internal static string TypeDataShouldNotBeNullOrEmpty {
        get {
            return ResourceManager.GetString("TypeDataShouldNotBeNullOrEmpty", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The type "{0}" was not found. The type name value must be the full name of the type. Verify the type name and run the command again.
    ///  
    /// </summary>
    internal static string TypeNotFound {
        get {
            return ResourceManager.GetString("TypeNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The TypeData must have "Members", "TypeConverters", "TypeAdapters" or "StandardMembers".
    ///  
    /// </summary>
    internal static string TypeDataShouldNotBeEmpty {
        get {
            return ResourceManager.GetString("TypeDataShouldNotBeEmpty", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A shared type table cannot be updated with more than one entry.
    ///  
    /// </summary>
    internal static string TypeTableCannotCoExist {
        get {
            return ResourceManager.GetString("TypeTableCannotCoExist", resourceCulture);
        }
    }

}

