
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class CmdletizationCoreResources {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal CmdletizationCoreResources() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.CmdletizationCoreResources", typeof(CmdletizationCoreResources).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cmdlets over '{0}' class
    ///    {0} is a placeholder for a name of a CIM class.  Example: "ROOT\cimv2\Win32_Process"
    ///  
    /// </summary>
    internal static string ScriptWriter_ModuleDescription {
        get {
            return ResourceManager.GetString("ScriptWriter_ModuleDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Two cmdlet parameters defined within the {0} element have the same name: {1}.  Resolve the conflict in the Cmdlet Definition XML and retry.
    ///    {StrContains="CmdletParameterMetadata"} {StrContains="PSName"}
    ///{0} is a placeholder for a name of an XML element.  Example: <GetCmdletParameters>
    ///{1} is a placeholder for a cmdlet parameter name.  Example: Name
    ///  
    /// </summary>
    internal static string ScriptWriter_DuplicateQueryParameterName {
        get {
            return ResourceManager.GetString("ScriptWriter_DuplicateQueryParameterName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process Cmdlet Definition XML for the following file: {0}. {1}
    ///    {0} is a placeholder for a file name.
    ///{1} is an exception message copied from an XmlException or XmlSchemaException
    ///  
    /// </summary>
    internal static string ExportCimCommand_ErrorInCmdletizationXmlFile {
        get {
            return ResourceManager.GetString("ExportCimCommand_ErrorInCmdletizationXmlFile", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The {0} cmdlet defines the {1} parameter set more than once.  Verify that the Cmdlet Definition XML does not have duplicate parameter set names and retry.
    ///    {StrContains="CmdletParameterSet"}
    ///{0} is a placeholder for a cmdlet name. Example: 'Get-Win32Process'
    ///{1} is a placeholder for a parameter set name.  Example: 'foo'
    ///  
    /// </summary>
    internal static string ScriptWriter_DuplicateParameterSetInStaticCmdlet {
        get {
            return ResourceManager.GetString("ScriptWriter_DuplicateParameterSetInStaticCmdlet", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the ObjectModelWrapper attribute. The {0} type defines multiple parameter sets. Verify that the Cmdlet Definition XML specifies a valid type in the ObjectModelWrapper attribute and retry.
    ///    {StrContains="ObjectModelWrapper"}
    ///{0} is a placeholder for a .NET class name.  Example: System.Net.IPAddress
    ///  
    /// </summary>
    internal static string ScriptWriter_ObjectModelWrapperDefinesMultipleParameterSets {
        get {
            return ResourceManager.GetString("ScriptWriter_ObjectModelWrapperDefinesMultipleParameterSets", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the ObjectModelWrapper attribute. The {0} type is an open generic type.  Verify that the Cmdlet Definition XML specifies a valid type in the ObjectModelWrapper attribute and retry.
    ///    {StrContains="ObjectModelWrapper"}
    ///{0} is a placeholder for a .NET class name.  Example: System.Net.IPAddress
    ///  
    /// </summary>
    internal static string ScriptWriter_ObjectModelWrapperIsStillGeneric {
        get {
            return ResourceManager.GetString("ScriptWriter_ObjectModelWrapperIsStillGeneric", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the ObjectModelWrapper attribute. The {0} type is not derived from the following class: {1}.  Verify that the Cmdlet Definition XML specifies a valid type in the ObjectModelWrapper attribute and retry.
    ///    {StrContains="ObjectModelWrapper"}
    ///{0} is a placeholder for a .NET class name.  Example: System.Net.IPAddress
    ///{1} is a placeholder for a .NET class name. Example: Microsoft.PowerShell.Cmdletization.ObjectModelWrapper
    ///  
    /// </summary>
    internal static string ScriptWriter_ObjectModelWrapperNotDerivedFromObjectModelWrapper {
        get {
            return ResourceManager.GetString("ScriptWriter_ObjectModelWrapperNotDerivedFromObjectModelWrapper", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the ObjectModelWrapper attribute. The {0} type defines the {1} cmdlet parameter with a {2} attribute parameter that is ignored.  Verify that the Cmdlet Definition XML specifies a valid type in the ObjectModelWrapper attribute and retry.
    ///    {StrContains="ObjectModelWrapper"}
    ///{0} is a placeholder for a .NET class name.  Example: System.Net.IPAddress
    ///{1} is a placeholder for a cmdlet parameter name.  Example: 'ProcessId'
    ///{2} is a placeholder for a name of a property of ParameterAttribute class. Example: ValueFromPipelineByPropertyName 
    ///  
    /// </summary>
    internal static string ScriptWriter_ObjectModelWrapperUsesIgnoredParameterMetadata {
        get {
            return ResourceManager.GetString("ScriptWriter_ObjectModelWrapperUsesIgnoredParameterMetadata", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot define the {0} parameter for the {1} cmdlet.  The parameter name is already defined by the {2} class.  Change the name of the parameter in Cmdlet Definition XML and retry.
    ///    {0} is a placeholder for a cmdlet parameter name.  Example: 'ProcessId'
    ///{1} is a placeholder for a cmdlet name. Example: 'Get-Win32Process'
    ///{2} is a placeholder for a .NET class name.  Example: 'Microsoft.PowerShell.Cmdletization.Cim.CimWrapper'
    ///  
    /// </summary>
    internal static string ScriptWriter_ParameterNameConflictsWithCommonParameters {
        get {
            return ResourceManager.GetString("ScriptWriter_ParameterNameConflictsWithCommonParameters", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot define the {0} parameter for the {1} cmdlet. The parameter name is already defined within the {2} XML element. Change the name of the parameter in the Cmdlet Definition XML, and then try again.
    ///    {0} is a placeholder for a cmdlet parameter name.  Example: 'ProcessId'
    ///{1} is a placeholder for a cmdlet name. Example: 'Get-Win32Process'
    ///{2} is a placeholder for an xml element name.  Example: <GetCmdletParameters>
    ///  
    /// </summary>
    internal static string ScriptWriter_ParameterNameConflictsWithQueryParameters {
        get {
            return ResourceManager.GetString("ScriptWriter_ParameterNameConflictsWithQueryParameters", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0} {1}
    ///    This is a resource string, to support locales where the order of placeholders might need to be reversed.
    ///{0} is a placeholder for a top-level exception message (i.e. "There is an error in XML document (2, 2).")
    ///{1} is a placeholder for a secondary exception message (i.e. "")
    ///  
    /// </summary>
    internal static string ScriptWriter_ConcatenationOfDeserializationExceptions {
        get {
            return ResourceManager.GetString("ScriptWriter_ConcatenationOfDeserializationExceptions", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The value of the EnumName attribute doesn't translate to a valid C# identifier: {0}. Verify the EnumName attribute in the Cmdlet Definition XML, and then try again.
    ///    {StrContains="EnumName"}
    ///  
    /// </summary>
    internal static string EnumWriter_InvalidEnumName {
        get {
            return ResourceManager.GetString("EnumWriter_InvalidEnumName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The value of the Name attribute is not a valid C# identifier: {0}.  Verify the Name attribute in the Cmdlet Definition XML, and then try again.
    ///    {StrContains="Enum"} {StrContains="Value"} {StrContains="Name"}
    ///  
    /// </summary>
    internal static string EnumWriter_InvalidValueName {
        get {
            return ResourceManager.GetString("EnumWriter_InvalidValueName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the <Enum EnumName="{0}" ...> element.  {1}
    ///    {StrContains="Enum"} {StrContains="EnumName"}
    ///  
    /// </summary>
    internal static string ScriptWriter_InvalidEnum {
        get {
            return ResourceManager.GetString("ScriptWriter_InvalidEnum", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The remote computer returned a CDXML file that is not valid. The following cmdlet adapter is not supported for importing a CDXML module from a remote computer: {0}
    ///    {0} is a placeholder for a fully qualified type name
    ///  
    /// </summary>
    internal static string ImportModule_UnsupportedCmdletAdapter {
        get {
            return ResourceManager.GetString("ImportModule_UnsupportedCmdletAdapter", resourceCulture);
        }
    }

}

