
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class InternalCommandStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal InternalCommandStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.InternalCommandStrings", typeof(InternalCommandStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Input name "{0}" is ambiguous. It can be resolved to multiple matched methods. Possible matches include:{1}.
    ///  
    /// </summary>
    internal static string AmbiguousMethodName {
        get {
            return ResourceManager.GetString("AmbiguousMethodName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Input name "{0}" is ambiguous. It can be resolved to multiple matched members. Possible matches include:{1}.
    ///  
    /// </summary>
    internal static string AmbiguousPropertyOrMethodName {
        get {
            return ResourceManager.GetString("AmbiguousPropertyOrMethodName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Retrieve the value for key '{0}'
    ///  
    /// </summary>
    internal static string ForEachObjectKeyAction {
        get {
            return ResourceManager.GetString("ForEachObjectKeyAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Invoke method '{0}' with arguments: {1}
    ///  
    /// </summary>
    internal static string ForEachObjectMethodActionWithArguments {
        get {
            return ResourceManager.GetString("ForEachObjectMethodActionWithArguments", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Invoke method '{0}'
    ///  
    /// </summary>
    internal static string ForEachObjectMethodActionWithoutArguments {
        get {
            return ResourceManager.GetString("ForEachObjectMethodActionWithoutArguments", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Retrieve the value for property '{0}'
    ///  
    /// </summary>
    internal static string ForEachObjectPropertyAction {
        get {
            return ResourceManager.GetString("ForEachObjectPropertyAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    InputObject: {0}
    ///  
    /// </summary>
    internal static string ForEachObjectTarget {
        get {
            return ResourceManager.GetString("ForEachObjectTarget", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot operate on a 'null' input object.
    ///  
    /// </summary>
    internal static string InputObjectIsNull {
        get {
            return ResourceManager.GetString("InputObjectIsNull", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Input name "{0}" cannot be resolved to a method.
    ///  
    /// </summary>
    internal static string MethodNotFound {
        get {
            return ResourceManager.GetString("MethodNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot invoke a method in the restricted language mode.
    ///  
    /// </summary>
    internal static string NoMethodInvocationInRestrictedLanguageMode {
        get {
            return ResourceManager.GetString("NoMethodInvocationInRestrictedLanguageMode", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The -WhatIf and -Confirm parameters are not supported for script blocks.
    ///  
    /// </summary>
    internal static string NoShouldProcessForScriptBlockSet {
        get {
            return ResourceManager.GetString("NoShouldProcessForScriptBlockSet", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The '{0}' operation is not allowed in the RestrictedLanguage mode.
    ///  
    /// </summary>
    internal static string OperationNotAllowedInRestrictedLanguageMode {
        get {
            return ResourceManager.GetString("OperationNotAllowedInRestrictedLanguageMode", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    An operator is required to compare the two specified values. Include a valid operator in the command, and then try the command again. For example, Get-Process | Where-Object -Property Name -eq Idle
    ///  
    /// </summary>
    internal static string OperatorNotSpecified {
        get {
            return ResourceManager.GetString("OperatorNotSpecified", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The input name "{0}" cannot be resolved to a property.
    ///  
    /// </summary>
    internal static string PropertyNotFound {
        get {
            return ResourceManager.GetString("PropertyNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The input name "{0}" cannot be resolved to a member.
    ///  
    /// </summary>
    internal static string PropertyOrMethodNotFound {
        get {
            return ResourceManager.GetString("PropertyOrMethodNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified operator requires both the -Property and -Value parameters. Provide values for both parameters, and then try the command again.
    ///  
    /// </summary>
    internal static string ValueNotSpecifiedForWhereObject {
        get {
            return ResourceManager.GetString("ValueNotSpecifiedForWhereObject", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    This method cannot be run on the current thread. It can only be called on the cmdlet thread.
    ///  
    /// </summary>
    internal static string PSTaskStreamWriterWrongThread {
        get {
            return ResourceManager.GetString("PSTaskStreamWriterWrongThread", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A ForEach-Object -Parallel using variable cannot be a script block. Passed-in script block variables are not supported with ForEach-Object -Parallel, and can result in undefined behavior.
    ///  
    /// </summary>
    internal static string ParallelUsingVariableCannotBeScriptBlock {
        get {
            return ResourceManager.GetString("ParallelUsingVariableCannotBeScriptBlock", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A ForEach-Object -Parallel piped input object cannot be a script block. Passed-in script block variables are not supported with ForEach-Object -Parallel, and can result in undefined behavior.
    ///  
    /// </summary>
    internal static string ParallelPipedInputObjectCannotBeScriptBlock {
        get {
            return ResourceManager.GetString("ParallelPipedInputObjectCannotBeScriptBlock", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The 'TimeoutSeconds' parameter cannot be used with the 'AsJob' parameter.
    ///  
    /// </summary>
    internal static string ParallelCannotUseTimeoutWithJob {
        get {
            return ResourceManager.GetString("ParallelCannotUseTimeoutWithJob", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The following common parameters are not currently supported in the Parallel parameter set:
    ///ErrorAction, WarningAction, InformationAction, PipelineVariable
    ///  
    /// </summary>
    internal static string ParallelCommonParametersNotSupported {
        get {
            return ResourceManager.GetString("ParallelCommonParametersNotSupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    An unexpected error has occurred while processing ForEach-Object -Parallel input. This may mean that some of the piped input did not get processed. Error: {0}.
    ///  
    /// </summary>
    internal static string ParallelPipedInputProcessingError {
        get {
            return ResourceManager.GetString("ParallelPipedInputProcessingError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ForEach-Object Cmdlet
    ///  
    /// </summary>
    internal static string WDACLogTitle {
        get {
            return ResourceManager.GetString("WDACLogTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Method invocation on type '{0}' will not be allowed when run in Constrained Language mode.
    ///  
    /// </summary>
    internal static string WDACLogMessage {
        get {
            return ResourceManager.GetString("WDACLogMessage", resourceCulture);
        }
    }

}

