
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class ExtendedTypeSystem {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal ExtendedTypeSystem() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.ExtendedTypeSystem", typeof(ExtendedTypeSystem).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The member "{0}" is already present.
    ///  
    /// </summary>
    internal static string MemberAlreadyPresent {
        get {
            return ResourceManager.GetString("MemberAlreadyPresent", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The member "{0}" is already present from the extended type data file.
    ///  
    /// </summary>
    internal static string MemberAlreadyPresentFromTypesXml {
        get {
            return ResourceManager.GetString("MemberAlreadyPresentFromTypesXml", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The member "{0}" is not present.
    ///  
    /// </summary>
    internal static string MemberNotPresent {
        get {
            return ResourceManager.GetString("MemberNotPresent", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Exception setting "{0}": "{1}"
    ///  
    /// </summary>
    internal static string ExceptionWhenSetting {
        get {
            return ResourceManager.GetString("ExceptionWhenSetting", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Exception getting "{0}": "{1}"
    ///  
    /// </summary>
    internal static string ExceptionWhenGetting {
        get {
            return ResourceManager.GetString("ExceptionWhenGetting", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The following exception occurred while trying to enumerate the collection: "{0}".
    ///  
    /// </summary>
    internal static string EnumerationException {
        get {
            return ResourceManager.GetString("EnumerationException", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot access member "{0}" outside of a PSObject.
    ///  
    /// </summary>
    internal static string AccessMemberOutsidePSObject {
        get {
            return ResourceManager.GetString("AccessMemberOutsidePSObject", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot change the member created from the type configuration: "{0}".
    ///  
    /// </summary>
    internal static string ChangeStaticMember {
        get {
            return ResourceManager.GetString("ChangeStaticMember", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The member name "{0}" is reserved.
    ///  
    /// </summary>
    internal static string ReservedMemberName {
        get {
            return ResourceManager.GetString("ReservedMemberName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    "{0}" cannot be changed.
    ///  
    /// </summary>
    internal static string CannotChangeReservedMember {
        get {
            return ResourceManager.GetString("CannotChangeReservedMember", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Exception calling "{0}" with "{1}" argument(s): "{2}"
    ///  
    /// </summary>
    internal static string MethodInvocationException {
        get {
            return ResourceManager.GetString("MethodInvocationException", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    An exception was thrown when trying to call "{0}" to extract the contents of an object of type "{1}": "{2}"
    ///  
    /// </summary>
    internal static string CopyToInvocationException {
        get {
            return ResourceManager.GetString("CopyToInvocationException", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot find an overload for "{0}" and the argument count: "{1}".
    ///  
    /// </summary>
    internal static string MethodArgumentCountException {
        get {
            return ResourceManager.GetString("MethodArgumentCountException", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Could not find a suitable generic method overload for "{0}" with "{1}" type parameters, and the argument count: "{2}".
    ///  
    /// </summary>
    internal static string MethodGenericArgumentCountException {
        get {
            return ResourceManager.GetString("MethodGenericArgumentCountException", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Multiple ambiguous overloads found for "{0}" and the argument count: "{1}".
    ///  
    /// </summary>
    internal static string MethodAmbiguousException {
        get {
            return ResourceManager.GetString("MethodAmbiguousException", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot convert argument "{0}", with value: "{1}", for "{2}" to type "{3}": "{4}"
    ///  
    /// </summary>
    internal static string MethodArgumentConversionException {
        get {
            return ResourceManager.GetString("MethodArgumentConversionException", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Get accessor for property "{0}" is unavailable.
    ///  
    /// </summary>
    internal static string GetWithoutGetterException {
        get {
            return ResourceManager.GetString("GetWithoutGetterException", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Set accessor for property "{0}" is unavailable.
    ///  
    /// </summary>
    internal static string SetWithoutSetterException {
        get {
            return ResourceManager.GetString("SetWithoutSetterException", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The setter method should be public, void, static, and have two parameters. The first parameter should be of the type PSObject. A second parameter is required if a getter method is also available, and should have the same type as the return type for the getter method.
    ///  
    /// </summary>
    internal static string CodePropertySetterFormat {
        get {
            return ResourceManager.GetString("CodePropertySetterFormat", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The getter method should be public, not void, static, and have one parameter of the type PSObject.
    ///  
    /// </summary>
    internal static string CodePropertyGetterFormat {
        get {
            return ResourceManager.GetString("CodePropertyGetterFormat", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    CodeProperty should use a getter or setter method.
    ///  
    /// </summary>
    internal static string CodePropertyGetterAndSetterNull {
        get {
            return ResourceManager.GetString("CodePropertyGetterAndSetterNull", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot create a code method because of the method format. The method should be public, static, and have one parameter of type PSObject.
    ///  
    /// </summary>
    internal static string CodeMethodMethodFormat {
        get {
            return ResourceManager.GetString("CodeMethodMethodFormat", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The alias with name "{0}" contains a cycle.
    ///  
    /// </summary>
    internal static string CycleInAlias {
        get {
            return ResourceManager.GetString("CycleInAlias", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot convert the "{0}" value of type "{1}" to type "{2}".
    ///  
    /// </summary>
    internal static string InvalidCastException {
        get {
            return ResourceManager.GetString("InvalidCastException", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot convert the value of type "{0}" to type "{1}".
    ///  
    /// </summary>
    internal static string InvalidCastExceptionWithoutValue {
        get {
            return ResourceManager.GetString("InvalidCastExceptionWithoutValue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot convert value "{0}" to type "{1}". Error: "{2}"
    ///  
    /// </summary>
    internal static string InvalidCastExceptionWithInnerException {
        get {
            return ResourceManager.GetString("InvalidCastExceptionWithInnerException", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot convert value "{0}" to type "{1}" because no commas are allowed for this enumeration.
    ///  
    /// </summary>
    internal static string InvalidCastExceptionEnumerationNoFlagAndComma {
        get {
            return ResourceManager.GetString("InvalidCastExceptionEnumerationNoFlagAndComma", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot convert value "{0}" to type "{1}" due to enumeration values that are not valid. Specify one of the following enumeration values and try again. The possible enumeration values are "{2}".
    ///  
    /// </summary>
    internal static string InvalidCastExceptionEnumerationNoValue {
        get {
            return ResourceManager.GetString("InvalidCastExceptionEnumerationNoValue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot convert null to type "{0}" due to enumeration values that are not valid. Specify one of the following enumeration values and try again. The possible enumeration values are "{1}".
    ///  
    /// </summary>
    internal static string InvalidCastExceptionEnumerationNull {
        get {
            return ResourceManager.GetString("InvalidCastExceptionEnumerationNull", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot convert null to type "{0}".
    ///  
    /// </summary>
    internal static string InvalidCastFromNull {
        get {
            return ResourceManager.GetString("InvalidCastFromNull", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot convert value to type "{0}".  Error: "{1}"
    ///  
    /// </summary>
    internal static string InvalidCastExceptionNoStringForConversion {
        get {
            return ResourceManager.GetString("InvalidCastExceptionNoStringForConversion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot convert value to type System.String.
    ///  
    /// </summary>
    internal static string InvalidCastCannotRetrieveString {
        get {
            return ResourceManager.GetString("InvalidCastCannotRetrieveString", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Reference type is expected in argument.
    ///  
    /// </summary>
    internal static string ReferenceTypeExpected {
        get {
            return ResourceManager.GetString("ReferenceTypeExpected", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot compare "{0}" because it is not IComparable.
    ///  
    /// </summary>
    internal static string NotIcomparable {
        get {
            return ResourceManager.GetString("NotIcomparable", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Could not compare "{0}" to "{1}". Error: "{2}"
    ///  
    /// </summary>
    internal static string ComparisonFailure {
        get {
            return ResourceManager.GetString("ComparisonFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot compare "{0}" to "{1}" because the objects are not the same type or the object "{0}" does not implement "{2}".
    ///  
    /// </summary>
    internal static string NotTheSameTypeOrNotIcomparable {
        get {
            return ResourceManager.GetString("NotTheSameTypeOrNotIcomparable", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot convert value "{0}" to type "{1}" because at least two matches were found ({2}, {3}) and only one match is allowed for this enumeration.
    ///  
    /// </summary>
    internal static string InvalidCastExceptionEnumerationMoreThanOneValue {
        get {
            return ResourceManager.GetString("InvalidCastExceptionEnumerationMoreThanOneValue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot convert value "{0}" to type "{1}". Boolean parameters accept only Boolean values and numbers, such as $True, $False, 1 or 0.
    ///  
    /// </summary>
    internal static string InvalidCastExceptionForBooleanArgumentValue {
        get {
            return ResourceManager.GetString("InvalidCastExceptionForBooleanArgumentValue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot get property value because "{0}" is a write-only property.
    ///  
    /// </summary>
    internal static string WriteOnlyProperty {
        get {
            return ResourceManager.GetString("WriteOnlyProperty", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    "{0}" is a ReadOnly property.
    ///  
    /// </summary>
    internal static string ReadOnlyProperty {
        get {
            return ResourceManager.GetString("ReadOnlyProperty", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot set "{0}" because only strings can be used as values to set XmlNode properties.
    ///  
    /// </summary>
    internal static string XmlNodeSetShouldBeAString {
        get {
            return ResourceManager.GetString("XmlNodeSetShouldBeAString", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot set "{0}" because only unique attributes or unique non-attributed leaf nodes can be set.
    ///  
    /// </summary>
    internal static string XmlNodeSetRestrictions {
        get {
            return ResourceManager.GetString("XmlNodeSetRestrictions", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A PSProperty or PSMethod object cannot be added to this collection.
    ///  
    /// </summary>
    internal static string CannotAddPropertyOrMethod {
        get {
            return ResourceManager.GetString("CannotAddPropertyOrMethod", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The following error occurred while loading the extended type data file: {0}
    ///  
    /// </summary>
    internal static string TypesXmlError {
        get {
            return ResourceManager.GetString("TypesXmlError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The following exception occurred while retrieving the string: "{0}"
    ///  
    /// </summary>
    internal static string ToStringException {
        get {
            return ResourceManager.GetString("ToStringException", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The field or property: "{0}" for type: "{1}" differs only in letter casing from the field or property: "{2}". The type must be Common Language Specification (CLS) compliant.
    ///  
    /// </summary>
    internal static string NotAClsCompliantFieldProperty {
        get {
            return ResourceManager.GetString("NotAClsCompliantFieldProperty", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The following exception occurred while retrieving the type name hierarchy: "{0}".
    ///  
    /// </summary>
    internal static string ExceptionRetrievingTypeNameHierarchy {
        get {
            return ResourceManager.GetString("ExceptionRetrievingTypeNameHierarchy", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The following exception occurred while retrieving member "{1}": "{0}"
    ///  
    /// </summary>
    internal static string ExceptionGettingMember {
        get {
            return ResourceManager.GetString("ExceptionGettingMember", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The following exception occurred while retrieving members: "{0}"
    ///  
    /// </summary>
    internal static string ExceptionGettingMembers {
        get {
            return ResourceManager.GetString("ExceptionGettingMembers", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The following exception occurred while retrieving the read state for property "{1}": "{0}"
    ///  
    /// </summary>
    internal static string ExceptionRetrievingPropertyReadState {
        get {
            return ResourceManager.GetString("ExceptionRetrievingPropertyReadState", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The following exception occurred while retrieving the write state for property "{1}": "{0}"
    ///  
    /// </summary>
    internal static string ExceptionRetrievingPropertyWriteState {
        get {
            return ResourceManager.GetString("ExceptionRetrievingPropertyWriteState", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The following exception occurred while retrieving the type for property "{1}": "{0}"
    ///  
    /// </summary>
    internal static string ExceptionRetrievingPropertyType {
        get {
            return ResourceManager.GetString("ExceptionRetrievingPropertyType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The following exception occurred while retrieving the string representation for property "{1}" : "{0}"
    ///  
    /// </summary>
    internal static string ExceptionRetrievingPropertyString {
        get {
            return ResourceManager.GetString("ExceptionRetrievingPropertyString", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The following exception occurred while retrieving the attributes for property "{1}": "{0}"
    ///  
    /// </summary>
    internal static string ExceptionRetrievingPropertyAttributes {
        get {
            return ResourceManager.GetString("ExceptionRetrievingPropertyAttributes", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The following exception occurred while retrieving the definitions for method "{1}": "{0}"
    ///  
    /// </summary>
    internal static string ExceptionRetrievingMethodDefinitions {
        get {
            return ResourceManager.GetString("ExceptionRetrievingMethodDefinitions", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The following exception occurred while retrieving the string representation for method "{1}": "{0}"
    ///  
    /// </summary>
    internal static string ExceptionRetrievingMethodString {
        get {
            return ResourceManager.GetString("ExceptionRetrievingMethodString", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The following exception occurred while retrieving the type for parameterized property "{1}": "{0}"
    ///  
    /// </summary>
    internal static string ExceptionRetrievingParameterizedPropertytype {
        get {
            return ResourceManager.GetString("ExceptionRetrievingParameterizedPropertytype", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The following exception occurred while retrieving the read state for parameterized property "{1}": "{0}"
    ///  
    /// </summary>
    internal static string ExceptionRetrievingParameterizedPropertyReadState {
        get {
            return ResourceManager.GetString("ExceptionRetrievingParameterizedPropertyReadState", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The following exception occurred while retrieving the write state for parameterized property "{1}": "{0}"
    ///  
    /// </summary>
    internal static string ExceptionRetrievingParameterizedPropertyWriteState {
        get {
            return ResourceManager.GetString("ExceptionRetrievingParameterizedPropertyWriteState", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The following exception occurred while retrieving the definitions for parameterized property "{1}": "{0}"
    ///  
    /// </summary>
    internal static string ExceptionRetrievingParameterizedPropertyDefinitions {
        get {
            return ResourceManager.GetString("ExceptionRetrievingParameterizedPropertyDefinitions", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The following exception occurred while retrieving the string representation for parameterized property "{1}": "{0}"
    ///  
    /// </summary>
    internal static string ExceptionRetrievingParameterizedPropertyString {
        get {
            return ResourceManager.GetString("ExceptionRetrievingParameterizedPropertyString", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot set the Value property for PSMemberInfo object of type "{0}".
    ///  
    /// </summary>
    internal static string CannotSetValueForMemberType {
        get {
            return ResourceManager.GetString("CannotSetValueForMemberType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Argument: '{0}' should be a {1}. Use {2}.
    ///  
    /// </summary>
    internal static string NonRefArgumentToRefParameter {
        get {
            return ResourceManager.GetString("NonRefArgumentToRefParameter", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Argument: '{0}' should not be a {1}. Do not use {2}.
    ///  
    /// </summary>
    internal static string RefArgumentToNonRefParameter {
        get {
            return ResourceManager.GetString("RefArgumentToNonRefParameter", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The property "{0}" was not found.
    ///  
    /// </summary>
    internal static string PropertyNotFoundInTypeDescriptor {
        get {
            return ResourceManager.GetString("PropertyNotFoundInTypeDescriptor", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot get or set the property value. The "{0}" argument should be of type "{1}" or "{2}".
    ///  
    /// </summary>
    internal static string InvalidComponent {
        get {
            return ResourceManager.GetString("InvalidComponent", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot set the value for property "{0}" because the object has type "{1}" instead of "{2}".
    ///  
    /// </summary>
    internal static string CannotSetNonManagementObject {
        get {
            return ResourceManager.GetString("CannotSetNonManagementObject", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Exception calling "{0}" : "{1}"
    ///  
    /// </summary>
    internal static string WMIMethodInvocationException {
        get {
            return ResourceManager.GetString("WMIMethodInvocationException", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0} is not a valid class path.
    ///  
    /// </summary>
    internal static string InvalidWMIClassPath {
        get {
            return ResourceManager.GetString("InvalidWMIClassPath", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0} is not a valid path.
    ///  
    /// </summary>
    internal static string InvalidWMIPath {
        get {
            return ResourceManager.GetString("InvalidWMIPath", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The adapter cannot determine whether property "{0}" can be changed.
    ///  
    /// </summary>
    internal static string PropertyIsSettableError {
        get {
            return ResourceManager.GetString("PropertyIsSettableError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The adapter cannot determine whether property "{0}" is gettable.
    ///  
    /// </summary>
    internal static string PropertyIsGettableError {
        get {
            return ResourceManager.GetString("PropertyIsGettableError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The adapter cannot get the value of property "{0}".
    ///  
    /// </summary>
    internal static string PropertyGetError {
        get {
            return ResourceManager.GetString("PropertyGetError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The adapter cannot set the value of property "{0}".
    ///  
    /// </summary>
    internal static string PropertySetError {
        get {
            return ResourceManager.GetString("PropertySetError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The adapter cannot get the type of property "{0}".
    ///  
    /// </summary>
    internal static string PropertyTypeError {
        get {
            return ResourceManager.GetString("PropertyTypeError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The adapter cannot get the type hierarchy of "{0}".
    ///  
    /// </summary>
    internal static string GetTypeNameHierarchyError {
        get {
            return ResourceManager.GetString("GetTypeNameHierarchyError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The adapter cannot get the properties of "{0}".
    ///  
    /// </summary>
    internal static string GetProperties {
        get {
            return ResourceManager.GetString("GetProperties", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The adapter cannot get property "{0}" for "{1}".
    ///  
    /// </summary>
    internal static string GetProperty {
        get {
            return ResourceManager.GetString("GetProperty", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    "{0}" returned a null value.
    ///  
    /// </summary>
    internal static string NullReturnValueError {
        get {
            return ResourceManager.GetString("NullReturnValueError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The property '{0}' was not found for the '{1}' object. The settable properties are: {2}.
    ///  
    /// </summary>
    internal static string PropertyNotFound {
        get {
            return ResourceManager.GetString("PropertyNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The property '{0}' was not found for the '{1}' object. There is no settable property available.
    ///  
    /// </summary>
    internal static string NoSettableProperty {
        get {
            return ResourceManager.GetString("NoSettableProperty", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot create object of type "{0}". {1}
    ///  
    /// </summary>
    internal static string ObjectCreationError {
        get {
            return ResourceManager.GetString("ObjectCreationError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot invoke static methods or access static properties on the open generic type {0}.  Specify the type parameters and retry.  For example, instead of [System.Collections.Generic.HashSet``1]::CreateSetComparer() use [System.Collections.Generic.HashSet[int]]::CreateSetComparer().
    ///    Error message shown when somebody tries to access a property or invoke a static method on an uninstantiated generic type:
    ///PS> [System.Collections.Generic.Comparer``1]::get_Default()
    ///
    ///{0} is a placeholder for a type name (for example: System.Collections.Generic.Comparer`1)
    ///  
    /// </summary>
    internal static string CannotInvokeStaticMethodOnUninstantiatedGenericType {
        get {
            return ResourceManager.GetString("CannotInvokeStaticMethodOnUninstantiatedGenericType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The following exception occurred while constructing the attribute "{1}": "{0}"
    ///  
    /// </summary>
    internal static string ExceptionConstructingAttribute {
        get {
            return ResourceManager.GetString("ExceptionConstructingAttribute", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The value "{0}" cannot be converted to a string array.
    ///  
    /// </summary>
    internal static string CannotConvertValueToStringArray {
        get {
            return ResourceManager.GetString("CannotConvertValueToStringArray", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot convert value to type "{0}". Only core types are supported in this language mode.
    ///  
    /// </summary>
    internal static string InvalidCastExceptionNonCoreType {
        get {
            return ResourceManager.GetString("InvalidCastExceptionNonCoreType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot convert to the ByRef-like type "{0}". ByRef-like types are not supported in PowerShell.
    ///  
    /// </summary>
    internal static string InvalidCastToByRefLikeType {
        get {
            return ResourceManager.GetString("InvalidCastToByRefLikeType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot get or set the property or field "{0}" of the ByRef-like type "{1}". ByRef-like types are not supported in PowerShell.
    ///  
    /// </summary>
    internal static string CannotAccessByRefLikePropertyOrField {
        get {
            return ResourceManager.GetString("CannotAccessByRefLikePropertyOrField", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot invoke the method "{0}" of the ByRef-like return type "{1}". ByRef-like types are not supported in PowerShell.
    ///  
    /// </summary>
    internal static string CannotCallMethodWithByRefLikeReturnType {
        get {
            return ResourceManager.GetString("CannotCallMethodWithByRefLikeReturnType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot create an instance of the ByRef-like type "{0}". ByRef-like types are not supported in PowerShell.
    ///  
    /// </summary>
    internal static string CannotInstantiateBoxedByRefLikeType {
        get {
            return ResourceManager.GetString("CannotInstantiateBoxedByRefLikeType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Extended Type System Hashtable Conversion
    ///  
    /// </summary>
    internal static string WDACHashTypeLogTitle {
        get {
            return ResourceManager.GetString("WDACHashTypeLogTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Type conversion from HashTable to '{0}' will not be allowed in ConstrainedLanguage mode.
    ///  
    /// </summary>
    internal static string WDACHashTypeLogMessage {
        get {
            return ResourceManager.GetString("WDACHashTypeLogMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Extended Type System Hashtable Conversion
    ///  
    /// </summary>
    internal static string WDACTypeConversionLogTitle {
        get {
            return ResourceManager.GetString("WDACTypeConversionLogTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Type conversion from '{0}' to '{1}' will not be allowed in ConstrainedLanguage mode.
    ///  
    /// </summary>
    internal static string WDACTypeConversionLogMessage {
        get {
            return ResourceManager.GetString("WDACTypeConversionLogMessage", resourceCulture);
        }
    }

}

