
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class ClipboardResources {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal ClipboardResources() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.PowerShell.Commands.Management.resources.ClipboardResources", typeof(ClipboardResources).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Set string '{0}' to the clipboard.
    ///  
    /// </summary>
    internal static string SetClipboardContent {
        get {
            return ResourceManager.GetString("SetClipboardContent", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Append string '{0}' to the clipboard.
    ///  
    /// </summary>
    internal static string AppendClipboardContent {
        get {
            return ResourceManager.GetString("AppendClipboardContent", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Set file '{0}' to the clipboard.
    ///  
    /// </summary>
    internal static string SetSingleFileToClipboard {
        get {
            return ResourceManager.GetString("SetSingleFileToClipboard", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Append file '{0}' to the clipboard.
    ///  
    /// </summary>
    internal static string AppendSingleFileToClipboard {
        get {
            return ResourceManager.GetString("AppendSingleFileToClipboard", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Set {0} files to the clipboard.
    ///  
    /// </summary>
    internal static string SetMultipleFilesToClipboard {
        get {
            return ResourceManager.GetString("SetMultipleFilesToClipboard", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Append {0} files to the clipboard.
    ///  
    /// </summary>
    internal static string AppendMultipleFilesToClipboard {
        get {
            return ResourceManager.GetString("AppendMultipleFilesToClipboard", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    There's no content in Clipboard or the content format is not compatible. Set input object to the Clipboard.
    ///  
    /// </summary>
    internal static string NoAppendableClipboardContent {
        get {
            return ResourceManager.GetString("NoAppendableClipboardContent", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Clipboard has been cleared.
    ///  
    /// </summary>
    internal static string ClipboardCleared {
        get {
            return ResourceManager.GetString("ClipboardCleared", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    TextFormatType can only be combined with Text format.
    ///  
    /// </summary>
    internal static string InvalidTypeCombine {
        get {
            return ResourceManager.GetString("InvalidTypeCombine", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Raw can only be combined with Text or FileDropList format.
    ///  
    /// </summary>
    internal static string InvalidRawCombine {
        get {
            return ResourceManager.GetString("InvalidRawCombine", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Html can only be combined with Html Text format.
    ///  
    /// </summary>
    internal static string InvalidHtmlCombine {
        get {
            return ResourceManager.GetString("InvalidHtmlCombine", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Only Text format is supported on this platform.
    ///  
    /// </summary>
    internal static string UnsupportedFormat {
        get {
            return ResourceManager.GetString("UnsupportedFormat", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The clipboard is not supported on this platform.
    ///  
    /// </summary>
    internal static string UnsupportedPlatform {
        get {
            return ResourceManager.GetString("UnsupportedPlatform", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The '-AsHtml' switch is not supported on this platform.
    ///  
    /// </summary>
    internal static string AsHtmlUnsupported {
        get {
            return ResourceManager.GetString("AsHtmlUnsupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The '-TextFormatType' parameter only supports 'Text' on this platform.
    ///  
    /// </summary>
    internal static string TextFormatUnsupported {
        get {
            return ResourceManager.GetString("TextFormatUnsupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The '-Path' parameter is not supported on this platform.
    ///  
    /// </summary>
    internal static string PathUnsupported {
        get {
            return ResourceManager.GetString("PathUnsupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The '-LiteralPath' parameter is not supported on this platform.
    ///  
    /// </summary>
    internal static string LiteralPathUnsupported {
        get {
            return ResourceManager.GetString("LiteralPathUnsupported", resourceCulture);
        }
    }

}

