
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class ServiceResources {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal ServiceResources() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.PowerShell.Commands.Management.resources.ServiceResources", typeof(ServiceResources).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0} ({1})
    ///  
    /// </summary>
    internal static string ServiceNameForConfirmation {
        get {
            return ResourceManager.GetString("ServiceNameForConfirmation", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot find any service with service name '{0}'.
    ///  
    /// </summary>
    internal static string NoServiceFoundForGivenName {
        get {
            return ResourceManager.GetString("NoServiceFoundForGivenName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot find any service with display name '{1}'.
    ///  
    /// </summary>
    internal static string NoServiceFoundForGivenDisplayName {
        get {
            return ResourceManager.GetString("NoServiceFoundForGivenDisplayName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot stop service '{1} ({0})' because it has dependent services. It can only be stopped if the Force flag is set.
    ///  
    /// </summary>
    internal static string ServiceHasDependentServices {
        get {
            return ResourceManager.GetString("ServiceHasDependentServices", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot stop service '{1} ({0})' because it has dependent services.
    ///  
    /// </summary>
    internal static string ServiceHasDependentServicesNoForce {
        get {
            return ResourceManager.GetString("ServiceHasDependentServicesNoForce", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Service '{1} ({0})' cannot be stopped due to the following error: {2}
    ///  
    /// </summary>
    internal static string CouldNotStopService {
        get {
            return ResourceManager.GetString("CouldNotStopService", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Service '{1} ({0})' cannot be started due to the following error: {2}
    ///  
    /// </summary>
    internal static string CouldNotStartService {
        get {
            return ResourceManager.GetString("CouldNotStartService", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Service '{1} ({0})' cannot be suspended due to the following error: {2}
    ///  
    /// </summary>
    internal static string CouldNotSuspendService {
        get {
            return ResourceManager.GetString("CouldNotSuspendService", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Service '{1} ({0})' cannot be suspended because the service does not support being suspended or resumed.
    ///  
    /// </summary>
    internal static string CouldNotSuspendServiceNotSupported {
        get {
            return ResourceManager.GetString("CouldNotSuspendServiceNotSupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Service '{1} ({0})' cannot be suspended because it is not currently running.
    ///  
    /// </summary>
    internal static string CouldNotSuspendServiceNotRunning {
        get {
            return ResourceManager.GetString("CouldNotSuspendServiceNotRunning", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Service '{1} ({0})' cannot be resumed due to the following error: {2}
    ///  
    /// </summary>
    internal static string CouldNotResumeService {
        get {
            return ResourceManager.GetString("CouldNotResumeService", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Service '{1} ({0})' cannot be resumed because the service does not support being suspended or resumed.
    ///  
    /// </summary>
    internal static string CouldNotResumeServiceNotSupported {
        get {
            return ResourceManager.GetString("CouldNotResumeServiceNotSupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Service '{1} ({0})' cannot be resumed because it is not currently running.
    ///  
    /// </summary>
    internal static string CouldNotResumeServiceNotRunning {
        get {
            return ResourceManager.GetString("CouldNotResumeServiceNotRunning", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Service '{1} ({0})' cannot be configured due to the following error: {2}
    ///  
    /// </summary>
    internal static string CouldNotSetService {
        get {
            return ResourceManager.GetString("CouldNotSetService", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Service '{1} ({0})' description cannot be configured due to the following error: {2}
    ///  
    /// </summary>
    internal static string CouldNotSetServiceDescription {
        get {
            return ResourceManager.GetString("CouldNotSetServiceDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Service '{1} ({0})' automatic (delayed start) cannot be configured due to the following error: {2}
    ///  
    /// </summary>
    internal static string CouldNotSetServiceDelayedAutoStart {
        get {
            return ResourceManager.GetString("CouldNotSetServiceDelayedAutoStart", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Service '{0}' security descriptor cannot be configured due to the following error: {1}
    ///  
    /// </summary>
    internal static string CouldNotSetServiceSecurityDescriptorSddl {
        get {
            return ResourceManager.GetString("CouldNotSetServiceSecurityDescriptorSddl", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Service '{1} ({0})' cannot be created due to the following error: {2}
    ///  
    /// </summary>
    internal static string CouldNotNewService {
        get {
            return ResourceManager.GetString("CouldNotNewService", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Service '{1} ({0})' was created, but its description cannot be configured due to the following error: {2}
    ///  
    /// </summary>
    internal static string CouldNotNewServiceDescription {
        get {
            return ResourceManager.GetString("CouldNotNewServiceDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Service '{1} ({0})' was created, but its StartupType 'Automatic (Delayed Start)' could not be configured due to the following error: {2}
    ///  
    /// </summary>
    internal static string CouldNotNewServiceDelayedAutoStart {
        get {
            return ResourceManager.GetString("CouldNotNewServiceDelayedAutoStart", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Service '{1} ({0})' cannot be removed due to the following error: {2}
    ///  
    /// </summary>
    internal static string CouldNotRemoveService {
        get {
            return ResourceManager.GetString("CouldNotRemoveService", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    'Cannot access dependent services of '{1} ({0})'
    ///  
    /// </summary>
    internal static string CouldNotAccessDependentServices {
        get {
            return ResourceManager.GetString("CouldNotAccessDependentServices", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Waiting for service '{1} ({0})' to start...
    ///  
    /// </summary>
    internal static string StartingService {
        get {
            return ResourceManager.GetString("StartingService", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Waiting for service '{1} ({0})' to stop...
    ///  
    /// </summary>
    internal static string StoppingService {
        get {
            return ResourceManager.GetString("StoppingService", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Waiting for service '{1} ({0})' to suspend...
    ///  
    /// </summary>
    internal static string SuspendingService {
        get {
            return ResourceManager.GetString("SuspendingService", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Waiting for service '{1} ({0})' to resume...
    ///  
    /// </summary>
    internal static string ResumingService {
        get {
            return ResourceManager.GetString("ResumingService", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Failed to start service '{1} ({0})'.
    ///  
    /// </summary>
    internal static string StartServiceFailed {
        get {
            return ResourceManager.GetString("StartServiceFailed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Service '{1} ({0})' stop failed.
    ///  
    /// </summary>
    internal static string StopServiceFailed {
        get {
            return ResourceManager.GetString("StopServiceFailed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Service '{1} ({0})' suspend failed.
    ///  
    /// </summary>
    internal static string SuspendServiceFailed {
        get {
            return ResourceManager.GetString("SuspendServiceFailed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Service '{1} ({0})' resume failed.
    ///  
    /// </summary>
    internal static string ResumeServiceFailed {
        get {
            return ResourceManager.GetString("ResumeServiceFailed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Failed to open SCManager due to the following error: {0}. Run PowerShell as admin and run your command again.
    ///  
    /// </summary>
    internal static string FailToOpenServiceControlManager {
        get {
            return ResourceManager.GetString("FailToOpenServiceControlManager", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The startup type '{0}' is not supported by {1}.
    ///  
    /// </summary>
    internal static string UnsupportedStartupType {
        get {
            return ResourceManager.GetString("UnsupportedStartupType", resourceCulture);
        }
    }

}

