
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

public class XamlLocalizableResources {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    public XamlLocalizableResources() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    public static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.Management.UI.Internal.resources.public.XamlLocalizableResources", typeof(XamlLocalizableResources).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    public static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Available Columns
    ///  
    /// </summary>
    public static string AutoResXGen_ColumnPicker_AutomationPropertiesName_49 {
        get {
            return ResourceManager.GetString("AutoResXGen_ColumnPicker_AutomationPropertiesName_49", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Add
    ///  
    /// </summary>
    public static string AutoResXGen_ColumnPicker_AutomationPropertiesName_75 {
        get {
            return ResourceManager.GetString("AutoResXGen_ColumnPicker_AutomationPropertiesName_75", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Remove
    ///  
    /// </summary>
    public static string AutoResXGen_ColumnPicker_AutomationPropertiesName_86 {
        get {
            return ResourceManager.GetString("AutoResXGen_ColumnPicker_AutomationPropertiesName_86", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Selected Columns
    ///  
    /// </summary>
    public static string AutoResXGen_ColumnPicker_AutomationPropertiesName_104 {
        get {
            return ResourceManager.GetString("AutoResXGen_ColumnPicker_AutomationPropertiesName_104", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Back
    ///    Localizable AutomationName for control that is used by accessibility screen readers.
    ///  
    /// </summary>
    public static string AutoResXGen_BackForwardHistory_AutomationPropertiesName_613 {
        get {
            return ResourceManager.GetString("AutoResXGen_BackForwardHistory_AutomationPropertiesName_613", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Forward
    ///    Localizable AutomationName for control that is used by accessibility screen readers.
    ///  
    /// </summary>
    public static string AutoResXGen_BackForwardHistory_AutomationPropertiesName_619 {
        get {
            return ResourceManager.GetString("AutoResXGen_BackForwardHistory_AutomationPropertiesName_619", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Find in this column
    ///    Background text shown in the search box.
    ///  
    /// </summary>
    public static string AutoResXGen_BreadcrumbItem_Text_144 {
        get {
            return ResourceManager.GetString("AutoResXGen_BreadcrumbItem_Text_144", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Expand
    ///  
    /// </summary>
    public static string AutoResXGen_DesignerStyleResources_Tooltip_148 {
        get {
            return ResourceManager.GetString("AutoResXGen_DesignerStyleResources_Tooltip_148", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Name
    ///  
    /// </summary>
    public static string AutoResXGen_ListOrganizer_AutomationPropertiesName_95 {
        get {
            return ResourceManager.GetString("AutoResXGen_ListOrganizer_AutomationPropertiesName_95", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    New Query
    ///  
    /// </summary>
    public static string AutoResXGen_ManagementList_Text_602 {
        get {
            return ResourceManager.GetString("AutoResXGen_ManagementList_Text_602", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Tasks
    ///    This is the title string for the Task Pane.
    ///  
    /// </summary>
    public static string AutoResXGen_TaskPane_Text_74 {
        get {
            return ResourceManager.GetString("AutoResXGen_TaskPane_Text_74", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Tasks
    ///    AutomationProperties.Name of a SeparatedList.
    ///  
    /// </summary>
    public static string AutoResXGen_Tile_AutomationPropertiesName_674 {
        get {
            return ResourceManager.GetString("AutoResXGen_Tile_AutomationPropertiesName_674", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Indeterminate Progress Icon
    ///  
    /// </summary>
    public static string AutoResXGen_WaitingRing_AutomationPropertiesName_74 {
        get {
            return ResourceManager.GetString("AutoResXGen_WaitingRing_AutomationPropertiesName_74", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Add criteria
    ///  
    /// </summary>
    public static string AutoResXGen_ManagementList2_Text_124 {
        get {
            return ResourceManager.GetString("AutoResXGen_ManagementList2_Text_124", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Overwrite the existing query or type a different name to save a new query. Each query consists of criteria, sorting, and column customizations.
    ///  
    /// </summary>
    public static string AutoResXGen_ManagementList2_Text_166 {
        get {
            return ResourceManager.GetString("AutoResXGen_ManagementList2_Text_166", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Ok
    ///  
    /// </summary>
    public static string AutoResXGen_ManagementList2_Content_186 {
        get {
            return ResourceManager.GetString("AutoResXGen_ManagementList2_Content_186", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cancel
    ///  
    /// </summary>
    public static string AutoResXGen_ManagementList2_Content_196 {
        get {
            return ResourceManager.GetString("AutoResXGen_ManagementList2_Content_196", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Click to save a search query
    ///  
    /// </summary>
    public static string AutoResXGen_ManagementList2_AutomationPropertiesName_314 {
        get {
            return ResourceManager.GetString("AutoResXGen_ManagementList2_AutomationPropertiesName_314", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Available columns
    ///  
    /// </summary>
    public static string AutoResXGen_ColumnPicker_Content_42 {
        get {
            return ResourceManager.GetString("AutoResXGen_ColumnPicker_Content_42", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    >>
    ///    The contents of a button which indicates that items will move from the left column to right column
    ///  
    /// </summary>
    public static string AutoResXGen_ColumnPicker_Content_73 {
        get {
            return ResourceManager.GetString("AutoResXGen_ColumnPicker_Content_73", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    <<
    ///    The contents of a button which indicates that items will move from the right column to left column
    ///  
    /// </summary>
    public static string AutoResXGen_ColumnPicker_Content_84 {
        get {
            return ResourceManager.GetString("AutoResXGen_ColumnPicker_Content_84", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Move up
    ///  
    /// </summary>
    public static string AutoResXGen_ColumnPicker_Content_127 {
        get {
            return ResourceManager.GetString("AutoResXGen_ColumnPicker_Content_127", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Move down
    ///  
    /// </summary>
    public static string AutoResXGen_ColumnPicker_Content_134 {
        get {
            return ResourceManager.GetString("AutoResXGen_ColumnPicker_Content_134", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    OK
    ///  
    /// </summary>
    public static string AutoResXGen_ColumnPicker_Content_189 {
        get {
            return ResourceManager.GetString("AutoResXGen_ColumnPicker_Content_189", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cancel
    ///  
    /// </summary>
    public static string AutoResXGen_ColumnPicker_Content_199 {
        get {
            return ResourceManager.GetString("AutoResXGen_ColumnPicker_Content_199", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Select columns
    ///  
    /// </summary>
    public static string AutoResXGen_ColumnPicker_Content_5 {
        get {
            return ResourceManager.GetString("AutoResXGen_ColumnPicker_Content_5", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Move selected column to list of visible columns
    ///  
    /// </summary>
    public static string AutoResXGen_ColumnPicker_Tooltip_76 {
        get {
            return ResourceManager.GetString("AutoResXGen_ColumnPicker_Tooltip_76", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Move selected column to list of hidden columns
    ///  
    /// </summary>
    public static string AutoResXGen_ColumnPicker_Tooltip_84 {
        get {
            return ResourceManager.GetString("AutoResXGen_ColumnPicker_Tooltip_84", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    This column may not be removed.
    ///  
    /// </summary>
    public static string AutoResXGen_ColumnPicker_Text_142 {
        get {
            return ResourceManager.GetString("AutoResXGen_ColumnPicker_Text_142", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The list must always display at least one column.
    ///  
    /// </summary>
    public static string AutoResXGen_ColumnPicker_Text_152 {
        get {
            return ResourceManager.GetString("AutoResXGen_ColumnPicker_Text_152", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Selected columns
    ///  
    /// </summary>
    public static string AutoResXGen_ColumnPicker_Content_93 {
        get {
            return ResourceManager.GetString("AutoResXGen_ColumnPicker_Content_93", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Find in this column
    ///  
    /// </summary>
    public static string ColumnsExplorer_Column_FindTextBox_BackgroundText {
        get {
            return ResourceManager.GetString("ColumnsExplorer_Column_FindTextBox_BackgroundText", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Expand
    ///  
    /// </summary>
    public static string AutoResXGen_DesignerStyleResources_ToolTip_97 {
        get {
            return ResourceManager.GetString("AutoResXGen_DesignerStyleResources_ToolTip_97", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Click to clear all filter criteria.
    ///  
    /// </summary>
    public static string AutoResXGen_ManagementList2_ToolTip_104 {
        get {
            return ResourceManager.GetString("AutoResXGen_ManagementList2_ToolTip_104", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Click to add search criteria.
    ///  
    /// </summary>
    public static string AutoResXGen_ManagementList2_ToolTip_132 {
        get {
            return ResourceManager.GetString("AutoResXGen_ManagementList2_ToolTip_132", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Click to expand search criteria.
    ///  
    /// </summary>
    public static string AutoResXGen_ManagementList2_ToolTip_32 {
        get {
            return ResourceManager.GetString("AutoResXGen_ManagementList2_ToolTip_32", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    There are currently no saved queries.
    ///  
    /// </summary>
    public static string AutoResXGen_ManagementList2_NoItemsText_50 {
        get {
            return ResourceManager.GetString("AutoResXGen_ManagementList2_NoItemsText_50", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Queries
    ///  
    /// </summary>
    public static string AutoResXGen_ManagementList2_AutomationPropertiesName_52 {
        get {
            return ResourceManager.GetString("AutoResXGen_ManagementList2_AutomationPropertiesName_52", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Delete
    ///  
    /// </summary>
    public static string AutoResXGen_ListOrganizer_AutomationPropertiesName_47 {
        get {
            return ResourceManager.GetString("AutoResXGen_ListOrganizer_AutomationPropertiesName_47", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Rename
    ///  
    /// </summary>
    public static string AutoResXGen_ListOrganizer_AutomationPropertiesName_72 {
        get {
            return ResourceManager.GetString("AutoResXGen_ListOrganizer_AutomationPropertiesName_72", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0} rule
    ///    The text representation of a rule in the filter panel, displayed to accessibility clients. {0} will be the name of the rule.
    ///  
    /// </summary>
    public static string FilterRule_AccessibleName {
        get {
            return ResourceManager.GetString("FilterRule_AccessibleName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Add
    ///  
    /// </summary>
    public static string AutoResXGen_AddFilterRulePicker_Content_214 {
        get {
            return ResourceManager.GetString("AutoResXGen_AddFilterRulePicker_Content_214", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cancel
    ///  
    /// </summary>
    public static string AutoResXGen_AddFilterRulePicker_Content_223 {
        get {
            return ResourceManager.GetString("AutoResXGen_AddFilterRulePicker_Content_223", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Add Filter Criteria
    ///  
    /// </summary>
    public static string AutoResXGen_AddFilterRulePicker_AutomationPropertiesName_293 {
        get {
            return ResourceManager.GetString("AutoResXGen_AddFilterRulePicker_AutomationPropertiesName_293", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Value
    ///    The name for text input fields
    ///  
    /// </summary>
    public static string AutoResXGen_FilterRulePanel_AutomationPropertiesName_199 {
        get {
            return ResourceManager.GetString("AutoResXGen_FilterRulePanel_AutomationPropertiesName_199", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    <Empty>
    ///  
    /// </summary>
    public static string AutoResXGen_FilterRulePanel_BackgroundText_200 {
        get {
            return ResourceManager.GetString("AutoResXGen_FilterRulePanel_BackgroundText_200", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Rules
    ///    The name of the panel which contains the filter rules
    ///  
    /// </summary>
    public static string AutoResXGen_FilterRulePanel_AutomationPropertiesName_257 {
        get {
            return ResourceManager.GetString("AutoResXGen_FilterRulePanel_AutomationPropertiesName_257", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Delete
    ///  
    /// </summary>
    public static string AutoResXGen_ManagementList_AutomationPropertiesName_302 {
        get {
            return ResourceManager.GetString("AutoResXGen_ManagementList_AutomationPropertiesName_302", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Query
    ///  
    /// </summary>
    public static string AutoResXGen_ManagementList2_Content_19 {
        get {
            return ResourceManager.GetString("AutoResXGen_ManagementList2_Content_19", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Queries
    ///  
    /// </summary>
    public static string AutoResXGen_ManagementList2_Content_33 {
        get {
            return ResourceManager.GetString("AutoResXGen_ManagementList2_Content_33", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Search
    ///  
    /// </summary>
    public static string AutoResXGen_SearchBox_AutomationPropertiesName_75 {
        get {
            return ResourceManager.GetString("AutoResXGen_SearchBox_AutomationPropertiesName_75", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ({0} of {1})
    ///    The text displayed in the management list title when the list has a filter applied. {0} will be the number of items shown in the list. {1} will be the total number of items in the list before filtering.
    ///  
    /// </summary>
    public static string ManagementListTitle_ListStatus_FilterApplied {
        get {
            return ResourceManager.GetString("ManagementListTitle_ListStatus_FilterApplied", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Searching...
    ///    The text displayed in the management list title when the list is processing a filter.
    ///  
    /// </summary>
    public static string ManagementListTitle_ListStatus_FilterInProgress {
        get {
            return ResourceManager.GetString("ManagementListTitle_ListStatus_FilterInProgress", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ({0})
    ///    The text displayed in the management list title when the list does not have a filter applied. {0} will be the number of items shown in the list.
    ///  
    /// </summary>
    public static string ManagementListTitle_ListStatus_FilterNotApplied {
        get {
            return ResourceManager.GetString("ManagementListTitle_ListStatus_FilterNotApplied", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Filter
    ///    Localizable AutomationName for control that is used by accessibility screen readers.
    ///  
    /// </summary>
    public static string AutoResXGen_BreadcrumbItem_AutomationPropertiesName_142 {
        get {
            return ResourceManager.GetString("AutoResXGen_BreadcrumbItem_AutomationPropertiesName_142", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Filter
    ///  
    /// </summary>
    public static string ColumnsExplorer_Column_FindTextBox_AutomationName {
        get {
            return ResourceManager.GetString("ColumnsExplorer_Column_FindTextBox_AutomationName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Shortcut Rules
    ///    The name used to indicate custom filter rules which are specific to a particular application.
    ///  
    /// </summary>
    public static string AutoResXGen_AddFilterRulePicker_AutomationPropertiesName_157 {
        get {
            return ResourceManager.GetString("AutoResXGen_AddFilterRulePicker_AutomationPropertiesName_157", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Columns Rules
    ///    The name used to indicate filter rules that are based upon the properties of the items in the list.
    ///  
    /// </summary>
    public static string AutoResXGen_AddFilterRulePicker_AutomationPropertiesName_180 {
        get {
            return ResourceManager.GetString("AutoResXGen_AddFilterRulePicker_AutomationPropertiesName_180", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Sort Glyph
    ///  
    /// </summary>
    public static string ManagementList_SortGlyph_Ascending_AutomationName {
        get {
            return ResourceManager.GetString("ManagementList_SortGlyph_Ascending_AutomationName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Sort Glyph
    ///  
    /// </summary>
    public static string ManagementList_SortGlyph_Descending_AutomationName {
        get {
            return ResourceManager.GetString("ManagementList_SortGlyph_Descending_AutomationName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Collapse
    ///  
    /// </summary>
    public static string AutoResXGen_DesignerStyleResources_ToolTip_119 {
        get {
            return ResourceManager.GetString("AutoResXGen_DesignerStyleResources_ToolTip_119", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Collapse
    ///  
    /// </summary>
    public static string AutoResXGen_DesignerStyleResources_ToolTip_160 {
        get {
            return ResourceManager.GetString("AutoResXGen_DesignerStyleResources_ToolTip_160", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Sorted ascending
    ///    The text used for the accessible ItemStatus property when a column is sorted ascending.
    ///  
    /// </summary>
    public static string InnerList_GridViewColumnHeader_ItemStatus_Ascending {
        get {
            return ResourceManager.GetString("InnerList_GridViewColumnHeader_ItemStatus_Ascending", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Sorted descending
    ///    The text used for the accessible ItemStatus property when a column is sorted descending.
    ///  
    /// </summary>
    public static string InnerList_GridViewColumnHeader_ItemStatus_Descending {
        get {
            return ResourceManager.GetString("InnerList_GridViewColumnHeader_ItemStatus_Descending", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Collapse
    ///  
    /// </summary>
    public static string CollapsingTabControl_ExpandButton_AutomationName {
        get {
            return ResourceManager.GetString("CollapsingTabControl_ExpandButton_AutomationName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Expand
    ///  
    /// </summary>
    public static string CollapsingTabControl_CollapseButton_AutomationName {
        get {
            return ResourceManager.GetString("CollapsingTabControl_CollapseButton_AutomationName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Search
    ///  
    /// </summary>
    public static string AutoResXGen_ManagementList_TextBlock_83 {
        get {
            return ResourceManager.GetString("AutoResXGen_ManagementList_TextBlock_83", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cancel
    ///  
    /// </summary>
    public static string AutoResXGen_ManagementList_TextBlock_106 {
        get {
            return ResourceManager.GetString("AutoResXGen_ManagementList_TextBlock_106", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Clear All
    ///  
    /// </summary>
    public static string AutoResXGen_ManagementList_TextBlock_129 {
        get {
            return ResourceManager.GetString("AutoResXGen_ManagementList_TextBlock_129", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Clear All
    ///  
    /// </summary>
    public static string AutoResXGen_ManagementList_AutomationPropertiesName_395 {
        get {
            return ResourceManager.GetString("AutoResXGen_ManagementList_AutomationPropertiesName_395", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Clear Search Text
    ///  
    /// </summary>
    public static string AutoResXGen_SearchBox_AutomationPropertiesName_85 {
        get {
            return ResourceManager.GetString("AutoResXGen_SearchBox_AutomationPropertiesName_85", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Tasks
    ///  
    /// </summary>
    public static string AutoResXGen_TaskPane_AutomationPropertiesName_133 {
        get {
            return ResourceManager.GetString("AutoResXGen_TaskPane_AutomationPropertiesName_133", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Search
    ///    The accessible name of the Search button in the filter panel.
    ///  
    /// </summary>
    public static string ManagementList_StartFilterButton_AutomationName {
        get {
            return ResourceManager.GetString("ManagementList_StartFilterButton_AutomationName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cancel
    ///    The accessible name of the Stop Search button in the filter panel.
    ///  
    /// </summary>
    public static string ManagementList_StopFilterButton_AutomationName {
        get {
            return ResourceManager.GetString("ManagementList_StopFilterButton_AutomationName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Expand or Collapse Filter Panel
    ///    The accessible name of the button that expands/collapses the filter panel.
    ///  
    /// </summary>
    public static string ManagementList_ToggleFilterPanelButton_AutomationName {
        get {
            return ResourceManager.GetString("ManagementList_ToggleFilterPanelButton_AutomationName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Filter
    ///    The background text of the list's search box when filtering is immediate.
    ///  
    /// </summary>
    public static string ManagementList_SearchBox_BackgroundText_Live {
        get {
            return ResourceManager.GetString("ManagementList_SearchBox_BackgroundText_Live", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Click to display saved search queries.
    ///  
    /// </summary>
    public static string AutoResXGen_ManagementList_ToolTip_314 {
        get {
            return ResourceManager.GetString("AutoResXGen_ManagementList_ToolTip_314", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Filter applied.
    ///  
    /// </summary>
    public static string AutoResXGen_ManagementList_Text_392 {
        get {
            return ResourceManager.GetString("AutoResXGen_ManagementList_Text_392", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    and
    ///    The first header operator indicates that it is the first item in the list of filter rules. The AND value is used to indicate that it is and'ed with the above SearchBox.
    ///  
    /// </summary>
    public static string FilterRulePanel_LogicalOperatorText_FirstHeader {
        get {
            return ResourceManager.GetString("FilterRulePanel_LogicalOperatorText_FirstHeader", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    and
    ///    The header operator indicates that it is the first item in a group of filter rules which are the same. The AND value is used to indicate that it is and'ed with the other groups in the panel.
    ///  
    /// </summary>
    public static string FilterRulePanel_LogicalOperatorText_Header {
        get {
            return ResourceManager.GetString("FilterRulePanel_LogicalOperatorText_Header", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    or
    ///    The Item operator indicates that it is NOT the first item in a group of filter rules which are the same. The OR value is used to indicate that it is or'ed with the other items in the same group.
    ///  
    /// </summary>
    public static string FilterRulePanel_LogicalOperatorText_Item {
        get {
            return ResourceManager.GetString("FilterRulePanel_LogicalOperatorText_Item", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    No matches found.
    ///    The text displayed in the ManagementList when the filter has been applied but matching items were found.
    ///  
    /// </summary>
    public static string ManagementList_NoMatchesFound_Message {
        get {
            return ResourceManager.GetString("ManagementList_NoMatchesFound_Message", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Collapse
    ///  
    /// </summary>
    public static string CollapsingTabControl_CollapseButton_ToolTip {
        get {
            return ResourceManager.GetString("CollapsingTabControl_CollapseButton_ToolTip", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Expand
    ///  
    /// </summary>
    public static string CollapsingTabControl_ExpandButton_ToolTip {
        get {
            return ResourceManager.GetString("CollapsingTabControl_ExpandButton_ToolTip", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Show Children
    ///  
    /// </summary>
    public static string NavigationList_ShownChildrenButton_ToolTip {
        get {
            return ResourceManager.GetString("NavigationList_ShownChildrenButton_ToolTip", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Show Children
    ///  
    /// </summary>
    public static string NavigationList_ShownChildrenButton_AutomationName {
        get {
            return ResourceManager.GetString("NavigationList_ShownChildrenButton_AutomationName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0}: {1}
    ///    The format string used for the ManagementList title when query has been applied. For example, "Users: My Fancy Query"
    ///  
    /// </summary>
    public static string ManagementListTitle_Title_WithViewName {
        get {
            return ResourceManager.GetString("ManagementListTitle_Title_WithViewName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cancel
    ///  
    /// </summary>
    public static string OutGridView_Button_Cancel {
        get {
            return ResourceManager.GetString("OutGridView_Button_Cancel", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    OK
    ///  
    /// </summary>
    public static string OutGridView_Button_OK {
        get {
            return ResourceManager.GetString("OutGridView_Button_OK", resourceCulture);
        }
    }

}

