
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class MshSnapinInfo {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal MshSnapinInfo() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.MshSnapinInfo", typeof(MshSnapinInfo).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unable to access PowerShell registry information.
    ///  
    /// </summary>
    internal static string MonadRootRegistryAccessFailed {
        get {
            return ResourceManager.GetString("MonadRootRegistryAccessFailed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unable to access PowerShell Engine registry information.
    ///  
    /// </summary>
    internal static string MonadEngineRegistryAccessFailed {
        get {
            return ResourceManager.GetString("MonadEngineRegistryAccessFailed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unable to access PublicKeyToken information.
    ///  
    /// </summary>
    internal static string PublicKeyTokenAccessFailed {
        get {
            return ResourceManager.GetString("PublicKeyTokenAccessFailed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Version {0} of PowerShell is not available on this computer.
    ///  
    /// </summary>
    internal static string SpecifiedVersionNotFound {
        get {
            return ResourceManager.GetString("SpecifiedVersionNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The PowerShell snap-in '{0}' is not installed on this computer.
    ///  
    /// </summary>
    internal static string MshSnapinDoesNotExist {
        get {
            return ResourceManager.GetString("MshSnapinDoesNotExist", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The mandatory value {0} is not specified for registry key {1}.
    ///  
    /// </summary>
    internal static string MandatoryValueNotPresent {
        get {
            return ResourceManager.GetString("MandatoryValueNotPresent", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The mandatory value {0} is not in the correct format for registry key {1}.  The expected format is 'string.'
    ///  
    /// </summary>
    internal static string MandatoryValueNotInCorrectFormat {
        get {
            return ResourceManager.GetString("MandatoryValueNotInCorrectFormat", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The mandatory value {0} is not in the correct format for registry key {1}.  The expected format is 'multistring.'
    ///  
    /// </summary>
    internal static string MandatoryValueNotInCorrectFormatMultiString {
        get {
            return ResourceManager.GetString("MandatoryValueNotInCorrectFormatMultiString", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot find required information in registry or missing key files.  Cannot load some cmdlets.
    ///  
    /// </summary>
    internal static string DefaultMshSnapinNotPresent {
        get {
            return ResourceManager.GetString("DefaultMshSnapinNotPresent", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    No snap-ins have been registered for PowerShell version {0}.
    ///  
    /// </summary>
    internal static string NoMshSnapinPresentForVersion {
        get {
            return ResourceManager.GetString("NoMshSnapinPresentForVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot retrieve the string resource because the reader has been disposed.
    ///  
    /// </summary>
    internal static string ResourceReaderDisposed {
        get {
            return ResourceManager.GetString("ResourceReaderDisposed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Version value {0} is not specified or is incorrect for registry key {1}.
    ///  
    /// </summary>
    internal static string VersionValueInCorrect {
        get {
            return ResourceManager.GetString("VersionValueInCorrect", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    No [PSVersion] attribute was found for PowerShell type {0}. Add a PSVersion attribute to the type using [PSVersion(PowerShell SnapinBase.PSEngineVersion)].
    ///  
    /// </summary>
    internal static string PSVersionAttributeNotExist {
        get {
            return ResourceManager.GetString("PSVersionAttributeNotExist", resourceCulture);
        }
    }

}

