
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class Modules {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal Modules() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.Modules", typeof(Modules).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified module '{0}' was not loaded because no valid module file was found in any module directory.
    ///  
    /// </summary>
    internal static string ModuleNotFound {
        get {
            return ResourceManager.GetString("ModuleNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified module '{0}' with version '{1}' was not loaded because no valid module file was found in any module directory.
    ///  
    /// </summary>
    internal static string ModuleWithVersionNotFound {
        get {
            return ResourceManager.GetString("ModuleWithVersionNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified MaximumVersion '{0}' was incorrect. If you are using '*', MaximumVersion only supports one '*' and should always be placed at the end of MaximumVersion.
    ///  
    /// </summary>
    internal static string MaximumVersionFormatIncorrect {
        get {
            return ResourceManager.GetString("MaximumVersionFormatIncorrect", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified module '{0}' with MaximumVersion '{1}' was not loaded because no valid module file was found in any module directory.
    ///  
    /// </summary>
    internal static string MaximumVersionNotFound {
        get {
            return ResourceManager.GetString("MaximumVersionNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified module '{0}' with MinimumVersion '{1}' and MaximumVersion '{2}' was not loaded because no valid module file was found in any module directory.
    ///  
    /// </summary>
    internal static string MinimumVersionAndMaximumVersionNotFound {
        get {
            return ResourceManager.GetString("MinimumVersionAndMaximumVersionNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The MinimumVersion '{0}' should not be greater than MaximumVersion '{1}'.
    ///  
    /// </summary>
    internal static string MinimumVersionAndMaximumVersionInvalidRange {
        get {
            return ResourceManager.GetString("MinimumVersionAndMaximumVersionInvalidRange", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The assembly '{0}' was not loaded because no assembly with that name was found. Verify the assembly name, and then try again.
    ///  
    /// </summary>
    internal static string ModuleAssemblyFound {
        get {
            return ResourceManager.GetString("ModuleAssemblyFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The module to process '{0}', listed in field '{1}' of module manifest '{2}' was not processed because no valid module was found in any module directory.
    ///  
    /// </summary>
    internal static string ManifestMemberNotFound {
        get {
            return ResourceManager.GetString("ManifestMemberNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    No custom object was returned for module '{0}' because the -AsCustomObject parameter can only be used with script modules.
    ///  
    /// </summary>
    internal static string CantUseAsCustomObjectWithBinaryModule {
        get {
            return ResourceManager.GetString("CantUseAsCustomObjectWithBinaryModule", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The module manifest '{0}' could not be processed because it is not a valid PowerShell module manifest file. Remove the elements that are not permitted: {1}
    ///  
    /// </summary>
    internal static string InvalidModuleManifest {
        get {
            return ResourceManager.GetString("InvalidModuleManifest", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Processing the module manifest file '{0}' did not result in a valid manifest object. Update the file to contain a valid PowerShell module manifest. A valid manifest can be created using the New-ModuleManifest cmdlet.
    ///  
    /// </summary>
    internal static string EmptyModuleManifest {
        get {
            return ResourceManager.GetString("EmptyModuleManifest", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The '{0}' module cannot be imported because its manifest contains one or more members that are not valid. The valid manifest members are ({1}). Remove the members that are not valid ({2}), then try to import the module again.
    ///  
    /// </summary>
    internal static string InvalidModuleManifestMember {
        get {
            return ResourceManager.GetString("InvalidModuleManifestMember", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The hashtable describing a module contains one or more members that are not valid. The valid members are ({0}). Remove the members that are not valid ({1}), then try again.
    ///  
    /// </summary>
    internal static string InvalidModuleSpecificationMember {
        get {
            return ResourceManager.GetString("InvalidModuleSpecificationMember", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot load the module '{0}' because the module nesting limit has been exceeded. Modules can only be nested to {1} levels. Evaluate and change the order in which you are loading modules to prevent exceeding the nesting limit, and then try running your script again.
    ///  
    /// </summary>
    internal static string ModuleTooDeeplyNested {
        get {
            return ResourceManager.GetString("ModuleTooDeeplyNested", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The member 'ModuleVersion' is not present in the module manifest. This member must exist and be assigned a version number of the form 'n.n.n.n'. Add the missing member to the file '{0}'.
    ///  
    /// </summary>
    internal static string ModuleManifestMissingModuleVersion {
        get {
            return ResourceManager.GetString("ModuleManifestMissingModuleVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The '{0}' member is not valid in the module manifest file '{2}': {1}
    ///  
    /// </summary>
    internal static string ModuleManifestInvalidValue {
        get {
            return ResourceManager.GetString("ModuleManifestInvalidValue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The version '{0}' of module '{1}' does not meet the required minimum version '{2}'. Verify that the version number is supported, and then try loading the module again.
    ///  
    /// </summary>
    internal static string ModuleManifestInsufficientModuleVersion {
        get {
            return ResourceManager.GetString("ModuleManifestInsufficientModuleVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The version of PowerShell on this computer is '{0}'. The module '{1}' requires a minimum PowerShell version of '{2}' to run. Verify that you have the minimum required version of PowerShell installed, and then try again.
    ///  
    /// </summary>
    internal static string ModuleManifestInsufficientPowerShellVersion {
        get {
            return ResourceManager.GetString("ModuleManifestInsufficientPowerShellVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The module manifest member 'NestedModules' cannot be used if the 'ModuleToProcess' member is a binary module. Edit the module manifest file at '{0}', and then try again.
    ///  
    /// </summary>
    internal static string ModuleManifestNestedModulesCantGoWithModuleToProcess {
        get {
            return ResourceManager.GetString("ModuleManifestNestedModulesCantGoWithModuleToProcess", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The member '{0}' in the module manifest is not valid: {1}. Verify that a valid value is specified for this field in the '{2}' file.
    ///  
    /// </summary>
    internal static string ModuleManifestInvalidManifestMember {
        get {
            return ResourceManager.GetString("ModuleManifestInvalidManifestMember", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The module manifest path '{0}' is not valid. The value of the Path argument must resolve to a single file that has a '.psd1' extension. Change the value of the Path argument to point to a valid psd1 file, and then try again.
    ///  
    /// </summary>
    internal static string InvalidModuleManifestPath {
        get {
            return ResourceManager.GetString("InvalidModuleManifestPath", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The ModuleVersion key in module manifest '{0}' specifies module version '{1}' which does not match its version folder name at '{2}'. Change the value of the ModuleVersion key to match the version folder name.
    ///  
    /// </summary>
    internal static string InvalidModuleManifestVersion {
        get {
            return ResourceManager.GetString("InvalidModuleManifestVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified NestedModule entry '{0}' in the module manifest '{1}' is invalid. Try again after updating this entry with valid values.
    ///  
    /// </summary>
    internal static string InvalidNestedModuleinModuleManifest {
        get {
            return ResourceManager.GetString("InvalidNestedModuleinModuleManifest", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified RequiredAssemblies entry '{0}' in the module manifest '{1}' is invalid. Try again after updating this entry with valid values.
    ///  
    /// </summary>
    internal static string InvalidRequiredAssembliesInModuleManifest {
        get {
            return ResourceManager.GetString("InvalidRequiredAssembliesInModuleManifest", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified FileList entry '{0}' in the module manifest '{1}' is invalid. Try again after updating this entry with valid values.
    ///  
    /// </summary>
    internal static string InvalidFilePathinModuleManifest {
        get {
            return ResourceManager.GetString("InvalidFilePathinModuleManifest", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified RequiredModules entry '{0}' in the module manifest '{1}' is invalid. Try again after updating this entry with valid values.
    ///  
    /// </summary>
    internal static string InvalidRequiredModulesinModuleManifest {
        get {
            return ResourceManager.GetString("InvalidRequiredModulesinModuleManifest", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified ModuleList entry '{0}' in the module manifest '{1}' is invalid. Try again after updating this entry with valid values.
    ///  
    /// </summary>
    internal static string InvalidModuleListinModuleManifest {
        get {
            return ResourceManager.GetString("InvalidModuleListinModuleManifest", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The module manifest '{0}' is specified with the CompatiblePSEditions key which is supported only on PowerShell version '5.1' or higher. Update the value of the PowerShellVersion key to '5.1' or higher, and try again.
    ///  
    /// </summary>
    internal static string InvalidPowerShellVersionInModuleManifest {
        get {
            return ResourceManager.GetString("InvalidPowerShellVersionInModuleManifest", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified value '{0}' for CompatiblePSEditions contains duplicate PowerShell Edition names. Try again after removing the duplicate PowerShell Edition names.
    ///  
    /// </summary>
    internal static string DuplicateEntriesInCompatiblePSEditions {
        get {
            return ResourceManager.GetString("DuplicateEntriesInCompatiblePSEditions", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Version specified in ModuleVersion key is equal to version folder name.
    ///  
    /// </summary>
    internal static string ModuleVersionEqualsToVersionFolder {
        get {
            return ResourceManager.GetString("ModuleVersionEqualsToVersionFolder", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Skipping the Version folder {0} under Module {1} as it does not have a valid module manifest file.
    ///  
    /// </summary>
    internal static string SkippingInvalidModuleVersionFolder {
        get {
            return ResourceManager.GetString("SkippingInvalidModuleVersionFolder", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The 'ModuleName' member does not exist in the hashtable that describes this module.
    ///  
    /// </summary>
    internal static string RequiredModuleMissingModuleName {
        get {
            return ResourceManager.GetString("RequiredModuleMissingModuleName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The 'ModuleVersion', 'MaximumVersion' and 'RequiredVersion' members do not exist in the hashtable that describes this module. One of these three members must exist, and be assigned a version number in the format 'n.n.n.n'.
    ///  
    /// </summary>
    internal static string RequiredModuleMissingModuleVersion {
        get {
            return ResourceManager.GetString("RequiredModuleMissingModuleVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The required module '{1}' is not loaded. Load the module or remove the module from 'RequiredModules' in the file '{0}'.
    ///  
    /// </summary>
    internal static string RequiredModuleNotLoaded {
        get {
            return ResourceManager.GetString("RequiredModuleNotLoaded", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The required module '{1}' with GUID '{2}' is not loaded. Load the module or remove the module from 'RequiredModules' in the file '{0}'.
    ///  
    /// </summary>
    internal static string RequiredModuleNotLoadedWrongGuid {
        get {
            return ResourceManager.GetString("RequiredModuleNotLoadedWrongGuid", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The required module '{1}' with version '{2}' is not loaded. Load the module or remove the module from 'RequiredModules' in the file '{0}'.
    ///  
    /// </summary>
    internal static string RequiredModuleNotLoadedWrongVersion {
        get {
            return ResourceManager.GetString("RequiredModuleNotLoadedWrongVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The required module '{1}' with MaximumVersion '{2}' is not loaded. Load the module or remove the module from 'RequiredModules' in the file '{0}'.
    ///  
    /// </summary>
    internal static string RequiredModuleNotLoadedWrongMaximumVersion {
        get {
            return ResourceManager.GetString("RequiredModuleNotLoadedWrongMaximumVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The required module '{1}' with MinimumVersion '{2}' and MaximumVersion '{3}' is not loaded. Load the module or remove the module from 'RequiredModules' in the file '{0}'.
    ///  
    /// </summary>
    internal static string RequiredModuleNotLoadedWrongMinimumVersionAndMaximumVersion {
        get {
            return ResourceManager.GetString("RequiredModuleNotLoadedWrongMinimumVersionAndMaximumVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The module '{0}' cannot be found with ModuleVersion '{1}'.
    ///  
    /// </summary>
    internal static string RequiredModuleNotFoundModuleVersion {
        get {
            return ResourceManager.GetString("RequiredModuleNotFoundModuleVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The module '{0}' cannot be found with RequiredVersion '{1}'.
    ///  
    /// </summary>
    internal static string RequiredModuleNotFoundRequiredVersion {
        get {
            return ResourceManager.GetString("RequiredModuleNotFoundRequiredVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The module '{0}' cannot be found with MaximumVersion '{1}'.
    ///  
    /// </summary>
    internal static string RequiredModuleNotFoundMaximumVersion {
        get {
            return ResourceManager.GetString("RequiredModuleNotFoundMaximumVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The module '{0}' cannot be found with ModuleVersion '{1}' and MaximumVersion '{2}'.
    ///  
    /// </summary>
    internal static string RequiredModuleNotFoundModuleAndMaximumVersion {
        get {
            return ResourceManager.GetString("RequiredModuleNotFoundModuleAndMaximumVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The module '{0}' cannot be found.
    ///  
    /// </summary>
    internal static string RequiredModuleNotFoundWithoutVersion {
        get {
            return ResourceManager.GetString("RequiredModuleNotFoundWithoutVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    No modules were removed. Verify that the specification of modules to remove is correct and those modules exist in the runspace.
    ///  
    /// </summary>
    internal static string NoModulesRemoved {
        get {
            return ResourceManager.GetString("NoModulesRemoved", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The '{0}' member, which was imported from module '{1}', cannot be removed for the following reason: {2}
    ///  
    /// </summary>
    internal static string UnableToRemoveModuleMember {
        get {
            return ResourceManager.GetString("UnableToRemoveModuleMember", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unable to remove the module '{0}' because it is read-only. Add the Force parameter to your command to remove read-only modules.
    ///  
    /// </summary>
    internal static string ModuleIsReadOnly {
        get {
            return ResourceManager.GetString("ModuleIsReadOnly", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unable to remove the module '{0}' because it is marked as 'constant.' A module cannot be removed if it is marked 'constant.'
    ///  
    /// </summary>
    internal static string ModuleIsConstant {
        get {
            return ResourceManager.GetString("ModuleIsConstant", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unable to remove the module '{0}' because it is required by '{1}'. Add the Force parameter to your command to remove the module.
    ///  
    /// </summary>
    internal static string ModuleIsRequired {
        get {
            return ResourceManager.GetString("ModuleIsRequired", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Export-ModuleMember cmdlet can only be called from inside a module.
    ///  
    /// </summary>
    internal static string CanOnlyBeUsedFromWithinAModule {
        get {
            return ResourceManager.GetString("CanOnlyBeUsedFromWithinAModule", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The extension '{0}' is not a valid module extension. The supported module extensions are '.dll', '.ps1', '.psm1', '.psd1' and '.cdxml'. Correct the extension then try adding the file '{1}' again.
    ///  
    /// </summary>
    internal static string InvalidModuleExtension {
        get {
            return ResourceManager.GetString("InvalidModuleExtension", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    This operation cannot be performed on a binary module. It can only be performed on a script module.
    ///  
    /// </summary>
    internal static string InvalidOperationOnBinaryModule {
        get {
            return ResourceManager.GetString("InvalidOperationOnBinaryModule", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The file '{0}' is not allowed because it does not have the extension '.ps1'.
    ///  
    /// </summary>
    internal static string ScriptsToProcessIncorrectExtension {
        get {
            return ResourceManager.GetString("ScriptsToProcessIncorrectExtension", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unknown
    ///  
    /// </summary>
    internal static string DefaultCompanyName {
        get {
            return ResourceManager.GetString("DefaultCompanyName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    (c) {0}. All rights reserved.
    ///  
    /// </summary>
    internal static string DefaultCopyrightMessage {
        get {
            return ResourceManager.GetString("DefaultCopyrightMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Removing the imported "{0}" function.
    ///  
    /// </summary>
    internal static string RemovingImportedFunction {
        get {
            return ResourceManager.GetString("RemovingImportedFunction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Removing the imported "{0}" alias.
    ///  
    /// </summary>
    internal static string RemovingImportedAlias {
        get {
            return ResourceManager.GetString("RemovingImportedAlias", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Removing the imported "{0}" variable.
    ///  
    /// </summary>
    internal static string RemovingImportedVariable {
        get {
            return ResourceManager.GetString("RemovingImportedVariable", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Loading module from path '{0}'.
    ///  
    /// </summary>
    internal static string LoadingModule {
        get {
            return ResourceManager.GetString("LoadingModule", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Loading '{0}' from path '{1}'.
    ///  
    /// </summary>
    internal static string LoadingFile {
        get {
            return ResourceManager.GetString("LoadingFile", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Dot-sourcing the script file '{0}'.
    ///  
    /// </summary>
    internal static string DottingScriptFile {
        get {
            return ResourceManager.GetString("DottingScriptFile", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Importing function '{0}'.
    ///  
    /// </summary>
    internal static string ImportingFunction {
        get {
            return ResourceManager.GetString("ImportingFunction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Importing cmdlet '{0}'.
    ///  
    /// </summary>
    internal static string ImportingCmdlet {
        get {
            return ResourceManager.GetString("ImportingCmdlet", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Importing alias '{0}'.
    ///  
    /// </summary>
    internal static string ImportingAlias {
        get {
            return ResourceManager.GetString("ImportingAlias", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Importing variable '{0}'.
    ///  
    /// </summary>
    internal static string ImportingVariable {
        get {
            return ResourceManager.GetString("ImportingVariable", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Exporting cmdlet '{0}'.
    ///  
    /// </summary>
    internal static string ExportingCmdlet {
        get {
            return ResourceManager.GetString("ExportingCmdlet", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Exporting function '{0}'.
    ///  
    /// </summary>
    internal static string ExportingFunction {
        get {
            return ResourceManager.GetString("ExportingFunction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Exporting alias '{0}'.
    ///  
    /// </summary>
    internal static string ExportingAlias {
        get {
            return ResourceManager.GetString("ExportingAlias", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Exporting variable '{0}'.
    ///  
    /// </summary>
    internal static string ExportingVariable {
        get {
            return ResourceManager.GetString("ExportingVariable", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The names of some imported commands from the module '{0}' include unapproved verbs that might make them less discoverable. To find the commands with unapproved verbs, run the Import-Module command again with the Verbose parameter. For a list of approved verbs, type Get-Verb.
    ///  
    /// </summary>
    internal static string ImportingNonStandardVerb {
        get {
            return ResourceManager.GetString("ImportingNonStandardVerb", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The '{0}' command in the {1}' module was imported, but because its name does not include an approved verb, it might be difficult to find. For a list of approved verbs, type Get-Verb.
    ///  
    /// </summary>
    internal static string ImportingNonStandardVerbVerbose {
        get {
            return ResourceManager.GetString("ImportingNonStandardVerbVerbose", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The '{0}' command in the {2}' module was imported, but because its name does not include an approved verb, it might be difficult to find. The suggested alternative verbs are "{1}".
    ///  
    /// </summary>
    internal static string ImportingNonStandardVerbVerboseSuggestion {
        get {
            return ResourceManager.GetString("ImportingNonStandardVerbVerboseSuggestion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Some imported command names contain one or more of the following restricted characters: # , ( ) {{ }} [ ] & - / \ $ ^ ; : " ' < > | ? @ ` * % + = ~
    ///  
    /// </summary>
    internal static string ImportingNonStandardNoun {
        get {
            return ResourceManager.GetString("ImportingNonStandardNoun", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The command name '{0}' from the module '{1}' contains one or more of the following restricted characters: # , ( ) {{ }} [ ] & - / \ $ ^ ; : " ' < > | ? @ ` * % + = ~
    ///  
    /// </summary>
    internal static string ImportingNonStandardNounVerbose {
        get {
            return ResourceManager.GetString("ImportingNonStandardNounVerbose", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Creating the "{0}" module manifest file.
    ///  
    /// </summary>
    internal static string CreatingModuleManifestFile {
        get {
            return ResourceManager.GetString("CreatingModuleManifestFile", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0} (Path: '{1}')
    ///  
    /// </summary>
    internal static string ConfirmRemoveModule {
        get {
            return ResourceManager.GetString("ConfirmRemoveModule", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The current processor architecture is: {0}. The module '{1}' requires the following architecture: {2}.
    ///  
    /// </summary>
    internal static string InvalidProcessorArchitecture {
        get {
            return ResourceManager.GetString("InvalidProcessorArchitecture", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The name of the current PowerShell host is: '{0}'. The module '{1}' requires the following PowerShell host: '{2}'.
    ///  
    /// </summary>
    internal static string InvalidPowerShellHostName {
        get {
            return ResourceManager.GetString("InvalidPowerShellHostName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The current PowerShell host is: '{0}' (version {1}). The module '{2}' requires a minimum PowerShell host version of '{3}' to run.
    ///  
    /// </summary>
    internal static string InvalidPowerShellHostVersion {
        get {
            return ResourceManager.GetString("InvalidPowerShellHostVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Module manifest for module '{0}'
    ///  
    /// </summary>
    internal static string ManifestHeaderLine1 {
        get {
            return ResourceManager.GetString("ManifestHeaderLine1", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Generated by: {0}
    ///  
    /// </summary>
    internal static string ManifestHeaderLine2 {
        get {
            return ResourceManager.GetString("ManifestHeaderLine2", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Generated on: {0}
    ///  
    /// </summary>
    internal static string ManifestHeaderLine3 {
        get {
            return ResourceManager.GetString("ManifestHeaderLine3", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Script module or binary module file associated with this manifest.
    ///  
    /// </summary>
    internal static string RootModule {
        get {
            return ResourceManager.GetString("RootModule", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Modules to import as nested modules of the module specified in RootModule/ModuleToProcess
    ///  
    /// </summary>
    internal static string NestedModules {
        get {
            return ResourceManager.GetString("NestedModules", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ID used to uniquely identify this module
    ///  
    /// </summary>
    internal static string GUID {
        get {
            return ResourceManager.GetString("GUID", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Author of this module
    ///  
    /// </summary>
    internal static string Author {
        get {
            return ResourceManager.GetString("Author", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Company or vendor of this module
    ///  
    /// </summary>
    internal static string CompanyName {
        get {
            return ResourceManager.GetString("CompanyName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Copyright statement for this module
    ///  
    /// </summary>
    internal static string Copyright {
        get {
            return ResourceManager.GetString("Copyright", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Version number of this module.
    ///  
    /// </summary>
    internal static string ModuleVersion {
        get {
            return ResourceManager.GetString("ModuleVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Description of the functionality provided by this module
    ///  
    /// </summary>
    internal static string Description {
        get {
            return ResourceManager.GetString("Description", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Minimum version of the PowerShell engine required by this module
    ///  
    /// </summary>
    internal static string PowerShellVersion {
        get {
            return ResourceManager.GetString("PowerShellVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Minimum version of the common language runtime (CLR) required by this module. {0}
    ///  
    /// </summary>
    internal static string CLRVersion {
        get {
            return ResourceManager.GetString("CLRVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Modules that must be imported into the global environment prior to importing this module
    ///  
    /// </summary>
    internal static string RequiredModules {
        get {
            return ResourceManager.GetString("RequiredModules", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Script files (.ps1) that are run in the caller's environment prior to importing this module.
    ///  
    /// </summary>
    internal static string ScriptsToProcess {
        get {
            return ResourceManager.GetString("ScriptsToProcess", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Type files (.ps1xml) to be loaded when importing this module
    ///  
    /// </summary>
    internal static string TypesToProcess {
        get {
            return ResourceManager.GetString("TypesToProcess", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Format files (.ps1xml) to be loaded when importing this module
    ///  
    /// </summary>
    internal static string FormatsToProcess {
        get {
            return ResourceManager.GetString("FormatsToProcess", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Assemblies that must be loaded prior to importing this module
    ///  
    /// </summary>
    internal static string RequiredAssemblies {
        get {
            return ResourceManager.GetString("RequiredAssemblies", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    List of all files packaged with this module
    ///  
    /// </summary>
    internal static string FileList {
        get {
            return ResourceManager.GetString("FileList", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Private data to pass to the module specified in RootModule/ModuleToProcess. This may also contain a PSData hashtable with additional module metadata used by PowerShell.
    ///  
    /// </summary>
    internal static string PrivateData {
        get {
            return ResourceManager.GetString("PrivateData", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Tags applied to this module. These help with module discovery in online galleries.
    ///  
    /// </summary>
    internal static string Tags {
        get {
            return ResourceManager.GetString("Tags", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A URL to the main website for this project.
    ///  
    /// </summary>
    internal static string ProjectUri {
        get {
            return ResourceManager.GetString("ProjectUri", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A URL to the license for this module.
    ///  
    /// </summary>
    internal static string LicenseUri {
        get {
            return ResourceManager.GetString("LicenseUri", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A URL to an icon representing this module.
    ///  
    /// </summary>
    internal static string IconUri {
        get {
            return ResourceManager.GetString("IconUri", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ReleaseNotes of this module
    ///  
    /// </summary>
    internal static string ReleaseNotes {
        get {
            return ResourceManager.GetString("ReleaseNotes", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Prerelease string of this module
    ///  
    /// </summary>
    internal static string Prerelease {
        get {
            return ResourceManager.GetString("Prerelease", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Flag to indicate whether the module requires explicit user acceptance for install/update/save
    ///  
    /// </summary>
    internal static string RequireLicenseAcceptance {
        get {
            return ResourceManager.GetString("RequireLicenseAcceptance", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    External dependent modules of this module
    ///  
    /// </summary>
    internal static string ExternalModuleDependencies {
        get {
            return ResourceManager.GetString("ExternalModuleDependencies", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    End of {0} hashtable
    ///  
    /// </summary>
    internal static string EndOfManifestHashTable {
        get {
            return ResourceManager.GetString("EndOfManifestHashTable", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The PrivateData parameter value must be a hash table to create the module manifest with the following parameter values Tags, ProjectUri, LicenseUri, IconUri or ReleaseNotes. Either remove the Tags, ProjectUri, LicenseUri, IconUri, or ReleaseNotes parameter values or wrap the contents of PrivateData in a hashtable.
    ///  
    /// </summary>
    internal static string PrivateDataValueTypeShouldBeHashTableError {
        get {
            return ResourceManager.GetString("PrivateDataValueTypeShouldBeHashTableError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The PrivateData should be defined as a hashtable, but this module manifest defines it as an object. Please consider wrapping the contents of PrivateData in a hashtable. This will enable you to add the Tags, ProjectUri, LicenseUri, IconUri, and ReleaseNotes properties to the module manifest at a later time.
    ///  
    /// </summary>
    internal static string PrivateDataValueTypeShouldBeHashTableWarning {
        get {
            return ResourceManager.GetString("PrivateDataValueTypeShouldBeHashTableWarning", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified value '{0}' is invalid, try again with a valid value.
    ///  
    /// </summary>
    internal static string InvalidParameterValue {
        get {
            return ResourceManager.GetString("InvalidParameterValue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Functions to export from this module, for best performance, do not use wildcards and do not delete the entry, use an empty array if there are no functions to export.
    ///  
    /// </summary>
    internal static string FunctionsToExport {
        get {
            return ResourceManager.GetString("FunctionsToExport", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Aliases to export from this module, for best performance, do not use wildcards and do not delete the entry, use an empty array if there are no aliases to export.
    ///  
    /// </summary>
    internal static string AliasesToExport {
        get {
            return ResourceManager.GetString("AliasesToExport", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cmdlets to export from this module, for best performance, do not use wildcards and do not delete the entry, use an empty array if there are no cmdlets to export.
    ///  
    /// </summary>
    internal static string CmdletsToExport {
        get {
            return ResourceManager.GetString("CmdletsToExport", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Variables to export from this module
    ///  
    /// </summary>
    internal static string VariablesToExport {
        get {
            return ResourceManager.GetString("VariablesToExport", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    DSC resources to export from this module
    ///  
    /// </summary>
    internal static string DscResourcesToExport {
        get {
            return ResourceManager.GetString("DscResourcesToExport", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Supported PSEditions
    ///  
    /// </summary>
    internal static string CompatiblePSEditions {
        get {
            return ResourceManager.GetString("CompatiblePSEditions", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Processor architecture (None, X86, Amd64) required by this module
    ///  
    /// </summary>
    internal static string ProcessorArchitecture {
        get {
            return ResourceManager.GetString("ProcessorArchitecture", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    List of all modules packaged with this module
    ///  
    /// </summary>
    internal static string ModuleList {
        get {
            return ResourceManager.GetString("ModuleList", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Minimum version of Microsoft .NET Framework required by this module. {0}
    ///  
    /// </summary>
    internal static string DotNetFrameworkVersion {
        get {
            return ResourceManager.GetString("DotNetFrameworkVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Name of the PowerShell host required by this module
    ///  
    /// </summary>
    internal static string PowerShellHostName {
        get {
            return ResourceManager.GetString("PowerShellHostName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Minimum version of the PowerShell host required by this module
    ///  
    /// </summary>
    internal static string PowerShellHostVersion {
        get {
            return ResourceManager.GetString("PowerShellHostVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    HelpInfo URI of this module
    ///  
    /// </summary>
    internal static string HelpInfoURI {
        get {
            return ResourceManager.GetString("HelpInfoURI", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Because the {0} module is providing the PSDrive in the current PowerShell session, no modules were removed. Change the current PSDrive provider, and then try removing modules again.
    ///  
    /// </summary>
    internal static string ModuleDriveInUse {
        get {
            return ResourceManager.GetString("ModuleDriveInUse", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The cmdlet '{0}' was not imported because there is a member with the same name in the current scope.
    ///  
    /// </summary>
    internal static string ImportModuleNoClobberForCmdlet {
        get {
            return ResourceManager.GetString("ImportModuleNoClobberForCmdlet", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The alias '{0}' was not imported because there is a member with the same name in the current scope.
    ///  
    /// </summary>
    internal static string ImportModuleNoClobberForAlias {
        get {
            return ResourceManager.GetString("ImportModuleNoClobberForAlias", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The function '{0}' was not imported because there is a member with the same name in the current scope.
    ///  
    /// </summary>
    internal static string ImportModuleNoClobberForFunction {
        get {
            return ResourceManager.GetString("ImportModuleNoClobberForFunction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The variable '{0}' was not imported because there is a member with the same name in the current scope.
    ///  
    /// </summary>
    internal static string ImportModuleNoClobberForVariable {
        get {
            return ResourceManager.GetString("ImportModuleNoClobberForVariable", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Wildcard characters are not allowed in members 'ModuleToProcess', 'RootModule', or 'NestedModules' in the module manifest '{0}'.
    ///  
    /// </summary>
    internal static string WildCardNotAllowedInModuleToProcessAndInNestedModules {
        get {
            return ResourceManager.GetString("WildCardNotAllowedInModuleToProcessAndInNestedModules", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The module '{0}' is a core module for PowerShell. Add the Force parameter to your command to remove core modules.
    ///  
    /// </summary>
    internal static string CoreModuleCannotBeRemoved {
        get {
            return ResourceManager.GetString("CoreModuleCannotBeRemoved", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The module manifest cannot contain both the 'ModuleToProcess' and 'RootModule' members. Change the module manifest file to remove one of these members at '{0}', and then try again.
    ///  
    /// </summary>
    internal static string ModuleManifestCannotContainBothModuleToProcessAndRootModule {
        get {
            return ResourceManager.GetString("ModuleManifestCannotContainBothModuleToProcessAndRootModule", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The module manifest member 'ModuleToProcess' has been deprecated. Use the 'RootModule' member instead.
    ///  
    /// </summary>
    internal static string ModuleToProcessFieldDeprecated {
        get {
            return ResourceManager.GetString("ModuleToProcessFieldDeprecated", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Default prefix for commands exported from this module. Override the default prefix using Import-Module -Prefix.
    ///  
    /// </summary>
    internal static string DefaultCommandPrefix {
        get {
            return ResourceManager.GetString("DefaultCommandPrefix", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The 'Global' and 'Scope' parameters cannot be specified together. Remove one of these parameters, and then try running the command again.
    ///  
    /// </summary>
    internal static string GlobalAndScopeParameterCannotBeSpecifiedTogether {
        get {
            return ResourceManager.GetString("GlobalAndScopeParameterCannotBeSpecifiedTogether", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The required module '{0}' is not loaded. The module '{0}' has a requiredModule '{1}' in its module manifest '{2}' that points to a cyclic dependency.
    ///  
    /// </summary>
    internal static string RequiredModulesCyclicDependency {
        get {
            return ResourceManager.GetString("RequiredModulesCyclicDependency", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The required module '{0}' was not loaded because no valid module file was found in any module directory.
    ///  
    /// </summary>
    internal static string RequiredModuleNotFound {
        get {
            return ResourceManager.GetString("RequiredModuleNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Some commands from module {0} cannot be imported over a CimSession. To get all the commands, verify that the remote server has PowerShell remote management enabled, and then try adding the PSSession parameter to an Import-Module cmdlet.
    ///  
    /// </summary>
    internal static string MixedModuleOverCimSessionWarning {
        get {
            return ResourceManager.GetString("MixedModuleOverCimSessionWarning", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Module {0} is loaded in Windows PowerShell using {1} remoting session; please note that all input and output of commands from this module will be deserialized objects. If you want to load this module into PowerShell please use 'Import-Module -SkipEditionCheck' syntax.
    ///  
    /// </summary>
    internal static string WinCompatModuleWarning {
        get {
            return ResourceManager.GetString("WinCompatModuleWarning", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Detected Windows PowerShell version {0}. Windows PowerShell 5.1 is required to load modules using Windows PowerShell compatibility feature. Install Windows Management Framework (WMF) 5.1 from https://aka.ms/WMF5Download to enable this feature.
    ///  
    /// </summary>
    internal static string WinCompatRequredVersionError {
        get {
            return ResourceManager.GetString("WinCompatRequredVersionError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Module '{0}' is blocked from loading using Windows PowerShell compatibility feature by a 'WindowsPowerShellCompatibilityModuleDenyList' setting in PowerShell configuration file.
    ///  
    /// </summary>
    internal static string WinCompatModuleInDenyList {
        get {
            return ResourceManager.GetString("WinCompatModuleInDenyList", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The module {0} cannot be imported over a CimSession. Try using the PSSession parameter of the Import-Module cmdlet.
    ///  
    /// </summary>
    internal static string PsModuleOverCimSessionError {
        get {
            return ResourceManager.GetString("PsModuleOverCimSessionError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The processor architecture value of {0} is not supported. Run the New-ModuleManifest command again, specifying one of the following supported enumeration values for processor architecture: None, MSIL, X86, Amd64, Arm
    ///  
    /// </summary>
    internal static string InvalidProcessorArchitectureInManifest {
        get {
            return ResourceManager.GetString("InvalidProcessorArchitectureInManifest", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Running the Get-Module cmdlet against a remote computer can only list available modules. Add the ListAvailable parameter to your command, and then try again.
    ///  
    /// </summary>
    internal static string RemoteDiscoveryWorksOnlyInListAvailableMode {
        get {
            return ResourceManager.GetString("RemoteDiscoveryWorksOnlyInListAvailableMode", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The '{0}' module was not imported because the '{0}' snap-in was already imported.
    ///  
    /// </summary>
    internal static string ModuleLoadedAsASnapin {
        get {
            return ResourceManager.GetString("ModuleLoadedAsASnapin", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Wildcard characters are not allowed in the member 'RequiredAssemblies' in the module manifest '{0}'.
    ///  
    /// </summary>
    internal static string WildCardNotAllowedInRequiredAssemblies {
        get {
            return ResourceManager.GetString("WildCardNotAllowedInRequiredAssemblies", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The value of the {0} key in {1} is {2} and the module has nested modules. When a CDXML file is the root module, the Import-Module command fails because the commands in nested modules cannot be exported. Move the CDXML file to the NestedModules key and try the command again.
    ///    {0} is equal to either ModuleToProcess or RootModule
    ///{1} is a placeholder for a file path to psd1 file
    ///{2} is a placeholder for a file path to cdxml file
    ///  
    /// </summary>
    internal static string CmdletizationDoesSupportRexportingNestedModules {
        get {
            return ResourceManager.GetString("CmdletizationDoesSupportRexportingNestedModules", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Failure from remote command: {0}: {{0}}
    ///  
    /// </summary>
    internal static string RemoteDiscoveryRemotePsrpCommandFailed {
        get {
            return ResourceManager.GetString("RemoteDiscoveryRemotePsrpCommandFailed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Failed to generate proxies for remote module '{0}'. {{0}}
    ///  
    /// </summary>
    internal static string RemoteDiscoveryFailedToGenerateProxyForRemoteModule {
        get {
            return ResourceManager.GetString("RemoteDiscoveryFailedToGenerateProxyForRemoteModule", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Failed to process the remote module {0}. {1}
    ///  
    /// </summary>
    internal static string RemoteDiscoveryFailedToProcessRemoteModule {
        get {
            return ResourceManager.GetString("RemoteDiscoveryFailedToProcessRemoteModule", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Failed to receive module data from the remote CimSession. {0}
    ///  
    /// </summary>
    internal static string RemoteDiscoveryFailureFromDiscoveryProvider {
        get {
            return ResourceManager.GetString("RemoteDiscoveryFailureFromDiscoveryProvider", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The required module '{0}' with GUID '{1}' and version '{2}' was not loaded because no valid module file was found in any module directory.
    ///  
    /// </summary>
    internal static string RequiredModuleNotFoundWrongGuidVersion {
        get {
            return ResourceManager.GetString("RequiredModuleNotFoundWrongGuidVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A CIM provider for module discovery was not found on the CIM server. {0}
    ///    {0} is a placeholder for a more detailed error message
    ///  
    /// </summary>
    internal static string RemoteDiscoveryProviderNotFound {
        get {
            return ResourceManager.GetString("RemoteDiscoveryProviderNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot verify the Microsoft .NET Framework version {0} because it is not included in the list of permitted versions.
    ///  
    /// </summary>
    internal static string CannotDetectNetFrameworkVersion {
        get {
            return ResourceManager.GetString("CannotDetectNetFrameworkVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Analyzing {0}.
    ///    {0} should not be localized, is used to contain a file path.
    ///  
    /// </summary>
    internal static string ScriptAnalysisModule {
        get {
            return ResourceManager.GetString("ScriptAnalysisModule", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Preparing modules for first use.
    ///  
    /// </summary>
    internal static string ScriptAnalysisPreparing {
        get {
            return ResourceManager.GetString("ScriptAnalysisPreparing", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Searching for available modules
    ///  
    /// </summary>
    internal static string DeterminingAvailableModules {
        get {
            return ResourceManager.GetString("DeterminingAvailableModules", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Searching UNC share {0}.
    ///    {0} should not be localized, is used to contain a file path.
    ///  
    /// </summary>
    internal static string SearchingUncShare {
        get {
            return ResourceManager.GetString("SearchingUncShare", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Running the Get-Module cmdlet against a remote computer can only be done for module names that do not include a path. Name parameter has this element '{0}' which resolves to a path. Update the Name parameter to not have path elements, and then try again.
    ///  
    /// </summary>
    internal static string RemoteDiscoveryWorksOnlyForUnQualifiedNames {
        get {
            return ResourceManager.GetString("RemoteDiscoveryWorksOnlyForUnQualifiedNames", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Running the Get-Module cmdlet without ListAvailable parameter is not supported for module names that include a path. Name parameter has this element '{0}' which resolves to a path. Update the Name parameter to not have path elements, and then try again.
    ///  
    /// </summary>
    internal static string ModuleDiscoveryForLoadedModulesWorksOnlyForUnQualifiedNames {
        get {
            return ResourceManager.GetString("ModuleDiscoveryForLoadedModulesWorksOnlyForUnQualifiedNames", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified module '{0}' was not found. Update the Name parameter to point to a valid path, and then try again. 
    ///  
    /// </summary>
    internal static string ModuleNotFoundForGetModule {
        get {
            return ResourceManager.GetString("ModuleNotFoundForGetModule", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Populating RepositorySourceLocation property for module {0}.
    ///  
    /// </summary>
    internal static string PopulatingRepositorySourceLocation {
        get {
            return ResourceManager.GetString("PopulatingRepositorySourceLocation", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The module to process '{0}', listed in field '{1}' of module manifest '{2}' was not processed. {3}
    ///  
    /// </summary>
    internal static string ManifestMemberNotValid {
        get {
            return ResourceManager.GetString("ManifestMemberNotValid", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    This prerequisite is valid for the PowerShell Desktop edition only.
    ///  
    /// </summary>
    internal static string PrerequisiteForDesktopEditionOnly {
        get {
            return ResourceManager.GetString("PrerequisiteForDesktopEditionOnly", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Module '{0}' does not support current PowerShell edition '{1}'. Its supported editions are '{2}'. Use 'Import-Module -SkipEditionCheck' to ignore the compatibility of this module.
    ///  
    /// </summary>
    internal static string PSEditionNotSupported {
        get {
            return ResourceManager.GetString("PSEditionNotSupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Module '{0}' supports PowerShell edition '{1}' and cannot be loaded implicitly using the Windows Compatibility feature because it is disabled in the settings file. Use 'Import-Module -UseWindowsPowerShell' to load this module with Windows PowerShell or 'Import-Module -SkipEditionCheck' to try to load the module with the current PowerShell.
    ///  
    /// </summary>
    internal static string ImplicitWinCompatDisabled {
        get {
            return ResourceManager.GetString("ImplicitWinCompatDisabled", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A non-empty string value should be specified for an experimental feature declared in the module manifest.
    ///  
    /// </summary>
    internal static string ExperimentalFeatureNameMissingOrEmpty {
        get {
            return ResourceManager.GetString("ExperimentalFeatureNameMissingOrEmpty", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    One or more invalid experimental feature names found: {0}. A module experimental feature name should follow this convention: 'ModuleName.FeatureName'.
    ///  
    /// </summary>
    internal static string InvalidExperimentalFeatureName {
        get {
            return ResourceManager.GetString("InvalidExperimentalFeatureName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The -SkipEditionCheck switch parameter cannot be used without the -ListAvailable switch parameter.
    ///  
    /// </summary>
    internal static string SkipEditionCheckNotSupportedWithoutListAvailable {
        get {
            return ResourceManager.GetString("SkipEditionCheckNotSupportedWithoutListAvailable", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Importing *.ps1 files as modules is not allowed in ConstrainedLanguage mode.
    ///  
    /// </summary>
    internal static string ImportPSFileNotAllowedInConstrainedLanguage {
        get {
            return ResourceManager.GetString("ImportPSFileNotAllowedInConstrainedLanguage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    An error has occurred while loading script module {0} because it has a different language mode than the module manifest. The manifest language mode is {1} and the module language mode is {2}. Ensure all module files are signed or otherwise part of your application allow list configuration.
    ///  
    /// </summary>
    internal static string MismatchedLanguageModes {
        get {
            return ResourceManager.GetString("MismatchedLanguageModes", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    This module uses the dot-source operator while exporting functions using wildcard characters, and this is disallowed when the system is under application verification enforcement.
    ///  
    /// </summary>
    internal static string CannotUseDotSourceWithWildCardFunctionExport {
        get {
            return ResourceManager.GetString("CannotUseDotSourceWithWildCardFunctionExport", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot export module members from a module that has a different language mode from the running session.
    ///  
    /// </summary>
    internal static string CannotExportMembersAccrossLanguageBoundaries {
        get {
            return ResourceManager.GetString("CannotExportMembersAccrossLanguageBoundaries", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot create new module while the session is in ConstrainedLanguage mode.
    ///  
    /// </summary>
    internal static string CannotCreateModuleWithScriptBlock {
        get {
            return ResourceManager.GetString("CannotCreateModuleWithScriptBlock", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot find the built-in module '{0}' that is compatible with the 'Core' edition. Please make sure the PowerShell built-in modules are available. They usually come with the PowerShell package under the $PSHOME module path, and are required for PowerShell to function properly.
    ///  
    /// </summary>
    internal static string CannotFindCoreCompatibleBuiltInModule {
        get {
            return ResourceManager.GetString("CannotFindCoreCompatibleBuiltInModule", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Export-ModuleMember Cmdlet
    ///  
    /// </summary>
    internal static string WDACExportModuleCommandLogTitle {
        get {
            return ResourceManager.GetString("WDACExportModuleCommandLogTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Export of module members will fail in Constrained Language mode because module '{0}', has a language mode '{1}' that is different from the current session '{2}'.
    ///  
    /// </summary>
    internal static string WDACExportModuleCommandLogMessage {
        get {
            return ResourceManager.GetString("WDACExportModuleCommandLogMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Module Implicit Function Export
    ///  
    /// </summary>
    internal static string WDACImplicitFunctionExportLogTitle {
        get {
            return ResourceManager.GetString("WDACImplicitFunctionExportLogTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Implicit function export for module '{0}' will be denied because it is trusted (runs in Full Language mode) but the session is not trusted (runs in Constrained Language mode). It is best practice to always export module functions individually by full name.
    ///  
    /// </summary>
    internal static string WDACImplicitFunctionExportLogMessage {
        get {
            return ResourceManager.GetString("WDACImplicitFunctionExportLogMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Importing Script File as Module
    ///  
    /// </summary>
    internal static string WDACScriptFileImportLogTitle {
        get {
            return ResourceManager.GetString("WDACScriptFileImportLogTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Importing the script file '{0}' as a module will be disallowed in ConstrainedLanguage mode.
    ///  
    /// </summary>
    internal static string WDACScriptFileImportLogMessage {
        get {
            return ResourceManager.GetString("WDACScriptFileImportLogMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Module Contains Dot-Source Operator
    ///  
    /// </summary>
    internal static string WDACModuleDotSourceLogTitle {
        get {
            return ResourceManager.GetString("WDACModuleDotSourceLogTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Module '{0}' import will in fail Constrained Language mode because it exports functions using wildcard characters while also using the dot-source operator.
    ///  
    /// </summary>
    internal static string WDACModuleDotSourceLogMessage {
        get {
            return ResourceManager.GetString("WDACModuleDotSourceLogMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    "Module Exporting Functions
    ///  
    /// </summary>
    internal static string WDACModuleFnExportWithNestedModulesLogTitle {
        get {
            return ResourceManager.GetString("WDACModuleFnExportWithNestedModulesLogTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Module '{0}' exports functions using name wildcard characters. Any nested module function names will be removed when running in Constrained Language mode.
    ///  
    /// </summary>
    internal static string WDACModuleFnExportWithNestedModulesLogMessage {
        get {
            return ResourceManager.GetString("WDACModuleFnExportWithNestedModulesLogMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    "New-Module Cmdlet
    ///  
    /// </summary>
    internal static string WDACNewModuleCommandLogTitle {
        get {
            return ResourceManager.GetString("WDACNewModuleCommandLogTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A new module from an untrusted Constrained Language session will be blocked from providing the FullLanguage script block.
    ///  
    /// </summary>
    internal static string WDACNewModuleCommandLogMessage {
        get {
            return ResourceManager.GetString("WDACNewModuleCommandLogMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    "Module Mismatched Language Modes
    ///  
    /// </summary>
    internal static string WDACMismatchedLanguageModesTitle {
        get {
            return ResourceManager.GetString("WDACMismatchedLanguageModesTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A dependent module is being loaded that has a different language mode than the parent. This will be disallowed when in Constrained Language mode.
    ///  
    /// </summary>
    internal static string WDACMismatchedLanguageModesMessage {
        get {
            return ResourceManager.GetString("WDACMismatchedLanguageModesMessage", resourceCulture);
        }
    }

}

