
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class DiscoveryExceptions {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal DiscoveryExceptions() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.DiscoveryExceptions", typeof(DiscoveryExceptions).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The cmdlet name "{0}" cannot be validated because it is not in the correct format. Cmdlet names must include a verb and a noun separated by a "-", such as "Get-Process".
    ///  
    /// </summary>
    internal static string CmdletFormatInvalid {
        get {
            return ResourceManager.GetString("CmdletFormatInvalid", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The parameter "{0}" is declared in parameter-set "{1}" multiple times.
    ///  
    /// </summary>
    internal static string ParameterDeclaredInParameterSetMultipleTimes {
        get {
            return ResourceManager.GetString("ParameterDeclaredInParameterSetMultipleTimes", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The alias "{0}" is declared multiple times.
    ///  
    /// </summary>
    internal static string AliasDeclaredMultipleTimes {
        get {
            return ResourceManager.GetString("AliasDeclaredMultipleTimes", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Parameter could not be declared. Parameters can be declared only on fields and properties.
    ///  
    /// </summary>
    internal static string CompiledCommandParameterMemberMustBeFieldOrProperty {
        get {
            return ResourceManager.GetString("CompiledCommandParameterMemberMustBeFieldOrProperty", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the cmdlet. A cmdlet name must consist of a verb and noun pair separated by '-'.
    ///  
    /// </summary>
    internal static string InvalidCmdletNameFormat {
        get {
            return ResourceManager.GetString("InvalidCmdletNameFormat", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The term '{0}' is not recognized as a name of a cmdlet, function, script file, or executable program.
    ///Check the spelling of the name, or if a path was included, verify that the path is correct and try again.
    ///  
    /// </summary>
    internal static string CommandNotFoundException {
        get {
            return ResourceManager.GetString("CommandNotFoundException", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Argument '{0}' is not recognized as a cmdlet: {1}
    ///  
    /// </summary>
    internal static string CmdletNotFoundException {
        get {
            return ResourceManager.GetString("CmdletNotFoundException", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The argument '{0}' is not recognized as a cmdlet, possibly because it does not derive from the Cmdlet or PSCmdlet classes: {1}
    ///  
    /// </summary>
    internal static string CmdletDoesNotDeriveFromCmdletType {
        get {
            return ResourceManager.GetString("CmdletDoesNotDeriveFromCmdletType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot resolve alias '{0}' because it refers to term '{1}', which is not recognized as a cmdlet, function, executable program, or script file. Verify the term and try again.
    ///  
    /// </summary>
    internal static string AliasNotResolvedException {
        get {
            return ResourceManager.GetString("AliasNotResolvedException", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Parameter '{0}' with value '{1}' cannot be processed because it is not a cmdlet and cannot be processed by the CommandProcessor.
    ///  
    /// </summary>
    internal static string CommandNameNotCmdlet {
        get {
            return ResourceManager.GetString("CommandNameNotCmdlet", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A cmdlet named '{0}' already exists. Cmdlets must have unique names.
    ///  
    /// </summary>
    internal static string DuplicateCmdletName {
        get {
            return ResourceManager.GetString("DuplicateCmdletName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A cmdlet provider named '{0}' already exists. Cmdlet providers must have unique names.
    ///  
    /// </summary>
    internal static string DuplicateCmdletProviderName {
        get {
            return ResourceManager.GetString("DuplicateCmdletProviderName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    An assembly named '{0}' already exists. Assemblies must have unique names.
    ///  
    /// </summary>
    internal static string DuplicateAssemblyName {
        get {
            return ResourceManager.GetString("DuplicateAssemblyName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A script named '{0}' already exists. Scripts must have unique names.
    ///  
    /// </summary>
    internal static string DuplicateScriptName {
        get {
            return ResourceManager.GetString("DuplicateScriptName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the #requires statement because it is not in the correct format.
    ///The #requires statement must be in one of the following formats:
    /// "#requires -shellid <shellID>"
    /// "#requires -version <major.minor>"
    /// "#requires -psedition <edition>"
    /// "#requires -pssnapin <psSnapInName> [-version <major.minor>]"
    /// "#requires -modules <ModuleSpecification>"
    /// "#requires -runasadministrator"
    ///  
    /// </summary>
    internal static string ScriptRequiresInvalidFormat {
        get {
            return ResourceManager.GetString("ScriptRequiresInvalidFormat", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The script '{0}' cannot be run because it contained a "#requires" statement with a shell ID of {1} that is incompatible with the current shell. To run this script you must use the shell located at '{2}'.
    ///  
    /// </summary>
    internal static string RequiresInterpreterNotCompatible {
        get {
            return ResourceManager.GetString("RequiresInterpreterNotCompatible", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The script '{0}' cannot be run because it contained a "#requires" statement with a shell ID of {1} that is incompatible with the current shell.
    ///  
    /// </summary>
    internal static string RequiresInterpreterNotCompatibleNoPath {
        get {
            return ResourceManager.GetString("RequiresInterpreterNotCompatibleNoPath", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The script '{0}' cannot be run because it contained a "#requires" statement for PowerShell {1}. The version of PowerShell that is required by the script does not match the currently running version of PowerShell {2}.
    ///  
    /// </summary>
    internal static string RequiresPSVersionNotCompatible {
        get {
            return ResourceManager.GetString("RequiresPSVersionNotCompatible", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The script '{0}' cannot be run because it contained a "#requires" statement for PowerShell editions '{1}'. The edition of PowerShell that is required by the script does not match the currently running PowerShell {2} edition.
    ///  
    /// </summary>
    internal static string RequiresPSEditionNotCompatible {
        get {
            return ResourceManager.GetString("RequiresPSEditionNotCompatible", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The script '{0}' cannot be run because the following snap-ins that are specified by the "#requires" statements of the script are missing: {1}.
    ///  
    /// </summary>
    internal static string RequiresMissingPSSnapIns {
        get {
            return ResourceManager.GetString("RequiresMissingPSSnapIns", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A #requires statement has specified only a shellID. #Requires statements must specify a required PowerShell snap-in when running in PowerShell.
    ///  
    /// </summary>
    internal static string RequiresShellIDInvalidForSingleShell {
        get {
            return ResourceManager.GetString("RequiresShellIDInvalidForSingleShell", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The script '{0}' cannot be run because it contains a "#requires" statement for running as Administrator. The current PowerShell session is not running as Administrator. Start PowerShell by using the Run as Administrator option, and then try running the script again.
    ///  
    /// </summary>
    internal static string RequiresElevation {
        get {
            return ResourceManager.GetString("RequiresElevation", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0} (Version {1})
    ///  
    /// </summary>
    internal static string PSSnapInNameVersion {
        get {
            return ResourceManager.GetString("PSSnapInNameVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The command could not be retrieved because the ArgumentList parameter can be specified only when retrieving a single cmdlet or script.
    ///  
    /// </summary>
    internal static string CommandArgsOnlyForSingleCmdlet {
        get {
            return ResourceManager.GetString("CommandArgsOnlyForSingleCmdlet", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The parameter name "{0}" is reserved for future use.
    ///  
    /// </summary>
    internal static string ReservedParameterName {
        get {
            return ResourceManager.GetString("ReservedParameterName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The script '{0}' cannot be run because the following modules that are specified by the "#requires" statements of the script are missing: {1}.
    ///  
    /// </summary>
    internal static string RequiresMissingModules {
        get {
            return ResourceManager.GetString("RequiresMissingModules", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The '{0}' command was found in the module '{1}', but the module could not be loaded. For more information, run 'Import-Module {1}'.
    ///  
    /// </summary>
    internal static string CouldNotAutoImportMatchingModule {
        get {
            return ResourceManager.GetString("CouldNotAutoImportMatchingModule", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The '{0}' command was found in the module '{1}', but the module could not be loaded due to the following error: [{2}]
    ///For more information, run 'Import-Module {1}'.
    ///  
    /// </summary>
    internal static string CouldNotAutoImportMatchingModuleWithErrorMessage {
        get {
            return ResourceManager.GetString("CouldNotAutoImportMatchingModuleWithErrorMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The module '{0}' could not be loaded. For more information, run 'Import-Module {0}'.
    ///  
    /// </summary>
    internal static string CouldNotAutoImportModule {
        get {
            return ResourceManager.GetString("CouldNotAutoImportModule", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    No matching commands include a parameter named '{0}'.  Check the spelling of the parameter name, and then try again.
    ///  
    /// </summary>
    internal static string CommandParameterNotFound {
        get {
            return ResourceManager.GetString("CommandParameterNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot dot-source this command because it was defined in a different language mode. To invoke this command without importing its contents, omit the '.' operator.
    ///  
    /// </summary>
    internal static string DotSourceNotSupported {
        get {
            return ResourceManager.GetString("DotSourceNotSupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The ShowCommandInfo and Syntax parameters cannot be specified together.
    ///  
    /// </summary>
    internal static string GetCommandShowCommandInfoParamError {
        get {
            return ResourceManager.GetString("GetCommandShowCommandInfoParamError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    This script command is disabled when the experimental feature '{0}' has been turned on.
    ///  
    /// </summary>
    internal static string ScriptDisabledWhenFeatureOn {
        get {
            return ResourceManager.GetString("ScriptDisabledWhenFeatureOn", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    This script command is disabled when the experimental feature '{0}' has been turned off.
    ///  
    /// </summary>
    internal static string ScriptDisabledWhenFeatureOff {
        get {
            return ResourceManager.GetString("ScriptDisabledWhenFeatureOff", resourceCulture);
        }
    }

}

