
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class ConsoleInfoErrorStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal ConsoleInfoErrorStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.ConsoleInfoErrorStrings", typeof(ConsoleInfoErrorStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Incorrect PowerShell version {0}. PowerShell version {1} is supported on this computer.
    ///  
    /// </summary>
    internal static string BadMonadVersion {
        get {
            return ResourceManager.GetString("BadMonadVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The following errors occurred when loading console {0}: {1}
    ///  
    /// </summary>
    internal static string ConsoleLoadFailure {
        get {
            return ResourceManager.GetString("ConsoleLoadFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot load PowerShell snap-in {0} because of the following error: {1}
    ///  
    /// </summary>
    internal static string PSSnapInLoadFailure {
        get {
            return ResourceManager.GetString("PSSnapInLoadFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    PowerShell snap-in "{0}" loaded with the following warnings: {1}
    ///  
    /// </summary>
    internal static string PSSnapInLoadWarning {
        get {
            return ResourceManager.GetString("PSSnapInLoadWarning", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The PowerShell snap-in module {0} does not have the required PowerShell snap-in strong name {1}.
    ///  
    /// </summary>
    internal static string PSSnapInAssemblyNameMismatch {
        get {
            return ResourceManager.GetString("PSSnapInAssemblyNameMismatch", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The cmdlet '{0}' should not occur more than once in PowerShell snap-in '{1}'.
    ///  
    /// </summary>
    internal static string PSSnapInDuplicateCmdlets {
        get {
            return ResourceManager.GetString("PSSnapInDuplicateCmdlets", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    PowerShell provider '{0}' should not occur more than once in PowerShell snap-in '{1}'.
    ///  
    /// </summary>
    internal static string PSSnapInDuplicateProviders {
        get {
            return ResourceManager.GetString("PSSnapInDuplicateProviders", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    PowerShell {0} is not supported in the current console. PowerShell {1} is supported in the current console.
    ///  
    /// </summary>
    internal static string AddPSSnapInBadMonadVersion {
        get {
            return ResourceManager.GetString("AddPSSnapInBadMonadVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    File {0} already exists and {1} was specified.
    ///  
    /// </summary>
    internal static string FileExistsNoClobber {
        get {
            return ResourceManager.GetString("FileExistsNoClobber", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The provided configuration file '{0}' does not exist.
    ///  
    /// </summary>
    internal static string ConfigurationFileDoesNotExist {
        get {
            return ResourceManager.GetString("ConfigurationFileDoesNotExist", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The provided configuration file '{0}' must have a .pssc file extension.
    ///  
    /// </summary>
    internal static string NotConfigurationFile {
        get {
            return ResourceManager.GetString("NotConfigurationFile", resourceCulture);
        }
    }

}

