
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class NewObjectStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal NewObjectStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.PowerShell.Commands.Utility.resources.NewObjectStrings", typeof(NewObjectStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A constructor was not found. Cannot find an appropriate constructor for type {0}.
    ///  
    /// </summary>
    internal static string CannotFindAppropriateCtor {
        get {
            return ResourceManager.GetString("CannotFindAppropriateCtor", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot find type [{0}]: verify that the assembly containing this type is loaded.
    ///  
    /// </summary>
    internal static string TypeNotFound {
        get {
            return ResourceManager.GetString("TypeNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot load COM type {0}.
    ///  
    /// </summary>
    internal static string CannotLoadComObjectType {
        get {
            return ResourceManager.GetString("CannotLoadComObjectType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The object written to the pipeline is an instance of the type "{0}" from the component's primary interoperability assembly. If this type exposes different members than the IDispatch members, scripts that are written to work with this object might not work if the primary interoperability assembly is not installed.
    ///  
    /// </summary>
    internal static string ComInteropLoaded {
        get {
            return ResourceManager.GetString("ComInteropLoaded", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The member "{1}" was not found for the specified {2} object.
    ///  
    /// </summary>
    internal static string MemberNotFound {
        get {
            return ResourceManager.GetString("MemberNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The value supplied is not valid, or the property is read-only. Change the value, and then try again.
    ///  
    /// </summary>
    internal static string InvalidValue {
        get {
            return ResourceManager.GetString("InvalidValue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Creating instances of attribute and delegated Windows RT types is not supported.
    ///  
    /// </summary>
    internal static string CannotInstantiateWinRTType {
        get {
            return ResourceManager.GetString("CannotInstantiateWinRTType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot create instances of the ByRef-like type "{0}". ByRef-like types are not supported in PowerShell.
    ///  
    /// </summary>
    internal static string CannotInstantiateBoxedByRefLikeType {
        get {
            return ResourceManager.GetString("CannotInstantiateBoxedByRefLikeType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot create type. Only core types are supported in this language mode.
    ///  
    /// </summary>
    internal static string CannotCreateTypeConstrainedLanguage {
        get {
            return ResourceManager.GetString("CannotCreateTypeConstrainedLanguage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot create type. Only core types are supported in {0} language mode on a policy locked down machine.
    ///  
    /// </summary>
    internal static string CannotCreateTypeLanguageMode {
        get {
            return ResourceManager.GetString("CannotCreateTypeLanguageMode", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    New-Object Cmdlet Type Creation
    ///  
    /// </summary>
    internal static string TypeWDACLogTitle {
        get {
            return ResourceManager.GetString("TypeWDACLogTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The type '{0}' will not be created in ConstrainedLanguage mode.
    ///  
    /// </summary>
    internal static string TypeWDACLogMessage {
        get {
            return ResourceManager.GetString("TypeWDACLogMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    New-Object Cmdlet COM Object Creation
    ///  
    /// </summary>
    internal static string ComWDACLogTitle {
        get {
            return ResourceManager.GetString("ComWDACLogTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The COM object '{0}' will not be created in ConstrainedLanguage mode.
    ///  
    /// </summary>
    internal static string ComWDACLogMessage {
        get {
            return ResourceManager.GetString("ComWDACLogMessage", resourceCulture);
        }
    }

}

