
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class TabCompletionStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal TabCompletionStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.TabCompletionStrings", typeof(TabCompletionStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The tab completion result cannot be properly deserialized because the remote runspace does not contain a TypeTable instance.
    ///  
    /// </summary>
    internal static string CannotDeserializeTabCompletionResult {
        get {
            return ResourceManager.GetString("CannotDeserializeTabCompletionResult", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot access properties on a null instance of the type CompletionResult.
    ///  
    /// </summary>
    internal static string NoAccessToProperties {
        get {
            return ResourceManager.GetString("NoAccessToProperties", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Bitwise NOT
    ///  
    /// </summary>
    internal static string bnotOperatorDescription {
        get {
            return ResourceManager.GetString("bnotOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Logical not. Negates the statement that follows it.
    ///  
    /// </summary>
    internal static string notOperatorDescription {
        get {
            return ResourceManager.GetString("notOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Equal to - case insensitive. When the left operand is a collection, returns values from the collection that equal the right operand, otherwise returns TRUE if the left operand equals the right operand.
    ///  
    /// </summary>
    internal static string eqOperatorDescription {
        get {
            return ResourceManager.GetString("eqOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Equal to - case insensitive. When the left operand is a collection, returns values from the collection that equal the right operand, otherwise returns TRUE if the left operand equals the right operand.
    ///  
    /// </summary>
    internal static string ieqOperatorDescription {
        get {
            return ResourceManager.GetString("ieqOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Equal to - case sensitive. When the left operand is a collection, returns values from the collection that equal the right operand, otherwise returns TRUE if the left operand equals the right operand.
    ///  
    /// </summary>
    internal static string ceqOperatorDescription {
        get {
            return ResourceManager.GetString("ceqOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Not equal to - case insensitive. When the left operand is a collection, returns values from the collection that do not equal the right operand, otherwise returns TRUE if the left operand does not equal the right operand.
    ///  
    /// </summary>
    internal static string neOperatorDescription {
        get {
            return ResourceManager.GetString("neOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Not equal to - case insensitive. When the left operand is a collection, returns values from the collection that do not equal the right operand, otherwise returns TRUE if the left operand does not equal the right operand.
    ///  
    /// </summary>
    internal static string ineOperatorDescription {
        get {
            return ResourceManager.GetString("ineOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Not equal to - case sensitive. When the left operand is a collection, returns values from the collection that do not equal the right operand, otherwise returns TRUE if the left operand does not equal the right operand.
    ///  
    /// </summary>
    internal static string cneOperatorDescription {
        get {
            return ResourceManager.GetString("cneOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Greater than or equal to - case insensitive. When the left operand is a collection, returns values from the collection that are greater than or equal to the right operand, otherwise returns TRUE if the left operand is greater than or equal to the right operand.
    ///  
    /// </summary>
    internal static string geOperatorDescription {
        get {
            return ResourceManager.GetString("geOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Greater than or equal to - case insensitive. When the left operand is a collection, returns values from the collection that are greater than or equal to the right operand, otherwise returns TRUE if the left operand is greater than or equal to the right operand.
    ///  
    /// </summary>
    internal static string igeOperatorDescription {
        get {
            return ResourceManager.GetString("igeOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Greater than or equal to - case sensitive. When the left operand is a collection, returns values from the collection that are greater than or equal to the right operand, otherwise returns TRUE if the left operand is greater than or equal to the right operand.
    ///  
    /// </summary>
    internal static string cgeOperatorDescription {
        get {
            return ResourceManager.GetString("cgeOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Greater than - case insensitive. When the left operand is a collection, returns values from the collection that are greater than the right operand, otherwise returns TRUE if the left operand is greater than the right operand.
    ///  
    /// </summary>
    internal static string gtOperatorDescription {
        get {
            return ResourceManager.GetString("gtOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Greater than - case insensitive. When the left operand is a collection, returns values from the collection that are greater than the right operand, otherwise returns TRUE if the left operand is greater than the right operand.
    ///  
    /// </summary>
    internal static string igtOperatorDescription {
        get {
            return ResourceManager.GetString("igtOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Greater than - case sensitive. When the left operand is a collection, returns values from the collection that are greater than the right operand, otherwise returns TRUE if the left operand is greater than the right operand.
    ///  
    /// </summary>
    internal static string cgtOperatorDescription {
        get {
            return ResourceManager.GetString("cgtOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Less than - case insensitive. When the left operand is a collection, returns values from the collection that are less than the right operand, otherwise returns TRUE if the left operand is less than the right operand.
    ///  
    /// </summary>
    internal static string ltOperatorDescription {
        get {
            return ResourceManager.GetString("ltOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Less than - case insensitive. When the left operand is a collection, returns values from the collection that are less than the right operand, otherwise returns TRUE if the left operand is less than the right operand.
    ///  
    /// </summary>
    internal static string iltOperatorDescription {
        get {
            return ResourceManager.GetString("iltOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Less than - case sensitive. When the left operand is a collection, returns values from the collection that are less than the right operand, otherwise returns TRUE if the left operand is less than the right operand.
    ///  
    /// </summary>
    internal static string cltOperatorDescription {
        get {
            return ResourceManager.GetString("cltOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Less than or equal to - case insensitive. When the left operand is a collection, returns values from the collection that are less than or equal to the right operand, otherwise returns TRUE if the left operand is less than or equal to the right operand.
    ///  
    /// </summary>
    internal static string leOperatorDescription {
        get {
            return ResourceManager.GetString("leOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Less than or equal to - case insensitive. When the left operand is a collection, returns values from the collection that are less than or equal to the right operand, otherwise returns TRUE if the left operand is less than or equal to the right operand.
    ///  
    /// </summary>
    internal static string ileOperatorDescription {
        get {
            return ResourceManager.GetString("ileOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Less than or equal to - case sensitive. When the left operand is a collection, returns values from the collection that are less than or equal to the right operand, otherwise returns TRUE if the left operand is less than or equal to the right operand.
    ///  
    /// </summary>
    internal static string cleOperatorDescription {
        get {
            return ResourceManager.GetString("cleOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Wildcard matching operator - case insensitive. When the left operand is a collection, returns values from the collection that match the right hand operand, otherwise returns TRUE if the left operand matches the right operand.
    ///  
    /// </summary>
    internal static string likeOperatorDescription {
        get {
            return ResourceManager.GetString("likeOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Wildcard matching operator - case insensitive. When the left operand is a collection, returns values from the collection that match the right hand operand, otherwise returns TRUE if the left operand matches the right operand.
    ///  
    /// </summary>
    internal static string ilikeOperatorDescription {
        get {
            return ResourceManager.GetString("ilikeOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Wildcard matching operator - case sensitive. When the left operand is a collection, returns values from the collection that match the right hand operand, otherwise returns TRUE if the left operand matches the right operand.
    ///  
    /// </summary>
    internal static string clikeOperatorDescription {
        get {
            return ResourceManager.GetString("clikeOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Wildcard matching operator - case insensitive. When the left operand is a collection, returns values from the collection that do not match the right hand operand, otherwise returns TRUE if the left operand does not match the right operand.
    ///  
    /// </summary>
    internal static string notlikeOperatorDescription {
        get {
            return ResourceManager.GetString("notlikeOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Wildcard matching operator - case insensitive. When the left operand is a collection, returns values from the collection that do not match the right hand operand, otherwise returns TRUE if the left operand does not match the right operand.
    ///  
    /// </summary>
    internal static string inotlikeOperatorDescription {
        get {
            return ResourceManager.GetString("inotlikeOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Wildcard matching operator - case sensitive. When the left operand is a collection, returns values from the collection that do not match the right hand operand, otherwise returns TRUE if the left operand does not match the right operand.
    ///  
    /// </summary>
    internal static string cnotlikeOperatorDescription {
        get {
            return ResourceManager.GetString("cnotlikeOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Regular expression matching operator - case insensitive. When the left operand is a collection, returns values from the collection that match the right hand operand, otherwise returns TRUE if the left operand matches the right operand.
    ///  
    /// </summary>
    internal static string matchOperatorDescription {
        get {
            return ResourceManager.GetString("matchOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Regular expression matching operator - case insensitive. When the left operand is a collection, returns values from the collection that match the right hand operand, otherwise returns TRUE if the left operand matches the right operand.
    ///  
    /// </summary>
    internal static string imatchOperatorDescription {
        get {
            return ResourceManager.GetString("imatchOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Regular expression matching operator - case sensitive. When the left operand is a collection, returns values from the collection that match the right hand operand, otherwise returns TRUE if the left operand matches the right operand.
    ///  
    /// </summary>
    internal static string cmatchOperatorDescription {
        get {
            return ResourceManager.GetString("cmatchOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Regular expression matching operator - case insensitive. When the left operand is a collection, returns values from the collection that do not match the right hand operand, otherwise returns TRUE if the left operand does not match the right operand.
    ///  
    /// </summary>
    internal static string notmatchOperatorDescription {
        get {
            return ResourceManager.GetString("notmatchOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Regular expression matching operator - case insensitive. When the left operand is a collection, returns values from the collection that do not match the right hand operand, otherwise returns TRUE if the left operand does not match the right operand.
    ///  
    /// </summary>
    internal static string inotmatchOperatorDescription {
        get {
            return ResourceManager.GetString("inotmatchOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Regular expression matching operator - case sensitive. When the left operand is a collection, returns values from the collection that do not match the right hand operand, otherwise returns TRUE if the left operand does not match the right operand.
    ///  
    /// </summary>
    internal static string cnotmatchOperatorDescription {
        get {
            return ResourceManager.GetString("cnotmatchOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Replace operator - case insensitive. Changes the left operand. Example: (dir *.ps1).FullName -replace '.ps1$','.ps1.bak'
    ///  
    /// </summary>
    internal static string replaceOperatorDescription {
        get {
            return ResourceManager.GetString("replaceOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Replace operator - case insensitive. Changes the left operand. Example: (dir *.ps1).FullName -replace '.ps1$','.ps1.bak'
    ///  
    /// </summary>
    internal static string ireplaceOperatorDescription {
        get {
            return ResourceManager.GetString("ireplaceOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Replace operator - case sensitive. Changes the left operand. Example: (dir *.ps1).FullName -replace '.ps1$','.ps1.bak'
    ///  
    /// </summary>
    internal static string creplaceOperatorDescription {
        get {
            return ResourceManager.GetString("creplaceOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Containment operator - case insensitive. Returns TRUE when the test value (right operand) exactly matches at least one of the values in the left operand.
    ///  
    /// </summary>
    internal static string containsOperatorDescription {
        get {
            return ResourceManager.GetString("containsOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Containment operator - case insensitive. Returns TRUE when the test value (right operand) exactly matches at least one of the values in the left operand.
    ///  
    /// </summary>
    internal static string icontainsOperatorDescription {
        get {
            return ResourceManager.GetString("icontainsOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Containment operator - case sensitive. Returns TRUE only when the test value (right operand) exactly matches at least one of the values in the left operand.
    ///  
    /// </summary>
    internal static string ccontainsOperatorDescription {
        get {
            return ResourceManager.GetString("ccontainsOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Containment operator - case insensitive. Returns TRUE when the test value (right operand) exactly matches none of the values in the left operand.
    ///  
    /// </summary>
    internal static string notcontainsOperatorDescription {
        get {
            return ResourceManager.GetString("notcontainsOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Containment operator - case insensitive. Returns TRUE when the test value (right operand) exactly matches none of the values in the left operand.
    ///  
    /// </summary>
    internal static string inotcontainsOperatorDescription {
        get {
            return ResourceManager.GetString("inotcontainsOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Containment operator - case sensitive. Returns TRUE when the test value (right operand) exactly matches none of the values in the left operand.
    ///  
    /// </summary>
    internal static string cnotcontainsOperatorDescription {
        get {
            return ResourceManager.GetString("cnotcontainsOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Containment operator - case insensitive. Returns TRUE when the test value (left operand) exactly matches at least one of the values in the right operand.
    ///  
    /// </summary>
    internal static string inOperatorDescription {
        get {
            return ResourceManager.GetString("inOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Containment operator - case insensitive. Returns TRUE when the test value (left operand) exactly matches at least one of the values in the right operand.
    ///  
    /// </summary>
    internal static string iinOperatorDescription {
        get {
            return ResourceManager.GetString("iinOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Containment operator - case sensitive. Returns TRUE when the test value (left operand) exactly matches at least one of the values in the right operand.
    ///  
    /// </summary>
    internal static string cinOperatorDescription {
        get {
            return ResourceManager.GetString("cinOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Containment operator - case sensitive. Returns TRUE when the test value (left operand) exactly matches none of the values in the right operand.
    ///  
    /// </summary>
    internal static string notinOperatorDescription {
        get {
            return ResourceManager.GetString("notinOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Containment operator - case insensitive. Returns TRUE when the test value (left operand) exactly matches none of the values in the right operand.
    ///  
    /// </summary>
    internal static string inotinOperatorDescription {
        get {
            return ResourceManager.GetString("inotinOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Containment operator - case sensitive. Returns TRUE when the test value (left operand) exactly matches none of the values in the right operand.
    ///  
    /// </summary>
    internal static string cnotinOperatorDescription {
        get {
            return ResourceManager.GetString("cnotinOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Split - case insensitive. Split one or more strings into substrings.
    ///-Split <String>
    ///
    ///<String> -Split <Delimiter>[,<Max-substrings>[,"<Options>"]]
    ///
    ///<String> -Split {<ScriptBlock>} [,<Max-substrings>]
    ///  
    /// </summary>
    internal static string splitOperatorDescription {
        get {
            return ResourceManager.GetString("splitOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Split - case insensitive. Split one or more strings into substrings.
    ///-Split <String>
    ///
    ///<String> -Split <Delimiter>[,<Max-substrings>[,"<Options>"]]
    ///
    ///<String> -Split {<ScriptBlock>} [,<Max-substrings>]
    ///  
    /// </summary>
    internal static string isplitOperatorDescription {
        get {
            return ResourceManager.GetString("isplitOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Split - case sensitive. Split one or more strings into substrings.
    ///-Split <String>
    ///
    ///<String> -Split <Delimiter>[,<Max-substrings>[,"<Options>"]]
    ///
    ///<String> -Split {<ScriptBlock>} [,<Max-substrings>]
    ///  
    /// </summary>
    internal static string csplitOperatorDescription {
        get {
            return ResourceManager.GetString("csplitOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Returns TRUE when the left operand is not an instance of the specified .NET Framework type (right operand).
    ///  
    /// </summary>
    internal static string isnotOperatorDescription {
        get {
            return ResourceManager.GetString("isnotOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Returns TRUE when the left operand is an instance of the specified .NET Framework type (right operand).
    ///  
    /// </summary>
    internal static string isOperatorDescription {
        get {
            return ResourceManager.GetString("isOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Converts the left operand to the specified .NET Framework type (right operand).
    ///  
    /// </summary>
    internal static string asOperatorDescription {
        get {
            return ResourceManager.GetString("asOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Formats strings by using the format method of string objects.
    ///  
    /// </summary>
    internal static string fOperatorDescription {
        get {
            return ResourceManager.GetString("fOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Logical and. Returns TRUE when both statements are TRUE.
    ///  
    /// </summary>
    internal static string andOperatorDescription {
        get {
            return ResourceManager.GetString("andOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Bitwise AND
    ///  
    /// </summary>
    internal static string bandOperatorDescription {
        get {
            return ResourceManager.GetString("bandOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Logical or. TRUE when either or both statements are TRUE.
    ///  
    /// </summary>
    internal static string orOperatorDescription {
        get {
            return ResourceManager.GetString("orOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Bitwise OR (inclusive)
    ///  
    /// </summary>
    internal static string borOperatorDescription {
        get {
            return ResourceManager.GetString("borOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Logical exclusive or. Returns TRUE when one of the statements is TRUE and the other is FALSE.
    ///  
    /// </summary>
    internal static string xorOperatorDescription {
        get {
            return ResourceManager.GetString("xorOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Bitwise OR (exclusive)
    ///  
    /// </summary>
    internal static string bxorOperatorDescription {
        get {
            return ResourceManager.GetString("bxorOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Join - combine multiple strings into a single string.
    ///-Join <String[]>
    ///<String[]> -Join <Delimiter>
    ///  
    /// </summary>
    internal static string joinOperatorDescription {
        get {
            return ResourceManager.GetString("joinOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Shift Left bit operator. Inserts zero in right-most bit position.
    ///  
    /// </summary>
    internal static string shlOperatorDescription {
        get {
            return ResourceManager.GetString("shlOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Shift Right bit operator. Inserts zero in the left-most bit position. For signed values, sign bit is preserved.
    ///  
    /// </summary>
    internal static string shrOperatorDescription {
        get {
            return ResourceManager.GetString("shrOperatorDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [string]
    ///Specifies the name of the property being created.
    ///  
    /// </summary>
    internal static string NameHashtableKeyDescription {
        get {
            return ResourceManager.GetString("NameHashtableKeyDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [string]
    ///Specifies the name of the property being created.
    ///  
    /// </summary>
    internal static string LabelHashtableKeyDescription {
        get {
            return ResourceManager.GetString("LabelHashtableKeyDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [scriptblock]
    ///A script block used to calculate the value of the new property.
    ///  
    /// </summary>
    internal static string ExpressionHashtableKeyDescription {
        get {
            return ResourceManager.GetString("ExpressionHashtableKeyDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [string]
    ///Define how the values are displayed in a column.
    ///Valid values are 'left', 'center', or 'right'.
    ///  
    /// </summary>
    internal static string AlignmentHashtableKeyDescription {
        get {
            return ResourceManager.GetString("AlignmentHashtableKeyDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [string]
    ///Specifies a format string that defines how the value is formatted for output.
    ///  
    /// </summary>
    internal static string FormatStringHashtableKeyDescription {
        get {
            return ResourceManager.GetString("FormatStringHashtableKeyDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [int]
    ///Specifies the maximum column width in a table when the value is displayed.
    ///The value must be greater than 0.
    ///  
    /// </summary>
    internal static string WidthHashtableKeyDescription {
        get {
            return ResourceManager.GetString("WidthHashtableKeyDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [int]
    ///The depth key specifies the depth of expansion per property.
    ///  
    /// </summary>
    internal static string DepthHashtableKeyDescription {
        get {
            return ResourceManager.GetString("DepthHashtableKeyDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [bool]
    ///Specifies the order of sorting for one or more properties.
    ///  
    /// </summary>
    internal static string AscendingHashtableKeyDescription {
        get {
            return ResourceManager.GetString("AscendingHashtableKeyDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [bool]
    ///Specifies the order of sorting for one or more properties.
    ///  
    /// </summary>
    internal static string DescendingHashtableKeyDescription {
        get {
            return ResourceManager.GetString("DescendingHashtableKeyDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [String[]]
    ///Specifies the log names to get events from.
    ///Supports wildcards.
    ///  
    /// </summary>
    internal static string LogNameHashtableKeyDescription {
        get {
            return ResourceManager.GetString("LogNameHashtableKeyDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [String[]]
    ///Specifies the event log providers to get events from.
    ///Supports wildcards.
    ///  
    /// </summary>
    internal static string ProviderNameHashtableKeyDescription {
        get {
            return ResourceManager.GetString("ProviderNameHashtableKeyDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [String[]]
    ///Specifies file paths to log files to get events from.
    ///Valid file formats are: .etl, .evt, and .evtx
    ///  
    /// </summary>
    internal static string PathHashtableKeyDescription {
        get {
            return ResourceManager.GetString("PathHashtableKeyDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [Long[]]
    ///Selects events with the specified keyword bitmasks.
    ///The following are standard keywords:
    ///4503599627370496: AuditFailure
    ///9007199254740992: AuditSuccess
    ///4503599627370496: CorrelationHint
    ///18014398509481984: CorrelationHint2
    ///36028797018963968: EventLogClassic
    ///281474976710656: ResponseTime
    ///2251799813685248: Sqm
    ///562949953421312: WdiContext
    ///1125899906842624: WdiDiagnostic
    ///  
    /// </summary>
    internal static string KeywordsHashtableKeyDescription {
        get {
            return ResourceManager.GetString("KeywordsHashtableKeyDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [int[]]
    ///Selects events with the specified event IDs.
    ///  
    /// </summary>
    internal static string IDHashtableKeyDescription {
        get {
            return ResourceManager.GetString("IDHashtableKeyDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [int[]]
    ///Selects events with the specified log levels.
    ///The following log levels are valid:
    ///1: Critical
    ///2: Error
    ///3: Warning
    ///4: Informational
    ///5: Verbose
    ///  
    /// </summary>
    internal static string LevelHashtableKeyDescription {
        get {
            return ResourceManager.GetString("LevelHashtableKeyDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [datetime]
    ///Selects events created after the specified date and time.
    ///  
    /// </summary>
    internal static string StartTimeHashtableKeyDescription {
        get {
            return ResourceManager.GetString("StartTimeHashtableKeyDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [datetime]
    ///Selects events created before the specified date and time.
    ///  
    /// </summary>
    internal static string EndTimeHashtableKeyDescription {
        get {
            return ResourceManager.GetString("EndTimeHashtableKeyDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [string]
    ///Selects events generated by the specified user.
    ///This can either be a string representation of a SID or a domain and username in the format DOMAIN\USERNAME or USERNAME@DOMAIN
    ///  
    /// </summary>
    internal static string UserIDHashtableKeyDescription {
        get {
            return ResourceManager.GetString("UserIDHashtableKeyDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [string[]]
    ///Selects events with any of the specified values in the EventData section.
    ///  
    /// </summary>
    internal static string DataHashtableKeyDescription {
        get {
            return ResourceManager.GetString("DataHashtableKeyDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [hashtable]
    ///Excludes events that match the values specified in the hashtable.
    ///  
    /// </summary>
    internal static string SuppressHashFilterHashtableKeyDescription {
        get {
            return ResourceManager.GetString("SuppressHashFilterHashtableKeyDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [string] or [hashtable]
    ///Specifies an array of PowerShell modules that the script requires.
    ///Each element can either be a string with the module name as value or a hashtable with the following keys:
    ///Name: Name of the module
    ///GUID: GUID of the module
    ///One of the following:
    ///ModuleVersion: Specifies a minimum acceptable version of the module.
    ///RequiredVersion: Specifies an exact, required version of the module.
    ///MaximumVersion: Specifies the maximum acceptable version of the module.
    ///  
    /// </summary>
    internal static string RequiresModulesParameterDescription {
        get {
            return ResourceManager.GetString("RequiresModulesParameterDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [string]
    ///Specifies a PowerShell edition that the script requires.
    ///Valid values are "Core" and "Desktop"
    ///  
    /// </summary>
    internal static string RequiresPSEditionParameterDescription {
        get {
            return ResourceManager.GetString("RequiresPSEditionParameterDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [switch]
    ///Specifies that PowerShell must be running as administrator on Windows.
    ///This must be the last parameter on the #requires statement line.
    ///  
    /// </summary>
    internal static string RequiresRunAsAdministratorParameterDescription {
        get {
            return ResourceManager.GetString("RequiresRunAsAdministratorParameterDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [version]
    ///Specifies the minimum version of PowerShell that the script requires.
    ///  
    /// </summary>
    internal static string RequiresVersionParameterDescription {
        get {
            return ResourceManager.GetString("RequiresVersionParameterDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Specifies that the script requires PowerShell Core to run.
    ///  
    /// </summary>
    internal static string RequiresPsEditionCoreDescription {
        get {
            return ResourceManager.GetString("RequiresPsEditionCoreDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Specifies that the script requires Windows PowerShell to run.
    ///  
    /// </summary>
    internal static string RequiresPsEditionDesktopDescription {
        get {
            return ResourceManager.GetString("RequiresPsEditionDesktopDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [string]
    ///Required. Specifies the module name.
    ///  
    /// </summary>
    internal static string RequiresModuleSpecModuleNameDescription {
        get {
            return ResourceManager.GetString("RequiresModuleSpecModuleNameDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [string]
    ///Optional. Specifies the GUID of the module.
    ///  
    /// </summary>
    internal static string RequiresModuleSpecGUIDDescription {
        get {
            return ResourceManager.GetString("RequiresModuleSpecGUIDDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [string]
    ///Specifies a minimum acceptable version of the module.
    ///  
    /// </summary>
    internal static string RequiresModuleSpecModuleVersionDescription {
        get {
            return ResourceManager.GetString("RequiresModuleSpecModuleVersionDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [string]
    ///Specifies an exact, required version of the module.
    ///  
    /// </summary>
    internal static string RequiresModuleSpecRequiredVersionDescription {
        get {
            return ResourceManager.GetString("RequiresModuleSpecRequiredVersionDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [string]
    ///Specifies the maximum acceptable version of the module.
    ///  
    /// </summary>
    internal static string RequiresModuleSpecMaximumVersionDescription {
        get {
            return ResourceManager.GetString("RequiresModuleSpecMaximumVersionDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A brief description of the function or script.
    ///This keyword can be used only once in each topic.
    ///  
    /// </summary>
    internal static string CommentHelpSYNOPSISKeywordDescription {
        get {
            return ResourceManager.GetString("CommentHelpSYNOPSISKeywordDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A detailed description of the function or script.
    ///This keyword can be used only once in each topic.
    ///  
    /// </summary>
    internal static string CommentHelpDESCRIPTIONKeywordDescription {
        get {
            return ResourceManager.GetString("CommentHelpDESCRIPTIONKeywordDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    .PARAMETER  <Parameter-Name>
    ///The description of a parameter.
    ///Add a .PARAMETER keyword for each parameter in the function or script syntax.
    ///  
    /// </summary>
    internal static string CommentHelpPARAMETERKeywordDescription {
        get {
            return ResourceManager.GetString("CommentHelpPARAMETERKeywordDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A sample command that uses the function or script, optionally followed by sample output and a description.
    ///Repeat this keyword for each example.
    ///  
    /// </summary>
    internal static string CommentHelpEXAMPLEKeywordDescription {
        get {
            return ResourceManager.GetString("CommentHelpEXAMPLEKeywordDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The .NET types of objects that can be piped to the function or script.
    ///You can also include a description of the input objects.
    ///  
    /// </summary>
    internal static string CommentHelpINPUTSKeywordDescription {
        get {
            return ResourceManager.GetString("CommentHelpINPUTSKeywordDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The .NET type of the objects that the cmdlet returns.
    ///You can also include a description of the returned objects.
    ///  
    /// </summary>
    internal static string CommentHelpOUTPUTSKeywordDescription {
        get {
            return ResourceManager.GetString("CommentHelpOUTPUTSKeywordDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Additional information about the function or script.
    ///  
    /// </summary>
    internal static string CommentHelpNOTESKeywordDescription {
        get {
            return ResourceManager.GetString("CommentHelpNOTESKeywordDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The name of a related topic.
    ///Repeat the .LINK keyword for each related topic.
    ///The .Link keyword content can also include a URI to an online version of the same help topic.
    ///  
    /// </summary>
    internal static string CommentHelpLINKKeywordDescription {
        get {
            return ResourceManager.GetString("CommentHelpLINKKeywordDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The name of the technology or feature that the function or script uses, or to which it is related.
    ///  
    /// </summary>
    internal static string CommentHelpCOMPONENTKeywordDescription {
        get {
            return ResourceManager.GetString("CommentHelpCOMPONENTKeywordDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The name of the user role for the help topic.
    ///  
    /// </summary>
    internal static string CommentHelpROLEKeywordDescription {
        get {
            return ResourceManager.GetString("CommentHelpROLEKeywordDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The keywords that describe the intended use of the function.
    ///  
    /// </summary>
    internal static string CommentHelpFUNCTIONALITYKeywordDescription {
        get {
            return ResourceManager.GetString("CommentHelpFUNCTIONALITYKeywordDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    .FORWARDHELPTARGETNAME <Command-Name>
    ///Redirects to the help topic for the specified command.
    ///  
    /// </summary>
    internal static string CommentHelpFORWARDHELPTARGETNAMEKeywordDescription {
        get {
            return ResourceManager.GetString("CommentHelpFORWARDHELPTARGETNAMEKeywordDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    .FORWARDHELPCATEGORY <Category>
    ///Specifies the help category of the item in .ForwardHelpTargetName
    ///  
    /// </summary>
    internal static string CommentHelpFORWARDHELPCATEGORYKeywordDescription {
        get {
            return ResourceManager.GetString("CommentHelpFORWARDHELPCATEGORYKeywordDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    .REMOTEHELPRUNSPACE <PSSession-variable>
    ///Specifies a session that contains the help topic.
    ///Enter a variable that contains a PSSession object.
    ///  
    /// </summary>
    internal static string CommentHelpREMOTEHELPRUNSPACEKeywordDescription {
        get {
            return ResourceManager.GetString("CommentHelpREMOTEHELPRUNSPACEKeywordDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    .EXTERNALHELP <XML Help File>
    ///The .ExternalHelp keyword is required when a function or script is documented in XML files.
    ///  
    /// </summary>
    internal static string CommentHelpEXTERNALHELPKeywordDescription {
        get {
            return ResourceManager.GetString("CommentHelpEXTERNALHELPKeywordDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Specifies the path to a .NET assembly to load.
    ///
    ///using assembly <.NET-assembly-path>
    ///  
    /// </summary>
    internal static string AssemblyKeywordDescription {
        get {
            return ResourceManager.GetString("AssemblyKeywordDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Specifies a PowerShell module to load classes from.
    ///
    ///using module <ModuleName or Path>
    ///
    ///using module <ModuleSpecification hashtable>
    ///  
    /// </summary>
    internal static string ModuleKeywordDescription {
        get {
            return ResourceManager.GetString("ModuleKeywordDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Specifies a .NET namespace to resolve types from or a namespace alias.
    ///
    ///using namespace <.NET-namespace>
    ///
    ///using namespace <AliasName> = <.NET-namespace>
    ///  
    /// </summary>
    internal static string NamespaceKeywordDescription {
        get {
            return ResourceManager.GetString("NamespaceKeywordDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Specifies an alias for a .NET Type.
    ///
    ///using type <AliasName> = <.NET-type>
    ///  
    /// </summary>
    internal static string TypeKeywordDescription {
        get {
            return ResourceManager.GetString("TypeKeywordDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A normal string.
    ///  
    /// </summary>
    internal static string RegistryStringToolTip {
        get {
            return ResourceManager.GetString("RegistryStringToolTip", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A string that contains unexpanded references to environment variables that are expanded when the value is retrieved.
    ///  
    /// </summary>
    internal static string RegistryExpandStringToolTip {
        get {
            return ResourceManager.GetString("RegistryExpandStringToolTip", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Binary data in any form.
    ///  
    /// </summary>
    internal static string RegistryBinaryToolTip {
        get {
            return ResourceManager.GetString("RegistryBinaryToolTip", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A 32-bit binary number.
    ///  
    /// </summary>
    internal static string RegistryDWordToolTip {
        get {
            return ResourceManager.GetString("RegistryDWordToolTip", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    An array of strings.
    ///  
    /// </summary>
    internal static string RegistryMultiStringToolTip {
        get {
            return ResourceManager.GetString("RegistryMultiStringToolTip", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A 64-bit binary number.
    ///  
    /// </summary>
    internal static string RegistryQWordToolTip {
        get {
            return ResourceManager.GetString("RegistryQWordToolTip", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    An unsupported registry data type.
    ///  
    /// </summary>
    internal static string RegistryUnknownToolTip {
        get {
            return ResourceManager.GetString("RegistryUnknownToolTip", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ',' - Comma
    ///  
    /// </summary>
    internal static string SeparatorCommaToolTip {
        get {
            return ResourceManager.GetString("SeparatorCommaToolTip", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ', ' - Comma-Space
    ///  
    /// </summary>
    internal static string SeparatorCommaSpaceToolTip {
        get {
            return ResourceManager.GetString("SeparatorCommaSpaceToolTip", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ';' - Semi-Colon
    ///  
    /// </summary>
    internal static string SeparatorSemiColonToolTip {
        get {
            return ResourceManager.GetString("SeparatorSemiColonToolTip", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    '; ' - Semi-Colon-Space
    ///  
    /// </summary>
    internal static string SeparatorSemiColonSpaceToolTip {
        get {
            return ResourceManager.GetString("SeparatorSemiColonSpaceToolTip", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0} - Newline
    ///  
    /// </summary>
    internal static string SeparatorNewlineToolTip {
        get {
            return ResourceManager.GetString("SeparatorNewlineToolTip", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    '-' - Dash
    ///  
    /// </summary>
    internal static string SeparatorDashToolTip {
        get {
            return ResourceManager.GetString("SeparatorDashToolTip", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ' ' - Space
    ///  
    /// </summary>
    internal static string SeparatorSpaceToolTip {
        get {
            return ResourceManager.GetString("SeparatorSpaceToolTip", resourceCulture);
        }
    }

}

