// Copyright (c) Microsoft Corporation.
// Licensed under the MIT License.

using System.Windows.Controls;

namespace Microsoft.PowerShell.Commands.ShowCommandInternal
{
    /// <summary>
    /// Interaction logic for NotImportedCmdletControl.xaml.
    /// </summary>
    public partial class NotImportedCmdletControl : UserControl
    {
        #region Construction and Destructor

        /// <summary>
        /// Initializes a new instance of the NotImportedCmdletControl class.
        /// </summary>
        public NotImportedCmdletControl()
        {
            InitializeComponent();
        }
        #endregion
    }
}
