
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class SecuritySupportStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal SecuritySupportStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.SecuritySupportStrings", typeof(SecuritySupportStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot load certificate. '{0}' must resolve to a file system path.
    ///  
    /// </summary>
    internal static string CertificatePathMustBeFileSystemPath {
        get {
            return ResourceManager.GetString("CertificatePathMustBeFileSystemPath", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Certificate '{0}' cannot be used for encryption. Encryption certificates must contain the Data Encipherment or Key Encipherment key usage, and include the Document Encryption Enhanced Key Usage ({1}).
    ///  
    /// </summary>
    internal static string CertificateCannotBeUsedForEncryption {
        get {
            return ResourceManager.GetString("CertificateCannotBeUsedForEncryption", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot load certificate. The identifier '{0}' matches multiple certificates. To encrypt to multiple recipients, provide multiple specific values to the '{1}' parameter, rather than a wildcard that matches multiple certificates.
    ///  
    /// </summary>
    internal static string IdentifierMustReferenceSingleCertificate {
        get {
            return ResourceManager.GetString("IdentifierMustReferenceSingleCertificate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot load encryption certificate. The certificate setting '{0}' does not represent a valid base-64 encoded certificate, nor does it represent a valid certificate by file, directory, thumbprint, or subject name.
    ///  
    /// </summary>
    internal static string NoCertificateFound {
        get {
            return ResourceManager.GetString("NoCertificateFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    WARNING: The certificate '{0}' contains a private key. Protected Event Logging certificates used for encryption should only contain the public key.
    ///  
    /// </summary>
    internal static string CertificateContainsPrivateKey {
        get {
            return ResourceManager.GetString("CertificateContainsPrivateKey", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ERROR: Could not protect event log message '{0}': {1}
    ///  
    /// </summary>
    internal static string CouldNotEncryptContent {
        get {
            return ResourceManager.GetString("CouldNotEncryptContent", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ERROR: Could not find or use certificate: {0}
    ///  
    /// </summary>
    internal static string CouldNotUseCertificate {
        get {
            return ResourceManager.GetString("CouldNotUseCertificate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Session key not available to encrypt secure string.
    ///  
    /// </summary>
    internal static string CannotEncryptSecureString {
        get {
            return ResourceManager.GetString("CannotEncryptSecureString", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Invalid buffer offset.
    ///  
    /// </summary>
    internal static string InvalidOffset {
        get {
            return ResourceManager.GetString("InvalidOffset", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Invalid public key data.
    ///  
    /// </summary>
    internal static string InvalidPublicKey {
        get {
            return ResourceManager.GetString("InvalidPublicKey", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot import public key.
    ///  
    /// </summary>
    internal static string CannotImportPublicKey {
        get {
            return ResourceManager.GetString("CannotImportPublicKey", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Invalid session key data.
    ///  
    /// </summary>
    internal static string InvalidSessionKey {
        get {
            return ResourceManager.GetString("InvalidSessionKey", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Script file, '{0}', is blocked from running by system policy.
    ///  
    /// </summary>
    internal static string ScriptFileBlockedBySystemPolicy {
        get {
            return ResourceManager.GetString("ScriptFileBlockedBySystemPolicy", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    An unknown script file policy enforcement value was returned: {0}.
    ///  
    /// </summary>
    internal static string UnknownSystemScriptFileEnforcement {
        get {
            return ResourceManager.GetString("UnknownSystemScriptFileEnforcement", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Script File Read
    ///  
    /// </summary>
    internal static string ExternalScriptWDACLogTitle {
        get {
            return ResourceManager.GetString("ExternalScriptWDACLogTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Script file '{0}' is not trusted by policy and will run in ConstrainedLanguage mode.
    ///  
    /// </summary>
    internal static string ExternalScriptWDACLogMessage {
        get {
            return ResourceManager.GetString("ExternalScriptWDACLogMessage", resourceCulture);
        }
    }

}

