
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class HelpDisplayStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal HelpDisplayStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.HelpDisplayStrings", typeof(HelpDisplayStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    NAME
    ///  
    /// </summary>
    internal static string Name {
        get {
            return ResourceManager.GetString("Name", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    SYNOPSIS
    ///  
    /// </summary>
    internal static string Synopsis {
        get {
            return ResourceManager.GetString("Synopsis", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    DESCRIPTION
    ///  
    /// </summary>
    internal static string DetailedDescription {
        get {
            return ResourceManager.GetString("DetailedDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    SYNTAX
    ///  
    /// </summary>
    internal static string Syntax {
        get {
            return ResourceManager.GetString("Syntax", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    PARAMETERS
    ///  
    /// </summary>
    internal static string Parameters {
        get {
            return ResourceManager.GetString("Parameters", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    INPUTS
    ///  
    /// </summary>
    internal static string InputType {
        get {
            return ResourceManager.GetString("InputType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    OUTPUTS
    ///  
    /// </summary>
    internal static string ReturnType {
        get {
            return ResourceManager.GetString("ReturnType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    TERMINATING ERRORS
    ///  
    /// </summary>
    internal static string TerminatingErrors {
        get {
            return ResourceManager.GetString("TerminatingErrors", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    NON-TERMINATING ERRORS
    ///  
    /// </summary>
    internal static string NonHyphenTerminatingErrors {
        get {
            return ResourceManager.GetString("NonHyphenTerminatingErrors", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    NOTES
    ///  
    /// </summary>
    internal static string Notes {
        get {
            return ResourceManager.GetString("Notes", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    EXAMPLES
    ///  
    /// </summary>
    internal static string Examples {
        get {
            return ResourceManager.GetString("Examples", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Example
    ///  
    /// </summary>
    internal static string Example {
        get {
            return ResourceManager.GetString("Example", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    EXAMPLE
    ///  
    /// </summary>
    internal static string ExampleUpperCase {
        get {
            return ResourceManager.GetString("ExampleUpperCase", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    OUTPUT
    ///  
    /// </summary>
    internal static string Output {
        get {
            return ResourceManager.GetString("Output", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    RELATED LINKS
    ///  
    /// </summary>
    internal static string RelatedLinks {
        get {
            return ResourceManager.GetString("RelatedLinks", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    SHORT DESCRIPTION
    ///  
    /// </summary>
    internal static string ShortDescription {
        get {
            return ResourceManager.GetString("ShortDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Title:
    ///  
    /// </summary>
    internal static string TitleColon {
        get {
            return ResourceManager.GetString("TitleColon", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Question:
    ///  
    /// </summary>
    internal static string QuestionColon {
        get {
            return ResourceManager.GetString("QuestionColon", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Answer
    ///  
    /// </summary>
    internal static string Answer {
        get {
            return ResourceManager.GetString("Answer", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Term:
    ///  
    /// </summary>
    internal static string TermColon {
        get {
            return ResourceManager.GetString("TermColon", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Definition:
    ///  
    /// </summary>
    internal static string DefinitionColon {
        get {
            return ResourceManager.GetString("DefinitionColon", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Content:
    ///  
    /// </summary>
    internal static string ContentColon {
        get {
            return ResourceManager.GetString("ContentColon", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    PROVIDER NAME
    ///  
    /// </summary>
    internal static string ProviderName {
        get {
            return ResourceManager.GetString("ProviderName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///        This cmdlet supports the common parameters: Verbose, Debug,
    ///    ErrorAction, ErrorVariable, WarningAction, WarningVariable,
    ///    OutBuffer, PipelineVariable, and OutVariable. For more information, see
    ///    about_CommonParameters (https://go.microsoft.com/fwlink/?LinkID=113216). 
    ///  
    /// </summary>
    internal static string BaseCmdletInformation {
        get {
            return ResourceManager.GetString("BaseCmdletInformation", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Required?                    
    ///  
    /// </summary>
    internal static string ParameterRequired {
        get {
            return ResourceManager.GetString("ParameterRequired", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Position?                    
    ///  
    /// </summary>
    internal static string ParameterPosition {
        get {
            return ResourceManager.GetString("ParameterPosition", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Type: 
    ///  
    /// </summary>
    internal static string TypeColon {
        get {
            return ResourceManager.GetString("TypeColon", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Target Object Type:   
    ///  
    /// </summary>
    internal static string TargetObjectTypeColon {
        get {
            return ResourceManager.GetString("TargetObjectTypeColon", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Default value                
    ///  
    /// </summary>
    internal static string ParameterDefaultValue {
        get {
            return ResourceManager.GetString("ParameterDefaultValue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Accept pipeline input?       
    ///  
    /// </summary>
    internal static string AcceptsPipelineInput {
        get {
            return ResourceManager.GetString("AcceptsPipelineInput", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Accept wildcard characters?  
    ///  
    /// </summary>
    internal static string AcceptsWildCardCharacters {
        get {
            return ResourceManager.GetString("AcceptsWildCardCharacters", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    (Category: 
    ///  
    /// </summary>
    internal static string Category {
        get {
            return ResourceManager.GetString("Category", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Suggested Action: 
    ///  
    /// </summary>
    internal static string SuggestedActionColon {
        get {
            return ResourceManager.GetString("SuggestedActionColon", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    For more information, type: 
    ///  
    /// </summary>
    internal static string VerboseHelpInfo {
        get {
            return ResourceManager.GetString("VerboseHelpInfo", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    For technical information, type: 
    ///  
    /// </summary>
    internal static string FullHelpInfo {
        get {
            return ResourceManager.GetString("FullHelpInfo", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    To see the examples, type: 
    ///  
    /// </summary>
    internal static string ExampleHelpInfo {
        get {
            return ResourceManager.GetString("ExampleHelpInfo", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    For online help, type: 
    ///  
    /// </summary>
    internal static string RelatedLinksHelpInfo {
        get {
            return ResourceManager.GetString("RelatedLinksHelpInfo", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    <CommonParameters>
    ///  
    /// </summary>
    internal static string CommonParameters {
        get {
            return ResourceManager.GetString("CommonParameters", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    REMARKS
    ///  
    /// </summary>
    internal static string RemarksSection {
        get {
            return ResourceManager.GetString("RemarksSection", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    true
    ///  
    /// </summary>
    internal static string TrueShort {
        get {
            return ResourceManager.GetString("TrueShort", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Named
    ///  
    /// </summary>
    internal static string NamedParameter {
        get {
            return ResourceManager.GetString("NamedParameter", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    DRIVES
    ///  
    /// </summary>
    internal static string Drives {
        get {
            return ResourceManager.GetString("Drives", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    CAPABILITIES
    ///  
    /// </summary>
    internal static string Capabilities {
        get {
            return ResourceManager.GetString("Capabilities", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    TASKS
    ///  
    /// </summary>
    internal static string Tasks {
        get {
            return ResourceManager.GetString("Tasks", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    TASK: 
    ///  
    /// </summary>
    internal static string Task {
        get {
            return ResourceManager.GetString("Task", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    FILTERS
    ///  
    /// </summary>
    internal static string Filters {
        get {
            return ResourceManager.GetString("Filters", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    DYNAMIC PARAMETERS
    ///  
    /// </summary>
    internal static string DynamicParameters {
        get {
            return ResourceManager.GetString("DynamicParameters", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cmdlets Supported: 
    ///  
    /// </summary>
    internal static string CmdletsSupported {
        get {
            return ResourceManager.GetString("CmdletsSupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ALIASES
    ///  
    /// </summary>
    internal static string AliasesSection {
        get {
            return ResourceManager.GetString("AliasesSection", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Get-Help cannot find the Help files for this cmdlet on this computer. It is displaying only partial help.
    ///    -- To download and install Help files for the module that includes this cmdlet, use Update-Help.
    ///    -- To view the Help topic for this cmdlet online, type: "Get-Help {0} -Online" or
    ///       go to {1}.
    ///  
    /// </summary>
    internal static string GetLatestHelpContent {
        get {
            return ResourceManager.GetString("GetLatestHelpContent", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    None
    ///  
    /// </summary>
    internal static string None {
        get {
            return ResourceManager.GetString("None", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Aliases                      
    ///  
    /// </summary>
    internal static string ParameterAliases {
        get {
            return ResourceManager.GetString("ParameterAliases", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Dynamic?                     
    ///  
    /// </summary>
    internal static string ParameterIsDynamic {
        get {
            return ResourceManager.GetString("ParameterIsDynamic", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Parameter set name           
    ///  
    /// </summary>
    internal static string ParameterSetName {
        get {
            return ResourceManager.GetString("ParameterSetName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unable to retrieve the HelpInfo XML file for UI culture {0}. Make sure the HelpInfoUri property in the module manifest is valid or check your network connection and then try the command again.
    ///  
    /// </summary>
    internal static string UnableToRetrieveHelpInfoXml {
        get {
            return ResourceManager.GetString("UnableToRetrieveHelpInfoXml", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ByPropertyName
    ///  
    /// </summary>
    internal static string PipelineByPropertyName {
        get {
            return ResourceManager.GetString("PipelineByPropertyName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ByValue
    ///  
    /// </summary>
    internal static string PipelineByValue {
        get {
            return ResourceManager.GetString("PipelineByValue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    FromRemainingArguments
    ///  
    /// </summary>
    internal static string PipelineFromRemainingArguments {
        get {
            return ResourceManager.GetString("PipelineFromRemainingArguments", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified culture is not supported: {0}. Specify a culture from the following list: {{{1}}}.
    ///  
    /// </summary>
    internal static string HelpCultureNotSupported {
        get {
            return ResourceManager.GetString("HelpCultureNotSupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Postponing error and trying fallback cultures, will show as error if none of fallbacks are supported:
    ///{0}
    ///  
    /// </summary>
    internal static string HelpCultureNotSupportedFallback {
        get {
            return ResourceManager.GetString("HelpCultureNotSupportedFallback", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The ModuleBase directory cannot be found. Verify the directory and try again.
    ///  
    /// </summary>
    internal static string ModuleBaseMustExist {
        get {
            return ResourceManager.GetString("ModuleBaseMustExist", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The path {0} is not a valid directory. Make sure the directory exists and retry.
    ///  
    /// </summary>
    internal static string PathMustBeValidContainers {
        get {
            return ResourceManager.GetString("PathMustBeValidContainers", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A Help URI cannot contain more than 10 redirections. Specify a valid Help URI.
    ///  
    /// </summary>
    internal static string TooManyRedirections {
        get {
            return ResourceManager.GetString("TooManyRedirections", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Updating Help
    ///  
    /// </summary>
    internal static string UpdateProgressActivity {
        get {
            return ResourceManager.GetString("UpdateProgressActivity", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Connecting to Help Content...
    ///  
    /// </summary>
    internal static string UpdateProgressConnecting {
        get {
            return ResourceManager.GetString("UpdateProgressConnecting", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Downloading Help Content...
    ///  
    /// </summary>
    internal static string UpdateProgressDownloading {
        get {
            return ResourceManager.GetString("UpdateProgressDownloading", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Installing Help content...
    ///  
    /// </summary>
    internal static string UpdateProgressInstalling {
        get {
            return ResourceManager.GetString("UpdateProgressInstalling", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Locating Help Content...
    ///  
    /// </summary>
    internal static string UpdateProgressLocating {
        get {
            return ResourceManager.GetString("UpdateProgressLocating", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    (All)
    ///  
    /// </summary>
    internal static string AllParameterSetsName {
        get {
            return ResourceManager.GetString("AllParameterSetsName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    No PowerShell modules were found that match the following pattern: {0}. Verify the pattern and then try the command again.
    ///  
    /// </summary>
    internal static string CannotMatchModulePattern {
        get {
            return ResourceManager.GetString("CannotMatchModulePattern", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    No PowerShell modules were found that match the specified FullyQualifiedModule {0}. Verify the FullyQualifiedModule value and then try the command again.
    ///  
    /// </summary>
    internal static string ModuleNotFoundWithFullyQualifiedName {
        get {
            return ResourceManager.GetString("ModuleNotFoundWithFullyQualifiedName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Help content cannot be found. Make sure the server is available and the help content location is properly defined in the HelpInfo XML.
    ///  
    /// </summary>
    internal static string HelpContentNotFound {
        get {
            return ResourceManager.GetString("HelpContentNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Update-Help command failed because the specified module does not support updatable help. Use Get-Help -Online or look online for help for the commands in this module.
    ///  
    /// </summary>
    internal static string HelpInfoUriNotFound {
        get {
            return ResourceManager.GetString("HelpInfoUriNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The following parameter must not be null or empty: Module.
    ///  
    /// </summary>
    internal static string ModuleNameNullOrEmpty {
        get {
            return ResourceManager.GetString("ModuleNameNullOrEmpty", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The following parameter must not be null or empty: Path.
    ///  
    /// </summary>
    internal static string PathNullOrEmpty {
        get {
            return ResourceManager.GetString("PathNullOrEmpty", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Update-Help has completed successfully.
    ///  
    /// </summary>
    internal static string UpdateHelpCompleted {
        get {
            return ResourceManager.GetString("UpdateHelpCompleted", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error extracting Help content.
    ///  
    /// </summary>
    internal static string UnzipFailure {
        get {
            return ResourceManager.GetString("UnzipFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unable to connect to Help content. The server on which Help content is stored might not be available. Verify that the server is available, or wait until the server is back online, and then try the command again.
    ///  
    /// </summary>
    internal static string UnableToConnect {
        get {
            return ResourceManager.GetString("UnableToConnect", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Help content at the specified location is not valid. Specify a location that contains valid Help Content.
    ///  
    /// </summary>
    internal static string HelpContentXmlValidationFailure {
        get {
            return ResourceManager.GetString("HelpContentXmlValidationFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The HelpInfo XML is not valid. Specify valid HelpInfo XML.
    ///  
    /// </summary>
    internal static string HelpInfoXmlValidationFailure {
        get {
            return ResourceManager.GetString("HelpInfoXmlValidationFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Help content was successfully saved to the following location: {0}
    ///  
    /// </summary>
    internal static string SaveHelpCompleted {
        get {
            return ResourceManager.GetString("SaveHelpCompleted", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Help content XSD file cannot be found in {0}. Verify that the XSD file exists at the specified location, and then retry the command.
    ///  
    /// </summary>
    internal static string HelpContentXsdNotFound {
        get {
            return ResourceManager.GetString("HelpContentXsdNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Failed to update Help for the module(s) :
    ///'{0}'
    ///{1}
    ///  
    /// </summary>
    internal static string FailedToUpdateHelpForModule {
        get {
            return ResourceManager.GetString("FailedToUpdateHelpForModule", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Saving Help
    ///  
    /// </summary>
    internal static string SaveProgressActivity {
        get {
            return ResourceManager.GetString("SaveProgressActivity", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Help content contains files that are not valid. Only .txt and .xml files are supported.
    ///  
    /// </summary>
    internal static string HelpContentContainsInvalidFiles {
        get {
            return ResourceManager.GetString("HelpContentContainsInvalidFiles", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Failed to save Help for the module(s) '{0}' : {1}
    ///  
    /// </summary>
    internal static string FailedToSaveHelpForModule {
        get {
            return ResourceManager.GetString("FailedToSaveHelpForModule", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Failed to save Help for the module(s) '{0}' with UI culture(s) {{{1}}} : {2}.
    ///English-US help content is available and can be saved using: Save-Help -UICulture en-US.
    ///  
    /// </summary>
    internal static string FailedToSaveHelpForModuleWithCulture {
        get {
            return ResourceManager.GetString("FailedToSaveHelpForModuleWithCulture", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Failed to update Help for the module(s) '{0}' with UI culture(s) {{{1}}} : {2}.
    ///English-US help content is available and can be installed using: Update-Help -UICulture en-US.
    ///  
    /// </summary>
    internal static string FailedToUpdateHelpForModuleWithCulture {
        get {
            return ResourceManager.GetString("FailedToUpdateHelpForModuleWithCulture", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Your current culture is ({0}), which is not associated with any language, consider changing your system culture or install the English-US help content using: Update-Help -UICulture en-US.
    ///  
    /// </summary>
    internal static string FailedToUpdateHelpWithLocaleNoUICulture {
        get {
            return ResourceManager.GetString("FailedToUpdateHelpWithLocaleNoUICulture", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    false
    ///  
    /// </summary>
    internal static string FalseShort {
        get {
            return ResourceManager.GetString("FalseShort", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The -Recurse parameter is only available if a source path is specified.
    ///  
    /// </summary>
    internal static string CannotSpecifyRecurseWithoutPath {
        get {
            return ResourceManager.GetString("CannotSpecifyRecurseWithoutPath", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The path {0} does not contain a FileSystem provider. Verify that the specified path contains the FileSystem provider, and then retry the command.
    ///  
    /// </summary>
    internal static string ProviderIsNotFileSystem {
        get {
            return ResourceManager.GetString("ProviderIsNotFileSystem", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Searching Help for {0} ...
    ///  
    /// </summary>
    internal static string SearchingForHelpContent {
        get {
            return ResourceManager.GetString("SearchingForHelpContent", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    No UI culture was found that matches the following pattern: {0}. Verify the pattern and then try the command again.
    ///  
    /// </summary>
    internal static string CannotMatchUICulturePattern {
        get {
            return ResourceManager.GetString("CannotMatchUICulturePattern", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Help was not saved for the module {0}, because the Save-Help command was run on this computer within the last 24 hours.
    ///To save help again, add the Force parameter to your command.
    ///  
    /// </summary>
    internal static string UseForceToSaveHelp {
        get {
            return ResourceManager.GetString("UseForceToSaveHelp", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Help was not updated for the module {0}, because the Update-Help command was run on this computer within the last 24 hours.
    ///To update help again, add the Force parameter to your command.
    ///  
    /// </summary>
    internal static string UseForceToUpdateHelp {
        get {
            return ResourceManager.GetString("UseForceToUpdateHelp", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The most current Help files are already installed.
    ///  
    /// </summary>
    internal static string NewestContentAlreadyInstalled {
        get {
            return ResourceManager.GetString("NewestContentAlreadyInstalled", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0}: {1}. Culture {2} Version {3}
    ///  
    /// </summary>
    internal static string SuccessfullyUpdatedHelpContent {
        get {
            return ResourceManager.GetString("SuccessfullyUpdatedHelpContent", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Updated {0}
    ///  
    /// </summary>
    internal static string UpdatedHelpContent {
        get {
            return ResourceManager.GetString("UpdatedHelpContent", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The value of the HelpInfoUri key in the module manifest must resolve to a container or root URL on a website where the help files are stored. The HelpInfoUri '{0}' does not resolve to a container.
    ///  
    /// </summary>
    internal static string InvalidHelpInfoUri {
        get {
            return ResourceManager.GetString("InvalidHelpInfoUri", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Help content must be in the namespace {0}.
    ///  
    /// </summary>
    internal static string HelpContentMustBeInTargetNamespace {
        get {
            return ResourceManager.GetString("HelpContentMustBeInTargetNamespace", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Get-Help cannot find the Help files for this cmdlet on this computer. It is displaying only partial help.
    ///    -- To download and install Help files for the module that includes this cmdlet, use Update-Help.
    ///  
    /// </summary>
    internal static string GetLatestHelpContentWithoutHelpUri {
        get {
            return ResourceManager.GetString("GetLatestHelpContentWithoutHelpUri", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The most current Help files are already downloaded.
    ///  
    /// </summary>
    internal static string NewestContentAlreadyDownloaded {
        get {
            return ResourceManager.GetString("NewestContentAlreadyDownloaded", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Saved {0}
    ///  
    /// </summary>
    internal static string SavedHelpContent {
        get {
            return ResourceManager.GetString("SavedHelpContent", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The HelpInfoURI {0} does not start with HTTP.
    ///  
    /// </summary>
    internal static string InvalidHelpInfoUriFormat {
        get {
            return ResourceManager.GetString("InvalidHelpInfoUriFormat", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The root level element of the help content must be "helpItems".
    ///  
    /// </summary>
    internal static string RootElementMustBeHelpItems {
        get {
            return ResourceManager.GetString("RootElementMustBeHelpItems", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Saving Help for module {0}
    ///  
    /// </summary>
    internal static string SaveProgressActivityForModule {
        get {
            return ResourceManager.GetString("SaveProgressActivityForModule", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Updating Help for module {0}
    ///  
    /// </summary>
    internal static string UpdateProgressActivityForModule {
        get {
            return ResourceManager.GetString("UpdateProgressActivityForModule", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Resolving URI: "{0}"
    ///  
    /// </summary>
    internal static string UpdateHelpResolveUriVerbose {
        get {
            return ResourceManager.GetString("UpdateHelpResolveUriVerbose", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Help URI: {0}
    ///  
    /// </summary>
    internal static string OnlineHelpUri {
        get {
            return ResourceManager.GetString("OnlineHelpUri", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0}, Current Version: {1}, Available Version: {2}, UICulture: {3}
    ///  
    /// </summary>
    internal static string UpdateHelpShouldProcessActionMessage {
        get {
            return ResourceManager.GetString("UpdateHelpShouldProcessActionMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    PROPERTIES
    ///  
    /// </summary>
    internal static string Properties {
        get {
            return ResourceManager.GetString("Properties", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    METHODS
    ///  
    /// </summary>
    internal static string Methods {
        get {
            return ResourceManager.GetString("Methods", resourceCulture);
        }
    }

}

