
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class UtilsStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal UtilsStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.PowerShell.Security.resources.UtilsStrings", typeof(UtilsStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot digitally sign file because file {0} is smaller than 4 bytes in size. Files must be at least 4 bytes in order to be digitally signed.
    ///  
    /// </summary>
    internal static string FileSmallerThan4Bytes {
        get {
            return ResourceManager.GetString("FileSmallerThan4Bytes", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot perform the operation because it is not supported on the object found in path {0}.
    ///  
    /// </summary>
    internal static string OperationNotSupportedOnPath {
        get {
            return ResourceManager.GetString("OperationNotSupportedOnPath", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot get the ACL because the necessary method, GetSecurityDescriptor, does not exist.
    ///  
    /// </summary>
    internal static string GetMethodNotFound {
        get {
            return ResourceManager.GetString("GetMethodNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot set the ACL because the method that it needs to invoke, SetSecurityDescriptor, does not exist.
    ///  
    /// </summary>
    internal static string SetMethodNotFound {
        get {
            return ResourceManager.GetString("SetMethodNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Could not perform operation because an exception was thrown during method invoke.
    ///  
    /// </summary>
    internal static string MethodInvokeFail {
        get {
            return ResourceManager.GetString("MethodInvokeFail", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Could not create a SACL with the specified central access policy.
    ///  
    /// </summary>
    internal static string GetSaclWithCapIdFail {
        get {
            return ResourceManager.GetString("GetSaclWithCapIdFail", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Could not create an empty SACL.
    ///  
    /// </summary>
    internal static string GetEmptySaclFail {
        get {
            return ResourceManager.GetString("GetEmptySaclFail", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Could not enable SeSecurityPrivilege.
    ///  
    /// </summary>
    internal static string GetTokenWithEnabledPrivilegeFail {
        get {
            return ResourceManager.GetString("GetTokenWithEnabledPrivilegeFail", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Could not set central access policy.
    ///  
    /// </summary>
    internal static string SetCentralAccessPolicyFail {
        get {
            return ResourceManager.GetString("SetCentralAccessPolicyFail", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ClearCentralAccessPolicy and CentralAccessPolicy parameters cannot be used at the same time.
    ///  
    /// </summary>
    internal static string InvalidCentralAccessPolicyParameters {
        get {
            return ResourceManager.GetString("InvalidCentralAccessPolicyParameters", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Central Access Policy identifier or name is not valid. If specifying an identifier, it must begin with S-1-17. If specifying a name, the policy must be applied on the target machine.
    ///  
    /// </summary>
    internal static string InvalidCentralAccessPolicyIdentifier {
        get {
            return ResourceManager.GetString("InvalidCentralAccessPolicyIdentifier", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    PowerShell credential request
    ///  
    /// </summary>
    internal static string PromptForCredential_DefaultCaption {
        get {
            return ResourceManager.GetString("PromptForCredential_DefaultCaption", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Enter your credentials.
    ///  
    /// </summary>
    internal static string PromptForCredential_DefaultMessage {
        get {
            return ResourceManager.GetString("PromptForCredential_DefaultMessage", resourceCulture);
        }
    }

}

