
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class FormatAndOut_out_gridview {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal FormatAndOut_out_gridview() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.PowerShell.Commands.Utility.resources.FormatAndOut_out_gridview", typeof(FormatAndOut_out_gridview).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The data format is not supported by Out-GridView.
    ///  
    /// </summary>
    internal static string DataNotQualifiedForGridView {
        get {
            return ResourceManager.GetString("DataNotQualifiedForGridView", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Microsoft .NET Framework 4.5 was installed while one or more PowerShell sessions were running. To use the {0} cmdlet, close all PowerShell windows, and then open a new PowerShell window.
    ///  
    /// </summary>
    internal static string RestartPowerShell {
        get {
            return ResourceManager.GetString("RestartPowerShell", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Type
    ///  
    /// </summary>
    internal static string TypeColumnName {
        get {
            return ResourceManager.GetString("TypeColumnName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Value
    ///  
    /// </summary>
    internal static string ValueColumnName {
        get {
            return ResourceManager.GetString("ValueColumnName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Index
    ///  
    /// </summary>
    internal static string IndexColumnName {
        get {
            return ResourceManager.GetString("IndexColumnName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A command named '{0}' was not found.
    ///  
    /// </summary>
    internal static string CommandNotFound {
        get {
            return ResourceManager.GetString("CommandNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    More than one command named '{0}' was found. Start '{1}' with no parameters, and then type '{0}' to filter the results.
    ///  
    /// </summary>
    internal static string MoreThanOneCommand {
        get {
            return ResourceManager.GetString("MoreThanOneCommand", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot write to console input buffer.
    ///  
    /// </summary>
    internal static string CannotWriteToConsoleInputBuffer {
        get {
            return ResourceManager.GetString("CannotWriteToConsoleInputBuffer", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0} should be smaller than {1}.
    ///    {0} is the property "Height"
    ///{1} is the maximum allowed value for the height property.
    ///  
    /// </summary>
    internal static string PropertyValidate {
        get {
            return ResourceManager.GetString("PropertyValidate", resourceCulture);
        }
    }

}

