// Copyright (c) Microsoft Corporation.
// Licensed under the MIT License.

using System.Windows.Controls;

namespace Microsoft.Management.UI.Internal
{
    /// <content>
    /// Partial class implementation for ManagementListTitle control.
    /// </content>
    [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.MSInternal", "CA903:InternalNamespaceShouldNotContainPublicTypes")]
    public partial class ManagementListTitle : Control
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="ManagementListTitle" /> class.
        /// </summary>
        public ManagementListTitle()
        {
            // This constructor intentionally left blank
        }
    }
}
