// Copyright (c) Microsoft Corporation.
// Licensed under the MIT License.
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a tool.
//     Runtime Version:4.0.30319.16598
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------

// namespace Microsoft.Management.UI.Internal {


    [global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]
    [global::System.CodeDom.Compiler.GeneratedCodeAttribute("Microsoft.VisualStudio.Editors.SettingsDesigner.SettingsSingleFileGenerator", "10.0.0.0")]
    internal sealed partial class HelpWindowSettings : global::System.Configuration.ApplicationSettingsBase {

        private static HelpWindowSettings defaultInstance = ((HelpWindowSettings)(global::System.Configuration.ApplicationSettingsBase.Synchronized(new HelpWindowSettings())));

        public static HelpWindowSettings Default {
            get {
                return defaultInstance;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("True")]
        public bool HelpRemarksDisplayed
        {
            get
            {
                return ((bool)(this["HelpRemarksDisplayed"]));
            }
            set
            {
                this["HelpRemarksDisplayed"] = value;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("True")]
        public bool HelpSyntaxDisplayed {
            get {
                return ((bool)(this["HelpSyntaxDisplayed"]));
            }
            set {
                this["HelpSyntaxDisplayed"] = value;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("True")]
        public bool HelpExamplesDisplayed {
            get {
                return ((bool)(this["HelpExamplesDisplayed"]));
            }
            set {
                this["HelpExamplesDisplayed"] = value;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("True")]
        public bool HelpSynopsysDisplayed {
            get {
                return ((bool)(this["HelpSynopsysDisplayed"]));
            }
            set {
                this["HelpSynopsysDisplayed"] = value;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("True")]
        public bool HelpDescriptionDisplayed {
            get {
                return ((bool)(this["HelpDescriptionDisplayed"]));
            }
            set {
                this["HelpDescriptionDisplayed"] = value;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("True")]
        public bool HelpParametersDisplayed {
            get {
                return ((bool)(this["HelpParametersDisplayed"]));
            }
            set {
                this["HelpParametersDisplayed"] = value;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("True")]
        public bool HelpInputsDisplayed {
            get {
                return ((bool)(this["HelpInputsDisplayed"]));
            }
            set {
                this["HelpInputsDisplayed"] = value;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("True")]
        public bool HelpOutputsDisplayed {
            get {
                return ((bool)(this["HelpOutputsDisplayed"]));
            }
            set {
                this["HelpOutputsDisplayed"] = value;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("True")]
        public bool HelpNotesDisplayed {
            get {
                return ((bool)(this["HelpNotesDisplayed"]));
            }
            set {
                this["HelpNotesDisplayed"] = value;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("True")]
        public bool HelpRelatedLinksDisplayed {
            get {
                return ((bool)(this["HelpRelatedLinksDisplayed"]));
            }
            set {
                this["HelpRelatedLinksDisplayed"] = value;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("False")]
        public bool HelpSearchMatchCase {
            get {
                return ((bool)(this["HelpSearchMatchCase"]));
            }
            set {
                this["HelpSearchMatchCase"] = value;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("False")]
        public bool HelpSearchWholeWord {
            get {
                return ((bool)(this["HelpSearchWholeWord"]));
            }
            set {
                this["HelpSearchWholeWord"] = value;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("400")]
        public double HelpWindowHeight {
            get {
                return ((double)(this["HelpWindowHeight"]));
            }
            set {
                this["HelpWindowHeight"] = value;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("600")]
        public double HelpWindowWidth {
            get {
                return ((double)(this["HelpWindowWidth"]));
            }
            set {
                this["HelpWindowWidth"] = value;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("-1")]
        public double HelpWindowTop {
            get {
                return ((double)(this["HelpWindowTop"]));
            }
            set {
                this["HelpWindowTop"] = value;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("-1")]
        public double HelpWindowLeft {
            get {
                return ((double)(this["HelpWindowLeft"]));
            }
            set {
                this["HelpWindowLeft"] = value;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("False")]
        public bool HelpWindowMaximized {
            get {
                return ((bool)(this["HelpWindowMaximized"]));
            }
            set {
                this["HelpWindowMaximized"] = value;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("100")]
        public double HelpZoom {
            get {
                return ((double)(this["HelpZoom"]));
            }
            set {
                this["HelpZoom"] = value;
            }
        }
    }
//}
