
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class Logging {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal Logging() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.Logging", typeof(Logging).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///        ExceptionClass=[ExceptionClass]
    ///    ErrorCategory=[ErrorCategory]
    ///    ErrorId=[ErrorId]
    ///    ErrorMessage=[ErrorMessage]
    ///
    ///    Severity=[Severity]
    ///
    ///    SequenceNumber=[SequenceNumber]
    ///
    ///    HostName=[HostName]
    ///    HostVersion=[HostVersion]
    ///    HostId=[HostId]
    ///    HostApplication=[HostApplication]
    ///    EngineVersion=[EngineVersion]
    ///    RunspaceId=[RunspaceId]
    ///    PipelineId=[PipelineId]
    ///    CommandName=[CommandName]
    ///    CommandType=[CommandType]
    ///    ScriptName=[ScriptName]
    ///    CommandPath=[CommandPath]
    ///    CommandLine=[CommandLine]
    ///
    ///AdditionalInfo:
    ///    Name=[AdditionalInfo_Name1];Value=[AdditionalInfo_Value1]
    ///    Name=[AdditionalInfo_Name2];Value=[AdditionalInfo_Value2]
    ///    Name=[AdditionalInfo_Name3];Value=[AdditionalInfo_Value3]
    ///  
    /// </summary>
    internal static string EngineHealthContext {
        get {
            return ResourceManager.GetString("EngineHealthContext", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///        ExceptionClass=[ExceptionClass]
    ///    ErrorCategory=[ErrorCategory]
    ///    ErrorId=[ErrorId]
    ///    ErrorMessage=[ErrorMessage]
    ///
    ///    Severity=[Severity]
    ///
    ///    SequenceNumber=[SequenceNumber]
    ///
    ///    HostName=[HostName]
    ///    HostVersion=[HostVersion]
    ///    HostId=[HostId]
    ///    HostApplication=[HostApplication]
    ///    EngineVersion=[EngineVersion]
    ///    RunspaceId=[RunspaceId]
    ///    PipelineId=[PipelineId]
    ///    CommandName=[CommandName]
    ///    CommandType=[CommandType]
    ///    ScriptName=[ScriptName]
    ///    CommandPath=[CommandPath]
    ///    CommandLine=[CommandLine]
    ///  
    /// </summary>
    internal static string CommandHealthContext {
        get {
            return ResourceManager.GetString("CommandHealthContext", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///        ProviderName=[ProviderName]
    ///    ExceptionClass=[ExceptionClass]
    ///    ErrorCategory=[ErrorCategory]
    ///    ErrorId=[ErrorId]
    ///    ErrorMessage=[ErrorMessage]
    ///
    ///    Severity=[Severity]
    ///
    ///    SequenceNumber=[SequenceNumber]
    ///
    ///    HostName=[HostName]
    ///    HostVersion=[HostVersion]
    ///    HostId=[HostId]
    ///    HostApplication=[HostApplication]
    ///    EngineVersion=[EngineVersion]
    ///    RunspaceId=[RunspaceId]
    ///    PipelineId=[PipelineId]
    ///    CommandName=[CommandName]
    ///    CommandType=[CommandType]
    ///    ScriptName=[ScriptName]
    ///    CommandPath=[CommandPath]
    ///    CommandLine=[CommandLine]
    ///  
    /// </summary>
    internal static string ProviderHealthContext {
        get {
            return ResourceManager.GetString("ProviderHealthContext", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///        NewEngineState=[NewEngineState]
    ///    PreviousEngineState=[PreviousEngineState]
    ///
    ///    SequenceNumber=[SequenceNumber]
    ///
    ///    HostName=[HostName]
    ///    HostVersion=[HostVersion]
    ///    HostId=[HostId]
    ///    HostApplication=[HostApplication]
    ///    EngineVersion=[EngineVersion]
    ///    RunspaceId=[RunspaceId]
    ///    PipelineId=[PipelineId]
    ///    CommandName=[CommandName]
    ///    CommandType=[CommandType]
    ///    ScriptName=[ScriptName]
    ///    CommandPath=[CommandPath]
    ///    CommandLine=[CommandLine]
    ///  
    /// </summary>
    internal static string EngineLifecycleContext {
        get {
            return ResourceManager.GetString("EngineLifecycleContext", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///        NewCommandState=[NewCommandState]
    ///
    ///    SequenceNumber=[SequenceNumber]
    ///
    ///    HostName=[HostName]
    ///    HostVersion=[HostVersion]
    ///    HostId=[HostId]
    ///    HostApplication=[HostApplication]
    ///    EngineVersion=[EngineVersion]
    ///    RunspaceId=[RunspaceId]
    ///    PipelineId=[PipelineId]
    ///    CommandName=[CommandName]
    ///    CommandType=[CommandType]
    ///    ScriptName=[ScriptName]
    ///    CommandPath=[CommandPath]
    ///    CommandLine=[CommandLine]
    ///  
    /// </summary>
    internal static string CommandLifecycleContext {
        get {
            return ResourceManager.GetString("CommandLifecycleContext", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///        ProviderName=[ProviderName]
    ///    NewProviderState=[NewProviderState]
    ///
    ///    SequenceNumber=[SequenceNumber]
    ///
    ///    HostName=[HostName]
    ///    HostVersion=[HostVersion]
    ///    HostId=[HostId]
    ///    HostApplication=[HostApplication]
    ///    EngineVersion=[EngineVersion]
    ///    RunspaceId=[RunspaceId]
    ///    PipelineId=[PipelineId]
    ///    CommandName=[CommandName]
    ///    CommandType=[CommandType]
    ///    ScriptName=[ScriptName]
    ///    CommandPath=[CommandPath]
    ///    CommandLine=[CommandLine]
    ///  
    /// </summary>
    internal static string ProviderLifecycleContext {
        get {
            return ResourceManager.GetString("ProviderLifecycleContext", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///        VariableName=[VariableName]
    ///    NewValue=[NewValue]
    ///    PreviousValue=[PreviousValue]
    ///
    ///    SequenceNumber=[SequenceNumber]
    ///
    ///    HostName=[HostName]
    ///    HostVersion=[HostVersion]
    ///    HostId=[HostId]
    ///    HostApplication=[HostApplication]
    ///    EngineVersion=[EngineVersion]
    ///    RunspaceId=[RunspaceId]
    ///    PipelineId=[PipelineId]
    ///    CommandName=[CommandName]
    ///    CommandType=[CommandType]
    ///    ScriptName=[ScriptName]
    ///    CommandPath=[CommandPath]
    ///    CommandLine=[CommandLine]
    ///  
    /// </summary>
    internal static string SettingsContext {
        get {
            return ResourceManager.GetString("SettingsContext", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///        DetailSequence=[DetailSequence]
    ///    DetailTotal=[DetailTotal]
    ///
    ///    SequenceNumber=[SequenceNumber]
    ///
    ///    UserId=[User]
    ///    HostName=[HostName]
    ///    HostVersion=[HostVersion]
    ///    HostId=[HostId]
    ///    HostApplication=[HostApplication]
    ///    EngineVersion=[EngineVersion]
    ///    RunspaceId=[RunspaceId]
    ///    PipelineId=[PipelineId]
    ///    ScriptName=[ScriptName]
    ///    CommandLine=[CommandLine]
    ///  
    /// </summary>
    internal static string PipelineExecutionDetailContext {
        get {
            return ResourceManager.GetString("PipelineExecutionDetailContext", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    UNKNOWN
    ///  
    /// </summary>
    internal static string UnknownUserName {
        get {
            return ResourceManager.GetString("UnknownUserName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The engine experimental feature '{0}' declared in the config file is not registered in the current PowerShell.
    ///  
    /// </summary>
    internal static string EngineExperimentalFeatureNotFound {
        get {
            return ResourceManager.GetString("EngineExperimentalFeatureNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The experimental feature '{0}' declared in the config file is invalid.
    ///The name of an experimental feature should follow the convention below: 
    ///    Engine Feature Name: 'PS[FeatureName]'
    ///    Module Feature Name: '[ModuleName].[FeatureName]'
    ///  
    /// </summary>
    internal static string InvalidExperimentalFeatureName {
        get {
            return ResourceManager.GetString("InvalidExperimentalFeatureName", resourceCulture);
        }
    }

}

