
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class CommandBaseStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal CommandBaseStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.CommandBaseStrings", typeof(CommandBaseStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Continue with this operation?
    ///  
    /// </summary>
    internal static string ShouldContinuePromptCaption {
        get {
            return ResourceManager.GetString("ShouldContinuePromptCaption", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    &Yes
    ///  
    /// </summary>
    internal static string ContinueOneLabel {
        get {
            return ResourceManager.GetString("ContinueOneLabel", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Continue with only the next step of the operation.
    ///  
    /// </summary>
    internal static string ContinueOneHelpMessage {
        get {
            return ResourceManager.GetString("ContinueOneHelpMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Yes to &All
    ///  
    /// </summary>
    internal static string ContinueAllLabel {
        get {
            return ResourceManager.GetString("ContinueAllLabel", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Continue with all the steps of the operation.
    ///  
    /// </summary>
    internal static string ContinueAllHelpMessage {
        get {
            return ResourceManager.GetString("ContinueAllHelpMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    &No
    ///  
    /// </summary>
    internal static string SkipOneLabel {
        get {
            return ResourceManager.GetString("SkipOneLabel", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Skip this operation and proceed with the next operation.
    ///  
    /// </summary>
    internal static string SkipOneHelpMessage {
        get {
            return ResourceManager.GetString("SkipOneHelpMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    No to A&ll
    ///  
    /// </summary>
    internal static string SkipAllLabel {
        get {
            return ResourceManager.GetString("SkipAllLabel", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Skip this operation and all subsequent operations.
    ///  
    /// </summary>
    internal static string SkipAllHelpMessage {
        get {
            return ResourceManager.GetString("SkipAllHelpMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Stop this command.
    ///  
    /// </summary>
    internal static string HaltHelpMessage {
        get {
            return ResourceManager.GetString("HaltHelpMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    &Halt Command
    ///  
    /// </summary>
    internal static string HaltLabel {
        get {
            return ResourceManager.GetString("HaltLabel", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    &Suspend
    ///  
    /// </summary>
    internal static string PauseLabel {
        get {
            return ResourceManager.GetString("PauseLabel", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Pause the current pipeline and return to the command prompt. Type "{0}" to resume the pipeline.
    ///  
    /// </summary>
    internal static string PauseHelpMessage {
        get {
            return ResourceManager.GetString("PauseHelpMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    
    ///    Program "{0}" ended with non-zero exit code: {1} ({2}).
    ///  
    /// </summary>
    internal static string ProgramExitedWithNonZeroCode {
        get {
            return ResourceManager.GetString("ProgramExitedWithNonZeroCode", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Performing the operation "{0}" on target "{1}".
    ///  
    /// </summary>
    internal static string ShouldProcessMessage {
        get {
            return ResourceManager.GetString("ShouldProcessMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    What if: {0}
    ///  
    /// </summary>
    internal static string ShouldProcessWhatIfMessage {
        get {
            return ResourceManager.GetString("ShouldProcessWhatIfMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Are you sure you want to perform this action?
    ///{0}
    ///  
    /// </summary>
    internal static string ShouldProcessWarningFallback {
        get {
            return ResourceManager.GetString("ShouldProcessWarningFallback", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Confirm
    ///  
    /// </summary>
    internal static string InquireCaptionDefault {
        get {
            return ResourceManager.GetString("InquireCaptionDefault", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The running command stopped because the preference variable "{0}" or common parameter is set to Stop: {1}
    ///  
    /// </summary>
    internal static string ErrorPreferenceStop {
        get {
            return ResourceManager.GetString("ErrorPreferenceStop", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The running command stopped because the preference variable "{0}" or common parameter is set to Stop.
    ///  
    /// </summary>
    internal static string PreferenceStop {
        get {
            return ResourceManager.GetString("PreferenceStop", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The running command stopped because the preference variable "{0}" or common parameter is set to the following value that is not valid: "{1}".
    ///  
    /// </summary>
    internal static string PreferenceInvalid {
        get {
            return ResourceManager.GetString("PreferenceInvalid", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The running command stopped because the user selected the Stop option.
    ///  
    /// </summary>
    internal static string InquireHalt {
        get {
            return ResourceManager.GetString("InquireHalt", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The running command stopped because the user interrupted the command.
    ///  
    /// </summary>
    internal static string InquireCtrlC {
        get {
            return ResourceManager.GetString("InquireCtrlC", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cmdlets derived from PSCmdlet cannot be invoked directly.  
    ///  
    /// </summary>
    internal static string CannotInvokePSCmdletsDirectly {
        get {
            return ResourceManager.GetString("CannotInvokePSCmdletsDirectly", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cmdlet '{0}' does not support parameter '{1}' in a remote session.
    ///  
    /// </summary>
    internal static string ParameterNotValidInRemoteRunspace {
        get {
            return ResourceManager.GetString("ParameterNotValidInRemoteRunspace", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Total count: {0}
    ///    {0} is a placeholder for an integer number.
    ///
    ///Reviewed by TArcher on 2010-07-20
    ///  
    /// </summary>
    internal static string PagingSupportAccurateTotalCountTemplate {
        get {
            return ResourceManager.GetString("PagingSupportAccurateTotalCountTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Estimated total count: {0}
    ///    {0} is a placeholder for an integer number
    ///
    ///Reviewed by TArcher on 2010-07-20
    ///
    ///  
    /// </summary>
    internal static string PagingSupportEstimatedTotalCountTemplate {
        get {
            return ResourceManager.GetString("PagingSupportEstimatedTotalCountTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unknown total count
    ///    Reviewed by TArcher on 2010-07-20
    ///  
    /// </summary>
    internal static string PagingSupportUnknownTotalCountTemplate {
        get {
            return ResourceManager.GetString("PagingSupportUnknownTotalCountTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    command '{0}'
    ///  
    /// </summary>
    internal static string ObsoleteCommand {
        get {
            return ResourceManager.GetString("ObsoleteCommand", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The {0} is obsolete. {1}
    ///  
    /// </summary>
    internal static string UseOfDeprecatedCommandWarning {
        get {
            return ResourceManager.GetString("UseOfDeprecatedCommandWarning", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Exec call failed with errorno {0} for command line: {1}
    ///  
    /// </summary>
    internal static string ExecFailed {
        get {
            return ResourceManager.GetString("ExecFailed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Command '{0}' was not found. The specified command must be an executable.
    ///  
    /// </summary>
    internal static string NativeCommandNotFound {
        get {
            return ResourceManager.GetString("NativeCommandNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Script Block Processing Dot-Source Check
    ///  
    /// </summary>
    internal static string WDACLogTitle {
        get {
            return ResourceManager.GetString("WDACLogTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Dot-Source processing for script block '{0}' will fail in Constrained Language mode because its language mode '{1}' does not match the current language mode '{2}'.
    ///  
    /// </summary>
    internal static string WDACLogMessage {
        get {
            return ResourceManager.GetString("WDACLogMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Command Searcher
    ///  
    /// </summary>
    internal static string SearcherWDACLogTitle {
        get {
            return ResourceManager.GetString("SearcherWDACLogTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Command '{0}' in module '{1}' is untrusted and will not be accessible in ConstrainedLanguage mode.
    ///  
    /// </summary>
    internal static string SearcherWDACLogMessage {
        get {
            return ResourceManager.GetString("SearcherWDACLogMessage", resourceCulture);
        }
    }

}

