
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class InternalHostUserInterfaceStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal InternalHostUserInterfaceStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.InternalHostUserInterfaceStrings", typeof(InternalHostUserInterfaceStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    WriteDebug stopped because the value of the DebugPreference variable was 'Stop'.
    ///  
    /// </summary>
    internal static string WriteDebugLineStoppedError {
        get {
            return ResourceManager.GetString("WriteDebugLineStoppedError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The value {0} is not a supported ActionPreference value.
    ///  
    /// </summary>
    internal static string UnsupportedPreferenceError {
        get {
            return ResourceManager.GetString("UnsupportedPreferenceError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The "{0}" parameter must contain at least one value.
    ///  
    /// </summary>
    internal static string PromptEmptyDescriptionsError {
        get {
            return ResourceManager.GetString("PromptEmptyDescriptionsError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    &Yes
    ///  
    /// </summary>
    internal static string ShouldContinueYesLabel {
        get {
            return ResourceManager.GetString("ShouldContinueYesLabel", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Continue.
    ///  
    /// </summary>
    internal static string ShouldContinueYesHelp {
        get {
            return ResourceManager.GetString("ShouldContinueYesHelp", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Yes to &All
    ///  
    /// </summary>
    internal static string ShouldContinueYesToAllLabel {
        get {
            return ResourceManager.GetString("ShouldContinueYesToAllLabel", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Continue, and do not ask again whether to continue in this session.
    ///  
    /// </summary>
    internal static string ShouldContinueYesToAllHelp {
        get {
            return ResourceManager.GetString("ShouldContinueYesToAllHelp", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    &No
    ///  
    /// </summary>
    internal static string ShouldContinueNoLabel {
        get {
            return ResourceManager.GetString("ShouldContinueNoLabel", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    End the operation with an error.
    ///  
    /// </summary>
    internal static string ShouldContinueNoHelp {
        get {
            return ResourceManager.GetString("ShouldContinueNoHelp", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    No to A&ll
    ///  
    /// </summary>
    internal static string ShouldContinueNoToAllLabel {
        get {
            return ResourceManager.GetString("ShouldContinueNoToAllLabel", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    End the operation with an error. Do not request to resume operation for this session.
    ///  
    /// </summary>
    internal static string ShouldContinueNoToAllHelp {
        get {
            return ResourceManager.GetString("ShouldContinueNoToAllHelp", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    &Suspend
    ///  
    /// </summary>
    internal static string ShouldContinueSuspendLabel {
        get {
            return ResourceManager.GetString("ShouldContinueSuspendLabel", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Pause the current operation and enter a command prompt. Type "exit" to resume the paused operation.
    ///  
    /// </summary>
    internal static string ShouldContinueSuspendHelp {
        get {
            return ResourceManager.GetString("ShouldContinueSuspendHelp", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Continue with this operation?
    ///  
    /// </summary>
    internal static string ShouldContinuePromptMessage {
        get {
            return ResourceManager.GetString("ShouldContinuePromptMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    (default is "{0}")
    ///  
    /// </summary>
    internal static string DefaultChoice {
        get {
            return ResourceManager.GetString("DefaultChoice", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    (default choices are {0})
    ///  
    /// </summary>
    internal static string DefaultChoicesForMultipleChoices {
        get {
            return ResourceManager.GetString("DefaultChoicesForMultipleChoices", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Choice[{0}]: 
    ///  
    /// </summary>
    internal static string ChoiceMessage {
        get {
            return ResourceManager.GetString("ChoiceMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    "{0}" should have at least one element.
    ///  
    /// </summary>
    internal static string EmptyChoicesError {
        get {
            return ResourceManager.GetString("EmptyChoicesError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    "{0}" must be a valid index into "{1}". "{2}" is not a valid index.
    ///  
    /// </summary>
    internal static string InvalidDefaultChoiceForMultipleSelection {
        get {
            return ResourceManager.GetString("InvalidDefaultChoiceForMultipleSelection", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the hot key because a question mark ("?") cannot be used as a hot key.
    ///  
    /// </summary>
    internal static string InvalidChoiceHotKeyError {
        get {
            return ResourceManager.GetString("InvalidChoiceHotKeyError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    VERBOSE: {0}
    ///  
    /// </summary>
    internal static string VerboseFormatString {
        get {
            return ResourceManager.GetString("VerboseFormatString", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    WARNING: {0}
    ///  
    /// </summary>
    internal static string WarningFormatString {
        get {
            return ResourceManager.GetString("WarningFormatString", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    DEBUG: {0}
    ///  
    /// </summary>
    internal static string DebugFormatString {
        get {
            return ResourceManager.GetString("DebugFormatString", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The host is not currently transcribing.
    ///  
    /// </summary>
    internal static string HostNotTranscribing {
        get {
            return ResourceManager.GetString("HostNotTranscribing", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Command start time: {0}
    ///  
    /// </summary>
    internal static string CommandStartTime {
        get {
            return ResourceManager.GetString("CommandStartTime", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    **********************
    ///PowerShell transcript start
    ///Start time: {0:yyyyMMddHHmmss}
    ///Username: {1}
    ///RunAs User: {2}
    ///Configuration Name: {3}
    ///Machine: {4} ({5})
    ///Host Application: {6}
    ///Process ID: {7}
    ///{8}
    ///**********************
    ///  
    /// </summary>
    internal static string TranscriptPrologue {
        get {
            return ResourceManager.GetString("TranscriptPrologue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    **********************
    ///PowerShell transcript start
    ///Start time: {0:yyyyMMddHHmmss}
    ///**********************
    ///  
    /// </summary>
    internal static string MinimalTranscriptPrologue {
        get {
            return ResourceManager.GetString("MinimalTranscriptPrologue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    **********************
    ///PowerShell transcript end
    ///End time: {0:yyyyMMddHHmmss}
    ///**********************
    ///  
    /// </summary>
    internal static string TranscriptEpilogue {
        get {
            return ResourceManager.GetString("TranscriptEpilogue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    File path {0} resolves to a directory.
    ///  
    /// </summary>
    internal static string InvalidTranscriptFilePath {
        get {
            return ResourceManager.GetString("InvalidTranscriptFilePath", resourceCulture);
        }
    }

}

