
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class PathUtilsStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal PathUtilsStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.PathUtilsStrings", typeof(PathUtilsStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Encoding 'UTF-7' is obsolete, please use UTF-8.
    ///  
    /// </summary>
    internal static string Utf7EncodingObsolete {
        get {
            return ResourceManager.GetString("Utf7EncodingObsolete", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    File {0} already exists and {1} was specified.
    ///  
    /// </summary>
    internal static string UtilityFileExistsNoClobber {
        get {
            return ResourceManager.GetString("UtilityFileExistsNoClobber", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot open file because the current provider ({0}) cannot open a file.
    ///  
    /// </summary>
    internal static string OutFile_ReadWriteFileNotFileSystemProvider {
        get {
            return ResourceManager.GetString("OutFile_ReadWriteFileNotFileSystemProvider", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot perform operation because the path resolved to more than one file. This command cannot operate on multiple files.
    ///  
    /// </summary>
    internal static string OutFile_MultipleFilesNotSupported {
        get {
            return ResourceManager.GetString("OutFile_MultipleFilesNotSupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot perform operation because the wildcard path {0} did not resolve to a file.
    ///  
    /// </summary>
    internal static string OutFile_DidNotResolveFile {
        get {
            return ResourceManager.GetString("OutFile_DidNotResolveFile", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unknown encoding {0}; valid values are {1}.
    ///  
    /// </summary>
    internal static string OutFile_WriteToFileEncodingUnknown {
        get {
            return ResourceManager.GetString("OutFile_WriteToFileEncodingUnknown", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The directory '{0}' already exists.  Use the -Force parameter if you want to overwrite the directory and files within the directory.
    ///  
    /// </summary>
    internal static string ExportPSSession_ErrorDirectoryExists {
        get {
            return ResourceManager.GetString("ExportPSSession_ErrorDirectoryExists", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The user module path does not exist, and hence a module folder cannot be created for the provided module name '{0}'.
    ///  
    /// </summary>
    internal static string ExportPSSession_ErrorModuleNameOrPath {
        get {
            return ResourceManager.GetString("ExportPSSession_ErrorModuleNameOrPath", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot create the module {0} due to the following: {1}. Use a different argument for the -OutputModule parameter and retry.
    ///    {StrContains="OutputModule"}
    ///{0} is a placeholder for the name of a directory
    ///{1} is a placeholder for an error message from the inner exception
    ///
    ///Reviewed by TArcher on 2010-07-21
    ///
    ///Example usage:
    ///PS C:\> $s = New-PSSession
    ///PS C:\> Export-PSSession -Session $s -OutputModule gibberish:here
    ///Export-PSSession : Cannot create the module 'gibberish:here' due to the following: Cannot find drive. A drive with the name 'gibberish' does not exist. Use a different argument for the -OutputModule parameter and try again.
    ///At line:1 char:1
    ///+ Export-PSSession -Session $s -OutputModule gibberish:here
    ///+ ^
    ///    + CategoryInfo          : ResourceExists: (gibberish:here:String) [Export-PSSession], ArgumentException
    ///    + FullyQualifiedErrorId : ExportProxyCommand_CannotCreateOutputDirectory,Microsoft.PowerShell.Commands.ExportPSSessionCommand
    ///    
    ///  
    /// </summary>
    internal static string ExportPSSession_CannotCreateOutputDirectory {
        get {
            return ResourceManager.GetString("ExportPSSession_CannotCreateOutputDirectory", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The module cannot be loaded because it has been generated with an incompatible version of the {0} cmdlet. Generate the module with the {0} cmdlet from the current session, and try loading the module again.
    ///  
    /// </summary>
    internal static string ExportPSSession_ScriptGeneratorVersionMismatch {
        get {
            return ResourceManager.GetString("ExportPSSession_ScriptGeneratorVersionMismatch", resourceCulture);
        }
    }

}

