
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class AddTypeStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal AddTypeStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.PowerShell.Commands.Utility.resources.AddTypeStrings", typeof(AddTypeStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The source code was already compiled and loaded.
    ///  
    /// </summary>
    internal static string AlreadyCompiledandLoaded {
        get {
            return ResourceManager.GetString("AlreadyCompiledandLoaded", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot add type. The "{0}" extension is not supported.
    ///  
    /// </summary>
    internal static string FileExtensionNotSupported {
        get {
            return ResourceManager.GetString("FileExtensionNotSupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot add type. Input files must all have the same file extension.
    ///  
    /// </summary>
    internal static string MultipleExtensionsNotSupported {
        get {
            return ResourceManager.GetString("MultipleExtensionsNotSupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot add type. The assembly '{0}' could not be found.
    ///  
    /// </summary>
    internal static string AssemblyNotFound {
        get {
            return ResourceManager.GetString("AssemblyNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot add type. The type name '{0}' already exists.
    ///  
    /// </summary>
    internal static string TypeAlreadyExists {
        get {
            return ResourceManager.GetString("TypeAlreadyExists", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot set output assembly. The path {0} did not resolve to a single file.
    ///  
    /// </summary>
    internal static string OutputAssemblyDidNotResolve {
        get {
            return ResourceManager.GetString("OutputAssemblyDidNotResolve", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot add type. Compilation errors occurred.
    ///  
    /// </summary>
    internal static string CompilerErrors {
        get {
            return ResourceManager.GetString("CompilerErrors", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot add type. The OutputType parameter requires that the OutputAssembly parameter be specified.
    ///  
    /// </summary>
    internal static string OutputTypeRequiresOutputAssembly {
        get {
            return ResourceManager.GetString("OutputTypeRequiresOutputAssembly", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot add type. Definition of new types is not supported in this language mode.
    ///  
    /// </summary>
    internal static string CannotDefineNewType {
        get {
            return ResourceManager.GetString("CannotDefineNewType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified reference assembly '{0}' is unnecessary and ignored.
    ///  
    /// </summary>
    internal static string ReferenceAssemblyIgnored {
        get {
            return ResourceManager.GetString("ReferenceAssemblyIgnored", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Both the assembly types 'ConsoleApplication' and 'WindowsApplication' are not currently supported.
    ///  
    /// </summary>
    internal static string AssemblyTypeNotSupported {
        get {
            return ResourceManager.GetString("AssemblyTypeNotSupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Add-Type Cmdlet
    ///  
    /// </summary>
    internal static string AddTypeLogTitle {
        get {
            return ResourceManager.GetString("AddTypeLogTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Add-Type cmdlet will not be allowed in ConstrainedLanguage mode.
    ///  
    /// </summary>
    internal static string AddTypeLogMessage {
        get {
            return ResourceManager.GetString("AddTypeLogMessage", resourceCulture);
        }
    }

}

