
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class SubsystemStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal SubsystemStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.SubsystemStrings", typeof(SubsystemStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The subsystem '{0}' does not allow more than one implementation to be registered.
    ///  
    /// </summary>
    internal static string MultipleRegistrationNotAllowed {
        get {
            return ResourceManager.GetString("MultipleRegistrationNotAllowed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The implementation with Id '{0}' was already registered for the subsystem '{1}'.
    ///  
    /// </summary>
    internal static string ImplementationAlreadyRegistered {
        get {
            return ResourceManager.GetString("ImplementationAlreadyRegistered", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The subsystem '{0}' does not allow the unregistration of an implementation.
    ///  
    /// </summary>
    internal static string UnregistrationNotAllowed {
        get {
            return ResourceManager.GetString("UnregistrationNotAllowed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    No implementation was registered for the subsystem '{0}'.
    ///  
    /// </summary>
    internal static string NoImplementationRegistered {
        get {
            return ResourceManager.GetString("NoImplementationRegistered", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A registered implementation with the Id '{0}' was not found.
    ///  
    /// </summary>
    internal static string ImplementationNotFound {
        get {
            return ResourceManager.GetString("ImplementationNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified subsystem type '{0}' is unknown.
    ///  
    /// </summary>
    internal static string SubsystemTypeUnknown {
        get {
            return ResourceManager.GetString("SubsystemTypeUnknown", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    You must specify a concrete subsystem type instead of the base interface 'ISubsystem'.
    ///  
    /// </summary>
    internal static string MustUseConcreteSubsystemType {
        get {
            return ResourceManager.GetString("MustUseConcreteSubsystemType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified subsystem kind '{0}' is unknown.
    ///  
    /// </summary>
    internal static string SubsystemKindUnknown {
        get {
            return ResourceManager.GetString("SubsystemKindUnknown", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    For the target subsystem kind '{0}', the specified subsystem instance needs to implement the corresponding concrete interface or abstract class '{1}'.
    ///  
    /// </summary>
    internal static string ConcreteSubsystemNotImplemented {
        get {
            return ResourceManager.GetString("ConcreteSubsystemNotImplemented", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The declared metadata for subsystem kind '{0}' is invalid. A subsystem that requires cmdlets or functions to be defined cannot allow multiple registrations because that would result in one implementation overwriting the commands defined by another implementation.
    ///  
    /// </summary>
    internal static string InvalidSubsystemInfo {
        get {
            return ResourceManager.GetString("InvalidSubsystemInfo", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The 'Id' property of an implementation for the subsystem '{0}' cannot be an empty GUID.
    ///  
    /// </summary>
    internal static string EmptyImplementationId {
        get {
            return ResourceManager.GetString("EmptyImplementationId", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The 'Name' property of an implementation for the subsystem '{0}' cannot be null or an empty string.
    ///  
    /// </summary>
    internal static string NullOrEmptyImplementationName {
        get {
            return ResourceManager.GetString("NullOrEmptyImplementationName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The 'Description' property of an implementation for the subsystem '{0}' cannot be null or an empty string.
    ///  
    /// </summary>
    internal static string NullOrEmptyImplementationDescription {
        get {
            return ResourceManager.GetString("NullOrEmptyImplementationDescription", resourceCulture);
        }
    }

}

