using System.Diagnostics.CodeAnalysis;

[assembly: SuppressMessage(
    "Style",
    "IDE0044:Add readonly modifier",
    Justification = "see src/System.Management.Automation/engine/ComInterop/README.md",
    Scope = "NamespaceAndDescendants",
    Target = "~N:System.Management.Automation.ComInterop")]
