
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class ConsoleControlStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal ConsoleControlStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.PowerShell.ConsoleHost.resources.ConsoleControlStrings", typeof(ConsoleControlStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Win32 internal error "{0}" 0x{1:X} occurred when adding a break handler. Contact Microsoft Support Services.
    ///  
    /// </summary>
    internal static string AddBreakHandlerExceptionMessage {
        get {
            return ResourceManager.GetString("AddBreakHandlerExceptionMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Win32 internal error "{0}" 0x{1:X} occurred while trying to remove a break handler. Contact Microsoft Customer Support Services.
    ///  
    /// </summary>
    internal static string RemoveBreakHandlerExceptionTemplate {
        get {
            return ResourceManager.GetString("RemoveBreakHandlerExceptionTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Win32 internal error "{0}" 0x{1:X} occurred while getting input about the console handle. Contact Microsoft Customer Support Services.
    ///  
    /// </summary>
    internal static string GetInputModeExceptionTemplate {
        get {
            return ResourceManager.GetString("GetInputModeExceptionTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Win32 internal error "{0}" 0x{1:X} occurred while retrieving the handle for the active console output buffer. Contact Microsoft Customer Support Services.
    ///  
    /// </summary>
    internal static string GetActiveScreenBufferHandleExceptionTemplate {
        get {
            return ResourceManager.GetString("GetActiveScreenBufferHandleExceptionTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Win32 internal error "{0}" 0x{1:X} occurred while getting the console mode. Contact Microsoft Customer Support Services.
    ///  
    /// </summary>
    internal static string GetModeExceptionTemplate {
        get {
            return ResourceManager.GetString("GetModeExceptionTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Win32 internal error "{0}" 0x{1:X} occurred while setting the console mode. Contact Microsoft Customer Support Services.
    ///  
    /// </summary>
    internal static string SetModeExceptionTemplate {
        get {
            return ResourceManager.GetString("SetModeExceptionTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Win32 internal error "{0}" 0x{1:X} occurred while reading characters from the console input buffer. Contact Microsoft Customer Support Services.
    ///  
    /// </summary>
    internal static string ReadConsoleExceptionTemplate {
        get {
            return ResourceManager.GetString("ReadConsoleExceptionTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Win32 internal error "{0}" 0x{1:X} occurred while reading input records from the console input buffer. Contact Microsoft Customer Support Services.
    ///  
    /// </summary>
    internal static string ReadConsoleInputExceptionTemplate {
        get {
            return ResourceManager.GetString("ReadConsoleInputExceptionTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Win32 internal error "{0}" 0x{1:X} occurred while reading the contents of the console input buffer. Contact Microsoft Customer Support Services.
    ///  
    /// </summary>
    internal static string PeekConsoleInputExceptionTemplate {
        get {
            return ResourceManager.GetString("PeekConsoleInputExceptionTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Win32 internal error "{0}" 0x{1:X} occurred while getting the number of events in the console input buffer. Contact Microsoft Customer Support Services.
    ///  
    /// </summary>
    internal static string GetNumberOfConsoleInputEventsExceptionTemplate {
        get {
            return ResourceManager.GetString("GetNumberOfConsoleInputEventsExceptionTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Win32 internal error "{0}" 0x{1:X} occurred while flushing the console input buffer. Contact Microsoft Customer Support Services.
    ///  
    /// </summary>
    internal static string FlushConsoleInputBufferExceptionTemplate {
        get {
            return ResourceManager.GetString("FlushConsoleInputBufferExceptionTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Win32 internal error "{0}" 0x{1:X} occurred while getting console output buffer information. Contact Microsoft Customer Support Services.
    ///  
    /// </summary>
    internal static string GetConsoleScreenBufferInfoExceptionTemplate {
        get {
            return ResourceManager.GetString("GetConsoleScreenBufferInfoExceptionTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Win32 internal error "{0}" 0x{1:X} occurred while setting the console output buffer size. Contact Microsoft Customer Support Services.
    ///  
    /// </summary>
    internal static string SetConsoleScreenBufferSizeExceptionTemplate {
        get {
            return ResourceManager.GetString("SetConsoleScreenBufferSizeExceptionTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Win32 internal error "{0}" 0x{1:X} occurred while writing to the console output buffer. Contact Microsoft Customer Support Services.
    ///  
    /// </summary>
    internal static string WriteConsoleOutputExceptionTemplate {
        get {
            return ResourceManager.GetString("WriteConsoleOutputExceptionTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Win32 internal error "{0}" 0x{1:X} occurred while reading the console output buffer. Contact Microsoft Customer Support Services.
    ///  
    /// </summary>
    internal static string ReadConsoleOutputExceptionTemplate {
        get {
            return ResourceManager.GetString("ReadConsoleOutputExceptionTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Win32 internal error "{0}" 0x{1:X} occurred while filling the console output buffer with characters. Contact Microsoft Customer Support Services.
    ///  
    /// </summary>
    internal static string FillConsoleOutputCharacterExceptionTemplate {
        get {
            return ResourceManager.GetString("FillConsoleOutputCharacterExceptionTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Win32 internal error "{0}" 0x{1:X} occurred while filling the console output buffer with attributes. Contact Microsoft Customer Support Services.
    ///  
    /// </summary>
    internal static string FillConsoleOutputAttributeExceptionTemplate {
        get {
            return ResourceManager.GetString("FillConsoleOutputAttributeExceptionTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Win32 internal error "{0}" 0x{1:X} occurred while scrolling the console output buffer. Contact Microsoft Customer Support Services.
    ///  
    /// </summary>
    internal static string ScrollConsoleScreenBufferExceptionTemplate {
        get {
            return ResourceManager.GetString("ScrollConsoleScreenBufferExceptionTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Win32 internal error "{0}" 0x{1:X} occurred while setting the console window information. Contact Microsoft Customer Support Services.
    ///  
    /// </summary>
    internal static string SetConsoleWindowInfoExceptionTemplate {
        get {
            return ResourceManager.GetString("SetConsoleWindowInfoExceptionTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Win32 internal error "{0}" 0x{1:X} occurred while getting the largest console window size. Contact Microsoft Customer Support Services.
    ///  
    /// </summary>
    internal static string GetLargestConsoleWindowSizeExceptionTemplate {
        get {
            return ResourceManager.GetString("GetLargestConsoleWindowSizeExceptionTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Win32 internal error "{0}" 0x{1:X} occurred while setting the console window title. Contact Microsoft Customer Support Services.
    ///  
    /// </summary>
    internal static string SetConsoleWindowTitleExceptionTemplate {
        get {
            return ResourceManager.GetString("SetConsoleWindowTitleExceptionTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Win32 internal error "{0}" 0x{1:X} occurred while writing to the console output buffer at the current cursor position. Contact Microsoft Customer Support Services.
    ///  
    /// </summary>
    internal static string WriteConsoleExceptionTemplate {
        get {
            return ResourceManager.GetString("WriteConsoleExceptionTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Win32 internal error "{0}" 0x{1:X} occurred while setting character attributes for the console output buffer. Contact Microsoft Customer Support Services.
    ///  
    /// </summary>
    internal static string SetConsoleTextAttributeExceptionTemplate {
        get {
            return ResourceManager.GetString("SetConsoleTextAttributeExceptionTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Win32 internal error "{0}" 0x{1:X} occurred while getting cursor information. Contact Microsoft Customer Support Services.
    ///  
    /// </summary>
    internal static string GetConsoleCursorInfoExceptionTemplate {
        get {
            return ResourceManager.GetString("GetConsoleCursorInfoExceptionTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Win32 internal error "{0}" 0x{1:X} occurred while setting cursor information.Contact Microsoft Customer Support Services.
    ///  
    /// </summary>
    internal static string SetConsoleCursorInfoExceptionTemplate {
        get {
            return ResourceManager.GetString("SetConsoleCursorInfoExceptionTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Win32 internal error "{0}" 0x{1:X} occurred while getting console font information. Contact Microsoft Customer Support Services.
    ///  
    /// </summary>
    internal static string GetConsoleFontInfoExceptionTemplate {
        get {
            return ResourceManager.GetString("GetConsoleFontInfoExceptionTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Win32 internal error "{0}" 0x{1:X} occurred while sending keyboard input. Contact Microsoft Customer Support Services.
    ///  
    /// </summary>
    internal static string SendKeyPressInputExceptionTemplate {
        get {
            return ResourceManager.GetString("SendKeyPressInputExceptionTemplate", resourceCulture);
        }
    }

}

