
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

public class ShowCommandResources {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    public ShowCommandResources() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    public static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.Management.UI.Internal.resources.public.ShowCommandResources", typeof(ShowCommandResources).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    public static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cancel
    ///  
    /// </summary>
    public static string ActionButtons_Button_Cancel {
        get {
            return ResourceManager.GetString("ActionButtons_Button_Cancel", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Co_py
    ///  
    /// </summary>
    public static string ActionButtons_Button_Copy {
        get {
            return ResourceManager.GetString("ActionButtons_Button_Copy", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    _Run
    ///  
    /// </summary>
    public static string ActionButtons_Button_Run {
        get {
            return ResourceManager.GetString("ActionButtons_Button_Run", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    All
    ///  
    /// </summary>
    public static string All {
        get {
            return ResourceManager.GetString("All", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Modules:
    ///  
    /// </summary>
    public static string AllModulesControl_Label_Modules {
        get {
            return ResourceManager.GetString("AllModulesControl_Label_Modules", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///     ? 
    ///  
    /// </summary>
    public static string CmdletControl_Button_GetHelp {
        get {
            return ResourceManager.GetString("CmdletControl_Button_GetHelp", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Help
    ///  
    /// </summary>
    public static string CmdletControl_Button_ToolTip_Help {
        get {
            return ResourceManager.GetString("CmdletControl_Button_ToolTip_Help", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Common Parameters
    ///  
    /// </summary>
    public static string CmdletControl_Header_CommonParameters {
        get {
            return ResourceManager.GetString("CmdletControl_Header_CommonParameters", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Errors
    ///  
    /// </summary>
    public static string CmdletControl_Header_Errors {
        get {
            return ResourceManager.GetString("CmdletControl_Header_Errors", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Parameters for "{0}":
    ///  
    /// </summary>
    public static string DetailsParameterTitleFormat {
        get {
            return ResourceManager.GetString("DetailsParameterTitleFormat", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Name: {0}
    ///Module: {1} ({2})
    ///  
    /// </summary>
    public static string CmdletTooltipFormat {
        get {
            return ResourceManager.GetString("CmdletTooltipFormat", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The following errors occurred running the command:
    ///{0}
    ///  
    /// </summary>
    public static string EndProcessingErrorMessage {
        get {
            return ResourceManager.GetString("EndProcessingErrorMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///     *
    ///    Used in MandatoryNameLabelFormat to designate a mandatory parameter with a *
    ///  
    /// </summary>
    public static string MandatoryLabelSegment {
        get {
            return ResourceManager.GetString("MandatoryLabelSegment", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0}:{1}
    ///    This is a label for a control, hence the colon. {0} is a parameter name, {1} is MandatoryLabelSegment or an empty string
    ///  
    /// </summary>
    public static string MandatoryNameLabelFormat {
        get {
            return ResourceManager.GetString("MandatoryNameLabelFormat", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0}:
    ///    This is a label for a control, hence the colon. {0} is a parameter name
    ///  
    /// </summary>
    public static string NameLabelFormat {
        get {
            return ResourceManager.GetString("NameLabelFormat", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Name:
    ///  
    /// </summary>
    public static string ShowModuleControl_Label_Name {
        get {
            return ResourceManager.GetString("ShowModuleControl_Label_Name", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    OK
    ///  
    /// </summary>
    public static string ActionButtons_Button_Ok {
        get {
            return ResourceManager.GetString("ActionButtons_Button_Ok", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ...
    ///  
    /// </summary>
    public static string MultiParameter_Button_Browse {
        get {
            return ResourceManager.GetString("MultiParameter_Button_Browse", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Command Name
    ///  
    /// </summary>
    public static string CommandNameAutomationName {
        get {
            return ResourceManager.GetString("CommandNameAutomationName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Modules
    ///  
    /// </summary>
    public static string ModulesAutomationName {
        get {
            return ResourceManager.GetString("ModulesAutomationName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    <No module name>
    ///  
    /// </summary>
    public static string NoModuleName {
        get {
            return ResourceManager.GetString("NoModuleName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Select multiple values for "{0}"
    ///  
    /// </summary>
    public static string SelectMultipleValuesForParameterFormat {
        get {
            return ResourceManager.GetString("SelectMultipleValuesForParameterFormat", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Can receive value from pipeline
    ///  
    /// </summary>
    public static string CanReceiveValueFromPipeline {
        get {
            return ResourceManager.GetString("CanReceiveValueFromPipeline", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Common to all parameter sets
    ///  
    /// </summary>
    public static string CommonToAllParameterSets {
        get {
            return ResourceManager.GetString("CommonToAllParameterSets", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Mandatory
    ///  
    /// </summary>
    public static string Mandatory {
        get {
            return ResourceManager.GetString("Mandatory", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Optional
    ///  
    /// </summary>
    public static string Optional {
        get {
            return ResourceManager.GetString("Optional", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Position: {0}
    ///  
    /// </summary>
    public static string PositionFormat {
        get {
            return ResourceManager.GetString("PositionFormat", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Type: {0}
    ///  
    /// </summary>
    public static string TypeFormat {
        get {
            return ResourceManager.GetString("TypeFormat", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Imported
    ///  
    /// </summary>
    public static string Imported {
        get {
            return ResourceManager.GetString("Imported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Not Imported
    ///  
    /// </summary>
    public static string NotImported {
        get {
            return ResourceManager.GetString("NotImported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Show Details
    ///  
    /// </summary>
    public static string ImportModuleButtonText {
        get {
            return ResourceManager.GetString("ImportModuleButtonText", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Failed to import the module required by command "{0}". Module name: "{1}". Error message: "{2}".
    ///  
    /// </summary>
    public static string ImportModuleFailedFormat {
        get {
            return ResourceManager.GetString("ImportModuleFailedFormat", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    To import the "{0}" module and its cmdlets, including "{1}", click {2}.
    ///  
    /// </summary>
    public static string NotImportedFormat {
        get {
            return ResourceManager.GetString("NotImportedFormat", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Show Command - Error
    ///  
    /// </summary>
    public static string ShowCommandError {
        get {
            return ResourceManager.GetString("ShowCommandError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Please Wait...
    ///  
    /// </summary>
    public static string PleaseWaitMessage {
        get {
            return ResourceManager.GetString("PleaseWaitMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Refresh
    ///  
    /// </summary>
    public static string ShowModuleControl_RefreshButton {
        get {
            return ResourceManager.GetString("ShowModuleControl_RefreshButton", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    There are no parameters.
    ///  
    /// </summary>
    public static string NoParameters {
        get {
            return ResourceManager.GetString("NoParameters", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Click after using "{0}" to see the new commands
    ///  
    /// </summary>
    public static string RefreshShowCommandTooltipFormat {
        get {
            return ResourceManager.GetString("RefreshShowCommandTooltipFormat", resourceCulture);
        }
    }

}

