<!--=================================================================
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT License.
==================================================================-->

<UserControl x:Class="Microsoft.PowerShell.Commands.ShowCommandInternal.MultipleSelectionControl"
             xmlns="http://schemas.microsoft.com/winfx/2006/xaml/presentation"
             xmlns:x="http://schemas.microsoft.com/winfx/2006/xaml"
             xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006"
             xmlns:d="http://schemas.microsoft.com/expression/blend/2008"
             xmlns:sc="clr-namespace:Microsoft.PowerShell.Commands.ShowCommandInternal"
             xmlns:default="clr-namespace:"
             mc:Ignorable="d">
    <Grid>
        <Grid.ColumnDefinitions>
            <ColumnDefinition Width="*"></ColumnDefinition>
            <ColumnDefinition Width="Auto"></ColumnDefinition>
        </Grid.ColumnDefinitions>
        <Grid.RowDefinitions>
            <RowDefinition Height="Auto"></RowDefinition>
            <RowDefinition></RowDefinition>
        </Grid.RowDefinitions>
        <ComboBox x:Name="comboxParameter" Margin="3" VerticalAlignment="Top" Grid.Column="0" IsEditable="True"  ItemsSource="{Binding ItemsSource}" AutomationProperties.AutomationId="comboxParameter">
        </ComboBox>
        <Button x:Name="multipleValueButton" Padding="3,0,3,0" Margin="3" Grid.Column="1" Click="ButtonBrowse_Click" Content="{x:Static default:ShowCommandResources.MultiParameter_Button_Browse}" AutomationProperties.AutomationId="ButtonBrowse"></Button>
    </Grid>
</UserControl>
