
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class NavigationResources {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal NavigationResources() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.PowerShell.Commands.Management.resources.NavigationResources", typeof(NavigationResources).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified path is a container that has child items. Do you want to delete this container and its child items?
    ///  
    /// </summary>
    internal static string DeleteHasChildrenPrompt {
        get {
            return ResourceManager.GetString("DeleteHasChildrenPrompt", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Do you want to delete the specified item?
    ///  
    /// </summary>
    internal static string DeletePrompt {
        get {
            return ResourceManager.GetString("DeletePrompt", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot copy because the specified destination already exists. Do you want to overwrite the existing content?
    ///  
    /// </summary>
    internal static string CopyToExistingPrompt {
        get {
            return ResourceManager.GetString("CopyToExistingPrompt", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    New drive
    ///  
    /// </summary>
    internal static string NewDriveConfirmAction {
        get {
            return ResourceManager.GetString("NewDriveConfirmAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Name: {0} Provider: {1} Root: {2}
    ///  
    /// </summary>
    internal static string NewDriveConfirmResourceTemplate {
        get {
            return ResourceManager.GetString("NewDriveConfirmResourceTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Remove Drive
    ///  
    /// </summary>
    internal static string RemoveDriveConfirmAction {
        get {
            return ResourceManager.GetString("RemoveDriveConfirmAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Name: {0} Provider: {1} Root: {2}
    ///  
    /// </summary>
    internal static string RemoveDriveConfirmResourceTemplate {
        get {
            return ResourceManager.GetString("RemoveDriveConfirmResourceTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot remove drive '{0}' because it is in use.
    ///  
    /// </summary>
    internal static string RemoveDriveInUse {
        get {
            return ResourceManager.GetString("RemoveDriveInUse", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The item at {0} has children and the Recurse parameter was not specified. If you continue, all children will be removed with the item. Are you sure you want to continue?
    ///  
    /// </summary>
    internal static string RemoveItemWithChildren {
        get {
            return ResourceManager.GetString("RemoveItemWithChildren", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot remove the item at '{0}' because it is in use.
    ///  
    /// </summary>
    internal static string RemoveItemInUse {
        get {
            return ResourceManager.GetString("RemoveItemInUse", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    An object at the specified path {0} does not exist, or has been filtered by the -Include or -Exclude parameter.
    ///  
    /// </summary>
    internal static string ItemNotFound {
        get {
            return ResourceManager.GetString("ItemNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Set Content
    ///  
    /// </summary>
    internal static string SetContentAction {
        get {
            return ResourceManager.GetString("SetContentAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Path: {0}
    ///  
    /// </summary>
    internal static string SetContentTarget {
        get {
            return ResourceManager.GetString("SetContentTarget", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Add Content
    ///  
    /// </summary>
    internal static string AddContentAction {
        get {
            return ResourceManager.GetString("AddContentAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Path: {0}
    ///  
    /// </summary>
    internal static string AddContentTarget {
        get {
            return ResourceManager.GetString("AddContentTarget", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot move item because the item at '{0}' does not exist.
    ///  
    /// </summary>
    internal static string MoveItemDoesntExist {
        get {
            return ResourceManager.GetString("MoveItemDoesntExist", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot move item because the item at '{0}' is in use.
    ///  
    /// </summary>
    internal static string MoveItemInUse {
        get {
            return ResourceManager.GetString("MoveItemInUse", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot rename because item at '{0}' does not exist.
    ///  
    /// </summary>
    internal static string RenameItemDoesntExist {
        get {
            return ResourceManager.GetString("RenameItemDoesntExist", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot rename the item at '{0}' because it is in use.
    ///  
    /// </summary>
    internal static string RenamedItemInUse {
        get {
            return ResourceManager.GetString("RenamedItemInUse", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot parse path because path '{0}' does not have a qualifier specified.
    ///  
    /// </summary>
    internal static string ParsePathFormatError {
        get {
            return ResourceManager.GetString("ParsePathFormatError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Begin
    ///  
    /// </summary>
    internal static string CreateAction {
        get {
            return ResourceManager.GetString("CreateAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Rollback
    ///  
    /// </summary>
    internal static string RollbackAction {
        get {
            return ResourceManager.GetString("RollbackAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Commit
    ///  
    /// </summary>
    internal static string CommitAction {
        get {
            return ResourceManager.GetString("CommitAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Current transaction
    ///  
    /// </summary>
    internal static string TransactionResource {
        get {
            return ResourceManager.GetString("TransactionResource", resourceCulture);
        }
    }

}

