
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class RegistryProviderStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal RegistryProviderStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.RegistryProviderStrings", typeof(RegistryProviderStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Set Item
    ///  
    /// </summary>
    internal static string SetItemAction {
        get {
            return ResourceManager.GetString("SetItemAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Item: {0} Value: {1}
    ///  
    /// </summary>
    internal static string SetItemResourceTemplate {
        get {
            return ResourceManager.GetString("SetItemResourceTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Clear Item
    ///  
    /// </summary>
    internal static string ClearItemAction {
        get {
            return ResourceManager.GetString("ClearItemAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Item: {0}
    ///  
    /// </summary>
    internal static string ClearItemResourceTemplate {
        get {
            return ResourceManager.GetString("ClearItemResourceTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    New Item
    ///  
    /// </summary>
    internal static string NewItemAction {
        get {
            return ResourceManager.GetString("NewItemAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Item: {0}
    ///  
    /// </summary>
    internal static string NewItemResourceTemplate {
        get {
            return ResourceManager.GetString("NewItemResourceTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Remove Key
    ///  
    /// </summary>
    internal static string RemoveKeyAction {
        get {
            return ResourceManager.GetString("RemoveKeyAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Item: {0}
    ///  
    /// </summary>
    internal static string RemoveKeyResourceTemplate {
        get {
            return ResourceManager.GetString("RemoveKeyResourceTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Copy Key
    ///  
    /// </summary>
    internal static string CopyKeyAction {
        get {
            return ResourceManager.GetString("CopyKeyAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Item: {0} Destination: {1}
    ///  
    /// </summary>
    internal static string CopyKeyResourceTemplate {
        get {
            return ResourceManager.GetString("CopyKeyResourceTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Rename Item
    ///  
    /// </summary>
    internal static string RenameItemAction {
        get {
            return ResourceManager.GetString("RenameItemAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Item: {0} NewName: {1}
    ///  
    /// </summary>
    internal static string RenameItemResourceTemplate {
        get {
            return ResourceManager.GetString("RenameItemResourceTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Move Item
    ///  
    /// </summary>
    internal static string MoveItemAction {
        get {
            return ResourceManager.GetString("MoveItemAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Item: {0} Destination: {1}
    ///  
    /// </summary>
    internal static string MoveItemResourceTemplate {
        get {
            return ResourceManager.GetString("MoveItemResourceTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Set Property
    ///  
    /// </summary>
    internal static string SetPropertyAction {
        get {
            return ResourceManager.GetString("SetPropertyAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Item: {0} Property: {1}
    ///  
    /// </summary>
    internal static string SetPropertyResourceTemplate {
        get {
            return ResourceManager.GetString("SetPropertyResourceTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Clear Property
    ///  
    /// </summary>
    internal static string ClearPropertyAction {
        get {
            return ResourceManager.GetString("ClearPropertyAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Item: {0} Property: {1}
    ///  
    /// </summary>
    internal static string ClearPropertyResourceTemplate {
        get {
            return ResourceManager.GetString("ClearPropertyResourceTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    New Property
    ///  
    /// </summary>
    internal static string NewPropertyAction {
        get {
            return ResourceManager.GetString("NewPropertyAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Item: {0} Property: {1}
    ///  
    /// </summary>
    internal static string NewPropertyResourceTemplate {
        get {
            return ResourceManager.GetString("NewPropertyResourceTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Remove Property
    ///  
    /// </summary>
    internal static string RemovePropertyAction {
        get {
            return ResourceManager.GetString("RemovePropertyAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Item: {0} Property: {1}
    ///  
    /// </summary>
    internal static string RemovePropertyResourceTemplate {
        get {
            return ResourceManager.GetString("RemovePropertyResourceTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Rename property.
    ///  
    /// </summary>
    internal static string RenamePropertyAction {
        get {
            return ResourceManager.GetString("RenamePropertyAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Item: {0} SourceProperty: {1} DestinationProperty: {2}
    ///  
    /// </summary>
    internal static string RenamePropertyResourceTemplate {
        get {
            return ResourceManager.GetString("RenamePropertyResourceTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Copy Property
    ///  
    /// </summary>
    internal static string CopyPropertyAction {
        get {
            return ResourceManager.GetString("CopyPropertyAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Item: {0} SourceProperty: {1} DestinationItem: {2} DestinationProperty: {3}
    ///  
    /// </summary>
    internal static string CopyPropertyResourceTemplate {
        get {
            return ResourceManager.GetString("CopyPropertyResourceTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Move Property
    ///  
    /// </summary>
    internal static string MovePropertyAction {
        get {
            return ResourceManager.GetString("MovePropertyAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Item: {0} SourceProperty: {1} DestinationItem: {2} DestinationProperty: {3}
    ///  
    /// </summary>
    internal static string MovePropertyResourceTemplate {
        get {
            return ResourceManager.GetString("MovePropertyResourceTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The operation was not processed. The location that was provided does not allow this operation.
    ///  
    /// </summary>
    internal static string ContainerInvalidOperationTemplate {
        get {
            return ResourceManager.GetString("ContainerInvalidOperationTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The operation is not allowed on the source location.
    ///  
    /// </summary>
    internal static string SourceContainerInvalidOperationTemplate {
        get {
            return ResourceManager.GetString("SourceContainerInvalidOperationTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The operation is not allowed on the destination location.
    ///  
    /// </summary>
    internal static string DestinationContainerInvalidOperationTemplate {
        get {
            return ResourceManager.GetString("DestinationContainerInvalidOperationTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The configuration settings for the local computer
    ///  
    /// </summary>
    internal static string HKLMDriveDescription {
        get {
            return ResourceManager.GetString("HKLMDriveDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The software settings for the current user
    ///  
    /// </summary>
    internal static string HKCUDriveDescription {
        get {
            return ResourceManager.GetString("HKCUDriveDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A key in this path already exists.
    ///  
    /// </summary>
    internal static string KeyAlreadyExists {
        get {
            return ResourceManager.GetString("KeyAlreadyExists", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The operation cannot be performed because the destination path is subordinate to the source path.
    ///  
    /// </summary>
    internal static string DestinationChildOfSource {
        get {
            return ResourceManager.GetString("DestinationChildOfSource", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The property already exists.
    ///  
    /// </summary>
    internal static string PropertyAlreadyExists {
        get {
            return ResourceManager.GetString("PropertyAlreadyExists", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Property {0} does not exist at path {1}.
    ///  
    /// </summary>
    internal static string PropertyNotAtPath {
        get {
            return ResourceManager.GetString("PropertyNotAtPath", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The registry key at the specified path does not exist.
    ///  
    /// </summary>
    internal static string KeyDoesNotExist {
        get {
            return ResourceManager.GetString("KeyDoesNotExist", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Could not bind parameter 'Type'. Could not convert "{0}" to "{1}". The possible enumeration values are "String, ExpandString, Binary, DWord, MultiString, QWord, Unknown".
    ///  
    /// </summary>
    internal static string TypeParameterBindingFailure {
        get {
            return ResourceManager.GetString("TypeParameterBindingFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Key {0} has been created, but a default value could not be set.
    ///  
    /// </summary>
    internal static string KeyCreatedValueFailed {
        get {
            return ResourceManager.GetString("KeyCreatedValueFailed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unable to create a drive with the specified root. The root path does not exist.
    ///  
    /// </summary>
    internal static string NewDriveRootDoesNotExist {
        get {
            return ResourceManager.GetString("NewDriveRootDoesNotExist", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The item cannot be renamed because an item with that name already exists in the same container.
    ///  
    /// </summary>
    internal static string RenameItemAlreadyExists {
        get {
            return ResourceManager.GetString("RenameItemAlreadyExists", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Registry key name must start with a valid base key name.
    ///  
    /// </summary>
    internal static string Arg_RegInvalidKeyName {
        get {
            return ResourceManager.GetString("Arg_RegInvalidKeyName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The subkey argument is not valid.
    ///  
    /// </summary>
    internal static string ArgRegKeyDelHive {
        get {
            return ResourceManager.GetString("ArgRegKeyDelHive", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot delete a subkey tree because the subkey does not exist.
    ///  
    /// </summary>
    internal static string Arg_RegSubKeyAbsent {
        get {
            return ResourceManager.GetString("Arg_RegSubKeyAbsent", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    No value exists with that name.
    ///  
    /// </summary>
    internal static string Arg_RegSubKeyValueAbsent {
        get {
            return ResourceManager.GetString("Arg_RegSubKeyValueAbsent", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The enum value {0} is not valid.
    ///  
    /// </summary>
    internal static string Arg_EnumIllegalVal {
        get {
            return ResourceManager.GetString("Arg_EnumIllegalVal", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A value argument must be specified.
    ///  
    /// </summary>
    internal static string Arg_Value {
        get {
            return ResourceManager.GetString("Arg_Value", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A name argument must be specified.
    ///  
    /// </summary>
    internal static string Arg_Name {
        get {
            return ResourceManager.GetString("Arg_Name", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified RegistryValueKind is a value that is not valid.
    ///  
    /// </summary>
    internal static string Arg_RegBadKeyKind {
        get {
            return ResourceManager.GetString("Arg_RegBadKeyKind", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    RegistryKey.SetValue does not allow a String[] that contains a null String reference.
    ///  
    /// </summary>
    internal static string Arg_RegSetStrArrNull {
        get {
            return ResourceManager.GetString("Arg_RegSetStrArrNull", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Registry subkeys should not be greater than 255 characters.
    ///  
    /// </summary>
    internal static string Arg_RegKeyStrLenBug {
        get {
            return ResourceManager.GetString("Arg_RegKeyStrLenBug", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A non-empty subkey name must be specified.
    ///  
    /// </summary>
    internal static string Arg_RegKeyStrEmpty {
        get {
            return ResourceManager.GetString("Arg_RegKeyStrEmpty", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The type of the value object did not match the specified RegistryValueKind or the object could not be properly converted.
    ///  
    /// </summary>
    internal static string Arg_RegSetMismatchedKind {
        get {
            return ResourceManager.GetString("Arg_RegSetMismatchedKind", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    RegistryKey.SetValue does not support arrays of type '{0}'. Only Byte[] and String[] are supported.
    ///  
    /// </summary>
    internal static string Arg_RegSetBadArrType {
        get {
            return ResourceManager.GetString("Arg_RegSetBadArrType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified registry key does not exist.
    ///  
    /// </summary>
    internal static string Arg_RegKeyNotFound {
        get {
            return ResourceManager.GetString("Arg_RegKeyNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The length of the specified value name exceeds the maximum of 16383 characters.
    ///  
    /// </summary>
    internal static string Arg_RegValueNameStrLenBug {
        get {
            return ResourceManager.GetString("Arg_RegValueNameStrLenBug", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The size of the specified value data exceeds the maximum of 1 MB.
    ///  
    /// </summary>
    internal static string Arg_ValueDataLenBug {
        get {
            return ResourceManager.GetString("Arg_ValueDataLenBug", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified registry subkey does not exist.
    ///  
    /// </summary>
    internal static string ArgumentException_RegSubKeyAbsent {
        get {
            return ResourceManager.GetString("ArgumentException_RegSubKeyAbsent", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified RegistryKeyPermissionCheck value is not valid.
    ///  
    /// </summary>
    internal static string Argument_InvalidRegistryKeyPermissionCheck {
        get {
            return ResourceManager.GetString("Argument_InvalidRegistryKeyPermissionCheck", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The registry key has subkeys; recursive removals are not supported by this method.
    ///  
    /// </summary>
    internal static string InvalidOperation_RegRemoveSubKey {
        get {
            return ResourceManager.GetString("InvalidOperation_RegRemoveSubKey", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot create a KTM handle without a Transaction.Current or specified transaction.
    ///  
    /// </summary>
    internal static string InvalidOperation_NeedTransaction {
        get {
            return ResourceManager.GetString("InvalidOperation_NeedTransaction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified transaction or Transaction.Current must match the transaction used to create or open this TransactedRegistryKey.
    ///  
    /// </summary>
    internal static string InvalidOperation_MustUseSameTransaction {
        get {
            return ResourceManager.GetString("InvalidOperation_MustUseSameTransaction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The TransactedRegistryKey object is not associated with a transaction because it is for a predefined key.
    ///  
    /// </summary>
    internal static string InvalidOperation_NotAssociatedWithTransaction {
        get {
            return ResourceManager.GetString("InvalidOperation_NotAssociatedWithTransaction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Requested registry access is not allowed.
    ///  
    /// </summary>
    internal static string Security_RegistryPermission {
        get {
            return ResourceManager.GetString("Security_RegistryPermission", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Access to the registry key '{0}' is denied.
    ///  
    /// </summary>
    internal static string UnauthorizedAccess_RegistryKeyGeneric_Key {
        get {
            return ResourceManager.GetString("UnauthorizedAccess_RegistryKeyGeneric_Key", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot write to the registry key.
    ///  
    /// </summary>
    internal static string UnauthorizedAccess_RegistryNoWrite {
        get {
            return ResourceManager.GetString("UnauthorizedAccess_RegistryNoWrite", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot access a closed registry key.
    ///  
    /// </summary>
    internal static string ObjectDisposed_RegKeyClosed {
        get {
            return ResourceManager.GetString("ObjectDisposed_RegKeyClosed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unknown error: {0}.
    ///  
    /// </summary>
    internal static string UnknownError_Num {
        get {
            return ResourceManager.GetString("UnknownError_Num", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Registry transactions are not supported on this platform.
    ///  
    /// </summary>
    internal static string NotSupported_KernelTransactions {
        get {
            return ResourceManager.GetString("NotSupported_KernelTransactions", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified handle is not valid.
    ///  
    /// </summary>
    internal static string AccessControl_InvalidHandle {
        get {
            return ResourceManager.GetString("AccessControl_InvalidHandle", resourceCulture);
        }
    }

}

