
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class CommandLineParameterParserStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal CommandLineParameterParserStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.PowerShell.ConsoleHost.resources.CommandLineParameterParserStrings", typeof(CommandLineParameterParserStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process command because a command is already specified with -Command, -CommandWithArgs, or -EncodedCommand.
    ///  
    /// </summary>
    internal static string CommandAlreadySpecified {
        get {
            return ResourceManager.GetString("CommandAlreadySpecified", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the command because of a missing parameter. A command must follow -Command.
    ///  
    /// </summary>
    internal static string MissingCommandParameter {
        get {
            return ResourceManager.GetString("MissingCommandParameter", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unrecognized parameter: '{0}'.
    ///  
    /// </summary>
    internal static string UnknownParameter {
        get {
            return ResourceManager.GetString("UnknownParameter", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    '-' was specified with the -Command parameter; no other arguments to -Command are permitted.
    ///  
    /// </summary>
    internal static string TooManyParametersToCommand {
        get {
            return ResourceManager.GetString("TooManyParametersToCommand", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    '-' was specified as the argument to -Command but standard input has not been redirected for this process.
    ///  
    /// </summary>
    internal static string StdinNotRedirected {
        get {
            return ResourceManager.GetString("StdinNotRedirected", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The command cannot be run because no argument has been supplied for the OutputFormat parameter.
    ///Specify one of the following formats for this parameter:
    ///{0}
    ///  
    /// </summary>
    internal static string MissingOutputFormatParameter {
        get {
            return ResourceManager.GetString("MissingOutputFormatParameter", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the command because the -InputFormat parameter requires an argument. Specify a valid format argument for this parameter.
    ///Valid formats are:
    ///{0}
    ///  
    /// </summary>
    internal static string MissingInputFormatParameter {
        get {
            return ResourceManager.GetString("MissingInputFormatParameter", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the command because of an incorrect parameter value. "{0}" is not a valid format.
    ///Valid formats are:
    ///{1}
    ///  
    /// </summary>
    internal static string BadFormatParameterValue {
        get {
            return ResourceManager.GetString("BadFormatParameterValue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the command because arguments to -Command or -EncodedCommand have already been specified with -EncodedArguments.
    ///  
    /// </summary>
    internal static string ArgsAlreadySpecified {
        get {
            return ResourceManager.GetString("ArgsAlreadySpecified", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the command because -EncodedArguments requires a value. Specify a value for the -EncodedArguments parameter.
    ///  
    /// </summary>
    internal static string MissingArgsValue {
        get {
            return ResourceManager.GetString("MissingArgsValue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The command cannot be run because the File parameter requires a file path. Supply a path for the File parameter and then try the command again.
    ///  
    /// </summary>
    internal static string MissingFileArgument {
        get {
            return ResourceManager.GetString("MissingFileArgument", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the command because -WindowStyle requires an argument that is normal, hidden, minimized or maximized. Specify one of these argument values and try again.
    ///  
    /// </summary>
    internal static string MissingWindowStyleArgument {
        get {
            return ResourceManager.GetString("MissingWindowStyleArgument", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Processing -File '{0}' failed: {1} Specify a valid path for the -File parameter.
    ///  
    /// </summary>
    internal static string InvalidFileArgument {
        get {
            return ResourceManager.GetString("InvalidFileArgument", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Processing -WindowStyle '{0}' failed: {1}.
    ///  
    /// </summary>
    internal static string InvalidWindowStyleArgument {
        get {
            return ResourceManager.GetString("InvalidWindowStyleArgument", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Processing -File '{0}' failed because the file does not have a '.ps1' extension. Specify a valid PowerShell script file name, and then try again.
    ///  
    /// </summary>
    internal static string InvalidFileArgumentExtension {
        get {
            return ResourceManager.GetString("InvalidFileArgumentExtension", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The argument '{0}' is not recognized as the name of a script file. Check the spelling of the name, or if a path was included, verify that the path is correct and try again.
    ///  
    /// </summary>
    internal static string ArgumentFileDoesNotExist {
        get {
            return ResourceManager.GetString("ArgumentFileDoesNotExist", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the command because the value specified with -EncodedArguments is not properly encoded. The value must be Base64 encoded.
    ///  
    /// </summary>
    internal static string BadArgsValue {
        get {
            return ResourceManager.GetString("BadArgsValue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the command because the value specified with -EncodedCommand is not properly encoded. The value must be Base64 encoded.
    ///  
    /// </summary>
    internal static string BadCommandValue {
        get {
            return ResourceManager.GetString("BadCommandValue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the execution policy because of a missing policy name. A policy name must follow -ExecutionPolicy.
    ///  
    /// </summary>
    internal static string MissingExecutionPolicyParameter {
        get {
            return ResourceManager.GetString("MissingExecutionPolicyParameter", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the command because -STA and -MTA are both specified. Specify either -STA or -MTA.
    ///  
    /// </summary>
    internal static string MtaStaMutuallyExclusive {
        get {
            return ResourceManager.GetString("MtaStaMutuallyExclusive", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the command because -ConfigurationName requires an argument that is a remote endpoint configuration name. Specify this argument and try again.
    ///  
    /// </summary>
    internal static string MissingConfigurationNameArgument {
        get {
            return ResourceManager.GetString("MissingConfigurationNameArgument", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the command because -ConfigurationFile requires an argument that is a session configuration (.pssc) file path. Specify this argument and try again.
    ///  
    /// </summary>
    internal static string MissingConfigurationFileArgument {
        get {
            return ResourceManager.GetString("MissingConfigurationFileArgument", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the command because -CustomPipeName requires an argument that is a name of the pipe you want to use. Specify this argument and try again.
    ///  
    /// </summary>
    internal static string MissingCustomPipeNameArgument {
        get {
            return ResourceManager.GetString("MissingCustomPipeNameArgument", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the command because -CustomPipeName specified is too long. Pipe names on this platform can be up to {0} characters long. Your pipe name '{1}' is {2} characters.
    ///  
    /// </summary>
    internal static string CustomPipeNameTooLong {
        get {
            return ResourceManager.GetString("CustomPipeNameTooLong", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the command because -SettingsFile requires an argument that is a file path.
    ///  
    /// </summary>
    internal static string MissingSettingsFileArgument {
        get {
            return ResourceManager.GetString("MissingSettingsFileArgument", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Processing -SettingsFile '{0}' failed: {1}. Specify a valid path for the -SettingsFile parameter.
    ///  
    /// </summary>
    internal static string InvalidSettingsFileArgument {
        get {
            return ResourceManager.GetString("InvalidSettingsFileArgument", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The argument '{0}' passed to the -SettingsFile does not exist. Provide the path to an existing json file as an argument to the -SettingsFile parameter.
    ///  
    /// </summary>
    internal static string SettingsFileNotExists {
        get {
            return ResourceManager.GetString("SettingsFileNotExists", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Invalid argument '{0}', did you mean:
    ///  
    /// </summary>
    internal static string InvalidArgument {
        get {
            return ResourceManager.GetString("InvalidArgument", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Parameter -WindowStyle is not implemented on this platform.
    ///  
    /// </summary>
    internal static string WindowStyleArgumentNotImplemented {
        get {
            return ResourceManager.GetString("WindowStyleArgumentNotImplemented", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the command because -WorkingDirectory requires an argument that is a directory path.
    ///  
    /// </summary>
    internal static string MissingWorkingDirectoryArgument {
        get {
            return ResourceManager.GetString("MissingWorkingDirectoryArgument", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Parameter -MTA is not supported on this platform.
    ///  
    /// </summary>
    internal static string MTANotImplemented {
        get {
            return ResourceManager.GetString("MTANotImplemented", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Parameter -STA is not supported on this platform.
    ///  
    /// </summary>
    internal static string STANotImplemented {
        get {
            return ResourceManager.GetString("STANotImplemented", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified arguments must not contain null elements.
    ///  
    /// </summary>
    internal static string NullElementInArgs {
        get {
            return ResourceManager.GetString("NullElementInArgs", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Invalid ExecutionPolicy value '{0}'.
    ///  
    /// </summary>
    internal static string InvalidExecutionPolicyArgument {
        get {
            return ResourceManager.GetString("InvalidExecutionPolicyArgument", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    An argument is required to be supplied to the '{0}' parameter.
    ///  
    /// </summary>
    internal static string MissingMandatoryArgument {
        get {
            return ResourceManager.GetString("MissingMandatoryArgument", resourceCulture);
        }
    }

}

