// Copyright (c) Microsoft Corporation.
// Licensed under the MIT License.
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a tool.
//     Runtime Version:4.0.30319.16808
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------

namespace Microsoft.Management.UI.Internal.ShowCommand {


    [global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]
    [global::System.CodeDom.Compiler.GeneratedCodeAttribute("Microsoft.VisualStudio.Editors.SettingsDesigner.SettingsSingleFileGenerator", "10.0.0.0")]
    internal sealed partial class ShowCommandSettings : global::System.Configuration.ApplicationSettingsBase {

        private static ShowCommandSettings defaultInstance = ((ShowCommandSettings)(global::System.Configuration.ApplicationSettingsBase.Synchronized(new ShowCommandSettings())));

        public static ShowCommandSettings Default {
            get {
                return defaultInstance;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("-1")]
        public double ShowOneCommandTop {
            get {
                return ((double)(this["ShowOneCommandTop"]));
            }
            set {
                this["ShowOneCommandTop"] = value;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("-1")]
        public double ShowOneCommandLeft {
            get {
                return ((double)(this["ShowOneCommandLeft"]));
            }
            set {
                this["ShowOneCommandLeft"] = value;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("-1")]
        public double ShowOneCommandWidth {
            get {
                return ((double)(this["ShowOneCommandWidth"]));
            }
            set {
                this["ShowOneCommandWidth"] = value;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("-1")]
        public double ShowOneCommandHeight {
            get {
                return ((double)(this["ShowOneCommandHeight"]));
            }
            set {
                this["ShowOneCommandHeight"] = value;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("-1")]
        public double ShowCommandsTop {
            get {
                return ((double)(this["ShowCommandsTop"]));
            }
            set {
                this["ShowCommandsTop"] = value;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("-1")]
        public double ShowCommandsLeft {
            get {
                return ((double)(this["ShowCommandsLeft"]));
            }
            set {
                this["ShowCommandsLeft"] = value;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("-1")]
        public double ShowCommandsWidth {
            get {
                return ((double)(this["ShowCommandsWidth"]));
            }
            set {
                this["ShowCommandsWidth"] = value;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("-1")]
        public double ShowCommandsHeight {
            get {
                return ((double)(this["ShowCommandsHeight"]));
            }
            set {
                this["ShowCommandsHeight"] = value;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("False")]
        public bool ShowCommandsWindowMaximized {
            get {
                return ((bool)(this["ShowCommandsWindowMaximized"]));
            }
            set {
                this["ShowCommandsWindowMaximized"] = value;
            }
        }

        [global::System.Configuration.UserScopedSettingAttribute()]
        [global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
        [global::System.Configuration.DefaultSettingValueAttribute("False")]
        public bool ShowOneCommandWindowMaximized {
            get {
                return ((bool)(this["ShowOneCommandWindowMaximized"]));
            }
            set {
                this["ShowOneCommandWindowMaximized"] = value;
            }
        }
    }
}
