
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class CertificateProviderStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal CertificateProviderStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.PowerShell.Security.resources.CertificateProviderStrings", typeof(CertificateProviderStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    X509 Certificate Provider
    ///  
    /// </summary>
    internal static string CertProvidername {
        get {
            return ResourceManager.GetString("CertProvidername", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot find the X509 certificate at path {0}.
    ///  
    /// </summary>
    internal static string CertificateNotFound {
        get {
            return ResourceManager.GetString("CertificateNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot find the X509 certificate store at path {0}.
    ///  
    /// </summary>
    internal static string CertificateStoreNotFound {
        get {
            return ResourceManager.GetString("CertificateStoreNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot find the certificate store because the specified X509 store location {0} is not valid.
    ///  
    /// </summary>
    internal static string CertificateStoreLocationNotFound {
        get {
            return ResourceManager.GetString("CertificateStoreLocationNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the path because path {0} is not a valid certificate provider path.
    ///  
    /// </summary>
    internal static string InvalidPath {
        get {
            return ResourceManager.GetString("InvalidPath", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Move certificate
    ///  
    /// </summary>
    internal static string Action_Move {
        get {
            return ResourceManager.GetString("Action_Move", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Remove certificate
    ///  
    /// </summary>
    internal static string Action_Remove {
        get {
            return ResourceManager.GetString("Action_Remove", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Remove certificate and its private key.
    ///  
    /// </summary>
    internal static string Action_RemoveAndDeleteKey {
        get {
            return ResourceManager.GetString("Action_RemoveAndDeleteKey", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Invoke Certificate Manager
    ///  
    /// </summary>
    internal static string Action_Invoke {
        get {
            return ResourceManager.GetString("Action_Invoke", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///      Item: {0} Destination: {1} 
    ///  
    /// </summary>
    internal static string MoveItemTemplate {
        get {
            return ResourceManager.GetString("MoveItemTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///      You cannot move a certificate container. 
    ///  
    /// </summary>
    internal static string CannotMoveContainer {
        get {
            return ResourceManager.GetString("CannotMoveContainer", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///      You cannot move a certificate from user store to or from machine. 
    ///  
    /// </summary>
    internal static string CannotMoveCrossContext {
        get {
            return ResourceManager.GetString("CannotMoveCrossContext", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    You cannot move a certificate to the same store. 
    ///  
    /// </summary>
    internal static string CannotMoveToSameStore {
        get {
            return ResourceManager.GetString("CannotMoveToSameStore", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    You cannot create an item other than certificate store. 
    ///  
    /// </summary>
    internal static string CannotCreateItem {
        get {
            return ResourceManager.GetString("CannotCreateItem", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Creating certificate stores under CurrentUser is not supported. 
    ///  
    /// </summary>
    internal static string CannotCreateUserStore {
        get {
            return ResourceManager.GetString("CannotCreateUserStore", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Deleting certificate stores under CurrentUser is not supported. 
    ///  
    /// </summary>
    internal static string CannotDeleteUserStore {
        get {
            return ResourceManager.GetString("CannotDeleteUserStore", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The destination is not a valid store. 
    ///  
    /// </summary>
    internal static string InvalidDestStore {
        get {
            return ResourceManager.GetString("InvalidDestStore", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Item: {0} 
    ///  
    /// </summary>
    internal static string RemoveItemTemplate {
        get {
            return ResourceManager.GetString("RemoveItemTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The store {0} is a built-in system store and cannot be deleted.
    ///  
    /// </summary>
    internal static string RemoveStoreTemplate {
        get {
            return ResourceManager.GetString("RemoveStoreTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    You cannot remove a certificate container. 
    ///  
    /// </summary>
    internal static string CannotRemoveContainer {
        get {
            return ResourceManager.GetString("CannotRemoveContainer", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Private key skipped. The certificate has no private key association. 
    ///  
    /// </summary>
    internal static string VerboseNoPrivateKey {
        get {
            return ResourceManager.GetString("VerboseNoPrivateKey", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The operation is on user root store and UI is not allowed. 
    ///  
    /// </summary>
    internal static string UINotAllowed {
        get {
            return ResourceManager.GetString("UINotAllowed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    . The following error may be a result of user credentials required on the remote machine. See Enable-WSManCredSSP Cmdlet help on how to enable and use CredSSP for delegation with PowerShell remoting. 
    ///  
    /// </summary>
    internal static string RemoteErrorMessage {
        get {
            return ResourceManager.GetString("RemoteErrorMessage", resourceCulture);
        }
    }

}

