
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class PowerShellStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal PowerShellStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.PowerShellStrings", typeof(PowerShellStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The state of the current PowerShell instance is not valid for this operation.
    ///  
    /// </summary>
    internal static string InvalidPowerShellStateGeneral {
        get {
            return ResourceManager.GetString("InvalidPowerShellStateGeneral", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The operation cannot be performed because a command has already been started. Wait for the command to complete, or stop it, and then try the operation again.
    ///  
    /// </summary>
    internal static string ExecutionAlreadyStarted {
        get {
            return ResourceManager.GetString("ExecutionAlreadyStarted", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    No commands are specified.
    ///  
    /// </summary>
    internal static string NoCommandToInvoke {
        get {
            return ResourceManager.GetString("NoCommandToInvoke", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The PowerShell instance is not in the correct state for creating a nested PowerShell instance. Nested PowerShell instances should only be created in a running PowerShell instance.
    ///  
    /// </summary>
    internal static string InvalidStateCreateNested {
        get {
            return ResourceManager.GetString("InvalidStateCreateNested", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot perform operation because the runspace is not in the '{0}' state. Current state of runspace is '{1}'.
    ///  
    /// </summary>
    internal static string InvalidRunspaceState {
        get {
            return ResourceManager.GetString("InvalidRunspaceState", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Nested PowerShell instances cannot be invoked asynchronously. Use the Invoke method.
    ///  
    /// </summary>
    internal static string NestedPowerShellInvokeAsync {
        get {
            return ResourceManager.GetString("NestedPowerShellInvokeAsync", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The {0} object was not created by calling {1} on this PowerShell instance.
    ///  
    /// </summary>
    internal static string AsyncResultNotOwned {
        get {
            return ResourceManager.GetString("AsyncResultNotOwned", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    When the runspace is set to reuse a thread, the apartment state in the invocation settings must match the runspace.
    ///  
    /// </summary>
    internal static string ApartmentStateMismatch {
        get {
            return ResourceManager.GetString("ApartmentStateMismatch", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    When the runspace is set to use the current thread, the apartment state in the invocation settings must match that of the current thread.
    ///  
    /// </summary>
    internal static string ApartmentStateMismatchCurrentThread {
        get {
            return ResourceManager.GetString("ApartmentStateMismatchCurrentThread", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A command is required to add a parameter. A command must be added to the PowerShell instance before adding a parameter.
    ///  
    /// </summary>
    internal static string ParameterRequiresCommand {
        get {
            return ResourceManager.GetString("ParameterRequiresCommand", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The keys in the dictionary must be strings.
    ///  
    /// </summary>
    internal static string KeyMustBeString {
        get {
            return ResourceManager.GetString("KeyMustBeString", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    There is no Runspace available to run commands in this thread. You can provide one in the DefaultRunspace property of the System.Management.Automation.Runspaces.Runspace type. The command you attempted to invoke was: {0}
    ///  
    /// </summary>
    internal static string CommandInvokedFromWrongThreadWithCommand {
        get {
            return ResourceManager.GetString("CommandInvokedFromWrongThreadWithCommand", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    This PowerShell object cannot be connected because it is not associated with a remote runspace or runspace pool.
    ///  
    /// </summary>
    internal static string CannotConnect {
        get {
            return ResourceManager.GetString("CannotConnect", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The running command has been disconnected but is still running on the remote server.  Reconnect to get command operation status and output data.
    ///  
    /// </summary>
    internal static string DiscOnSyncCommand {
        get {
            return ResourceManager.GetString("DiscOnSyncCommand", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The operation cannot be performed because the current PowerShell session is in the Disconnected state.  Connect this PowerShell session, and then either wait for the command to finish, or stop the command.
    ///  
    /// </summary>
    internal static string ExecutionDisconnected {
        get {
            return ResourceManager.GetString("ExecutionDisconnected", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The operation cannot be performed because the current PowerShell session is in the Disconnected state.  Connect this PowerShell session, and then try again.
    ///  
    /// </summary>
    internal static string IsDisconnected {
        get {
            return ResourceManager.GetString("IsDisconnected", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The connection attempt to the remote command failed.
    ///  
    /// </summary>
    internal static string ConnectFailed {
        get {
            return ResourceManager.GetString("ConnectFailed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The operation cannot be performed because a command is currently stopping. Wait for the command to complete stopping, and then try the operation again.
    ///  
    /// </summary>
    internal static string ExecutionStopping {
        get {
            return ResourceManager.GetString("ExecutionStopping", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    There is no Runspace available to run commands in this thread. You can provide one in the DefaultRunspace property of the System.Management.Automation.Runspaces.Runspace type. The current PowerShell instance contains no command to invoke.
    ///  
    /// </summary>
    internal static string CommandInvokedFromWrongThreadWithoutCommand {
        get {
            return ResourceManager.GetString("CommandInvokedFromWrongThreadWithoutCommand", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A PowerShell object cannot be created that uses the current runspace because there is no current runspace available.  The current runspace might be starting, such as when it is created with an Initial Session State.
    ///  
    /// </summary>
    internal static string NoDefaultRunspaceForPSCreate {
        get {
            return ResourceManager.GetString("NoDefaultRunspaceForPSCreate", resourceCulture);
        }
    }

}

