
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class ParameterBinderStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal ParameterBinderStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.ParameterBinderStrings", typeof(ParameterBinderStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A parameter cannot be found that matches parameter name '{1}'.
    ///  
    /// </summary>
    internal static string NamedParameterNotFound {
        get {
            return ResourceManager.GetString("NamedParameterNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A positional parameter cannot be found that accepts argument '{1}'.
    ///  
    /// </summary>
    internal static string PositionalParameterNotFound {
        get {
            return ResourceManager.GetString("PositionalParameterNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Missing an argument for parameter '{1}'. Specify a parameter of type '{2}' and try again.
    ///  
    /// </summary>
    internal static string MissingArgument {
        get {
            return ResourceManager.GetString("MissingArgument", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Parameter cannot be processed because the parameter name '{1}' is ambiguous. Possible matches include:{6}.
    ///  
    /// </summary>
    internal static string AmbiguousParameter {
        get {
            return ResourceManager.GetString("AmbiguousParameter", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot convert '{6}' to the type '{2}' required by parameter '{1}'. {7}
    ///  
    /// </summary>
    internal static string CannotConvertArgument {
        get {
            return ResourceManager.GetString("CannotConvertArgument", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot bind parameter '{1}'. {6}
    ///  
    /// </summary>
    internal static string CannotConvertArgumentNoMessage {
        get {
            return ResourceManager.GetString("CannotConvertArgumentNoMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot bind positional parameters '{1}'.
    ///  
    /// </summary>
    internal static string AmbiguousPositionalParameter {
        get {
            return ResourceManager.GetString("AmbiguousPositionalParameter", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot bind positional parameters because no names were given.
    ///  
    /// </summary>
    internal static string AmbiguousPositionalParameterNoName {
        get {
            return ResourceManager.GetString("AmbiguousPositionalParameterNoName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Parameter set cannot be resolved using the specified named parameters. One or more parameters issued cannot be used together or an insufficient number of parameters were provided.
    ///  
    /// </summary>
    internal static string AmbiguousParameterSet {
        get {
            return ResourceManager.GetString("AmbiguousParameterSet", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process command because of one or more missing mandatory parameters:{1}.
    ///  
    /// </summary>
    internal static string MissingMandatoryParameter {
        get {
            return ResourceManager.GetString("MissingMandatoryParameter", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Parameter '{1}' cannot be specified in parameter set '{6}'.
    ///  
    /// </summary>
    internal static string ParameterNotInParameterSet {
        get {
            return ResourceManager.GetString("ParameterNotInParameterSet", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot bind parameter because parameter '{1}' is specified more than once. To provide multiple values to parameters that can accept multiple values, use the array syntax. For example, "-parameter value1,value2,value3".
    ///  
    /// </summary>
    internal static string ParameterAlreadyBound {
        get {
            return ResourceManager.GetString("ParameterAlreadyBound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot evaluate parameter '{1}' because its argument is specified as a script block and there is no input. A script block cannot be evaluated without input.
    ///  
    /// </summary>
    internal static string ScriptBlockArgumentNoInput {
        get {
            return ResourceManager.GetString("ScriptBlockArgumentNoInput", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The input to the script block for parameter '{1}' failed. {6}
    ///  
    /// </summary>
    internal static string ScriptBlockArgumentInvocationFailed {
        get {
            return ResourceManager.GetString("ScriptBlockArgumentInvocationFailed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot evaluate parameter '{1}' because its argument input did not produce any output.
    ///  
    /// </summary>
    internal static string ScriptBlockArgumentNoOutput {
        get {
            return ResourceManager.GetString("ScriptBlockArgumentNoOutput", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The input object cannot be bound to any parameters for the command either because the command does not take pipeline input or the input and its properties do not match any of the parameters that take pipeline input.
    ///  
    /// </summary>
    internal static string InputObjectNotBound {
        get {
            return ResourceManager.GetString("InputObjectNotBound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The input object cannot be bound because it did not contain the information required to bind all mandatory parameters: {6}
    ///  
    /// </summary>
    internal static string InputObjectMissingMandatory {
        get {
            return ResourceManager.GetString("InputObjectMissingMandatory", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Pipeline input cannot be processed because the default value of parameter '{1}' cannot be retrieved. {6}
    ///  
    /// </summary>
    internal static string GetDefaultValueFailed {
        get {
            return ResourceManager.GetString("GetDefaultValueFailed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot retrieve the dynamic parameters for the cmdlet. {6}
    ///  
    /// </summary>
    internal static string GetDynamicParametersException {
        get {
            return ResourceManager.GetString("GetDynamicParametersException", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Supply values for the following parameters:
    ///  
    /// </summary>
    internal static string PromptMessage {
        get {
            return ResourceManager.GetString("PromptMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    cmdlet {0} at command pipeline position {1}
    ///  
    /// </summary>
    internal static string PromptCaption {
        get {
            return ResourceManager.GetString("PromptCaption", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process argument transformation on parameter '{1}'. {6}
    ///  
    /// </summary>
    internal static string ParameterArgumentTransformationError {
        get {
            return ResourceManager.GetString("ParameterArgumentTransformationError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {6}
    ///  
    /// </summary>
    internal static string ParameterArgumentTransformationErrorMessageOnly {
        get {
            return ResourceManager.GetString("ParameterArgumentTransformationErrorMessageOnly", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot validate argument on parameter '{1}'. {6}
    ///  
    /// </summary>
    internal static string ParameterArgumentValidationError {
        get {
            return ResourceManager.GetString("ParameterArgumentValidationError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot bind parameter '{1}' to the target. {6}
    ///  
    /// </summary>
    internal static string ParameterBindingFailed {
        get {
            return ResourceManager.GetString("ParameterBindingFailed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot bind argument to parameter '{1}' because it is null.
    ///  
    /// </summary>
    internal static string ParameterArgumentValidationErrorNullNotAllowed {
        get {
            return ResourceManager.GetString("ParameterArgumentValidationErrorNullNotAllowed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot bind argument to parameter '{1}' because it is an empty string.
    ///  
    /// </summary>
    internal static string ParameterArgumentValidationErrorEmptyStringNotAllowed {
        get {
            return ResourceManager.GetString("ParameterArgumentValidationErrorEmptyStringNotAllowed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot bind argument to parameter '{1}' because it is an empty collection.
    ///  
    /// </summary>
    internal static string ParameterArgumentValidationErrorEmptyCollectionNotAllowed {
        get {
            return ResourceManager.GetString("ParameterArgumentValidationErrorEmptyCollectionNotAllowed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot bind argument to parameter '{1}' because it is an empty array.
    ///  
    /// </summary>
    internal static string ParameterArgumentValidationErrorEmptyArrayNotAllowed {
        get {
            return ResourceManager.GetString("ParameterArgumentValidationErrorEmptyArrayNotAllowed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process command. The parameter '{0}' is defined multiple times.
    ///  
    /// </summary>
    internal static string DuplicateParameterDefinition {
        get {
            return ResourceManager.GetString("DuplicateParameterDefinition", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot bind cmdlet {0} because parameter '{1}' is of type '{2}' and the Add() method cannot be identified, or multiple Add() methods exist. {6}
    ///  
    /// </summary>
    internal static string CannotExtractAddMethod {
        get {
            return ResourceManager.GetString("CannotExtractAddMethod", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot bind cmdlet {0} because runtime-defined parameter '{1}' was added to the RuntimeDefinedParameterDictionary with key '{6}'. The key must be the same as RuntimeDefinedParameter.Name.
    ///  
    /// </summary>
    internal static string RuntimeDefinedParameterNameMismatch {
        get {
            return ResourceManager.GetString("RuntimeDefinedParameterNameMismatch", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot bind argument to parameter '{1}', because PSTypeNames of the argument do not match the PSTypeName required by the parameter: {6}.
    ///  
    /// </summary>
    internal static string MismatchedPSTypeName {
        get {
            return ResourceManager.GetString("MismatchedPSTypeName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Multiple different default values are defined in $PSDefaultParameterValues for the parameter matching the following name or alias: {0}. These defaults have been ignored.
    ///  
    /// </summary>
    internal static string DifferentValuesAssignedToSingleParameter {
        get {
            return ResourceManager.GetString("DifferentValuesAssignedToSingleParameter", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The following name or alias defined in $PSDefaultParameterValues for this cmdlet resolves to multiple parameters: {0}. The default has been ignored.
    ///  
    /// </summary>
    internal static string MultipleParametersMatched {
        get {
            return ResourceManager.GetString("MultipleParametersMatched", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {6} This error might have been caused by applying the default parameter binding. You can disable the default parameter binding in $PSDefaultParameterValues by setting $PSDefaultParameterValues["Disabled"] to be $true, and then trying again. The following default parameters were successfully bound for this cmdlet when the error occurred:{7}
    ///  
    /// </summary>
    internal static string DefaultBindingErrorElaborationMultiple {
        get {
            return ResourceManager.GetString("DefaultBindingErrorElaborationMultiple", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {6} This failure might be caused by applying the default parameter binding. You can disable the default parameter binding in $PSDefaultParameterValues by setting $PSDefaultParameterValues["Disabled"] to be $true, and retry. The following default parameter was successfully bound for this cmdlet when the error occurred:{7}
    ///  
    /// </summary>
    internal static string DefaultBindingErrorElaborationSingle {
        get {
            return ResourceManager.GetString("DefaultBindingErrorElaborationSingle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The binding of default value '{0}' to parameter '{1}' failed: {2}
    ///  
    /// </summary>
    internal static string FailToBindDefaultParameter {
        get {
            return ResourceManager.GetString("FailToBindDefaultParameter", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The key '{0}' does not have a valid format. For information about the correct format, see about_Parameters_Default_Values at https://go.microsoft.com/fwlink/?LinkId=228266.
    ///  
    /// </summary>
    internal static string SingleKeyInBadFormat {
        get {
            return ResourceManager.GetString("SingleKeyInBadFormat", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The keys '{0}' do not have valid formats. For information about the correct format, see about_Parameters_Default_Values at https://go.microsoft.com/fwlink/?LinkId=228266.
    ///  
    /// </summary>
    internal static string MultipleKeysInBadFormat {
        get {
            return ResourceManager.GetString("MultipleKeysInBadFormat", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Parameter '{0}' is obsolete. {1}
    ///  
    /// </summary>
    internal static string UseOfDeprecatedParameterWarning {
        get {
            return ResourceManager.GetString("UseOfDeprecatedParameterWarning", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The key '{0}' of type '{1}' is not a string value. DefaultParameterDictionary only accepts string value keys.
    ///  
    /// </summary>
    internal static string StringValueKeyExpected {
        get {
            return ResourceManager.GetString("StringValueKeyExpected", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The key '{0}' has already been added to the dictionary.
    ///  
    /// </summary>
    internal static string KeyAlreadyAdded {
        get {
            return ResourceManager.GetString("KeyAlreadyAdded", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Method or Property Invocation Not Allowed
    ///  
    /// </summary>
    internal static string WDACBinderInvocationLogTitle {
        get {
            return ResourceManager.GetString("WDACBinderInvocationLogTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Invocation of Method or Property '{0}' on type '{1}' will not be allowed in Constrained Language mode for untrusted scripts.
    ///  
    /// </summary>
    internal static string WDACBinderInvocationLogMessage {
        get {
            return ResourceManager.GetString("WDACBinderInvocationLogMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Type Creation Not Allowed
    ///  
    /// </summary>
    internal static string WDACBinderTypeCreationLogTitle {
        get {
            return ResourceManager.GetString("WDACBinderTypeCreationLogTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Creation of Type '{0}' will not be allowed during parameter binding in Constrained Language mode for untrusted scripts.
    ///  
    /// </summary>
    internal static string WDACBinderTypeCreationLogMessage {
        get {
            return ResourceManager.GetString("WDACBinderTypeCreationLogMessage", resourceCulture);
        }
    }

}

