
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class EventResource {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal EventResource() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.EventResource", typeof(EventResource).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A message was not found for event id PowerShell.Core.Instrumentation.man.
    ///
    /// </summary>
    internal static string MissingEventIdMessage {
        get {
            return ResourceManager.GetString("MissingEventIdMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Scheduled Job {0} started at {1}
    ///
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_ScheduledJobStarted {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_ScheduledJobStarted", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Scheduled Job {0} completed at {1} with state {2}
    ///
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_ScheduledJobCompleted {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_ScheduledJobCompleted", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Scheduled Job Exception {0}:
    /// Message: {1}
    /// StackTrace: {2}
    /// InnerException: {3}
    ///
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_ScheduledJobException {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_ScheduledJobException", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Experimental Feature Initialization: Ignore the experimental feature '{0}' from the config file. {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_ExperimentalFeatureInvalidName {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_ExperimentalFeatureInvalidName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Experimental Feature Initialization: Failed to read the config file.
    /// Exception: {0}
    /// Message: {1}
    /// StackTrace: {2}
    ///
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_ExperimentalFeatureReadConfigError {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_ExperimentalFeatureReadConfigError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Workflow plugin loaded.
    /// 	 EndpointName: {0}
    /// 	 User: {1}
    /// 	 HostingMode: {2}
    /// 	 Protocol: {3}
    /// 	 Configuration:
    /// {4}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowPluginStarted {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowPluginStarted", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Workflow execution started.
    /// 	 WorkflowId: {0}
    /// 	 ManagedNodes: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowExecutionStarted {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowExecutionStarted", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Workflow state changed.
    /// 	 WorkflowId: {0}
    /// 	 NewState: {1}
    /// 	 OldState: {2}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowStateChanged {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowStateChanged", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Workflow plugin has been requested for a shutdown.
    /// 	 EndpointName: {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowPluginRequestedToShutdown {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowPluginRequestedToShutdown", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Workflow plugin restarted.
    /// 	 EndpointName: {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowPluginRestarted {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowPluginRestarted", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Workflow is resuming.
    /// 	 WorkflowId: {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowWorkflowsResuming {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowWorkflowsResuming", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A quota limit that was set for the endpoint was exceeded.
    /// 	 EndpointName: {0}
    /// 	 ConfigName: {1}
    /// 	 AllowedValue: {2}
    /// 	 ValueInQuestion: {3}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowQuotaViolationDetected {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowQuotaViolationDetected", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Workflow has resumed.
    /// 	 WorkflowId: {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowWorkflowsResumed {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowWorkflowsResumed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Workflow runspace pool was created.
    /// 	 WorkflowId: {0}
    /// 	 ManagedNode: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowRunspacePoolCreated {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowRunspacePoolCreated", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Activity was queued for execution.
    /// 	 WorkflowId: {0}
    /// 	 ActivityName: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowActivityExecutionQueued {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowActivityExecutionQueued", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Activity execution started.
    /// 	 ActivityName: {0}
    /// 	 ActivityTypeName: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowActivityExecutionStarted {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowActivityExecutionStarted", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Workflow is being imported from a XAML file.
    /// 	 WorkflowId: {0}
    /// 	 XamlFile: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowImportingFromXaml {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowImportingFromXaml", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Workflow has been imported from a XAML file.
    /// 	 WorkflowId: {0}
    /// 	 XamlFile: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowImportedFromXaml {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowImportedFromXaml", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Workflow could not be imported from a XAML file because of an error.
    /// 	 WorkflowId: {0}
    /// 	 ErrorDescription: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowImportFromXamlError {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowImportFromXamlError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Workflow validation started.
    /// 	 WorkflowId: {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowImportFromXamlValidationStarted {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowImportFromXamlValidationStarted", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Workflow validation succeeded.
    /// 	 WorkflowId: {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowImportFromXamlValidationFinishedSuccessfully {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowImportFromXamlValidationFinishedSuccessfully", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Workflow validation failed with error.
    /// 	 WorkflowId: {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowImportFromXamlValidationFinishedWithError {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowImportFromXamlValidationFinishedWithError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Workflow activity validated.
    /// 	 WorkflowId: {0}
    /// 	 ActivityDisplayName: {1}
    /// 	 ActivityTypeName: {2}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowImportFromXamlActivityValidated {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowImportFromXamlActivityValidated", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Workflow activity could not be validated.
    /// 	 WorkflowId: {0}
    /// 	 ActivityDisplayName: {1}
    /// 	 ActivityTypeName: {2}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowImportFromXamlActivityValidationFailed {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowImportFromXamlActivityValidationFailed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Activity execution failed.
    /// 	 WorkflowId: {0}
    /// 	 ActivityName: {1}
    /// 	 FailureDescription: {2}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowActivityExecutionFailed {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowActivityExecutionFailed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Runspace availability changed.
    /// 	 RunspaceId: {0}
    /// 	 Availability: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowRunspaceAvailabilityChanged {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowRunspaceAvailabilityChanged", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Runspace state changed.
    /// 	 RunspaceId: {0}
    /// 	 NewState: {1}
    /// 	 OldState: {2}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowRunspaceStateChanged {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowRunspaceStateChanged", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Workflow loaded for execution.
    /// 	 WorkflowId: {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowLoadedForExecution {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowLoadedForExecution", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Workflow unloaded.
    /// 	 WorkflowId: {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowUnloaded {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowUnloaded", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Workflow execution cancelled.
    /// 	 WorkflowId: {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowCancelled {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowCancelled", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Workflow execution aborted.
    /// 	 WorkflowId: {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowAborted {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowAborted", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Workflow cleanup operation executed.
    /// 	 WorkflowId: {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowCleanup {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowCleanup", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Persisted workflow loaded from disk.
    /// 	 WorkflowId: {0}
    /// 	 Path: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowLoadedFromDisk {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowLoadedFromDisk", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Workflow data was deleted from disk.
    /// 	 WorkflowId: {0}
    /// 	 Path: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowDeletedFromDisk {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowDeletedFromDisk", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Starting remove job.
    /// 	 JobId: {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PRemoveJobStarted {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PRemoveJobStarted", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Job state changed.
    /// 	 JobId: {0}
    /// 	 WorkflowId: {1}
    /// 	 NewState: {2}
    /// 	 OldState: {3}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PJobStateChanged {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PJobStateChanged", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Job error.
    /// 	 JobId: {0}
    /// 	 WorkflowId: {1}
    /// 	 ErrorDescription: {2}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PJobError {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PJobError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Job created for workflow (child job).
    /// 	 ParentJobId: {0}
    /// 	 ChildJobId: {1}
    /// 	 ChildWorkflowId: {2}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowJobCreated {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowJobCreated", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Parent job created for workflow.
    /// 	 JobId: {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PParentJobCreated {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PParentJobCreated", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    All required jobs were created for workflow execution.
    /// 	 JobId: {0}
    /// 	 WorkflowId: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PJobCreationCompleted {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PJobCreationCompleted", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Child job removed for workflow.
    /// 	 ParentJobId: {0}
    /// 	 ChildJobId: {1}
    /// 	 WorkflowId: {2}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PJobRemoved {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PJobRemoved", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    An error occurred while removing job.
    /// 	 ParentJobId: {0}
    /// 	 ChildJobId: {1}
    /// 	 WorkflowId: {2}
    /// 	 Error: {3}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PJobRemoveError {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PJobRemoveError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Loading workflow for execution.
    /// 	 WorkflowId: {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PLoadingWorkflowForExecution {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PLoadingWorkflowForExecution", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Workflow execution finished.
    /// 	 WorkflowId: {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowExecutionFinished {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowExecutionFinished", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cancelling workflow execution.
    /// 	 WorkflowId: {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PCancellingWorkflowExecution {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PCancellingWorkflowExecution", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Aborting workflow execution.
    /// 	 WorkflowId: {0}
    /// 	 Reason: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PAbortingWorkflowExecution {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PAbortingWorkflowExecution", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unloading workflow.
    /// 	 WorkflowId: {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PUnloadingWorkflow {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PUnloadingWorkflow", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Forced workflow shutdown started.
    /// 	 WorkflowId: {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PForcedWorkflowShutdownStarted {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PForcedWorkflowShutdownStarted", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Forced workflow shutdown finished.
    /// 	 WorkflowId: {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PForcedWorkflowShutdownFinished {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PForcedWorkflowShutdownFinished", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    An error occurred while forcefully shutting down a workflow.
    /// 	 WorkflowId: {0}
    /// 	 ErrorDescription: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PForcedWorkflowShutdownError {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PForcedWorkflowShutdownError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Persisting workflow to disk.
    /// 	 WorkflowId: {0}
    /// 	 PersistPath: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PPersistingWorkflow {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PPersistingWorkflow", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Workflow persisted to disk.
    /// 	 WorkflowId: {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowPersisted {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowPersisted", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Activity execution finished.
    /// 	 ActivityName: {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowActivityExecutionFinished {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowActivityExecutionFinished", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Workflow execution error.
    /// 	 WorkflowId: {0}
    /// 	 ErrorDescription: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowExecutionError {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowExecutionError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A new PowerShell endpoint was registered.
    /// 	 EndpointName: {0}
    /// 	 EndpointType: {1}
    /// 	 RegisteredBy: {2}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PEndpointRegistered {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PEndpointRegistered", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Endpoint configuration modified.
    /// 	 EndpointName: {0}
    /// 	 ModifiedBy: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PEndpointModified {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PEndpointModified", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Endpoint configuration unregistered.
    /// 	 EndpointName: {0}
    /// 	 UnregisteredBy: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PEndpointUnregistered {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PEndpointUnregistered", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Endpoint configuration disabled.
    /// 	 EndpointName: {0}
    /// 	 DisabledBy: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PEndpointDisabled {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PEndpointDisabled", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Endpoint configuration enabled.
    /// 	 EndpointName: {0}
    /// 	 EnabledBy: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PEndpointEnabled {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PEndpointEnabled", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Out of process runspace started.
    /// 	 Command: {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3POutOfProcessRunspaceStarted {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3POutOfProcessRunspaceStarted", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Parameter splatting was performed during workflow execution.
    /// 	 Parameters: {0}
    /// 	 Computers: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PParameterSplattingWasPerformed {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PParameterSplattingWasPerformed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Workflow engine started.
    /// 	 EndpointName: {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PWorkflowEngineStarted {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PWorkflowEngineStarted", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Workflow manager instantiated with
    /// 	 CheckpointPath: {0}
    /// 	 ConfigProviderId: {1}
    /// 	 UserName: {2}
    /// 	 Path: {3}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_M3PWORKFLOW_MANAGER_CHECKPOINTPATH {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_M3PWORKFLOW_MANAGER_CHECKPOINTPATH", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Computer Name $null or . resolve to LocalHost
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_CMDLETS_HOSTNAMERESOLVE {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_CMDLETS_HOSTNAMERESOLVE", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Resolving to default scheme http
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_CMDLETS_SCHEMERESOLVE {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_CMDLETS_SCHEMERESOLVE", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Remote shell name resolved to default PowerShellCore
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_CMDLETS_SHELLRESOLVE {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_CMDLETS_SHELLRESOLVE", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {2}
    ///
    ///Context:
    ///{0}
    ///
    ///User Data:
    ///{1}
    ///
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_COMMAND_HEALTH {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_COMMAND_HEALTH", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {2}
    ///
    ///Context:
    ///{0}
    ///
    ///User Data:
    ///{1}
    ///
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_ENGINE_HEALTH {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_ENGINE_HEALTH", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {2}
    ///
    ///Context:
    ///{0}
    ///
    ///User Data:
    ///{1}
    ///
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_PROVIDER_HEALTH {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_PROVIDER_HEALTH", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {2}
    ///
    ///Context:
    ///{0}
    ///
    ///User Data:
    ///{1}
    ///
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_PIPELINE_DETAIL {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_PIPELINE_DETAIL", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Creating Scriptblock text ({0} of {1}):
    ///{2}
    ///
    ///ScriptBlock ID: {3}
    ///Path: {4}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_SCRIPTBLOCK_CREATE_DETAIL {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_SCRIPTBLOCK_CREATE_DETAIL", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Started invocation of ScriptBlock ID: {0}
    ///Runspace ID: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_SCRIPTBLOCK_INVOKE_START_DETAIL {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_SCRIPTBLOCK_INVOKE_START_DETAIL", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Completed invocation of ScriptBlock ID: {0}
    ///Runspace ID: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_SCRIPTBLOCK_INVOKE_COMPLETE_DETAIL {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_SCRIPTBLOCK_INVOKE_COMPLETE_DETAIL", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {2}
    ///
    ///Context:
    ///{0}
    ///
    ///User Data:
    ///{1}
    ///
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_COMMAND_LIFECYCLE {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_COMMAND_LIFECYCLE", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {2}
    ///
    ///Context:
    ///{0}
    ///
    ///User Data:
    ///{1}
    ///
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_ENGINE_LIFECYCLE {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_ENGINE_LIFECYCLE", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {2}
    ///
    ///Context:
    ///{0}
    ///
    ///User Data:
    ///{1}
    ///
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_PROVIDER_LIFECYCLE {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_PROVIDER_LIFECYCLE", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {2}
    ///
    ///Context:
    ///{0}
    ///
    ///User Data:
    ///{1}
    ///
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_SETTINGS {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_SETTINGS", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Correlating activity id's.
    /// 	 CurrentActivityId: {0}
    /// 	 ParentActivityId: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_WriteTransferEvent {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_WriteTransferEvent", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Class Name = {0}
    ///Method Name = {1}
    ///Workflow GUID = {2}
    ///Message = {3}
    ///{4}
    ///Activity Name = {5}
    ///Activity GUID = {6}
    ///Parameters = {7}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_ENGINE_TRACE {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_ENGINE_TRACE", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Creating Runspace object
    /// 	 Instance Id: {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_RUNSPACE_CONSTRUCTOR {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_RUNSPACE_CONSTRUCTOR", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Creating RunspacePool object
    /// 	 InstanceId {0}
    /// 	 MinRunspaces {1}
    /// 	 MaxRunspaces {2}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_RUNSPACEPOOL_CONSTRUCTOR {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_RUNSPACEPOOL_CONSTRUCTOR", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Opening RunspacePool
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_RUNSPACEPOOL_OPEN {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_RUNSPACEPOOL_OPEN", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Modifying activity Id and correlating
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_RUNSPACEPOOL_TRANSFER {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_RUNSPACEPOOL_TRANSFER", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Runspace state changed to {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_RUNSPACE_STATE_CHANGE {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_RUNSPACE_STATE_CHANGE", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Attempting session creation retry {0} for error code {1} on session Id {2}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_REMOTE_RUNSPACE_CREATE_RETRY {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_REMOTE_RUNSPACE_CREATE_RETRY", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    PowerShell has started an IPC listening thread on process: {0} in AppDomain: {1}.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_REMOTE_NAMEDPIPE_LISTENER_START {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_REMOTE_NAMEDPIPE_LISTENER_START", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    PowerShell has ended an IPC listening thread on process: {0} in AppDomain: {1}.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_REMOTE_NAMEDPIPE_LISTENER_END {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_REMOTE_NAMEDPIPE_LISTENER_END", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    An error has occurred in PowerShell IPC listening thread on process: {0} in AppDomain: {1}.  Error Message: {2}.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_REMOTE_NAMEDPIPE_LISTENER_ERROR {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_REMOTE_NAMEDPIPE_LISTENER_ERROR", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    PowerShell IPC connect on process: {0} in AppDomain: {1} for User: {2}.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_REMOTE_NAMEDPIPE_CONNECT {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_REMOTE_NAMEDPIPE_CONNECT", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    PowerShell IPC disconnect on process: {0} in AppDomain: {1} for User: {2}.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_REMOTE_NAMEDPIPE_DISCONNECT {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_REMOTE_NAMEDPIPE_DISCONNECT", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Port resolved to {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_RUNSPACE_PORT {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_RUNSPACE_PORT", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    AppName resolved to {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_RUNSPACE_APPNAME {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_RUNSPACE_APPNAME", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ComputerName resolved to {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_RUNSPACE_COMPUTERNAME {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_RUNSPACE_COMPUTERNAME", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Scheme is {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_RUNSPACE_SCHEME {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_RUNSPACE_SCHEME", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Test analytic message
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_RUNSPACE_TEST {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_RUNSPACE_TEST", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Connection Parameters are
    /// Connection URI: {0}
    /// Resource URI: {1}
    /// User: {2}
    /// OpenTimeout: {3}
    /// IdleTimeout: {4}
    /// CancelTimeout: {5}
    /// AuthenticationMechanism: {6}
    /// Thumb Print: {7}
    /// MaxUriRedirectionCount: {8}
    /// MaxReceivedDataSizePerCommand: {0}0
    /// MaxReceivedObjectSize: {0}1
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_RUNSPACE_WSMANCONNECTIONINFO {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_RUNSPACE_WSMANCONNECTIONINFO", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Modifying activity Id and correlating
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_RUNSPACEPOOL_TRANSFER {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_RUNSPACEPOOL_TRANSFER", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Received object with Runspace Id: {0} Command Id: {1} Destination: {2} DataType: {3} TargetInterface: {4}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_TRANSPORT_RCVDOBJ {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_TRANSPORT_RCVDOBJ", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    An unhandled exception occurred in the appdomain.
    ///Exception Type: {0}
    ///Exception Message: {1}
    ///Exception StackTrace: {2}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_APPDOMAIN_UNHANDLED_EXCEPTION {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_APPDOMAIN_UNHANDLED_EXCEPTION", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Runspace Id: {0} Pipeline Id: {1}. WSMan reported an error with error code: {2}.
    /// Error message: {3}
    /// StackTrace: {4}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_TRANSPORT_ERROR {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_TRANSPORT_ERROR", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    An unhandled exception occurred in the appdomain.
    ///Exception Type: {0}
    ///Exception Message: {1}
    ///Exception StackTrace: {2}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_APPDOMAIN_UNHANDLED_EXCEPTION {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_APPDOMAIN_UNHANDLED_EXCEPTION", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Runspace Id: {0} Pipeline Id: {1}. WSMan reported an error with error code: {2}.
    /// Error message: {3}
    /// StackTrace: {4}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_TRANSPORT_ERROR {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_TRANSPORT_ERROR", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Runspace Id {0}. Establishing a connection using WSMan Create Shell
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_TRANSPORT_CONNECT {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_TRANSPORT_CONNECT", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Runspace Id {0}. Callback received for WSMan Create Shell
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_TRANSPORT_SHELL_CONNECT_CALLBACK {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_TRANSPORT_SHELL_CONNECT_CALLBACK", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Runspace Id: {0}. Closing shell using WSManCloseShell
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_TRANSPORT_SHELL_CLOSE {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_TRANSPORT_SHELL_CLOSE", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Runspace Id: {0}. Callback received for WSManCloseShell
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_TRANSPORT_SHELL_CLOSE_CALLBACK {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_TRANSPORT_SHELL_CLOSE_CALLBACK", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Runspace Id: {0} Pipeline Id: {1}. Sending data of size {2}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_TRANSPORT_SEND_DATA {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_TRANSPORT_SEND_DATA", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Runspace Id: {0} Pipeline Id: {1}. Callback received for WSManSendShellInputEx
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_TRANSPORT_SEND_DATA_CALLBACK {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_TRANSPORT_SEND_DATA_CALLBACK", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Runspace Id: {0} Pipeline Id: {1}. Placing Receive request using WSManReceiveShellOutputEx
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_TRANSPORT_RECEIVE_DATA {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_TRANSPORT_RECEIVE_DATA", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Runspace Id: {0} Pipeline Id: {1}. Received Data of size {2}.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_TRANSPORT_RECEIVE_DATA_CALLBACK {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_TRANSPORT_RECEIVE_DATA_CALLBACK", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Runspace Id {0} Pipeline Id {1}. Establishing a command connection using WSManRunShellCommandEx
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_TRANSPORT_CMD_CONNECT {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_TRANSPORT_CMD_CONNECT", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Runspace Id {0} Pipeline Id {1}. Callback received for command connection
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_TRANSPORT_CMD_CONNECT_CALLBACK {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_TRANSPORT_CMD_CONNECT_CALLBACK", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Runspace Id: {0} Pipeline Id {1}. Closing transport for command
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_TRANSPORT_CMD_CLOSE {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_TRANSPORT_CMD_CLOSE", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Runspace Id: {0} Pipeline Id {1}. Callback received for command close
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_TRANSPORT_CMD_CLOSE_CALLBACK {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_TRANSPORT_CMD_CLOSE_CALLBACK", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Runspace Id: {0} Pipeline Id {1}. Sending signal with code {2} using WSManSignalShellEx
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_TRANSPORT_SIGNAL {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_TRANSPORT_SIGNAL", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Runspace Id: {0} Pipeline Id {1}. Callback received for WSManSignalShellEx
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_TRANSPORT_SIGNAL_CALLBACK {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_TRANSPORT_SIGNAL_CALLBACK", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Runspace Id: {0}. Connection is getting redirected to Uri: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_TRANSPORT_URI_REDIRECTION {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_TRANSPORT_URI_REDIRECTION", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Runspace Id: {0} Pipeline Id: {1}. Server is sending data of size {2} to client. DataType: {3} TargetInterface: {4}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_TRANSPORT_SERVER_SEND_DATA {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_TRANSPORT_SERVER_SEND_DATA", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Request {0}. Creating a server remote session. UserName: {1} Custom Shell Id: {2}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_CREATE_SERVER_REMOTESESSION {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_CREATE_SERVER_REMOTESESSION", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Reporting context for request: {0} Context Reported: {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_REPORT_CONTEXT {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_REPORT_CONTEXT", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Reporting operation complete for request: {0}
    /// Error Code: {1}
    /// Error Message: {2}
    /// StackTrace: {3}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_REPORT_OPERATION_COMPLETE {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_REPORT_OPERATION_COMPLETE", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Shell Context {0}. Request Id {1}. Creating a commonad session for running a command.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_CREATE_COMMAND_REMOTESESSION {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_CREATE_COMMAND_REMOTESESSION", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Shell Context {0} Command Context {1} Request Id {2}. Stopping command.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_STOP_COMMAND {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_STOP_COMMAND", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Shell Context {0} Command Context {1} Request Id {2}. Received data from client.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_SERVER_RECEIVED_DATA {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_SERVER_RECEIVED_DATA", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Shell Context {0} Command Context {1} Request Id {2}. Client sent a receive request so that server can send data.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_SERVER_RECEIVE_REQUEST {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_SERVER_RECEIVE_REQUEST", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Shell Context {0} Command Context {1} IsReceiveOperation {2}. Got close operation request.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_SERVER_CLOSE_OPERATION {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_SERVER_CLOSE_OPERATION", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Loading assembly {0} for custom shell with shell Id {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_LOAD_PSCUSTOMSHELL_ASSEMBLY {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_LOAD_PSCUSTOMSHELL_ASSEMBLY", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Loading type {0} for custom shell with shell Id {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_LOAD_PSCUSTOMSHELL_TYPE {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_LOAD_PSCUSTOMSHELL_TYPE", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Received remoting fragment.
    /// 	 Object Id: {0}
    /// 	 Fragment Id: {1}
    /// 	 Start Flag: {2}
    /// 	 End Flag: {3}
    /// 	 Payload Length: {4}
    /// 	 Payload Data: {5}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_RECEIVED_FRAGMENT {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_RECEIVED_FRAGMENT", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Sent remoting fragment.
    /// 	 Object Id: {0}
    /// 	 Fragment Id: {1}
    /// 	 Start Flag: {2}
    /// 	 End Flag: {3}
    /// 	 Payload Length: {4}
    /// 	 Payload Data: {5}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_SENT_FRAGMENT {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_SENT_FRAGMENT", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Shutting down winrm service.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_SHUTTING_DOWN {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_SHUTTING_DOWN", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Successfully rehydrated an object.
    /// 	 Deserialized type name: {0}
    /// 	 Rehydrated by casting to type: {1}
    /// 	 Rehydrated object is of type: {2}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_SERIALIZER_REHYDRATION_SUCCESS {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_SERIALIZER_REHYDRATION_SUCCESS", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Failed to rehydrated an object.
    /// 	 Deserialized type name: {0}
    /// 	 Rehydrated by casting to type: {1}
    /// 	 Type cast exception: {2}
    /// 	 Type cast inner exception: {3}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_SERIALIZER_REHYDRATION_FAILURE {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_SERIALIZER_REHYDRATION_FAILURE", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Serialization depth has been overridden.
    /// 	 Serialized type name: {0}
    /// 	 Original depth: {1}
    /// 	 Overridden depth: {2}
    /// 	 Current depth below top level: {3}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_SERIALIZER_DEPTH_OVERRIDE {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_SERIALIZER_DEPTH_OVERRIDE", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Serialization mode has been overridden.
    /// 	 Serialized type name: {0}
    /// 	 Overridden mode: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_SERIALIZER_MODE_OVERRIDE {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_SERIALIZER_MODE_OVERRIDE", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Serialization of a script property has been skipped, because there is no runspace to use for evaluation of the property.
    /// 	 Property name: {0}
    /// 	 Property owner's type name: {1}
    /// 	 Getter script: {2}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_SERIALIZER_SCRIPT_PROPERTY_WITHOUT_RUNSPACE {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_SERIALIZER_SCRIPT_PROPERTY_WITHOUT_RUNSPACE", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Serialization of a property has been skipped, because property getter failed.
    /// 	 Property name: {0}
    /// 	 Property owner's type name: {1}
    /// 	 Exception from property getter: {2}
    /// 	 Inner exception from property getter: {3}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_SERIALIZER_PROPERTY_GETTER_FAILED {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_SERIALIZER_PROPERTY_GETTER_FAILED", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Serialization of an enumerable object might not be complete, because object being enumerated threw an exception.
    /// 	 Type of object being enumerated: {0}
    /// 	 Exception: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_SERIALIZER_ENUMERATION_FAILED {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_SERIALIZER_ENUMERATION_FAILED", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Serialization called object's ToString method which failed.
    /// 	 Type of object: {0}
    /// 	 Exception: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_SERIALIZER_TOSTRING_FAILED {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_SERIALIZER_TOSTRING_FAILED", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Maximum depth below top level has been reached, forcing object to be serialized as strings.
    /// 	 Object type at max depth: {0}
    /// 	 Property name at max depth: {1}
    /// 	 Depth: {2}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_SERIALIZER_MAX_DEPTH_WHEN_SERIALIZING {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_SERIALIZER_MAX_DEPTH_WHEN_SERIALIZING", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    XmlException has been thrown by the deserializer (most likely indicating incorrect clixml format).
    /// 	 Line number: {0} Line position: {1}
    /// 	 Exception: {2}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_SERIALIZER_XMLEXCEPTION_WHEN_DESERIALIZING {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_SERIALIZER_XMLEXCEPTION_WHEN_DESERIALIZING", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Serialization of specified properties failed, because one of the specified properties was missing.
    /// 	 Type of object: {0}
    /// 	 Property name: {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_A_SERIALIZER_SPECIFIC_PROPERTY_MISSING {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_A_SERIALIZER_SPECIFIC_PROPERTY_MISSING", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    PowerShell console is starting up
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_PowershellConsoleStartupStart {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_PowershellConsoleStartupStart", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    PowerShell console is ready for user input
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_PowershellConsoleStartupStop {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_PowershellConsoleStartupStop", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_DebugMessage {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_DebugMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Tracing ErrorRecord:
    /// Message: {0}
    /// CategoryInfo.Category: {1}
    /// CategoryInfo.Reason : {2}
    /// CategoryInfo.TargetName : {3}
    /// FullyQualifiedErrorId: {4}
    /// Exception Details:
    /// Message : {5}
    /// Stack Trace: {6}
    /// InnerException {7}
    ///
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_Powershell_ErrorRecord {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_Powershell_ErrorRecord", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Exception:
    /// Message: {0}
    /// StackTrace: {1}
    /// InnerException : {2}
    ///
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_Powershell_Exception {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_Powershell_Exception", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Tracing PSObject
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_Powershell_PSObject {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_Powershell_PSObject", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Tracing Job:
    /// Id: {0}
    /// InstanceId: {1}
    /// Name: {2}
    /// Location: {3}
    /// State: {4}
    /// Command: {5}
    ///
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_Powershell_Job {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_Powershell_Job", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Trace Information:
    /// {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_MESSAGE {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_MESSAGE", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Trace Information:
    /// {0} {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_MESSAGE2 {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_MESSAGE2", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    BEGIN ImportWorkflowCommand::StartWorkflowApplication. Starting invocation of workflow function. Tracking Guid {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_M3PBeginStartWorkflowApplication {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_M3PBeginStartWorkflowApplication", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    END ImportWorkflowCommand::StartWorkflowApplication. Ending invocation of workflow function. Tracking Guid {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_M3PEndStartWorkflowApplication {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_M3PEndStartWorkflowApplication", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    BEGIN Creating new job in ImportWorkflowCommand::StartWorkflowApplication. Tracking Guid {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_M3PBeginCreateNewJob {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_M3PBeginCreateNewJob", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    END Creating new job in ImportWorkflowCommand::StartWorkflowApplication. Tracking Guid {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_M3PEndCreateNewJob {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_M3PEndCreateNewJob", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    END Creating new job in ImportWorkflowCommand::StartWorkflowApplication. Tracking Guid {0} : ContainerParentJob Guid {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_M3PTrackingGuidContainerParentJobCorrelation {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_M3PTrackingGuidContainerParentJobCorrelation", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    BEGIN JobLogic ContainerParentJob Guid {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_M3PBeginJobLogic {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_M3PBeginJobLogic", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    END JobLogic ContainerParentJob Guid {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_M3PEndJobLogic {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_M3PEndJobLogic", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    BEGIN WorkflowExecution ContainerParentJob Guid {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_M3PBeginWorkflowExecution {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_M3PBeginWorkflowExecution", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    END WorkflowExecution ContainerParentJob Guid {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_M3PEndWorkflowExecution {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_M3PEndWorkflowExecution", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    WorkflowJob with Guid {0} added to ContainerParentJob with Guid {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_M3PChildWorkflowJobAddition {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_M3PChildWorkflowJobAddition", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ProxyJob with Guid {0} associated with remote ContainerParentJob with Guid {1}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_M3PProxyJobRemoteJobAssociation {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_M3PProxyJobRemoteJobAssociation", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    BEGIN Execution of ContainerParentJob with Guid {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_M3PBeginContainerParentJobExecution {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_M3PBeginContainerParentJobExecution", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    END Execution of ContainerParentJob with Guid {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_M3PEndContainerParentJobExecution {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_M3PEndContainerParentJobExecution", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    BEGIN Execution of Proxy Job with Guid {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_M3PBeginProxyJobExecution {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_M3PBeginProxyJobExecution", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    END Execution of Proxy Job with Guid {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_M3PEndProxyJobExecution {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_M3PEndProxyJobExecution", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    BEGIN StateChanged event handler for Proxy Job with Guid {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_M3PBeginProxyJobEventHandler {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_M3PBeginProxyJobEventHandler", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    END StateChanged event handler for Proxy Job with Guid {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_M3PEndProxyJobEventHandler {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_M3PEndProxyJobEventHandler", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    BEGIN StateChanged event handler for Proxy Child Job with Guid {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_M3PBeginProxyChildJobEventHandler {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_M3PBeginProxyChildJobEventHandler", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    END StateChanged event handler for Proxy Child Job with Guid {0}
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_M3PEndProxyChildJobEventHandler {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_M3PEndProxyChildJobEventHandler", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    BEGIN Running garbage collection
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_M3PBeginRunGarbageCollection {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_M3PBeginRunGarbageCollection", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    END Running garbage collection
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_D_M3PEndRunGarbageCollection {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_D_M3PEndRunGarbageCollection", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Persistence store has reached its maximum specified size
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_M3PPERSISTENCE_STORE_MAXSIZE_REACHED {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_M3PPERSISTENCE_STORE_MAXSIZE_REACHED", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Windows PowerShell ISE has started to run script file {0}.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_ISEExecuteScript {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_ISEExecuteScript", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Windows PowerShell ISE has started to run a user-selected script from file {0}.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_ISEExecuteSelection {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_ISEExecuteSelection", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Windows PowerShell ISE is stopping the current command.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_ISEStopCommand {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_ISEStopCommand", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Windows PowerShell ISE is resuming the debugger.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_ISEResumeDebugger {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_ISEResumeDebugger", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Windows PowerShell ISE is stopping the debugger.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_ISEStopDebugger {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_ISEStopDebugger", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Windows PowerShell ISE is stepping into debugging.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_ISEDebuggerStepInto {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_ISEDebuggerStepInto", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Windows PowerShell ISE is stepping over debugging.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_ISEDebuggerStepOver {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_ISEDebuggerStepOver", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Windows PowerShell ISE is stepping out of debugging.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_ISEDebuggerStepOut {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_ISEDebuggerStepOut", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Windows PowerShell ISE is enabling all breakpoints.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_ISEEnableAllBreakpoints {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_ISEEnableAllBreakpoints", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Windows PowerShell ISE is disabling all breakpoints.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_ISEDisableAllBreakpoints {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_ISEDisableAllBreakpoints", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Windows PowerShell ISE is removing all breakpoints.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_ISERemoveAllBreakpoints {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_ISERemoveAllBreakpoints", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Windows PowerShell ISE is setting the breakpoint at line #: {0} of file {1}.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_ISESetBreakpoint {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_ISESetBreakpoint", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Windows PowerShell ISE is removing the breakpoint on line #: {0} of file {1}.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_ISERemoveBreakpoint {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_ISERemoveBreakpoint", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Windows PowerShell ISE is enabling the breakpoint on line #: {0} of file {1}.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_ISEEnableBreakpoint {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_ISEEnableBreakpoint", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Windows PowerShell ISE is disabling the breakpoint on line #: {0} of file {1}.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_ISEDisableBreakpoint {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_ISEDisableBreakpoint", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Windows PowerShell ISE has hit a breakpoint on line #: {0} of file {1}.
    ///
    /// </summary>
    internal static string PS_PROVIDEReventE_O_ISEHitBreakpoint {
        get {
            return ResourceManager.GetString("PS_PROVIDEReventE_O_ISEHitBreakpoint", resourceCulture);
        }
    }

}

