
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class ConsoleHostStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal ConsoleHostStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.PowerShell.ConsoleHost.resources.ConsoleHostStrings", typeof(ConsoleHostStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot display prompt because too many nested prompts are already running.
    ///  
    /// </summary>
    internal static string TooManyNestedPromptsError {
        get {
            return ResourceManager.GetString("TooManyNestedPromptsError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process input loop. ExitCurrentLoop was called when no InputLoops were running.
    ///  
    /// </summary>
    internal static string InputExitCurrentLoopOutOfSyncError {
        get {
            return ResourceManager.GetString("InputExitCurrentLoopOutOfSyncError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    PS>
    ///  
    /// </summary>
    internal static string DefaultPrompt {
        get {
            return ResourceManager.GetString("DefaultPrompt", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The shell cannot be started. A failure occurred during initialization:
    ///  
    /// </summary>
    internal static string ShellCannotBeStarted {
        get {
            return ResourceManager.GetString("ShellCannotBeStarted", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The shell cannot be started. An InitialSessionState object has been provided along with a -ConfigurationFile argument. Both configuration directives cannot be used at the same time.
    ///  
    /// </summary>
    internal static string ShellCannotBeStartedWithConfigConflict {
        get {
            return ResourceManager.GetString("ShellCannotBeStartedWithConfigConflict", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    An error has occurred that was not properly handled. Additional information is shown below. The PowerShell process will exit.
    ///  
    /// </summary>
    internal static string UnhandledExceptionShutdownMessage {
        get {
            return ResourceManager.GetString("UnhandledExceptionShutdownMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    **********************
    ///PowerShell transcript start
    ///Start time: {0:yyyyMMddHHmmss}
    ///Username  : {1}\{2}
    ///Machine   : {3} ({4})
    ///**********************
    ///  
    /// </summary>
    internal static string TranscriptPrologue {
        get {
            return ResourceManager.GetString("TranscriptPrologue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    **********************
    ///PowerShell transcript end
    ///End time: {0:yyyyMMddHHmmss}
    ///**********************
    ///  
    /// </summary>
    internal static string TranscriptEpilogue {
        get {
            return ResourceManager.GetString("TranscriptEpilogue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Command '{0}' could not be run because some PowerShell Snap-Ins did not load.
    ///  
    /// </summary>
    internal static string InitialCommandNotExecuted {
        get {
            return ResourceManager.GetString("InitialCommandNotExecuted", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Command '{0}' was not run as the session in which it was intended to run was either closed or broken
    ///  
    /// </summary>
    internal static string CommandNotExecuted {
        get {
            return ResourceManager.GetString("CommandNotExecuted", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Entering debug mode. Use h or ? for help. 
    ///  
    /// </summary>
    internal static string EnteringDebugger {
        get {
            return ResourceManager.GetString("EnteringDebugger", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Hit {0}
    ///  
    /// </summary>
    internal static string HitBreakpoint {
        get {
            return ResourceManager.GetString("HitBreakpoint", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0}:{1,-3} {2}
    ///  
    /// </summary>
    internal static string DebuggerSourceCodeFormat {
        get {
            return ResourceManager.GetString("DebuggerSourceCodeFormat", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    
    ///The current session does not support debugging; execution will continue.
    ///
    ///
    ///  
    /// </summary>
    internal static string SessionDoesNotSupportDebugger {
        get {
            return ResourceManager.GetString("SessionDoesNotSupportDebugger", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot load PSReadline module.  Console is running without PSReadline.
    ///  
    /// </summary>
    internal static string CannotLoadPSReadline {
        get {
            return ResourceManager.GetString("CannotLoadPSReadline", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    More than one server mode parameter was specified. Server mode parameters must be used exclusively.
    ///  
    /// </summary>
    internal static string ConflictingServerModeParameters {
        get {
            return ResourceManager.GetString("ConflictingServerModeParameters", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Loading personal and system profiles took {0}ms.
    ///  
    /// </summary>
    internal static string SlowProfileLoadingMessage {
        get {
            return ResourceManager.GetString("SlowProfileLoadingMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Run as Administrator
    ///  
    /// </summary>
    internal static string RunAsAdministrator {
        get {
            return ResourceManager.GetString("RunAsAdministrator", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    PushRunspace can only push a remote runspace.
    ///  
    /// </summary>
    internal static string PushRunspaceNotRemote {
        get {
            return ResourceManager.GetString("PushRunspaceNotRemote", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The '{0}' parameter is mandatory and must be specified when using the '{1}' parameter.
    ///  
    /// </summary>
    internal static string MissingMandatoryParameter {
        get {
            return ResourceManager.GetString("MissingMandatoryParameter", resourceCulture);
        }
    }

}

