
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class CsvCommandStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal CsvCommandStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.PowerShell.Commands.Utility.resources.CsvCommandStrings", typeof(CsvCommandStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot append CSV content to the following file: {1}. The appended object does not have a property that corresponds to the following column: {0}. To continue with mismatched properties, add the -Force parameter, and then retry the command.
    ///    {0} is a placeholder for property name (i.e. ProcessId)
    ///        {1} is a placeholder for filename (i.e. c:\users\lukasza\foo.csv)
    ///
    ///        {StrContains="Force"}
    ///
    ///        Reviewed by TArcher on 2010-06-29.
    ///    
    ///  
    /// </summary>
    internal static string CannotAppendCsvWithMismatchedPropertyNames {
        get {
            return ResourceManager.GetString("CannotAppendCsvWithMismatchedPropertyNames", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    You must specify either the -UseQuotes or -QuoteFields parameters, but not both.
    ///  
    /// </summary>
    internal static string CannotSpecifyQuoteFieldsAndUseQuotes {
        get {
            return ResourceManager.GetString("CannotSpecifyQuoteFieldsAndUseQuotes", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    You must specify either the -IncludeTypeInformation or -NoTypeInformation parameters, but not both.
    ///  
    /// </summary>
    internal static string CannotSpecifyIncludeTypeInformationAndNoTypeInformation {
        get {
            return ResourceManager.GetString("CannotSpecifyIncludeTypeInformationAndNoTypeInformation", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    You must specify either the -Path or -LiteralPath parameters, but not both.
    ///  
    /// </summary>
    internal static string CannotSpecifyPathAndLiteralPath {
        get {
            return ResourceManager.GetString("CannotSpecifyPathAndLiteralPath", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    One or more headers were not specified. Default names starting with "H" have been used in place of any missing headers.
    ///  
    /// </summary>
    internal static string UseDefaultNameForUnspecifiedHeader {
        get {
            return ResourceManager.GetString("UseDefaultNameForUnspecifiedHeader", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    FileName is a mandatory parameter.
    ///  
    /// </summary>
    internal static string FileNameIsAMandatoryParameter {
        get {
            return ResourceManager.GetString("FileNameIsAMandatoryParameter", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ReconcilePreexistingPropertyNames method should only get called when appending.
    ///  
    /// </summary>
    internal static string ReconcilePreexistingPropertyNamesMethodShouldOnlyGetCalledWhenAppending {
        get {
            return ResourceManager.GetString("ReconcilePreexistingPropertyNamesMethodShouldOnlyGetCalledWhenAppending", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ReconcilePreexistingPropertyNames method should only get called when preexisting property names have been read successfully.
    ///  
    /// </summary>
    internal static string ReconcilePreexistingPropertyNamesMethodShouldOnlyGetCalledWhenPreexistingPropertyNamesHaveBeenReadSuccessfully {
        get {
            return ResourceManager.GetString("ReconcilePreexistingPropertyNamesMethodShouldOnlyGetCalledWhenPreexistingPropertyNamesHaveBeenReadSuccessfully", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    BuildPropertyNames method should be called only once per cmdlet instance.
    ///  
    /// </summary>
    internal static string BuildPropertyNamesMethodShouldBeCalledOnlyOncePerCmdletInstance {
        get {
            return ResourceManager.GetString("BuildPropertyNamesMethodShouldBeCalledOnlyOncePerCmdletInstance", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Type hierarchy should not have null values.
    ///  
    /// </summary>
    internal static string TypeHierarchyShouldNotHaveNullValues {
        get {
            return ResourceManager.GetString("TypeHierarchyShouldNotHaveNullValues", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    EOF is reached.
    ///  
    /// </summary>
    internal static string EOFIsReached {
        get {
            return ResourceManager.GetString("EOFIsReached", resourceCulture);
        }
    }

}

