
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

public class UICultureResources {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    public UICultureResources() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    public static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.Management.UI.Internal.resources.public.UICultureResources", typeof(UICultureResources).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    public static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Select Columns...
    ///  
    /// </summary>
    public static string ColumnPicker {
        get {
            return ResourceManager.GetString("ColumnPicker", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    (none)
    ///    The group title for items within a column whose value is empty/null.
    ///  
    /// </summary>
    public static string GroupTitleNone {
        get {
            return ResourceManager.GetString("GroupTitleNone", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Value should be of Type {0}.
    ///    This text represents the error which will be shown when the entered type does not match the type we are expecting. {0} is the expected type.
    ///  
    /// </summary>
    public static string ErrorTextBoxTypeConversionErrorText {
        get {
            return ResourceManager.GetString("ErrorTextBoxTypeConversionErrorText", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The current selection is empty.
    ///    The error validation string to present to the user when they have selected a value out of bounds.
    ///  
    /// </summary>
    public static string ValidatingSelectorValueOutOfBounds {
        get {
            return ResourceManager.GetString("ValidatingSelectorValueOutOfBounds", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    contains
    ///    A filter rule that indicates a field must contain the specified value.
    ///  
    /// </summary>
    public static string FilterRule_Contains {
        get {
            return ResourceManager.GetString("FilterRule_Contains", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    does not contain
    ///    A filter rule that indicates a field must not contain the specified value.
    ///  
    /// </summary>
    public static string FilterRule_DoesNotContain {
        get {
            return ResourceManager.GetString("FilterRule_DoesNotContain", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    does not equal
    ///    A filter rule that indicates a field must not equal the specified value.
    ///  
    /// </summary>
    public static string FilterRule_DoesNotEqual {
        get {
            return ResourceManager.GetString("FilterRule_DoesNotEqual", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    equals
    ///    A filter rule that indicates a field must equal the specified value.
    ///  
    /// </summary>
    public static string FilterRule_Equals {
        get {
            return ResourceManager.GetString("FilterRule_Equals", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    is greater than or equal to
    ///    A filter rule that indicates a field must be greater than or equal to the specified value.
    ///  
    /// </summary>
    public static string FilterRule_GreaterThanOrEqual {
        get {
            return ResourceManager.GetString("FilterRule_GreaterThanOrEqual", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    is between
    ///    A filter rule that indicates a field must be between the specified values.
    ///  
    /// </summary>
    public static string FilterRule_IsBetween {
        get {
            return ResourceManager.GetString("FilterRule_IsBetween", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    is empty
    ///    A filter rule that indicates a field must be empty.
    ///  
    /// </summary>
    public static string FilterRule_IsEmpty {
        get {
            return ResourceManager.GetString("FilterRule_IsEmpty", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    is not empty
    ///    A filter rule that indicates a field must not be empty.
    ///  
    /// </summary>
    public static string FilterRule_IsNotEmpty {
        get {
            return ResourceManager.GetString("FilterRule_IsNotEmpty", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    is less than or equal to
    ///    A filter rule that indicates a field must be less than or equal to the specified value.
    ///  
    /// </summary>
    public static string FilterRule_LessThanOrEqual {
        get {
            return ResourceManager.GetString("FilterRule_LessThanOrEqual", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ends with
    ///    A filter rule that indicates a field must end with the specified value.
    ///  
    /// </summary>
    public static string FilterRule_TextEndsWith {
        get {
            return ResourceManager.GetString("FilterRule_TextEndsWith", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    starts with
    ///    A filter rule that indicates a field must start with the specified value.
    ///  
    /// </summary>
    public static string FilterRule_TextStartsWith {
        get {
            return ResourceManager.GetString("FilterRule_TextStartsWith", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Back
    ///    The text representing the tool tip and help text for the Back Button in the Back Forward History control when the button is disabled
    ///  
    /// </summary>
    public static string BackForwardHistory_BrowseBack_Disabled_Helptext {
        get {
            return ResourceManager.GetString("BackForwardHistory_BrowseBack_Disabled_Helptext", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Forward
    ///    The text representing the tool tip and help text for the Forward Button in the Back Forward History control when the button is disabled
    ///  
    /// </summary>
    public static string BackForwardHistory_BrowseForward_Disabled_Helptext {
        get {
            return ResourceManager.GetString("BackForwardHistory_BrowseForward_Disabled_Helptext", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The value must be a valid date in the following format: {0}.
    ///    {0} will be filled in with the culture appropriate ShortDatePattern
    ///  
    /// </summary>
    public static string ErrorMessageForUnparsableDateTimeType {
        get {
            return ResourceManager.GetString("ErrorMessageForUnparsableDateTimeType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The value must be a valid number.
    ///  
    /// </summary>
    public static string ErrorMessageForUnparsableNumericType {
        get {
            return ResourceManager.GetString("ErrorMessageForUnparsableNumericType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Search
    ///    The default background text of the search box.
    ///  
    /// </summary>
    public static string SearchBox_BackgroundText {
        get {
            return ResourceManager.GetString("SearchBox_BackgroundText", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    LeftToRight
    ///    This value will be loaded at runtime to define the flow direction of WPF application.  This value should be set to "RightToLeft" for mirrored language and "LeftToRight" for others.
    ///  
    /// </summary>
    public static string FlowDirection {
        get {
            return ResourceManager.GetString("FlowDirection", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    …
    ///    An ellipsis character.
    ///  
    /// </summary>
    public static string Ellipsis {
        get {
            return ResourceManager.GetString("Ellipsis", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Ctrl+Add
    ///  
    /// </summary>
    public static string ZoomIn1Shortcut {
        get {
            return ResourceManager.GetString("ZoomIn1Shortcut", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Ctrl+Shift+Add
    ///  
    /// </summary>
    public static string ZoomIn2Shortcut {
        get {
            return ResourceManager.GetString("ZoomIn2Shortcut", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Ctrl+Plus
    ///  
    /// </summary>
    public static string ZoomIn3Shortcut {
        get {
            return ResourceManager.GetString("ZoomIn3Shortcut", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Ctrl+Shift+Plus
    ///  
    /// </summary>
    public static string ZoomIn4Shortcut {
        get {
            return ResourceManager.GetString("ZoomIn4Shortcut", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Ctrl+Subtract
    ///  
    /// </summary>
    public static string ZoomOut1Shortcut {
        get {
            return ResourceManager.GetString("ZoomOut1Shortcut", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Ctrl+Shift+Subtract
    ///  
    /// </summary>
    public static string ZoomOut2Shortcut {
        get {
            return ResourceManager.GetString("ZoomOut2Shortcut", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Ctrl+Minus
    ///  
    /// </summary>
    public static string ZoomOut3Shortcut {
        get {
            return ResourceManager.GetString("ZoomOut3Shortcut", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Ctrl+Shift+Minus
    ///  
    /// </summary>
    public static string ZoomOut4Shortcut {
        get {
            return ResourceManager.GetString("ZoomOut4Shortcut", resourceCulture);
        }
    }

}

