
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class RunspacePoolStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal RunspacePoolStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.RunspacePoolStrings", typeof(RunspacePoolStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The maximum pool size cannot be less than 1.
    ///  
    /// </summary>
    internal static string MaxPoolLessThan1 {
        get {
            return ResourceManager.GetString("MaxPoolLessThan1", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The minimum pool size cannot be less than 1.
    ///  
    /// </summary>
    internal static string MinPoolLessThan1 {
        get {
            return ResourceManager.GetString("MinPoolLessThan1", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The minimum pool size cannot be greater than the maximum pool size.
    ///  
    /// </summary>
    internal static string MinPoolGreaterThanMaxPool {
        get {
            return ResourceManager.GetString("MinPoolGreaterThanMaxPool", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The state of the runspace pool is not valid for this operation.
    ///  
    /// </summary>
    internal static string InvalidRunspacePoolStateGeneral {
        get {
            return ResourceManager.GetString("InvalidRunspacePoolStateGeneral", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot perform the operation because the runspace pool is not in the '{0}' state. The current state is '{1}'.
    ///  
    /// </summary>
    internal static string InvalidRunspacePoolState {
        get {
            return ResourceManager.GetString("InvalidRunspacePoolState", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot open the runspace pool because it is not in the 'BeforeOpen' state. The current state is '{0}'.
    ///  
    /// </summary>
    internal static string CannotOpenAgain {
        get {
            return ResourceManager.GetString("CannotOpenAgain", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The {0} object was not created by calling {1} on the current RunspacePool instance.
    ///  
    /// </summary>
    internal static string AsyncResultNotOwned {
        get {
            return ResourceManager.GetString("AsyncResultNotOwned", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot release the runspace to the current pool because the runspace does not belong to the current pool.
    ///  
    /// </summary>
    internal static string RunspaceNotBelongsToPool {
        get {
            return ResourceManager.GetString("RunspaceNotBelongsToPool", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    This property cannot be changed after the runspace pool has been opened.
    ///  
    /// </summary>
    internal static string ChangePropertyAfterOpen {
        get {
            return ResourceManager.GetString("ChangePropertyAfterOpen", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    This runspace does not support disconnect and connect operations.
    ///  
    /// </summary>
    internal static string RunspaceDisconnectConnectNotSupported {
        get {
            return ResourceManager.GetString("RunspaceDisconnectConnectNotSupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot perform the operation because the runspace pool is in the Disconnected state.
    ///  
    /// </summary>
    internal static string CannotWhileDisconnected {
        get {
            return ResourceManager.GetString("CannotWhileDisconnected", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Disconnect operation is not supported on the server.  The server must be running PowerShell 3.0 or greater for remote runspace pool disconnection support.
    ///  
    /// </summary>
    internal static string DisconnectNotSupportedOnServer {
        get {
            return ResourceManager.GetString("DisconnectNotSupportedOnServer", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    This runspace pool {0} is not configured to provide disconnected PowerShell objects for commands running on the remote server.  Use the RunspacePool class GetRunspacePools() static method to query the server and return runspace pool objects that are configured to do this.
    ///  
    /// </summary>
    internal static string CannotReconstructCommands {
        get {
            return ResourceManager.GetString("CannotReconstructCommands", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    This runspace pool cannot be connected because the corresponding server side runspace pool is connected to another client.
    ///  
    /// </summary>
    internal static string CannotConnect {
        get {
            return ResourceManager.GetString("CannotConnect", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ResetRunspaceState is not supported on the server.  The server must be running PowerShell 5.0 or greater.
    ///  
    /// </summary>
    internal static string ResetRunspaceStateNotSupportedOnServer {
        get {
            return ResourceManager.GetString("ResetRunspaceStateNotSupportedOnServer", resourceCulture);
        }
    }

}

