
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class ConsoleHostUserInterfaceStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal ConsoleHostUserInterfaceStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.PowerShell.ConsoleHost.resources.ConsoleHostUserInterfaceStrings", typeof(ConsoleHostUserInterfaceStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The "{0}" collection must have at least one element.
    ///  
    /// </summary>
    internal static string PromptEmptyDescriptionsErrorTemplate {
        get {
            return ResourceManager.GetString("PromptEmptyDescriptionsErrorTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot recognize "{1}" as a {0} due to an overflow error.
    ///  
    /// </summary>
    internal static string PromptParseOverflowErrorTemplate {
        get {
            return ResourceManager.GetString("PromptParseOverflowErrorTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot recognize "{1}" as a {0} due to a format error.
    ///  
    /// </summary>
    internal static string PromptParseFormatErrorTemplate {
        get {
            return ResourceManager.GetString("PromptParseFormatErrorTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    "{0}" cannot be recognized as a valid Prompt command.
    ///  
    /// </summary>
    internal static string PromptUnrecognizedCommandErrorTemplate {
        get {
            return ResourceManager.GetString("PromptUnrecognizedCommandErrorTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    No help is available for {0}.
    ///  
    /// </summary>
    internal static string PromptNoHelpAvailableErrorTemplate {
        get {
            return ResourceManager.GetString("PromptNoHelpAvailableErrorTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0}: 
    ///  
    /// </summary>
    internal static string PromptFieldPromptInputSeparatorTemplate {
        get {
            return ResourceManager.GetString("PromptFieldPromptInputSeparatorTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The field "{0}" is a zero rank array.
    ///  
    /// </summary>
    internal static string RankZeroArrayErrorTemplate {
        get {
            return ResourceManager.GetString("RankZeroArrayErrorTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    "{0}" should have at least one element.
    ///  
    /// </summary>
    internal static string EmptyChoicesErrorTemplate {
        get {
            return ResourceManager.GetString("EmptyChoicesErrorTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    "{0}" must be a valid index into "{1}" or -1 for no default choice.
    ///  
    /// </summary>
    internal static string InvalidDefaultChoiceErrorTemplate {
        get {
            return ResourceManager.GetString("InvalidDefaultChoiceErrorTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    "{0}" must be a valid index into "{1}". "{2}" is not a valid index.
    ///  
    /// </summary>
    internal static string InvalidDefaultChoiceForMultipleSelection {
        get {
            return ResourceManager.GetString("InvalidDefaultChoiceForMultipleSelection", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    [?] Help 
    ///  
    /// </summary>
    internal static string PromptForChoiceHelp {
        get {
            return ResourceManager.GetString("PromptForChoiceHelp", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The prompt was canceled.
    ///  
    /// </summary>
    internal static string PromptCanceledError {
        get {
            return ResourceManager.GetString("PromptCanceledError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the hot key because a question mark ("?") cannot be used as a hot key.
    ///  
    /// </summary>
    internal static string InvalidChoiceHotKeyError {
        get {
            return ResourceManager.GetString("InvalidChoiceHotKeyError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot display the prompt for "{0}" because type "{1}" cannot be loaded.
    ///  
    /// </summary>
    internal static string PromptTypeLoadErrorTemplate {
        get {
            return ResourceManager.GetString("PromptTypeLoadErrorTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    "{0}" cannot be null or empty.
    ///  
    /// </summary>
    internal static string NullOrEmptyErrorTemplate {
        get {
            return ResourceManager.GetString("NullOrEmptyErrorTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    "{0}" cannot be null.
    ///  
    /// </summary>
    internal static string NullErrorTemplate {
        get {
            return ResourceManager.GetString("NullErrorTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    (Type !? for Help.)
    ///  
    /// </summary>
    internal static string PromptHelp {
        get {
            return ResourceManager.GetString("PromptHelp", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    (default is "{0}"): 
    ///  
    /// </summary>
    internal static string DefaultChoicePrompt {
        get {
            return ResourceManager.GetString("DefaultChoicePrompt", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    (default is "{0}")
    ///  
    /// </summary>
    internal static string DefaultChoiceForMultipleChoices {
        get {
            return ResourceManager.GetString("DefaultChoiceForMultipleChoices", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    (default choices are {0})
    ///  
    /// </summary>
    internal static string DefaultChoicesForMultipleChoices {
        get {
            return ResourceManager.GetString("DefaultChoicesForMultipleChoices", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Choice[{0}]: 
    ///  
    /// </summary>
    internal static string ChoiceMessage {
        get {
            return ResourceManager.GetString("ChoiceMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    DEBUG: {0}
    ///  
    /// </summary>
    internal static string DebugFormatString {
        get {
            return ResourceManager.GetString("DebugFormatString", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    VERBOSE: {0}
    ///  
    /// </summary>
    internal static string VerboseFormatString {
        get {
            return ResourceManager.GetString("VerboseFormatString", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    WARNING: {0}
    ///  
    /// </summary>
    internal static string WarningFormatString {
        get {
            return ResourceManager.GetString("WarningFormatString", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    PowerShell is in NonInteractive mode. Read and Prompt functionality is not available.
    ///  
    /// </summary>
    internal static string ReadFailsOnNonInteractiveFlag {
        get {
            return ResourceManager.GetString("ReadFailsOnNonInteractiveFlag", resourceCulture);
        }
    }

}

