// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.
// See the LICENSE file in the project root for more information.

namespace Reporting
{
    public class Os
    {
        public string Locale { get; set; }

        public string Architecture { get; set; }

        public string Name { get; set; }
    }
}
