
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class Serialization {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal Serialization() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.Serialization", typeof(Serialization).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0} attribute was expected.
    ///  
    /// </summary>
    internal static string AttributeExpected {
        get {
            return ResourceManager.GetString("AttributeExpected", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0} XML tag is not recognized.
    ///  
    /// </summary>
    internal static string InvalidElementTag {
        get {
            return ResourceManager.GetString("InvalidElementTag", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    No object found for referenceId {0}
    ///  
    /// </summary>
    internal static string InvalidReferenceId {
        get {
            return ResourceManager.GetString("InvalidReferenceId", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Name attribute for dictionary key is incorrectly specified.
    ///  
    /// </summary>
    internal static string InvalidDictionaryKeyName {
        get {
            return ResourceManager.GetString("InvalidDictionaryKeyName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Name attribute for dictionary value is incorrectly specified.
    ///  
    /// </summary>
    internal static string InvalidDictionaryValueName {
        get {
            return ResourceManager.GetString("InvalidDictionaryValueName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Version of PSObject is not valid.
    ///  
    /// </summary>
    internal static string InvalidVersion {
        get {
            return ResourceManager.GetString("InvalidVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Version of incoming PSObject is {0}. Expected value is 1.
    ///  
    /// </summary>
    internal static string UnexpectedVersion {
        get {
            return ResourceManager.GetString("UnexpectedVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process names because no TypeNames were found for referenceId {0}.
    ///  
    /// </summary>
    internal static string InvalidTypeHierarchyReferenceId {
        get {
            return ResourceManager.GetString("InvalidTypeHierarchyReferenceId", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Value of depth parameter must be greater than or equal to 1.
    ///  
    /// </summary>
    internal static string DepthOfOneRequired {
        get {
            return ResourceManager.GetString("DepthOfOneRequired", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Current Node type is {0}. Expected type is {1}.
    ///  
    /// </summary>
    internal static string InvalidNodeType {
        get {
            return ResourceManager.GetString("InvalidNodeType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Key for dictionary entry is not specified.
    ///  
    /// </summary>
    internal static string DictionaryKeyNotSpecified {
        get {
            return ResourceManager.GetString("DictionaryKeyNotSpecified", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Value for dictionary entry is not specified.
    ///  
    /// </summary>
    internal static string DictionaryValueNotSpecified {
        get {
            return ResourceManager.GetString("DictionaryValueNotSpecified", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    There are no more objects to deserialize.
    ///  
    /// </summary>
    internal static string ReadCalledAfterDone {
        get {
            return ResourceManager.GetString("ReadCalledAfterDone", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Null is specified as dictionary key.
    ///  
    /// </summary>
    internal static string NullAsDictionaryKey {
        get {
            return ResourceManager.GetString("NullAsDictionaryKey", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The contents of the {0} primitive type are not valid.
    ///  
    /// </summary>
    internal static string InvalidPrimitiveType {
        get {
            return ResourceManager.GetString("InvalidPrimitiveType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Serialized XML is nested too deeply.
    ///  
    /// </summary>
    internal static string DeserializationTooDeep {
        get {
            return ResourceManager.GetString("DeserializationTooDeep", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Serializer was closed.
    ///  
    /// </summary>
    internal static string Stopping {
        get {
            return ResourceManager.GetString("Stopping", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The data in the command exceeded the maximum size that is allowed by the session configuration. The allowed maximum is {0} MB. Change the input, use a different session configuration, or change the "{1}" and "{2}" properties of the session configuration on the remote computer.
    ///  
    /// </summary>
    internal static string DeserializationMemoryQuota {
        get {
            return ResourceManager.GetString("DeserializationMemoryQuota", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Deserialization of encrypted secure string failed
    ///  
    /// </summary>
    internal static string DeserializeSecureStringFailed {
        get {
            return ResourceManager.GetString("DeserializeSecureStringFailed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The key type {0} is not valid. The PSPrimitiveDictionary class accepts only keys of the type System.String.
    ///  
    /// </summary>
    internal static string PrimitiveHashtableInvalidKey {
        get {
            return ResourceManager.GetString("PrimitiveHashtableInvalidKey", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The type of the value {0} is not valid. The PSPrimitiveDictionary class accepts only values of types that are fully serializable over PowerShell remoting. See the Help topic about_Remoting for a list of fully-serializable types.
    ///  
    /// </summary>
    internal static string PrimitiveHashtableInvalidValue {
        get {
            return ResourceManager.GetString("PrimitiveHashtableInvalidValue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Could not decrypt data. The data was not encrypted with this key.
    ///  
    /// </summary>
    internal static string InvalidKey {
        get {
            return ResourceManager.GetString("InvalidKey", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The parameter value "{0}" is not a valid encrypted string.
    ///  
    /// </summary>
    internal static string InvalidEncryptedString {
        get {
            return ResourceManager.GetString("InvalidEncryptedString", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified {0} is not valid. Valid {0} length settings are either 128 bits, 192 bits, or 256 bits.
    ///  
    /// </summary>
    internal static string InvalidKeyLength {
        get {
            return ResourceManager.GetString("InvalidKeyLength", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Deserialization of SecureString is currently only supported on Windows.
    ///  
    /// </summary>
    internal static string DeserializeSecureStringNotSupported {
        get {
            return ResourceManager.GetString("DeserializeSecureStringNotSupported", resourceCulture);
        }
    }

}

