
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class AliasCommandStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal AliasCommandStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.PowerShell.Commands.Utility.resources.AliasCommandStrings", typeof(AliasCommandStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Set Alias
    ///  
    /// </summary>
    internal static string SetAliasAction {
        get {
            return ResourceManager.GetString("SetAliasAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Name: {0} Value: {1}
    ///  
    /// </summary>
    internal static string SetAliasTarget {
        get {
            return ResourceManager.GetString("SetAliasTarget", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    New Alias
    ///  
    /// </summary>
    internal static string NewAliasAction {
        get {
            return ResourceManager.GetString("NewAliasAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Name: {0} Value: {1}
    ///  
    /// </summary>
    internal static string NewAliasTarget {
        get {
            return ResourceManager.GetString("NewAliasTarget", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Import Alias
    ///  
    /// </summary>
    internal static string ImportAliasAction {
        get {
            return ResourceManager.GetString("ImportAliasAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Name: {0} Value: {1}
    ///  
    /// </summary>
    internal static string ImportAliasTarget {
        get {
            return ResourceManager.GetString("ImportAliasTarget", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot open file {0} to export the alias. {1}
    ///  
    /// </summary>
    internal static string ExportAliasFileOpenFailed {
        get {
            return ResourceManager.GetString("ExportAliasFileOpenFailed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Alias File
    ///  
    /// </summary>
    internal static string ExportAliasHeaderTitle {
        get {
            return ResourceManager.GetString("ExportAliasHeaderTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Exported by : {0}
    ///  
    /// </summary>
    internal static string ExportAliasHeaderUser {
        get {
            return ResourceManager.GetString("ExportAliasHeaderUser", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Date/Time : {0:F}
    ///  
    /// </summary>
    internal static string ExportAliasHeaderDate {
        get {
            return ResourceManager.GetString("ExportAliasHeaderDate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Computer : {0}
    ///  
    /// </summary>
    internal static string ExportAliasHeaderMachine {
        get {
            return ResourceManager.GetString("ExportAliasHeaderMachine", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot import the alias because the specified path '{0}' referred to a '{1}' provider path. Change the value of the Path parameter to a file system path.
    ///  
    /// </summary>
    internal static string ImportAliasFromFileSystemOnly {
        get {
            return ResourceManager.GetString("ImportAliasFromFileSystemOnly", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot import alias because path '{0}' contains wildcard characters that resolve to multiple paths. Aliases can be imported from only one file. Change the value of the Path parameter to a path that resolves to a single file.
    ///  
    /// </summary>
    internal static string ImportAliasPathResolvedToMultiple {
        get {
            return ResourceManager.GetString("ImportAliasPathResolvedToMultiple", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot open file {0} to import the alias. {1}
    ///  
    /// </summary>
    internal static string ImportAliasFileOpenFailed {
        get {
            return ResourceManager.GetString("ImportAliasFileOpenFailed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot import an alias. Line number {1} in the file '{0}' is not a properly-formatted, comma-separated values (CSV) line for aliases. Change the line to contain four values separated by commas. If the value text itself contains a comma, then the value must be contained in quotation marks.
    ///  
    /// </summary>
    internal static string ImportAliasFileInvalidFormat {
        get {
            return ResourceManager.GetString("ImportAliasFileInvalidFormat", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot import the alias because line number {1} in the file '{0}' contains an option that is not recognized for aliases. Change the file to contain valid options.
    ///  
    /// </summary>
    internal static string ImportAliasOptionsError {
        get {
            return ResourceManager.GetString("ImportAliasOptionsError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    This command cannot find a matching alias because an alias with the {0} '{1}' does not exist.
    ///  
    /// </summary>
    internal static string NoAliasFound {
        get {
            return ResourceManager.GetString("NoAliasFound", resourceCulture);
        }
    }

}

