
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class CmdletizationResources {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal CmdletizationResources() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.PowerShell.Commands.Management.resources.CmdletizationResources", typeof(CmdletizationResources).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot find the {0} class on the {1} CIM server.  Verify the value of the ClassName xml attribute in Cmdlet Definition XML and retry. Valid class name example: ROOT\cimv2\Win32_Process.
    ///    {StrContains="ClassName"} {StrContains="ROOT\cimv2\Win32_Process"}
    ///{0} is a placeholder for a name of a (potentially misspelled) CIM class.  Example: "Win32_Process".
    ///{1} is a placeholder for a server name.  Example: "localhost".
    ///  
    /// </summary>
    internal static string CimJob_InvalidClassName {
        get {
            return ResourceManager.GetString("CimJob_InvalidClassName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    CIM method {1} on the {0} CIM object
    ///    {0} is a placeholder for a CIM path. Example: \\SERVER1\ROOT\cimv2:Win32_Process.Handle="11828"
    ///{1} is a placeholder for a CIM method name. Example: Create
    ///  
    /// </summary>
    internal static string CimJob_MethodDescription {
        get {
            return ResourceManager.GetString("CimJob_MethodDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Failed to run {1}.  {0}
    ///    {0} is a placeholder for a generic CIM failure.  Example: 'Invalid namespace' or '9'
    ///{1} is a placeholder for a description of CIM operation.  This most likely comes from CimJob_MethodDescription or CimJob_QueryDescription
    ///  
    /// </summary>
    internal static string CimJob_GenericCimFailure {
        get {
            return ResourceManager.GetString("CimJob_GenericCimFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Running the following operation: {0}.
    ///    {0} is a placeholder for a description of CIM operation.  This most likely comes from CimJob_MethodDescription or CimJob_QueryDescription
    ///  
    /// </summary>
    internal static string CimJob_VerboseExecutionMessage {
        get {
            return ResourceManager.GetString("CimJob_VerboseExecutionMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    CIM cmdlets do not support the {0} parameter together with the AsJob parameter.  Remove one of these parameters and retry.
    ///    {StrContains="AsJob"}
    ///{0} is a placeholder for 'WhatIf' or 'Confirm' cmdlet parameters
    ///  
    /// </summary>
    internal static string SessionBasedWrapper_ShouldProcessVsJobConflict {
        get {
            return ResourceManager.GetString("SessionBasedWrapper_ShouldProcessVsJobConflict", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    CIM query for instances of the {0} class on the {1} CIM server: {2}
    ///    {0} is a placeholder for a name of a CIM class.  Example: "Win32_Process".
    ///{1} is a placeholder for a server name.  Example: "localhost".
    ///  
    /// </summary>
    internal static string CimJob_SafeQueryDescription {
        get {
            return ResourceManager.GetString("CimJob_SafeQueryDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The CIM method returned the following error code: {0}
    ///    {0} is a placeholder for an error code returned from a CIM method.  Example: 123
    ///  
    /// </summary>
    internal static string CimJob_ErrorCodeFromMethod {
        get {
            return ResourceManager.GetString("CimJob_ErrorCodeFromMethod", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The {2} CIM method exposed by the {0} class on the {1} CIM server
    ///    {0} is a placeholder for a name of a CIM class.  Example: "Win32_Process".
    ///{1} is a placeholder for a server name.  Example: "localhost".
    ///{2} is a placeholder for a CIM method name. Example: Create
    ///  
    /// </summary>
    internal static string CimJob_SafeMethodDescription {
        get {
            return ResourceManager.GetString("CimJob_SafeMethodDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    CIM intrinsic type
    ///  
    /// </summary>
    internal static string CimConversion_CimIntrinsicValue {
        get {
            return ResourceManager.GetString("CimConversion_CimIntrinsicValue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    WQL literal
    ///  
    /// </summary>
    internal static string CimConversion_WqlQuery {
        get {
            return ResourceManager.GetString("CimConversion_WqlQuery", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot find the {2} output parameter of the {1} method of the {0} CIM object.  Verify the value of the ParameterName attribute in Cmdlet Definition XML and retry.
    ///    {StrContains="ParameterName"}
    ///{0} is a placeholder for a CIM path. Example: \\SERVER1\ROOT\cimv2:Win32_Process.Handle="11828"
    ///{1} is a placeholder for a name of a (potentially misspelled) CIM method.  Example: "Terminate".
    ///{2} is a placeholder for a name of a (potentially misspelled) method parameter.
    ///
    ///  
    /// </summary>
    internal static string CimJob_InvalidOutputParameterName {
        get {
            return ResourceManager.GetString("CimJob_InvalidOutputParameterName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    No matching {1} objects found by {0}. Verify query parameters and retry.
    ///  
    /// </summary>
    internal static string CimJob_NotFound_ComplexCase {
        get {
            return ResourceManager.GetString("CimJob_NotFound_ComplexCase", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    No {2} objects found with property '{0}' equal to '{1}'.  Verify the value of the property and retry.
    ///  
    /// </summary>
    internal static string CimJob_NotFound_SimpleGranularCase_Equality {
        get {
            return ResourceManager.GetString("CimJob_NotFound_SimpleGranularCase_Equality", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Type of {0} property ({1}) doesn't match the CIM type ({2}) associated with the type declared in Cmdlet Definition XML.
    ///  
    /// </summary>
    internal static string CimJob_MismatchedTypeOfPropertyReturnedByQuery {
        get {
            return ResourceManager.GetString("CimJob_MismatchedTypeOfPropertyReturnedByQuery", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    CIM query for enumerating associated instance of the {0} class on the {1} CIM server
    ///    {0} is a placeholder for a name of a CIM class.  Example: "Win32_Process".
    ///{1} is a placeholder for a server name.  Example: "localhost".
    ///  
    /// </summary>
    internal static string CimJob_SafeAssociationDescription {
        get {
            return ResourceManager.GetString("CimJob_SafeAssociationDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    CIM query for enumerating instances of the {0} class on the {1} CIM server, that are associated with the following instance: {2}
    ///    {0} is a placeholder for a name of a CIM class.  Example: "Win32_Process".
    ///{1} is a placeholder for a server name.  Example: "localhost".
    ///{2} is a placeholder for a string describing a CimInstance.  Example: "Win32_Process[Handle=123]".
    ///  
    /// </summary>
    internal static string CimJob_AssociationDescription {
        get {
            return ResourceManager.GetString("CimJob_AssociationDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The {0} command cannot complete, because the {1} server is currently busy.  The command will be automatically resumed in {2:f2} seconds.
    ///    {0} is a placeholder for a command name.  Example: "Get-NetAdapter"
    ///{1} is a placeholder for a computer name.  Example: "localhost"
    ///{2} is a placeholder for a number of seconds.  Example: 1.23
    ///  
    /// </summary>
    internal static string CimJob_SleepAndRetryVerboseMessage {
        get {
            return ResourceManager.GetString("CimJob_SleepAndRetryVerboseMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot connect to CIM server. {0}
    ///    {0} is a placeholder for a more detailed error message.
    ///  
    /// </summary>
    internal static string CimJob_BrokenSession {
        get {
            return ResourceManager.GetString("CimJob_BrokenSession", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The cmdlet does not fully support the Inquire action for debug messages.  Cmdlet operation will continue during the prompt.  Select a different action preference via -Debug switch or $DebugPreference variable, and try again.
    ///    {StrContains="Debug"} {StrContains="DebugPreference"} {StrContains="Inquire"}
    ///  
    /// </summary>
    internal static string CimCmdletAdapter_DebugInquire {
        get {
            return ResourceManager.GetString("CimCmdletAdapter_DebugInquire", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The cmdlet does not fully support the Inquire action for warnings.  Cmdlet operation will continue during the prompt.  Select a different action preference via -WarningAction parameter or $WarningPreference variable, and try again.
    ///    {StrContains="WarningAction"} {StrContains="WarningPreference"} {StrContains="Inquire"}
    ///  
    /// </summary>
    internal static string CimCmdletAdapter_WarningInquire {
        get {
            return ResourceManager.GetString("CimCmdletAdapter_WarningInquire", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The cmdlet does not fully support the Stop action for warnings.  Cmdlet operation will be stopped with a delay.  Select a different action preference via -WarningAction parameter or $WarningPreference variable, and try again.
    ///    {StrContains="WarningAction"} {StrContains="WarningPreference"} {StrContains="Stop"}
    ///  
    /// </summary>
    internal static string CimCmdletAdapter_WarningStop {
        get {
            return ResourceManager.GetString("CimCmdletAdapter_WarningStop", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0}: {1}
    ///    {0} is a placeholder for a computername.  Example: "localhost"
    ///{1} is a placeholder for the original message.  Example: "Deleting managed resource"
    ///  
    /// </summary>
    internal static string CimJob_ComputerNameConcatenationTemplate {
        get {
            return ResourceManager.GetString("CimJob_ComputerNameConcatenationTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0}: A CimSession to the CIM server uses the DCOM protocol, which does not support the {1} switch.
    ///    {0} is a placeholder for a name of a computer
    ///{1} is a placeholder for 'Confirm' or 'WhatIf'
    ///  
    /// </summary>
    internal static string CimCmdletAdapter_RemoteDcomDoesntSupportExtendedSemantics {
        get {
            return ResourceManager.GetString("CimCmdletAdapter_RemoteDcomDoesntSupportExtendedSemantics", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    No {2} objects found with property '{0}' matching '{1}'.  Verify the value of the property and retry.
    ///  
    /// </summary>
    internal static string CimJob_NotFound_SimpleGranularCase_Wildcard {
        get {
            return ResourceManager.GetString("CimJob_NotFound_SimpleGranularCase_Wildcard", resourceCulture);
        }
    }

}

