
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class ImplicitRemotingStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal ImplicitRemotingStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.PowerShell.Commands.Utility.resources.ImplicitRemotingStrings", typeof(ImplicitRemotingStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Data returned by the remote {0} command is not in the expected format.
    ///  
    /// </summary>
    internal static string ErrorMalformedDataFromRemoteCommand {
        get {
            return ResourceManager.GetString("ErrorMalformedDataFromRemoteCommand", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The {0} cmdlet requires the following commands in the remote session: Get-Command, Get-FormatData, and Select-Object. The following commands are used, but optional: Get-Help, and Measure-Object. Verify that the remote session includes the required commands, and then try again.
    ///  
    /// </summary>
    internal static string ErrorRequiredRemoteCommandNotFound {
        get {
            return ResourceManager.GetString("ErrorRequiredRemoteCommandNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Running the {0} command in a remote session reported the following error: {1}.
    ///  
    /// </summary>
    internal static string ErrorFromRemoteCommand {
        get {
            return ResourceManager.GetString("ErrorFromRemoteCommand", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Running the {0} command in a remote session returned no results.
    ///  
    /// </summary>
    internal static string ErrorNoResultsFromRemoteEnd {
        get {
            return ResourceManager.GetString("ErrorNoResultsFromRemoteEnd", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    No session has been associated with this implicit remoting module.
    ///  
    /// </summary>
    internal static string ErrorNoRunspaceForThisModule {
        get {
            return ResourceManager.GetString("ErrorNoRunspaceForThisModule", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Could not resolve remote alias '{0}'.
    ///  
    /// </summary>
    internal static string ErrorCouldntResolveAlias {
        get {
            return ResourceManager.GetString("ErrorCouldntResolveAlias", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Proxy creation has been skipped for the '{0}' command, because the name did not match the value of the Name parameter.
    ///  
    /// </summary>
    internal static string ErrorSkippedNonRequestedCommand {
        get {
            return ResourceManager.GetString("ErrorSkippedNonRequestedCommand", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Extended type definition has been skipped for the '{0}' type because its name did not match the value of the FormatTypeName parameter.
    ///  
    /// </summary>
    internal static string ErrorSkippedNonRequestedTypeDefinition {
        get {
            return ResourceManager.GetString("ErrorSkippedNonRequestedTypeDefinition", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Proxy creation has been skipped for the '{0}' command, because PowerShell could not verify the safety of the command name.
    ///  
    /// </summary>
    internal static string ErrorSkippedUnsafeCommandName {
        get {
            return ResourceManager.GetString("ErrorSkippedUnsafeCommandName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Proxy creation has been skipped for the following command: '{0}', because it would shadow an existing local command.  Use the AllowClobber parameter if you want to shadow existing local commands.
    ///  
    /// </summary>
    internal static string ErrorCommandSkippedBecauseOfShadowing {
        get {
            return ResourceManager.GetString("ErrorCommandSkippedBecauseOfShadowing", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    No command proxies have been created, because all of the requested remote commands would shadow existing local commands.  Use the AllowClobber parameter if you want to shadow existing local commands.
    ///  
    /// </summary>
    internal static string ErrorNoCommandsImportedBecauseOfSkipping {
        get {
            return ResourceManager.GetString("ErrorNoCommandsImportedBecauseOfSkipping", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Implicit remoting for {0}
    ///  
    /// </summary>
    internal static string ProxyModuleDescription {
        get {
            return ResourceManager.GetString("ProxyModuleDescription", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Implicit remoting event (session id: {0}; event handler id: {1})
    ///  
    /// </summary>
    internal static string EventSourceIdentifier {
        get {
            return ResourceManager.GetString("EventSourceIdentifier", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Implicit remoting module
    ///  
    /// </summary>
    internal static string ModuleHeaderTitle {
        get {
            return ResourceManager.GetString("ModuleHeaderTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    generated on {0}
    ///  
    /// </summary>
    internal static string ModuleHeaderDate {
        get {
            return ResourceManager.GetString("ModuleHeaderDate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    by {0} cmdlet
    ///  
    /// </summary>
    internal static string ModuleHeaderCommand {
        get {
            return ResourceManager.GetString("ModuleHeaderCommand", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Invoked with the following command line: {0}
    ///  
    /// </summary>
    internal static string ModuleHeaderCommandLine {
        get {
            return ResourceManager.GetString("ModuleHeaderCommandLine", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Optional parameter that can be used to specify the session on which this proxy module works
    ///  
    /// </summary>
    internal static string ModuleHeaderRunspaceOverrideParameter {
        get {
            return ResourceManager.GetString("ModuleHeaderRunspaceOverrideParameter", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Creating a new session for implicit remoting of "{{0}}" command...
    ///  
    /// </summary>
    internal static string CreateNewRunspaceMessageTemplate {
        get {
            return ResourceManager.GetString("CreateNewRunspaceMessageTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Session for implicit remoting module at {{0}}
    ///  
    /// </summary>
    internal static string ProxyRunspaceNameTemplate {
        get {
            return ResourceManager.GetString("ProxyRunspaceNameTemplate", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Creating implicit remoting module ...
    ///  
    /// </summary>
    internal static string ProgressActivity {
        get {
            return ResourceManager.GetString("ProgressActivity", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Getting command information from remote session ...
    ///  
    /// </summary>
    internal static string ProgressStatusGetCommandStart {
        get {
            return ResourceManager.GetString("ProgressStatusGetCommandStart", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Getting command information from remote session ... {0} commands received
    ///  
    /// </summary>
    internal static string ProgressStatusGetCommandProgress {
        get {
            return ResourceManager.GetString("ProgressStatusGetCommandProgress", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Getting formatting and output information from remote session ...
    ///  
    /// </summary>
    internal static string ProgressStatusGetFormatDataStart {
        get {
            return ResourceManager.GetString("ProgressStatusGetFormatDataStart", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Getting formatting and output information from remote session ... {0} objects received
    ///  
    /// </summary>
    internal static string ProgressStatusGetFormatDataProgress {
        get {
            return ResourceManager.GetString("ProgressStatusGetFormatDataProgress", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Completed.
    ///  
    /// </summary>
    internal static string ProgressStatusCompleted {
        get {
            return ResourceManager.GetString("ProgressStatusCompleted", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    PowerShell Credential Request
    ///  
    /// </summary>
    internal static string CredentialRequestTitle {
        get {
            return ResourceManager.GetString("CredentialRequestTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Enter your credentials for {0}.
    ///  
    /// </summary>
    internal static string CredentialRequestBody {
        get {
            return ResourceManager.GetString("CredentialRequestBody", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Enter the HTTP proxy credentials that are used for the following connection: {0}
    ///  
    /// </summary>
    internal static string ProxyCredentialRequestBody {
        get {
            return ResourceManager.GetString("ProxyCredentialRequestBody", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Commands that are available in the new remote session are different than those available when the implicit remoting module was created.  Consider creating the module again by using the Export-PSSession cmdlet.
    ///  
    /// </summary>
    internal static string WarningMismatchedImplicitRemotingHash {
        get {
            return ResourceManager.GetString("WarningMismatchedImplicitRemotingHash", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Files cannot be loaded because running scripts is disabled on this system. Provide a valid certificate with which to sign the files.
    ///  
    /// </summary>
    internal static string CertificateNeeded {
        get {
            return ResourceManager.GetString("CertificateNeeded", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The file {0} could not be signed.
    ///  
    /// </summary>
    internal static string InvalidSigningOperation {
        get {
            return ResourceManager.GetString("InvalidSigningOperation", resourceCulture);
        }
    }

}

