
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class Metadata {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal Metadata() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.Metadata", typeof(Metadata).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot initialize attributes for "{0}": "{1}"
    ///  
    /// </summary>
    internal static string MetadataMemberInitialization {
        get {
            return ResourceManager.GetString("MetadataMemberInitialization", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The argument cannot be validated because its type "{0}" is not the same type ({1}) as the maximum and minimum limits of the parameter. Make sure the argument is of type {1} and then try the command again.
    ///  
    /// </summary>
    internal static string ValidateRangeElementType {
        get {
            return ResourceManager.GetString("ValidateRangeElementType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The argument "{0}" cannot be validated because its value is not greater than zero.
    ///  
    /// </summary>
    internal static string ValidateRangePositiveFailure {
        get {
            return ResourceManager.GetString("ValidateRangePositiveFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The argument "{0}" cannot be validated because its value is not greater than or equal to zero.
    ///  
    /// </summary>
    internal static string ValidateRangeNonNegativeFailure {
        get {
            return ResourceManager.GetString("ValidateRangeNonNegativeFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The argument "{0}" cannot be validated because its value is not less than zero.
    ///  
    /// </summary>
    internal static string ValidateRangeNegativeFailure {
        get {
            return ResourceManager.GetString("ValidateRangeNegativeFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The argument "{0}" cannot be validated because its value is not less than or equal to zero.
    ///  
    /// </summary>
    internal static string ValidateRangeNonPositiveFailure {
        get {
            return ResourceManager.GetString("ValidateRangeNonPositiveFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified minimum range ({0}) cannot be accepted because it is not the same type as the specified maximum range ({1}). Update the ValidateRange attribute for the parameter.
    ///  
    /// </summary>
    internal static string ValidateRangeMinRangeMaxRangeType {
        get {
            return ResourceManager.GetString("ValidateRangeMinRangeMaxRangeType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot accept the MaxRange and MinRange parameter types. Both parameters must be objects that implement an IComparable interface.
    ///  
    /// </summary>
    internal static string ValidateRangeNotIComparable {
        get {
            return ResourceManager.GetString("ValidateRangeNotIComparable", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified maximum range cannot be accepted because it is less than the specified minimum range. Update the ValidateRange attribute for the parameter.
    ///  
    /// </summary>
    internal static string ValidateRangeMaxRangeSmallerThanMinRange {
        get {
            return ResourceManager.GetString("ValidateRangeMaxRangeSmallerThanMinRange", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The {0} argument is greater than the maximum allowed range of {1}. Supply an argument that is less than or equal to {1} and then try the command again.
    ///  
    /// </summary>
    internal static string ValidateRangeGreaterThanMaxRangeFailure {
        get {
            return ResourceManager.GetString("ValidateRangeGreaterThanMaxRangeFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The {0} argument is less than the minimum allowed range of {1}. Supply an argument that is greater than or equal to {1} and then try the command again.
    ///  
    /// </summary>
    internal static string ValidateRangeSmallerThanMinRangeFailure {
        get {
            return ResourceManager.GetString("ValidateRangeSmallerThanMinRangeFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The argument "{0}" does not match the "{1}" pattern. Supply an argument that matches "{1}" and try the command again.
    ///  
    /// </summary>
    internal static string ValidatePatternFailure {
        get {
            return ResourceManager.GetString("ValidatePatternFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The ValidateCount attribute cannot be applied to a non-array parameter. Either remove the attribute from the parameter or make the parameter an array parameter.
    ///  
    /// </summary>
    internal static string ValidateCountNotInArray {
        get {
            return ResourceManager.GetString("ValidateCountNotInArray", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The parameter requires exactly {0} value(s) - {1} value(s) were provided.
    ///  
    /// </summary>
    internal static string ValidateCountExactFailure {
        get {
            return ResourceManager.GetString("ValidateCountExactFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The parameter requires at least {0} value(s) and no more than {1} value(s) - {2} value(s) were provided.
    ///  
    /// </summary>
    internal static string ValidateCountMinMaxFailure {
        get {
            return ResourceManager.GetString("ValidateCountMinMaxFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified maximum number of arguments for a parameter is fewer than the specified minimum number of arguments. Update the ValidateCount attribute for the parameter.
    ///  
    /// </summary>
    internal static string ValidateCountMaxLengthSmallerThanMinLength {
        get {
            return ResourceManager.GetString("ValidateCountMaxLengthSmallerThanMinLength", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified maximum character length of the argument is shorter than the specified minimum argument character length. Update the ValidateLength attribute for the parameter.
    ///  
    /// </summary>
    internal static string ValidateLengthMaxLengthSmallerThanMinLength {
        get {
            return ResourceManager.GetString("ValidateLengthMaxLengthSmallerThanMinLength", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The ValidateLength attribute cannot be applied to a parameter that is not a string or string[] parameter. Make the parameter a string or string[] parameter.
    ///  
    /// </summary>
    internal static string ValidateLengthNotString {
        get {
            return ResourceManager.GetString("ValidateLengthNotString", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The character length ({1}) of the argument is too short. Specify an argument with a length that is greater than or equal to "{0}", and then try the command again.
    ///  
    /// </summary>
    internal static string ValidateLengthMinLengthFailure {
        get {
            return ResourceManager.GetString("ValidateLengthMinLengthFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The character length of the {1} argument is too long. Shorten the character length of the argument so it is fewer than or equal to "{0}" characters, and then try the command again.
    ///  
    /// </summary>
    internal static string ValidateLengthMaxLengthFailure {
        get {
            return ResourceManager.GetString("ValidateLengthMaxLengthFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The argument "{0}" does not belong to the set "{1}" specified by the ValidateSet attribute. Supply an argument that is in the set and then try the command again.
    ///  
    /// </summary>
    internal static string ValidateSetFailure {
        get {
            return ResourceManager.GetString("ValidateSetFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Valid values generator return a null value.
    ///  
    /// </summary>
    internal static string ValidateSetGeneratedValidValuesListIsNull {
        get {
            return ResourceManager.GetString("ValidateSetGeneratedValidValuesListIsNull", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    "{0}" failed on property "{1}" {2}
    ///  
    /// </summary>
    internal static string ValidateFailureResult {
        get {
            return ResourceManager.GetString("ValidateFailureResult", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot get or run the command. The maximum number of parameter sets for this command has been exceeded.
    ///  
    /// </summary>
    internal static string ParsingTooManyParameterSets {
        get {
            return ResourceManager.GetString("ParsingTooManyParameterSets", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the argument because the argument value is not a string. The values of parameter arguments that have the ArgumentTransformationAttribute specified should be strings.
    ///  
    /// </summary>
    internal static string ArgumentTransformationArgumentsShouldBeStrings {
        get {
            return ResourceManager.GetString("ArgumentTransformationArgumentsShouldBeStrings", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The variable cannot be validated because the value {1} is not a valid value for the {0} variable.
    ///  
    /// </summary>
    internal static string InvalidValueFailure {
        get {
            return ResourceManager.GetString("InvalidValueFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The attribute cannot be added because variable {0} with value {1} would no longer be valid.
    ///  
    /// </summary>
    internal static string InvalidMetadataForCurrentValue {
        get {
            return ResourceManager.GetString("InvalidMetadataForCurrentValue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The argument is null. Provide a valid value for the argument, and then try running the command again.
    ///  
    /// </summary>
    internal static string ValidateNotNullFailure {
        get {
            return ResourceManager.GetString("ValidateNotNullFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The argument has a null value, or an element of the argument collection contains a null value. Provide a collection that does not contain any null values, and then try the command again.
    ///  
    /// </summary>
    internal static string ValidateNotNullCollectionFailure {
        get {
            return ResourceManager.GetString("ValidateNotNullCollectionFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The argument is null or empty. Provide an argument that is not null or empty, and then try the command again.
    ///  
    /// </summary>
    internal static string ValidateNotNullOrEmptyFailure {
        get {
            return ResourceManager.GetString("ValidateNotNullOrEmptyFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The argument is null, empty, or an element of the argument collection contains a null value. Supply a collection that does not contain any null values and then try the command again.
    ///  
    /// </summary>
    internal static string ValidateNotNullOrEmptyCollectionFailure {
        get {
            return ResourceManager.GetString("ValidateNotNullOrEmptyCollectionFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The argument is null, empty, or consists of only white-space characters. Provide an argument that contains non white-space characters, and then try the command again.
    ///  
    /// </summary>
    internal static string ValidateNotNullOrWhiteSpaceFailure {
        get {
            return ResourceManager.GetString("ValidateNotNullOrWhiteSpaceFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    An element of the argument collection is null, empty, or consists of only white-space characters. Supply a collection that does not contain any those values and then try the command again.
    ///  
    /// </summary>
    internal static string ValidateNotNullOrWhiteSpaceCollectionFailure {
        get {
            return ResourceManager.GetString("ValidateNotNullOrWhiteSpaceCollectionFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A parameter with the name '{0}' was defined multiple times for the command.
    ///  
    /// </summary>
    internal static string ParameterNameAlreadyExistsForCommand {
        get {
            return ResourceManager.GetString("ParameterNameAlreadyExistsForCommand", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The parameter alias cannot be specified because an alias with the name '{0}' was already defined multiple times for the command.
    ///  
    /// </summary>
    internal static string AliasParameterNameAlreadyExistsForCommand {
        get {
            return ResourceManager.GetString("AliasParameterNameAlreadyExistsForCommand", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The parameter '{0}' cannot be specified because it conflicts with the parameter alias of the same name for parameter '{1}'.
    ///  
    /// </summary>
    internal static string ParameterNameConflictsWithAlias {
        get {
            return ResourceManager.GetString("ParameterNameConflictsWithAlias", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The "{1}" validation script for the argument with value "{0}" did not return a result of True. Determine why the validation script failed, and then try the command again.
    ///  
    /// </summary>
    internal static string ValidateScriptFailure {
        get {
            return ResourceManager.GetString("ValidateScriptFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The "{0}" argument does not contain a valid PowerShell version. Supply a valid version number and then try the command again.
    ///  
    /// </summary>
    internal static string ValidateVersionFailure {
        get {
            return ResourceManager.GetString("ValidateVersionFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot validate argument '{0}' because it is not a valid variable name.
    ///  
    /// </summary>
    internal static string ValidateVariableName {
        get {
            return ResourceManager.GetString("ValidateVariableName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The job conversion type must derive from IAstToScriptBlockConverter.
    ///  
    /// </summary>
    internal static string JobDefinitionMustDeriveFromIJobConverter {
        get {
            return ResourceManager.GetString("JobDefinitionMustDeriveFromIJobConverter", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The path argument is invalid. Supply a path argument that is a string type.
    ///  
    /// </summary>
    internal static string ValidateDrivePathArgNotString {
        get {
            return ResourceManager.GetString("ValidateDrivePathArgNotString", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The path argument drive {0} does not belong to the set of approved drives: {1}. Supply a path argument with an approved drive.
    ///  
    /// </summary>
    internal static string ValidateDrivePathFailure {
        get {
            return ResourceManager.GetString("ValidateDrivePathFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The path argument contains invalid characters.
    ///  
    /// </summary>
    internal static string ValidateDrivePathInvalidChar {
        get {
            return ResourceManager.GetString("ValidateDrivePathInvalidChar", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The path argument has no root drive.  Supply a full path argument with a root drive.
    ///  
    /// </summary>
    internal static string ValidateDrivePathNoRoot {
        get {
            return ResourceManager.GetString("ValidateDrivePathNoRoot", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The argument value for the parameter '{0}' cannot be null or an empty string.
    ///  
    /// </summary>
    internal static string ArgumentNullOrEmpty {
        get {
            return ResourceManager.GetString("ArgumentNullOrEmpty", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Enum member '{0}' is not a valid value for the parameter '{1}'. Specify one of the following members and try again: {2}.
    ///  
    /// </summary>
    internal static string InvalidEnumArgument {
        get {
            return ResourceManager.GetString("InvalidEnumArgument", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process input. The argument "{0}" is not trusted.
    ///  
    /// </summary>
    internal static string ValidateTrustedDataFailure {
        get {
            return ResourceManager.GetString("ValidateTrustedDataFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ValidateTrustedData Attribute Check Failure
    ///  
    /// </summary>
    internal static string WDACParameterArgNotTrustedLogTitle {
        get {
            return ResourceManager.GetString("WDACParameterArgNotTrustedLogTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The parameter argument '{0}' is not trusted and will fail the ValidateTrustedData parameter attribute check in Constrained Language mode.
    ///  
    /// </summary>
    internal static string WDACParameterArgNotTrustedMessage {
        get {
            return ResourceManager.GetString("WDACParameterArgNotTrustedMessage", resourceCulture);
        }
    }

}

