
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

public class HelpWindowResources {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    public HelpWindowResources() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    public static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.Management.UI.Internal.resources.public.HelpWindowResources", typeof(HelpWindowResources).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    public static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cancel
    ///  
    /// </summary>
    public static string CancelText {
        get {
            return ResourceManager.GetString("CancelText", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Match Case
    ///  
    /// </summary>
    public static string CaseSensitiveTitle {
        get {
            return ResourceManager.GetString("CaseSensitiveTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    CommonParameters
    ///    Name of a group of parameters common to all cmdlets
    ///  
    /// </summary>
    public static string CommonParameters {
        get {
            return ResourceManager.GetString("CommonParameters", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Description
    ///  
    /// </summary>
    public static string DescriptionTitle {
        get {
            return ResourceManager.GetString("DescriptionTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Examples
    ///  
    /// </summary>
    public static string ExamplesTitle {
        get {
            return ResourceManager.GetString("ExamplesTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    _Find:
    ///  
    /// </summary>
    public static string FindText {
        get {
            return ResourceManager.GetString("FindText", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Help Sections
    ///  
    /// </summary>
    public static string HelpSectionsTitle {
        get {
            return ResourceManager.GetString("HelpSectionsTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0} Help
    ///  
    /// </summary>
    public static string HelpTitleFormat {
        get {
            return ResourceManager.GetString("HelpTitleFormat", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Inputs
    ///  
    /// </summary>
    public static string InputsTitle {
        get {
            return ResourceManager.GetString("InputsTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0} {1}
    ///  
    /// </summary>
    public static string LinkTextFormat {
        get {
            return ResourceManager.GetString("LinkTextFormat", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Methods
    ///  
    /// </summary>
    public static string MethodsTitle {
        get {
            return ResourceManager.GetString("MethodsTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    _Next
    ///  
    /// </summary>
    public static string NextText {
        get {
            return ResourceManager.GetString("NextText", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    No matches found
    ///  
    /// </summary>
    public static string NoMatches {
        get {
            return ResourceManager.GetString("NoMatches", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Notes
    ///  
    /// </summary>
    public static string NotesTitle {
        get {
            return ResourceManager.GetString("NotesTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    OK
    ///  
    /// </summary>
    public static string OKText {
        get {
            return ResourceManager.GetString("OKText", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    1 match
    ///  
    /// </summary>
    public static string OneMatch {
        get {
            return ResourceManager.GetString("OneMatch", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Outputs
    ///  
    /// </summary>
    public static string OutputsTitle {
        get {
            return ResourceManager.GetString("OutputsTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Accept wildcard characters?
    ///  
    /// </summary>
    public static string ParameterAcceptWildcard {
        get {
            return ResourceManager.GetString("ParameterAcceptWildcard", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Default value
    ///  
    /// </summary>
    public static string ParameterDefaultValue {
        get {
            return ResourceManager.GetString("ParameterDefaultValue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Accept pipeline input?
    ///  
    /// </summary>
    public static string ParameterPipelineInput {
        get {
            return ResourceManager.GetString("ParameterPipelineInput", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Position?
    ///  
    /// </summary>
    public static string ParameterPosition {
        get {
            return ResourceManager.GetString("ParameterPosition", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Required?
    ///  
    /// </summary>
    public static string ParameterRequired {
        get {
            return ResourceManager.GetString("ParameterRequired", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Parameters
    ///  
    /// </summary>
    public static string ParametersTitle {
        get {
            return ResourceManager.GetString("ParametersTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    _Previous
    ///  
    /// </summary>
    public static string PreviousText {
        get {
            return ResourceManager.GetString("PreviousText", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Properties
    ///  
    /// </summary>
    public static string PropertiesTitle {
        get {
            return ResourceManager.GetString("PropertiesTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    RelatedLinks
    ///  
    /// </summary>
    public static string RelatedLinksTitle {
        get {
            return ResourceManager.GetString("RelatedLinksTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Remarks
    ///  
    /// </summary>
    public static string RemarksTitle {
        get {
            return ResourceManager.GetString("RemarksTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Search Options
    ///  
    /// </summary>
    public static string SearchOptionsTitle {
        get {
            return ResourceManager.GetString("SearchOptionsTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Settings
    ///  
    /// </summary>
    public static string SettingsText {
        get {
            return ResourceManager.GetString("SettingsText", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0} matches
    ///  
    /// </summary>
    public static string SomeMatchesFormat {
        get {
            return ResourceManager.GetString("SomeMatchesFormat", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Synopsis
    ///  
    /// </summary>
    public static string SynopsisTitle {
        get {
            return ResourceManager.GetString("SynopsisTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Syntax
    ///  
    /// </summary>
    public static string SyntaxTitle {
        get {
            return ResourceManager.GetString("SyntaxTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Help for {0}
    ///    {0} is the name of a cmdlet
    ///  
    /// </summary>
    public static string Title {
        get {
            return ResourceManager.GetString("Title", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Whole Word
    ///  
    /// </summary>
    public static string WholeWordTitle {
        get {
            return ResourceManager.GetString("WholeWordTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0}%
    ///  
    /// </summary>
    public static string ZoomLabelTextFormat {
        get {
            return ResourceManager.GetString("ZoomLabelTextFormat", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Zoom
    ///  
    /// </summary>
    public static string ZoomSlider {
        get {
            return ResourceManager.GetString("ZoomSlider", resourceCulture);
        }
    }

}

