
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class FormatAndOut_format_xxx {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal FormatAndOut_format_xxx() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.FormatAndOut_format_xxx", typeof(FormatAndOut_format_xxx).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cmdlet parameters View and Property are mutually exclusive.
    ///  
    /// </summary>
    internal static string CannotSpecifyViewAndPropertyError {
        get {
            return ResourceManager.GetString("CannotSpecifyViewAndPropertyError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cmdlet parameters AutoSize and Column are mutually exclusive.
    ///  
    /// </summary>
    internal static string CannotSpecifyAutosizeAndColumnsError {
        get {
            return ResourceManager.GetString("CannotSpecifyAutosizeAndColumnsError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The view name {0} cannot be found.
    ///  
    /// </summary>
    internal static string UnknownViewNameError {
        get {
            return ResourceManager.GetString("UnknownViewNameError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///     The view name {0} cannot be found in the {1} formatting.
    ///    {0} indicates one of the valid formating types such as Table, List, Wide or Custom.
    ///  
    /// </summary>
    internal static string UnknownViewNameErrorSuffix {
        get {
            return ResourceManager.GetString("UnknownViewNameErrorSuffix", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///     There are no existing {0} views for {1} objects.
    ///  
    /// </summary>
    internal static string NonExistingViewNameError {
        get {
            return ResourceManager.GetString("NonExistingViewNameError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The view name {0} cannot be found. Specify one of the following {1} views and try again: {2}.
    ///  
    /// </summary>
    internal static string InvalidViewNameError {
        get {
            return ResourceManager.GetString("InvalidViewNameError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///     Try using one of these other format cmdlets: 
    ///    Prefix text to suggest user to use one of the valid view names.
    ///  
    /// </summary>
    internal static string SuggestValidViewNamePrefix {
        get {
            return ResourceManager.GetString("SuggestValidViewNamePrefix", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///       {0}: 
    ///  
    /// </summary>
    internal static string GroupStartDataIndentedAutoGeneratedLabel {
        get {
            return ResourceManager.GetString("GroupStartDataIndentedAutoGeneratedLabel", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The following object supports IEnumerable:
    ///  
    /// </summary>
    internal static string IEnum_Header {
        get {
            return ResourceManager.GetString("IEnum_Header", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The IEnumerable contains no objects.
    ///  
    /// </summary>
    internal static string IEnum_NoObjects {
        get {
            return ResourceManager.GetString("IEnum_NoObjects", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The IEnumerable contains the following object:
    ///  
    /// </summary>
    internal static string IEnum_OneObject {
        get {
            return ResourceManager.GetString("IEnum_OneObject", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The IEnumerable contains the following {0} objects:
    ///  
    /// </summary>
    internal static string IEnum_ManyObjects {
        get {
            return ResourceManager.GetString("IEnum_ManyObjects", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unknown class Id {0}.
    ///  
    /// </summary>
    internal static string FOD_ClassIdInvalid {
        get {
            return ResourceManager.GetString("FOD_ClassIdInvalid", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The type {0} for property {1} is not valid.
    ///  
    /// </summary>
    internal static string FOD_InvalidPropertyType {
        get {
            return ResourceManager.GetString("FOD_InvalidPropertyType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The value of the {0} data member cannot be null.
    ///  
    /// </summary>
    internal static string FOD_NullDataMember {
        get {
            return ResourceManager.GetString("FOD_NullDataMember", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The object type is not recognized.
    ///  
    /// </summary>
    internal static string FOD_InvalidClassidProperty {
        get {
            return ResourceManager.GetString("FOD_InvalidClassidProperty", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Failed to create object with class Id {0}.
    ///  
    /// </summary>
    internal static string FOD_InvalidClassid {
        get {
            return ResourceManager.GetString("FOD_InvalidClassid", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The {0} property is recursive.
    ///  
    /// </summary>
    internal static string FOD_RecursiveProperty {
        get {
            return ResourceManager.GetString("FOD_RecursiveProperty", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Failed to evaluate expression "{0}".
    ///  
    /// </summary>
    internal static string PSPropertyExpressionError {
        get {
            return ResourceManager.GetString("PSPropertyExpressionError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Failed to interpret format string "{0}".
    ///  
    /// </summary>
    internal static string FormattingError {
        get {
            return ResourceManager.GetString("FormattingError", resourceCulture);
        }
    }

}

