
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class ProcessResources {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal ProcessResources() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.PowerShell.Commands.Management.resources.ProcessResources", typeof(ProcessResources).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot find a process with the name "{0}". Verify the process name and call the cmdlet again.
    ///  
    /// </summary>
    internal static string NoProcessFoundForGivenName {
        get {
            return ResourceManager.GetString("NoProcessFoundForGivenName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot find a process with the name "{0}". Try running with -Id to search by Id of processes.
    ///  
    /// </summary>
    internal static string RecommendIdTagForGivenName {
        get {
            return ResourceManager.GetString("RecommendIdTagForGivenName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    This command cannot be run because the debugger cannot be attached to the process "{0} ({1})". Specify another process and Run your command.
    ///  
    /// </summary>
    internal static string NoDebuggerFound {
        get {
            return ResourceManager.GetString("NoDebuggerFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot find a process with the process identifier {1}.
    ///  
    /// </summary>
    internal static string NoProcessFoundForGivenId {
        get {
            return ResourceManager.GetString("NoProcessFoundForGivenId", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot stop process "{0} ({1})" because of the following error: {2}
    ///  
    /// </summary>
    internal static string CouldNotStopProcess {
        get {
            return ResourceManager.GetString("CouldNotStopProcess", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0} ({1})
    ///  
    /// </summary>
    internal static string ProcessNameForConfirmation {
        get {
            return ResourceManager.GetString("ProcessNameForConfirmation", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0} {1}
    ///  
    /// </summary>
    internal static string StartProcessTarget {
        get {
            return ResourceManager.GetString("StartProcessTarget", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot enumerate the modules of the "{0}" process.
    ///  
    /// </summary>
    internal static string CouldNotEnumerateModules {
        get {
            return ResourceManager.GetString("CouldNotEnumerateModules", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot enumerate the file version information of the "{0}" process.
    ///  
    /// </summary>
    internal static string CouldNotEnumerateFileVer {
        get {
            return ResourceManager.GetString("CouldNotEnumerateFileVer", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot enumerate the modules and the file version information of the "{0}" process.
    ///  
    /// </summary>
    internal static string CouldNotEnumerateModuleFileVer {
        get {
            return ResourceManager.GetString("CouldNotEnumerateModuleFileVer", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Are you sure you want to perform the Stop-Process operation on the following item: {0}({1})?
    ///  
    /// </summary>
    internal static string ConfirmStopProcess {
        get {
            return ResourceManager.GetString("ConfirmStopProcess", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified path is not a valid win32 application. Try again with the UseShellExecute.
    ///  
    /// </summary>
    internal static string UseShell {
        get {
            return ResourceManager.GetString("UseShell", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    This command stopped operation of "{0} ({1})" because of the following error: {2}.
    ///  
    /// </summary>
    internal static string ProcessIsNotTerminated {
        get {
            return ResourceManager.GetString("ProcessIsNotTerminated", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    This command cannot be run because Redirection parameters cannot be used with UseShellExecute parameter
    ///  
    /// </summary>
    internal static string RedirectionParams {
        get {
            return ResourceManager.GetString("RedirectionParams", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Exception getting "Modules" or "FileVersion": "This feature is not supported for remote computers.".
    ///  
    /// </summary>
    internal static string NoComputerNameWithFileVersion {
        get {
            return ResourceManager.GetString("NoComputerNameWithFileVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    This command cannot attach the debugger to the process due to {0} because no default debugger is available.
    ///  
    /// </summary>
    internal static string DebuggerError {
        get {
            return ResourceManager.GetString("DebuggerError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    This command stopped operation because it cannot wait on 'System Idle' process. Specify another process and Run your command again.
    ///  
    /// </summary>
    internal static string WaitOnIdleProcess {
        get {
            return ResourceManager.GetString("WaitOnIdleProcess", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    This command stopped operation because it cannot wait on itself. Specify another process and Run your command again.
    ///  
    /// </summary>
    internal static string WaitOnItself {
        get {
            return ResourceManager.GetString("WaitOnItself", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    This command stopped operation because process "{0} ({1})" is not stopped in the specified time-out.
    ///  
    /// </summary>
    internal static string ProcessNotTerminated {
        get {
            return ResourceManager.GetString("ProcessNotTerminated", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    This command cannot be run due to the error: {0}
    ///  
    /// </summary>
    internal static string InvalidStartProcess {
        get {
            return ResourceManager.GetString("InvalidStartProcess", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    This command cannot be run because the input "{0}" is not a valid Application.  Give a valid application and run your command again.
    ///  
    /// </summary>
    internal static string InvalidApplication {
        get {
            return ResourceManager.GetString("InvalidApplication", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    This command cannot be run because either the parameter "{0}" has a value that is not valid or cannot be used with this command. Give a valid input and Run your command again.
    ///  
    /// </summary>
    internal static string InvalidInput {
        get {
            return ResourceManager.GetString("InvalidInput", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    This command cannot be run because "{0}" and "{1}" are same. Give different inputs and Run your command again.
    ///  
    /// </summary>
    internal static string DuplicateEntry {
        get {
            return ResourceManager.GetString("DuplicateEntry", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    This command cannot be run completely because the system cannot find all the information required.
    ///  
    /// </summary>
    internal static string CannotStarttheProcess {
        get {
            return ResourceManager.GetString("CannotStarttheProcess", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Failed to retrieve the new process handle: "{0}". The Process object outputted may have some properties and methods that do not work properly.
    ///  
    /// </summary>
    internal static string FailedToCreateProcessObject {
        get {
            return ResourceManager.GetString("FailedToCreateProcessObject", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    This command cannot be run due to error 1783. The possible cause of this error can be using of a non-existing user "{0}". Please give a valid user and run your command again.
    ///  
    /// </summary>
    internal static string InvalidUserError {
        get {
            return ResourceManager.GetString("InvalidUserError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error adding '{0}' to the network: {1}
    ///  
    /// </summary>
    internal static string JoinNetworkFailed {
        get {
            return ResourceManager.GetString("JoinNetworkFailed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error removing '{0}' from the network: {1}
    ///  
    /// </summary>
    internal static string RemoveFailed {
        get {
            return ResourceManager.GetString("RemoveFailed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error renaming '{0}': {1}
    ///  
    /// </summary>
    internal static string RenameFailed {
        get {
            return ResourceManager.GetString("RenameFailed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Parameters "{0}" and "{1}" cannot be specified at the same time.
    ///  
    /// </summary>
    internal static string ContradictParametersSpecified {
        get {
            return ResourceManager.GetString("ContradictParametersSpecified", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot debug process "{0} ({1})" because of the following error: {2}
    ///  
    /// </summary>
    internal static string CouldNotDebugProcess {
        get {
            return ResourceManager.GetString("CouldNotDebugProcess", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The user does not have access to the requested information.
    ///  
    /// </summary>
    internal static string AttachDebuggerReturnCode2 {
        get {
            return ResourceManager.GetString("AttachDebuggerReturnCode2", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified parameter is not valid.
    ///  
    /// </summary>
    internal static string AttachDebuggerReturnCode21 {
        get {
            return ResourceManager.GetString("AttachDebuggerReturnCode21", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The user does not have sufficient privilege.
    ///  
    /// </summary>
    internal static string AttachDebuggerReturnCode3 {
        get {
            return ResourceManager.GetString("AttachDebuggerReturnCode3", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unknown failure.
    ///  
    /// </summary>
    internal static string AttachDebuggerReturnCode8 {
        get {
            return ResourceManager.GetString("AttachDebuggerReturnCode8", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The path specified does not exist.
    ///  
    /// </summary>
    internal static string AttachDebuggerReturnCode9 {
        get {
            return ResourceManager.GetString("AttachDebuggerReturnCode9", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The parameter '{0}' is not supported for the cmdlet '{1}' on this edition of Windows.
    ///  
    /// </summary>
    internal static string ParameterNotSupported {
        get {
            return ResourceManager.GetString("ParameterNotSupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The parameter '{0}' is not supported for the cmdlet '{1}' on this edition of PowerShell.
    ///  
    /// </summary>
    internal static string ParameterNotSupportedOnPSEdition {
        get {
            return ResourceManager.GetString("ParameterNotSupportedOnPSEdition", resourceCulture);
        }
    }

}

