
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class EventlogResources {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal EventlogResources() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.PowerShell.Commands.Management.resources.EventlogResources", typeof(EventlogResources).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Log "{0}" could not be read to completion due to the following error. This may have occurred because the log was cleared while still being read. {1}
    ///  
    /// </summary>
    internal static string LogReadError {
        get {
            return ResourceManager.GetString("LogReadError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    No matches found
    ///  
    /// </summary>
    internal static string NoEntriesFound {
        get {
            return ResourceManager.GetString("NoEntriesFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Do you want to clear the "{0}" log on the computer "{1}"?
    ///  
    /// </summary>
    internal static string ClearEventLogWarning {
        get {
            return ResourceManager.GetString("ClearEventLogWarning", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Log name "{0}" does not exist in the computer "{1}".
    ///  
    /// </summary>
    internal static string LogDoesNotExist {
        get {
            return ResourceManager.GetString("LogDoesNotExist", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The path to the "{1}" computer cannot be found.
    ///  
    /// </summary>
    internal static string PathDoesNotExist {
        get {
            return ResourceManager.GetString("PathDoesNotExist", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The source name "{2}" does not exist on computer "{1}".
    ///  
    /// </summary>
    internal static string SourceDoesNotExist {
        get {
            return ResourceManager.GetString("SourceDoesNotExist", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The registry key for the log "{0}" for source "{2}" could not be opened.
    ///  
    /// </summary>
    internal static string AccessDenied {
        get {
            return ResourceManager.GetString("AccessDenied", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The operating system reported an error when writing the event entry to the event log. A Windows error code is not available.
    ///  
    /// </summary>
    internal static string OSWritingError {
        get {
            return ResourceManager.GetString("OSWritingError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Do you want to change the properties of the "{0}" log on the "{1}" computer?
    ///  
    /// </summary>
    internal static string LimitEventLogWarning {
        get {
            return ResourceManager.GetString("LimitEventLogWarning", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The registry key for the log "{0}" could not be opened on the computer "{1}".
    ///  
    /// </summary>
    internal static string PermissionDenied {
        get {
            return ResourceManager.GetString("PermissionDenied", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The value supplied for MaximumSize parameter has to be in the range of 64 KB to 4GB with an increment of 64 KB. Please enter a proper value and then retry.
    ///  
    /// </summary>
    internal static string ValueOutofRange {
        get {
            return ResourceManager.GetString("ValueOutofRange", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Access to the "{1}" computer is denied.
    ///  
    /// </summary>
    internal static string NoAccess {
        get {
            return ResourceManager.GetString("NoAccess", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The "{2}" source is already registered with the "{0}" log.
    ///  
    /// </summary>
    internal static string SourceExistWithLog {
        get {
            return ResourceManager.GetString("SourceExistWithLog", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The "{2}" source is already registered on the "{1}" computer.
    ///  
    /// </summary>
    internal static string SourceExistInComp {
        get {
            return ResourceManager.GetString("SourceExistInComp", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Do you want to remove the "{0}" log from the "{1}" computer?
    ///  
    /// </summary>
    internal static string RemoveEventLogWarning {
        get {
            return ResourceManager.GetString("RemoveEventLogWarning", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Do you want to remove the "{0}" source from the "{1}" computer?
    ///  
    /// </summary>
    internal static string RemoveSourceWarning {
        get {
            return ResourceManager.GetString("RemoveSourceWarning", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Retention days is valid only if the overflow action is "OverwriteOlder". Please change and try again.
    ///  
    /// </summary>
    internal static string InvalidOverflowAction {
        get {
            return ResourceManager.GetString("InvalidOverflowAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Specify a valid value for the number of retention days.
    ///  
    /// </summary>
    internal static string InvalidArgument {
        get {
            return ResourceManager.GetString("InvalidArgument", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Access is denied. Try running the command again in a session that has been opened with elevated user rights (that is, Run as Administrator).
    ///  
    /// </summary>
    internal static string AccessIsDenied {
        get {
            return ResourceManager.GetString("AccessIsDenied", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The command is not supported in this version of the operating system.
    ///  
    /// </summary>
    internal static string NotSupported {
        get {
            return ResourceManager.GetString("NotSupported", resourceCulture);
        }
    }

}

