
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class ErrorCategoryStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal ErrorCategoryStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.ErrorCategoryStrings", typeof(ErrorCategoryStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    CloseError: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string CloseError {
        get {
            return ResourceManager.GetString("CloseError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Deadlock detected: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string DeadlockDetected {
        get {
            return ResourceManager.GetString("DeadlockDetected", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    DeviceError: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string DeviceError {
        get {
            return ResourceManager.GetString("DeviceError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    InvalidArgument: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string InvalidArgument {
        get {
            return ResourceManager.GetString("InvalidArgument", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    InvalidData: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string InvalidData {
        get {
            return ResourceManager.GetString("InvalidData", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    InvalidOperation: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string InvalidOperation {
        get {
            return ResourceManager.GetString("InvalidOperation", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    InvalidResult: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string InvalidResult {
        get {
            return ResourceManager.GetString("InvalidResult", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    InvalidType: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string InvalidType {
        get {
            return ResourceManager.GetString("InvalidType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    MetadataError: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string MetadataError {
        get {
            return ResourceManager.GetString("MetadataError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    NotImplemented: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string NotImplemented {
        get {
            return ResourceManager.GetString("NotImplemented", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    NotInstalled: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string NotInstalled {
        get {
            return ResourceManager.GetString("NotInstalled", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ObjectNotFound: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string ObjectNotFound {
        get {
            return ResourceManager.GetString("ObjectNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    OpenError: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string OpenError {
        get {
            return ResourceManager.GetString("OpenError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    OperationStopped: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string OperationStopped {
        get {
            return ResourceManager.GetString("OperationStopped", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    OperationTimeout: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string OperationTimeout {
        get {
            return ResourceManager.GetString("OperationTimeout", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ParserError: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string ParserError {
        get {
            return ResourceManager.GetString("ParserError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    PermissionDenied: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string PermissionDenied {
        get {
            return ResourceManager.GetString("PermissionDenied", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ReadError: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string ReadError {
        get {
            return ResourceManager.GetString("ReadError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ResourceBusy: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string ResourceBusy {
        get {
            return ResourceManager.GetString("ResourceBusy", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ResourceExists: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string ResourceExists {
        get {
            return ResourceManager.GetString("ResourceExists", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ResourceUnavailable: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string ResourceUnavailable {
        get {
            return ResourceManager.GetString("ResourceUnavailable", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    SyntaxError: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string SyntaxError {
        get {
            return ResourceManager.GetString("SyntaxError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    WriteError: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string WriteError {
        get {
            return ResourceManager.GetString("WriteError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    FromStdErr: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string FromStdErr {
        get {
            return ResourceManager.GetString("FromStdErr", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    SecurityError: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string SecurityError {
        get {
            return ResourceManager.GetString("SecurityError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ProtocolError: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string ProtocolError {
        get {
            return ResourceManager.GetString("ProtocolError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ConnectionError: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string ConnectionError {
        get {
            return ResourceManager.GetString("ConnectionError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    AuthenticationError: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string AuthenticationError {
        get {
            return ResourceManager.GetString("AuthenticationError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    LimitsExceeded: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string LimitsExceeded {
        get {
            return ResourceManager.GetString("LimitsExceeded", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    QuotaExceeded: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string QuotaExceeded {
        get {
            return ResourceManager.GetString("QuotaExceeded", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    NotEnabled: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string NotEnabled {
        get {
            return ResourceManager.GetString("NotEnabled", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    NotSpecified: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string NotSpecified {
        get {
            return ResourceManager.GetString("NotSpecified", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unrecognized error category {4}: ({1}:{2}) [{0}], {3}
    ///  
    /// </summary>
    internal static string InvalidErrorCategory {
        get {
            return ResourceManager.GetString("InvalidErrorCategory", resourceCulture);
        }
    }

}

