
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class FlashExtractStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal FlashExtractStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.PowerShell.Commands.Utility.resources.FlashExtractStrings", typeof(FlashExtractStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    There are two instances of non-sequence Property '{0}' at {1} and {2} in parent Span {3}
    ///  
    /// </summary>
    internal static string ChildNamesMustBeUnique {
        get {
            return ResourceManager.GetString("ChildNamesMustBeUnique", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Property '{0}' at {1} has no value and is not optional or a containing span
    ///  
    /// </summary>
    internal static string NonOptionalNonStructMustHaveValue {
        get {
            return ResourceManager.GetString("NonOptionalNonStructMustHaveValue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    No program can be found for the given input
    ///  
    /// </summary>
    internal static string NoProgramFound {
        get {
            return ResourceManager.GetString("NoProgramFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Internal error: Property cannot be null
    ///  
    /// </summary>
    internal static string NullProperty {
        get {
            return ResourceManager.GetString("NullProperty", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unexpected Span ending bracket found at {0}
    ///  
    /// </summary>
    internal static string UnexpectedEndingBracket {
        get {
            return ResourceManager.GetString("UnexpectedEndingBracket", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Missing Span value at {0} while processing {1}
    ///  
    /// </summary>
    internal static string MissingSpan {
        get {
            return ResourceManager.GetString("MissingSpan", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Invalid format at {0} while processing {1}
    ///  
    /// </summary>
    internal static string InvalidFormat {
        get {
            return ResourceManager.GetString("InvalidFormat", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Invalid type name '{0}' at {1} while processing {2}
    ///  
    /// </summary>
    internal static string InvalidTypeName {
        get {
            return ResourceManager.GetString("InvalidTypeName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Missing Span name at {0} while processing {1}
    ///  
    /// </summary>
    internal static string InvalidSpanName {
        get {
            return ResourceManager.GetString("InvalidSpanName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    First character of a name must be a letter or underscore at {0} while processing {1}
    ///  
    /// </summary>
    internal static string InvalidFirstNameChar {
        get {
            return ResourceManager.GetString("InvalidFirstNameChar", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Characters of a name must be letters, digits, or underscores at {0} while processing {1}
    ///  
    /// </summary>
    internal static string InvalidNameChar {
        get {
            return ResourceManager.GetString("InvalidNameChar", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot embed attribute suffixes (e.g. optional, sequence) in names at {0} while processing {1}
    ///  
    /// </summary>
    internal static string AttributeSuffixWithinName {
        get {
            return ResourceManager.GetString("AttributeSuffixWithinName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot use reserved word '{2}' at {0} while processing {1}
    ///  
    /// </summary>
    internal static string ReservedWord {
        get {
            return ResourceManager.GetString("ReservedWord", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Expected value indicator at {0} while processing {1}
    ///  
    /// </summary>
    internal static string ValueIndicatorNotFound {
        get {
            return ResourceManager.GetString("ValueIndicatorNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unable to convert string value to specified type while processing {0}
    ///  
    /// </summary>
    internal static string CannotConvertValue {
        get {
            return ResourceManager.GetString("CannotConvertValue", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unexpected EOF while processing {0}
    ///  
    /// </summary>
    internal static string UnexpectedEOF {
        get {
            return ResourceManager.GetString("UnexpectedEOF", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Span starting at line {0} column {1}
    ///  
    /// </summary>
    internal static string SpanStartPosition {
        get {
            return ResourceManager.GetString("SpanStartPosition", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Internal error: Property '{0}' at {1} has no containing parent
    ///  
    /// </summary>
    internal static string NoParent {
        get {
            return ResourceManager.GetString("NoParent", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Property '{0}' definition at {1} is inconsistent with earlier definition(s)
    ///  
    /// </summary>
    internal static string InconsistentPropertyDefinition {
        get {
            return ResourceManager.GetString("InconsistentPropertyDefinition", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Internal error: no parent region found.
    ///  
    /// </summary>
    internal static string NoParentRegionFound {
        get {
            return ResourceManager.GetString("NoParentRegionFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Internal error: parent regions cannot be null or empty.
    ///  
    /// </summary>
    internal static string ParentRegionsCannotBeEmpty {
        get {
            return ResourceManager.GetString("ParentRegionsCannotBeEmpty", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    line {0} column {1}
    ///  
    /// </summary>
    internal static string ParserSpanStartPosition {
        get {
            return ResourceManager.GetString("ParserSpanStartPosition", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    at {0}
    ///  
    /// </summary>
    internal static string AtPosition {
        get {
            return ResourceManager.GetString("AtPosition", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    '{0}'
    ///  
    /// </summary>
    internal static string SingleNameString {
        get {
            return ResourceManager.GetString("SingleNameString", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    line {0} column {1}
    ///  
    /// </summary>
    internal static string MarkupCharPosition {
        get {
            return ResourceManager.GetString("MarkupCharPosition", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    No template text file start for Span {0}
    ///  
    /// </summary>
    internal static string NoStrippedTemplateFileStart {
        get {
            return ResourceManager.GetString("NoStrippedTemplateFileStart", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Internal error: Property '{0}' not found
    ///  
    /// </summary>
    internal static string PropertyNotFound {
        get {
            return ResourceManager.GetString("PropertyNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Internal error: Invalid parser operation
    ///  
    /// </summary>
    internal static string InvalidOperation {
        get {
            return ResourceManager.GetString("InvalidOperation", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cached program was not found in template file for property: {0}
    ///  
    /// </summary>
    internal static string CachedProgramNotFound {
        get {
            return ResourceManager.GetString("CachedProgramNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Template text does not match cached programs
    ///  
    /// </summary>
    internal static string DifferentTemplateHash {
        get {
            return ResourceManager.GetString("DifferentTemplateHash", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The template text contains no example to parse
    ///  
    /// </summary>
    internal static string NoExampleFound {
        get {
            return ResourceManager.GetString("NoExampleFound", resourceCulture);
        }
    }

}

