<!--=================================================================
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT License.
==================================================================-->

<UserControl x:Class="Microsoft.PowerShell.Commands.ShowCommandInternal.NotImportedCmdletControl"
             xmlns="http://schemas.microsoft.com/winfx/2006/xaml/presentation"
             xmlns:x="http://schemas.microsoft.com/winfx/2006/xaml"
             xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006"
             xmlns:d="http://schemas.microsoft.com/expression/blend/2008"
             xmlns:this="clr-namespace:Microsoft.PowerShell.Commands.ShowCommandInternal"
             xmlns:default="clr-namespace:"
             xmlns:sc="clr-namespace:Microsoft.Management.UI.Internal"
             FocusManager.FocusedElement="{Binding ElementName=CommandName}"
             mc:Ignorable="d"
             d:DesignHeight="300" d:DesignWidth="300">
    <Border BorderThickness="1" Padding="0,3,3,0" BorderBrush="{x:Static SystemColors.ControlDarkBrush}">
        <Grid Margin="3">
            <Grid.RowDefinitions>
                <RowDefinition Height="Auto"></RowDefinition>
                <RowDefinition Height="Auto"></RowDefinition>
            </Grid.RowDefinitions>
            <TextBlock x:Name="ImportModuleMessage" Grid.Row="0" Margin="3" Text="{Binding ImportModuleMessage, Mode=OneWay}" Focusable="True" TextWrapping="Wrap"></TextBlock>
            <Button x:Name="ImportModuleButton" Grid.Row="1" Margin="0,3,0,3" Padding="2" HorizontalAlignment="Right" Content="{x:Static default:ShowCommandResources.ImportModuleButtonText}"
                IsEnabled="{Binding Path='ParentModule.AllModules.ExtraViewModel.CanRefresh'}"></Button>
        </Grid>
    </Border>
</UserControl>
