
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class AutomationExceptions {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal AutomationExceptions() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.AutomationExceptions", typeof(AutomationExceptions).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process argument because the value of argument "{0}" is not valid. Change the value of the "{0}" argument and run the operation again.
    ///  
    /// </summary>
    internal static string Argument {
        get {
            return ResourceManager.GetString("Argument", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process argument because the value of parameter "{0}" is not valid. Valid values are "Global", "Local", or "Script", or a number relative to the current scope (0 through the number of scopes where 0 is the current scope and 1 is its parent). Change the value of the "{0}" parameter and run the operation again.
    ///  
    /// </summary>
    internal static string InvalidScopeIdArgument {
        get {
            return ResourceManager.GetString("InvalidScopeIdArgument", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process argument because the value of argument "{0}" is null. Change the value of argument "{0}" to a non-null value.
    ///  
    /// </summary>
    internal static string ArgumentNull {
        get {
            return ResourceManager.GetString("ArgumentNull", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process argument because the value of argument "{0}" is out of range. Change argument "{0}" to a value that is within range.
    ///  
    /// </summary>
    internal static string ArgumentOutOfRange {
        get {
            return ResourceManager.GetString("ArgumentOutOfRange", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot perform operation because operation "{0}" is not valid. Remove operation "{0}", or investigate why it is not valid.
    ///  
    /// </summary>
    internal static string InvalidOperation {
        get {
            return ResourceManager.GetString("InvalidOperation", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot perform operation because operation "{0}" is not implemented.
    ///  
    /// </summary>
    internal static string NotImplemented {
        get {
            return ResourceManager.GetString("NotImplemented", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot perform operation because operation "{0}" is not supported.
    ///  
    /// </summary>
    internal static string NotSupported {
        get {
            return ResourceManager.GetString("NotSupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot perform operation because object "{0}" has already been disposed.
    ///  
    /// </summary>
    internal static string ObjectDisposed {
        get {
            return ResourceManager.GetString("ObjectDisposed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The script block cannot be invoked because it contains more than one clause. The Invoke() method can only be used on script blocks that contain a single clause.
    ///  
    /// </summary>
    internal static string ScriptBlockInvokeOnOneClauseOnly {
        get {
            return ResourceManager.GetString("ScriptBlockInvokeOnOneClauseOnly", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The script block cannot be converted because it contains more than one clause. Expressions or control structures are not permitted. Verify that the script block contains exactly one pipeline or command.
    ///  
    /// </summary>
    internal static string CanConvertOneClauseOnly {
        get {
            return ResourceManager.GetString("CanConvertOneClauseOnly", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    An empty script block cannot be converted. Verify that the script block contains exactly one pipeline or command.
    ///  
    /// </summary>
    internal static string CantConvertEmptyPipeline {
        get {
            return ResourceManager.GetString("CantConvertEmptyPipeline", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Only a script block that contains exactly one pipeline or command can be converted. Expressions or control structures are not permitted. Verify that the script block contains exactly one pipeline or command.
    ///  
    /// </summary>
    internal static string CanOnlyConvertOnePipeline {
        get {
            return ResourceManager.GetString("CanOnlyConvertOnePipeline", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A script block that contains a top-level trap statement cannot be converted.
    ///  
    /// </summary>
    internal static string CantConvertScriptBlockWithTrap {
        get {
            return ResourceManager.GetString("CantConvertScriptBlockWithTrap", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot generate a PowerShell object for a ScriptBlock dereferencing variables undeclared in the param(...) block.  Name of undeclared variable: {0}.
    ///  
    /// </summary>
    internal static string CantConvertWithUndeclaredVariables {
        get {
            return ResourceManager.GetString("CantConvertWithUndeclaredVariables", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot generate a PowerShell object for a ScriptBlock evaluating non-constant expressions. Non-constant expression: {0}.
    ///  
    /// </summary>
    internal static string CantConvertWithNonConstantExpression {
        get {
            return ResourceManager.GetString("CantConvertWithNonConstantExpression", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot generate a PowerShell object for a ScriptBlock evaluating dynamic expressions. Dynamic expression: {0}.
    ///  
    /// </summary>
    internal static string CantConvertWithDynamicExpression {
        get {
            return ResourceManager.GetString("CantConvertWithDynamicExpression", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot generate a PowerShell object for a ScriptBlock that tries to pass other script blocks inside argument values.
    ///  
    /// </summary>
    internal static string CantConvertWithScriptBlocks {
        get {
            return ResourceManager.GetString("CantConvertWithScriptBlocks", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot generate a PowerShell object for a ScriptBlock which invokes pipelines, commands or functions to evaluate arguments of the main pipeline.
    ///  
    /// </summary>
    internal static string CantConvertWithCommandInvocations {
        get {
            return ResourceManager.GetString("CantConvertWithCommandInvocations", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot generate a PowerShell object for a ScriptBlock that uses dot sourcing.
    ///  
    /// </summary>
    internal static string CantConvertWithDotSourcing {
        get {
            return ResourceManager.GetString("CantConvertWithDotSourcing", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot generate a PowerShell object for a ScriptBlock that invokes other script blocks.
    ///  
    /// </summary>
    internal static string CantConvertWithScriptBlockInvocation {
        get {
            return ResourceManager.GetString("CantConvertWithScriptBlockInvocation", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The script block cannot be converted to a PowerShell object because it contains forbidden redirection operators.
    ///  
    /// </summary>
    internal static string CanConvertOneOutputErrorRedir {
        get {
            return ResourceManager.GetString("CanConvertOneOutputErrorRedir", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot generate a PowerShell object for a ScriptBlock that does not have an associated operation context.
    ///  
    /// </summary>
    internal static string CantConvertScriptBlockWithNoContext {
        get {
            return ResourceManager.GetString("CantConvertScriptBlockWithNoContext", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The command was stopped by the user.
    ///  
    /// </summary>
    internal static string HaltCommandException {
        get {
            return ResourceManager.GetString("HaltCommandException", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Object "{0}" is the wrong type to return from the dynamicparam block. The dynamicparam block must return either $null, or an object with type [System.Management.Automation.RuntimeDefinedParameterDictionary].
    ///  
    /// </summary>
    internal static string DynamicParametersWrongType {
        get {
            return ResourceManager.GetString("DynamicParametersWrongType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The script block cannot be converted to an open generic type. Define an appropriate closed generic type, and then retry.
    ///  
    /// </summary>
    internal static string CantConvertScriptBlockToOpenGenericType {
        get {
            return ResourceManager.GetString("CantConvertScriptBlockToOpenGenericType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot generate a PowerShell object for a ScriptBlock that starts a pipeline with an expression.
    ///  
    /// </summary>
    internal static string CantConvertPipelineStartsWithExpression {
        get {
            return ResourceManager.GetString("CantConvertPipelineStartsWithExpression", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The value of the using variable '$using:{0}' cannot be retrieved because it has not been set in the local session.
    ///  
    /// </summary>
    internal static string UsingVariableIsUndefined {
        get {
            return ResourceManager.GetString("UsingVariableIsUndefined", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot get the value of the Using expression '{0}' in the specified variable dictionary. When creating a PowerShell instance from a script block, the Using expression cannot contain an indexing operation or member-accessing operation.
    ///  
    /// </summary>
    internal static string CantGetUsingExpressionValueWithSpecifiedVariableDictionary {
        get {
            return ResourceManager.GetString("CantGetUsingExpressionValueWithSpecifiedVariableDictionary", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Compiled Script Block Dot Source
    ///  
    /// </summary>
    internal static string WDACCompiledScriptBlockLogTitle {
        get {
            return ResourceManager.GetString("WDACCompiledScriptBlockLogTitle", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Script block '{0}' invocation into current scope will be disallowed in Constrained Language mode. Script language mode: {1}, Context language mode: {2}.
    ///  
    /// </summary>
    internal static string WDACCompiledScriptBlockLogMessage {
        get {
            return ResourceManager.GetString("WDACCompiledScriptBlockLogMessage", resourceCulture);
        }
    }

}

