
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class VerbDescriptionStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal VerbDescriptionStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.VerbDescriptionStrings", typeof(VerbDescriptionStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Adds a resource to a container, or attaches an item to another item
    ///  
    /// </summary>
    internal static string Add {
        get {
            return ResourceManager.GetString("Add", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Confirms or agrees to the status of a resource or process
    ///  
    /// </summary>
    internal static string Approve {
        get {
            return ResourceManager.GetString("Approve", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Affirms the state of a resource
    ///  
    /// </summary>
    internal static string Assert {
        get {
            return ResourceManager.GetString("Assert", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Stores data by replicating it
    ///  
    /// </summary>
    internal static string Backup {
        get {
            return ResourceManager.GetString("Backup", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Restricts access to a resource
    ///  
    /// </summary>
    internal static string Block {
        get {
            return ResourceManager.GetString("Block", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Creates an artifact (usually a binary or document) out of some set of input files (usually source code or declarative documents)
    ///  
    /// </summary>
    internal static string Build {
        get {
            return ResourceManager.GetString("Build", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Creates a snapshot of the current state of the data or of its configuration
    ///  
    /// </summary>
    internal static string Checkpoint {
        get {
            return ResourceManager.GetString("Checkpoint", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Removes all the resources from a container but does not delete the container
    ///  
    /// </summary>
    internal static string Clear {
        get {
            return ResourceManager.GetString("Clear", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Changes the state of a resource to make it inaccessible, unavailable, or unusable
    ///  
    /// </summary>
    internal static string Close {
        get {
            return ResourceManager.GetString("Close", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Evaluates the data from one resource against the data from another resource
    ///  
    /// </summary>
    internal static string Compare {
        get {
            return ResourceManager.GetString("Compare", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Concludes an operation
    ///  
    /// </summary>
    internal static string Complete {
        get {
            return ResourceManager.GetString("Complete", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Compacts the data of a resource
    ///  
    /// </summary>
    internal static string Compress {
        get {
            return ResourceManager.GetString("Compress", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Acknowledges, verifies, or validates the state of a resource or process
    ///  
    /// </summary>
    internal static string Confirm {
        get {
            return ResourceManager.GetString("Confirm", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Creates a link between a source and a destination
    ///  
    /// </summary>
    internal static string Connect {
        get {
            return ResourceManager.GetString("Connect", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Changes the data from one representation to another when the cmdlet supports bidirectional conversion or when the cmdlet supports conversion between multiple data types
    ///  
    /// </summary>
    internal static string Convert {
        get {
            return ResourceManager.GetString("Convert", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Converts one primary type of input (the cmdlet noun indicates the input) to one or more supported output types
    ///  
    /// </summary>
    internal static string ConvertFrom {
        get {
            return ResourceManager.GetString("ConvertFrom", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Converts from one or more types of input to a primary output type (the cmdlet noun indicates the output type)
    ///  
    /// </summary>
    internal static string ConvertTo {
        get {
            return ResourceManager.GetString("ConvertTo", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Copies a resource to another name or to another container
    ///  
    /// </summary>
    internal static string Copy {
        get {
            return ResourceManager.GetString("Copy", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Examines a resource to diagnose operational problems
    ///  
    /// </summary>
    internal static string Debug {
        get {
            return ResourceManager.GetString("Debug", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Refuses, objects, blocks, or opposes the state of a resource or process
    ///  
    /// </summary>
    internal static string Deny {
        get {
            return ResourceManager.GetString("Deny", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Sends an application, website, or solution to a remote target[s] in such a way that a consumer of that solution can access it after deployment is complete
    ///  
    /// </summary>
    internal static string Deploy {
        get {
            return ResourceManager.GetString("Deploy", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Configures a resource to an unavailable or inactive state
    ///  
    /// </summary>
    internal static string Disable {
        get {
            return ResourceManager.GetString("Disable", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Breaks the link between a source and a destination
    ///  
    /// </summary>
    internal static string Disconnect {
        get {
            return ResourceManager.GetString("Disconnect", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Detaches a named entity from a location
    ///  
    /// </summary>
    internal static string Dismount {
        get {
            return ResourceManager.GetString("Dismount", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Modifies existing data by adding or removing content
    ///  
    /// </summary>
    internal static string Edit {
        get {
            return ResourceManager.GetString("Edit", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Configures a resource to an available or active state
    ///  
    /// </summary>
    internal static string Enable {
        get {
            return ResourceManager.GetString("Enable", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Specifies an action that allows the user to move into a resource
    ///  
    /// </summary>
    internal static string Enter {
        get {
            return ResourceManager.GetString("Enter", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Sets the current environment or context to the most recently used context
    ///  
    /// </summary>
    internal static string Exit {
        get {
            return ResourceManager.GetString("Exit", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Restores the data of a resource that has been compressed to its original state
    ///  
    /// </summary>
    internal static string Expand {
        get {
            return ResourceManager.GetString("Expand", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Encapsulates the primary input into a persistent data store, such as a file, or into an interchange format
    ///  
    /// </summary>
    internal static string Export {
        get {
            return ResourceManager.GetString("Export", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Looks for an object in a container that is unknown, implied, optional, or specified
    ///  
    /// </summary>
    internal static string Find {
        get {
            return ResourceManager.GetString("Find", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Arranges objects in a specified form or layout
    ///  
    /// </summary>
    internal static string Format {
        get {
            return ResourceManager.GetString("Format", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Specifies an action that retrieves a resource
    ///  
    /// </summary>
    internal static string Get {
        get {
            return ResourceManager.GetString("Get", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Allows access to a resource
    ///  
    /// </summary>
    internal static string Grant {
        get {
            return ResourceManager.GetString("Grant", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Arranges or associates one or more resources
    ///  
    /// </summary>
    internal static string Group {
        get {
            return ResourceManager.GetString("Group", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Makes a resource undetectable
    ///  
    /// </summary>
    internal static string Hide {
        get {
            return ResourceManager.GetString("Hide", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Creates a resource from data that is stored in a persistent data store (such as a file) or in an interchange format
    ///  
    /// </summary>
    internal static string Import {
        get {
            return ResourceManager.GetString("Import", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Prepares a resource for use, and sets it to a default state
    ///  
    /// </summary>
    internal static string Initialize {
        get {
            return ResourceManager.GetString("Initialize", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Places a resource in a location, and optionally initializes it
    ///  
    /// </summary>
    internal static string Install {
        get {
            return ResourceManager.GetString("Install", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Performs an action, such as running a command or a method
    ///  
    /// </summary>
    internal static string Invoke {
        get {
            return ResourceManager.GetString("Invoke", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Combines resources into one resource
    ///  
    /// </summary>
    internal static string Join {
        get {
            return ResourceManager.GetString("Join", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Applies constraints to a resource
    ///  
    /// </summary>
    internal static string Limit {
        get {
            return ResourceManager.GetString("Limit", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Secures a resource
    ///  
    /// </summary>
    internal static string Lock {
        get {
            return ResourceManager.GetString("Lock", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Identifies resources that are consumed by a specified operation, or retrieves statistics about a resource
    ///  
    /// </summary>
    internal static string Measure {
        get {
            return ResourceManager.GetString("Measure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Creates a single resource from multiple resources
    ///  
    /// </summary>
    internal static string Merge {
        get {
            return ResourceManager.GetString("Merge", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Attaches a named entity to a location
    ///  
    /// </summary>
    internal static string Mount {
        get {
            return ResourceManager.GetString("Mount", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Moves a resource from one location to another
    ///  
    /// </summary>
    internal static string Move {
        get {
            return ResourceManager.GetString("Move", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Creates a resource
    ///  
    /// </summary>
    internal static string New {
        get {
            return ResourceManager.GetString("New", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Changes the state of a resource to make it accessible, available, or usable
    ///  
    /// </summary>
    internal static string Open {
        get {
            return ResourceManager.GetString("Open", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Increases the effectiveness of a resource
    ///  
    /// </summary>
    internal static string Optimize {
        get {
            return ResourceManager.GetString("Optimize", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Sends data out of the environment
    ///  
    /// </summary>
    internal static string Out {
        get {
            return ResourceManager.GetString("Out", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Use the Test verb
    ///  
    /// </summary>
    internal static string Ping {
        get {
            return ResourceManager.GetString("Ping", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Removes an item from the top of a stack
    ///  
    /// </summary>
    internal static string Pop {
        get {
            return ResourceManager.GetString("Pop", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Safeguards a resource from attack or loss
    ///  
    /// </summary>
    internal static string Protect {
        get {
            return ResourceManager.GetString("Protect", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Makes a resource available to others
    ///  
    /// </summary>
    internal static string Publish {
        get {
            return ResourceManager.GetString("Publish", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Adds an item to the top of a stack
    ///  
    /// </summary>
    internal static string Push {
        get {
            return ResourceManager.GetString("Push", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Acquires information from a source
    ///  
    /// </summary>
    internal static string Read {
        get {
            return ResourceManager.GetString("Read", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Accepts information sent from a source
    ///  
    /// </summary>
    internal static string Receive {
        get {
            return ResourceManager.GetString("Receive", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Resets a resource to the state that was undone
    ///  
    /// </summary>
    internal static string Redo {
        get {
            return ResourceManager.GetString("Redo", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Creates an entry for a resource in a repository such as a database
    ///  
    /// </summary>
    internal static string Register {
        get {
            return ResourceManager.GetString("Register", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Deletes a resource from a container
    ///  
    /// </summary>
    internal static string Remove {
        get {
            return ResourceManager.GetString("Remove", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Changes the name of a resource
    ///  
    /// </summary>
    internal static string Rename {
        get {
            return ResourceManager.GetString("Rename", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Restores a resource to a usable condition
    ///  
    /// </summary>
    internal static string Repair {
        get {
            return ResourceManager.GetString("Repair", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Asks for a resource or asks for permissions
    ///  
    /// </summary>
    internal static string Request {
        get {
            return ResourceManager.GetString("Request", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Sets a resource back to its original state
    ///  
    /// </summary>
    internal static string Reset {
        get {
            return ResourceManager.GetString("Reset", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Changes the size of a resource
    ///  
    /// </summary>
    internal static string Resize {
        get {
            return ResourceManager.GetString("Resize", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Maps a shorthand representation of a resource to a more complete representation
    ///  
    /// </summary>
    internal static string Resolve {
        get {
            return ResourceManager.GetString("Resolve", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Stops an operation and then starts it again
    ///  
    /// </summary>
    internal static string Restart {
        get {
            return ResourceManager.GetString("Restart", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Sets a resource to a predefined state, such as a state set by Checkpoint
    ///  
    /// </summary>
    internal static string Restore {
        get {
            return ResourceManager.GetString("Restore", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Starts an operation that has been suspended
    ///  
    /// </summary>
    internal static string Resume {
        get {
            return ResourceManager.GetString("Resume", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Specifies an action that does not allow access to a resource
    ///  
    /// </summary>
    internal static string Revoke {
        get {
            return ResourceManager.GetString("Revoke", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Preserves data to avoid loss
    ///  
    /// </summary>
    internal static string Save {
        get {
            return ResourceManager.GetString("Save", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Creates a reference to a resource in a container
    ///  
    /// </summary>
    internal static string Search {
        get {
            return ResourceManager.GetString("Search", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Locates a resource in a container
    ///  
    /// </summary>
    internal static string Select {
        get {
            return ResourceManager.GetString("Select", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Delivers information to a destination
    ///  
    /// </summary>
    internal static string Send {
        get {
            return ResourceManager.GetString("Send", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Replaces data on an existing resource or creates a resource that contains some data
    ///  
    /// </summary>
    internal static string Set {
        get {
            return ResourceManager.GetString("Set", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Makes a resource visible to the user
    ///  
    /// </summary>
    internal static string Show {
        get {
            return ResourceManager.GetString("Show", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Assures that two or more resources are in the same state
    ///  
    /// </summary>
    internal static string Sync {
        get {
            return ResourceManager.GetString("Sync", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Bypasses one or more resources or points in a sequence
    ///  
    /// </summary>
    internal static string Skip {
        get {
            return ResourceManager.GetString("Skip", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Separates parts of a resource
    ///  
    /// </summary>
    internal static string Split {
        get {
            return ResourceManager.GetString("Split", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Initiates an operation
    ///  
    /// </summary>
    internal static string Start {
        get {
            return ResourceManager.GetString("Start", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Moves to the next point or resource in a sequence
    ///  
    /// </summary>
    internal static string Step {
        get {
            return ResourceManager.GetString("Step", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Discontinues an activity
    ///  
    /// </summary>
    internal static string Stop {
        get {
            return ResourceManager.GetString("Stop", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Presents a resource for approval
    ///  
    /// </summary>
    internal static string Submit {
        get {
            return ResourceManager.GetString("Submit", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Pauses an activity
    ///  
    /// </summary>
    internal static string Suspend {
        get {
            return ResourceManager.GetString("Suspend", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Specifies an action that alternates between two resources, such as to change between two locations, responsibilities, or states
    ///  
    /// </summary>
    internal static string Switch {
        get {
            return ResourceManager.GetString("Switch", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Verifies the operation or consistency of a resource
    ///  
    /// </summary>
    internal static string Test {
        get {
            return ResourceManager.GetString("Test", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Tracks the activities of a resource
    ///  
    /// </summary>
    internal static string Trace {
        get {
            return ResourceManager.GetString("Trace", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Removes restrictions to a resource
    ///  
    /// </summary>
    internal static string Unblock {
        get {
            return ResourceManager.GetString("Unblock", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Sets a resource to its previous state
    ///  
    /// </summary>
    internal static string Undo {
        get {
            return ResourceManager.GetString("Undo", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Removes a resource from an indicated location
    ///  
    /// </summary>
    internal static string Uninstall {
        get {
            return ResourceManager.GetString("Uninstall", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Releases a resource that was locked
    ///  
    /// </summary>
    internal static string Unlock {
        get {
            return ResourceManager.GetString("Unlock", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Removes safeguards from a resource that were added to prevent it from attack or loss
    ///  
    /// </summary>
    internal static string Unprotect {
        get {
            return ResourceManager.GetString("Unprotect", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Makes a resource unavailable to others
    ///  
    /// </summary>
    internal static string Unpublish {
        get {
            return ResourceManager.GetString("Unpublish", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Removes the entry for a resource from a repository
    ///  
    /// </summary>
    internal static string Unregister {
        get {
            return ResourceManager.GetString("Unregister", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Brings a resource up-to-date to maintain its state, accuracy, conformance, or compliance
    ///  
    /// </summary>
    internal static string Update {
        get {
            return ResourceManager.GetString("Update", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Uses or includes a resource to do something
    ///  
    /// </summary>
    internal static string Use {
        get {
            return ResourceManager.GetString("Use", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Pauses an operation until a specified event occurs
    ///  
    /// </summary>
    internal static string Wait {
        get {
            return ResourceManager.GetString("Wait", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Continually inspects or monitors a resource for changes
    ///  
    /// </summary>
    internal static string Watch {
        get {
            return ResourceManager.GetString("Watch", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Adds information to a target
    ///  
    /// </summary>
    internal static string Write {
        get {
            return ResourceManager.GetString("Write", resourceCulture);
        }
    }

}

