
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class EtwLoggingStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal EtwLoggingStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.EtwLoggingStrings", typeof(EtwLoggingStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Command {0} is {1}.
    ///  
    /// </summary>
    internal static string CommandStateChange {
        get {
            return ResourceManager.GetString("CommandStateChange", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Engine state changed from {0} to {1}.
    ///  
    /// </summary>
    internal static string EngineStateChange {
        get {
            return ResourceManager.GetString("EngineStateChange", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Fully Qualified Error ID = {0}
    ///  
    /// </summary>
    internal static string ErrorRecordId {
        get {
            return ResourceManager.GetString("ErrorRecordId", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Error Message = {0}
    ///  
    /// </summary>
    internal static string ErrorRecordMessage {
        get {
            return ResourceManager.GetString("ErrorRecordMessage", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Recommended Action = {0}
    ///  
    /// </summary>
    internal static string ErrorRecordRecommendedAction {
        get {
            return ResourceManager.GetString("ErrorRecordRecommendedAction", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Execution Policy
    ///  
    /// </summary>
    internal static string ExecutionPolicyName {
        get {
            return ResourceManager.GetString("ExecutionPolicyName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Job Command = {0}
    ///  
    /// </summary>
    internal static string JobCommand {
        get {
            return ResourceManager.GetString("JobCommand", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Job Id = {0}
    ///  
    /// </summary>
    internal static string JobId {
        get {
            return ResourceManager.GetString("JobId", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Job Instance Id = {0}
    ///  
    /// </summary>
    internal static string JobInstanceId {
        get {
            return ResourceManager.GetString("JobInstanceId", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Job Location = {0}
    ///  
    /// </summary>
    internal static string JobLocation {
        get {
            return ResourceManager.GetString("JobLocation", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Job Name = {0}
    ///  
    /// </summary>
    internal static string JobName {
        get {
            return ResourceManager.GetString("JobName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Job State = {0}
    ///  
    /// </summary>
    internal static string JobState {
        get {
            return ResourceManager.GetString("JobState", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///            Command Name = 
    ///  
    /// </summary>
    internal static string LogContextCommandName {
        get {
            return ResourceManager.GetString("LogContextCommandName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///            Command Path = 
    ///  
    /// </summary>
    internal static string LogContextCommandPath {
        get {
            return ResourceManager.GetString("LogContextCommandPath", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///            Command Type = 
    ///  
    /// </summary>
    internal static string LogContextCommandType {
        get {
            return ResourceManager.GetString("LogContextCommandType", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///            Engine Version = 
    ///  
    /// </summary>
    internal static string LogContextEngineVersion {
        get {
            return ResourceManager.GetString("LogContextEngineVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///            Host ID = 
    ///  
    /// </summary>
    internal static string LogContextHostId {
        get {
            return ResourceManager.GetString("LogContextHostId", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///            Host Name = 
    ///  
    /// </summary>
    internal static string LogContextHostName {
        get {
            return ResourceManager.GetString("LogContextHostName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///            Host Application = 
    ///  
    /// </summary>
    internal static string LogContextHostApplication {
        get {
            return ResourceManager.GetString("LogContextHostApplication", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///            Host Version = 
    ///  
    /// </summary>
    internal static string LogContextHostVersion {
        get {
            return ResourceManager.GetString("LogContextHostVersion", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///            Pipeline ID = 
    ///  
    /// </summary>
    internal static string LogContextPipelineId {
        get {
            return ResourceManager.GetString("LogContextPipelineId", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///            Runspace ID = 
    ///  
    /// </summary>
    internal static string LogContextRunspaceId {
        get {
            return ResourceManager.GetString("LogContextRunspaceId", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///            Script Name = 
    ///  
    /// </summary>
    internal static string LogContextScriptName {
        get {
            return ResourceManager.GetString("LogContextScriptName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///            Sequence Number = 
    ///  
    /// </summary>
    internal static string LogContextSequenceNumber {
        get {
            return ResourceManager.GetString("LogContextSequenceNumber", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///            Severity = 
    ///  
    /// </summary>
    internal static string LogContextSeverity {
        get {
            return ResourceManager.GetString("LogContextSeverity", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///            Shell ID = 
    ///  
    /// </summary>
    internal static string LogContextShellId {
        get {
            return ResourceManager.GetString("LogContextShellId", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///            Time = 
    ///  
    /// </summary>
    internal static string LogContextTime {
        get {
            return ResourceManager.GetString("LogContextTime", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///            User = 
    ///  
    /// </summary>
    internal static string LogContextUser {
        get {
            return ResourceManager.GetString("LogContextUser", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///            Connected User = 
    ///  
    /// </summary>
    internal static string LogContextConnectedUser {
        get {
            return ResourceManager.GetString("LogContextConnectedUser", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    NULL Job
    ///  
    /// </summary>
    internal static string NullJobName {
        get {
            return ResourceManager.GetString("NullJobName", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Provider name
    ///  
    /// </summary>
    internal static string ProviderNameString {
        get {
            return ResourceManager.GetString("ProviderNameString", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Provider {0} changed state to {1}.
    ///  
    /// </summary>
    internal static string ProviderStateChange {
        get {
            return ResourceManager.GetString("ProviderStateChange", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Script execution is {0}.
    ///  
    /// </summary>
    internal static string ScriptStateChange {
        get {
            return ResourceManager.GetString("ScriptStateChange", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Variable {0} changed from {1} to {2}.
    ///  
    /// </summary>
    internal static string SettingChange {
        get {
            return ResourceManager.GetString("SettingChange", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Variable {0} changed to {1}.
    ///  
    /// </summary>
    internal static string SettingChangeNoPrevious {
        get {
            return ResourceManager.GetString("SettingChangeNoPrevious", resourceCulture);
        }
    }

}

