
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class WebCmdletStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal WebCmdletStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("Microsoft.PowerShell.Commands.Utility.resources.WebCmdletStrings", typeof(WebCmdletStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Access to the path '{0}' is denied.
    ///  
    /// </summary>
    internal static string AccessDenied {
        get {
            return ResourceManager.GetString("AccessDenied", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The cmdlet cannot protect plain text secrets sent over unencrypted connections. To suppress this warning and send plain text secrets over unencrypted networks, reissue the command specifying the AllowUnencryptedAuthentication parameter.
    ///  
    /// </summary>
    internal static string AllowUnencryptedAuthenticationRequired {
        get {
            return ResourceManager.GetString("AllowUnencryptedAuthenticationRequired", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The cmdlet cannot run because the following conflicting parameters are specified: Authentication and UseDefaultCredentials. Authentication does not support Default Credentials. Specify either Authentication or UseDefaultCredentials, then retry.
    ///  
    /// </summary>
    internal static string AuthenticationConflict {
        get {
            return ResourceManager.GetString("AuthenticationConflict", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The cmdlet cannot run because the following parameter is not specified: Credential. The supplied Authentication type requires a Credential. Specify Credential, then retry.
    ///  
    /// </summary>
    internal static string AuthenticationCredentialNotSupplied {
        get {
            return ResourceManager.GetString("AuthenticationCredentialNotSupplied", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The cmdlet cannot run because the following parameter is not specified: Token. The supplied Authentication type requires a Token. Specify Token, then retry.
    ///  
    /// </summary>
    internal static string AuthenticationTokenNotSupplied {
        get {
            return ResourceManager.GetString("AuthenticationTokenNotSupplied", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The cmdlet cannot run because the following conflicting parameters are specified: Credential and Token. Specify either Credential or Token, then retry.
    ///  
    /// </summary>
    internal static string AuthenticationTokenConflict {
        get {
            return ResourceManager.GetString("AuthenticationTokenConflict", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The cmdlet cannot run because the following conflicting parameters are specified: Body and InFile. Specify either Body or Infile, then retry.  
    ///  
    /// </summary>
    internal static string BodyConflict {
        get {
            return ResourceManager.GetString("BodyConflict", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The cmdlet cannot run because the following conflicting parameters are specified: Body and Form. Specify either Body or Form, then retry.  
    ///  
    /// </summary>
    internal static string BodyFormConflict {
        get {
            return ResourceManager.GetString("BodyFormConflict", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The cmdlet cannot run because the following conflicting parameters are specified: InFile and Form. Specify either InFile or Form, then retry.  
    ///  
    /// </summary>
    internal static string FormInFileConflict {
        get {
            return ResourceManager.GetString("FormInFileConflict", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The cmdlet cannot run because the -ContentType parameter is not a valid Content-Type header. Specify a valid Content-Type for -ContentType, then retry. To suppress header validation, supply the -SkipHeaderValidation parameter.
    ///  
    /// </summary>
    internal static string ContentTypeException {
        get {
            return ResourceManager.GetString("ContentTypeException", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The cmdlet cannot run because the following conflicting parameters are specified: Credential and UseDefaultCredentials. Specify either Credential or UseDefaultCredentials, then retry.
    ///  
    /// </summary>
    internal static string CredentialConflict {
        get {
            return ResourceManager.GetString("CredentialConflict", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Path '{0}' resolves to a directory. Specify a path including a file name, and then retry the command.
    ///  
    /// </summary>
    internal static string DirectoryPathSpecified {
        get {
            return ResourceManager.GetString("DirectoryPathSpecified", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The provided JSON includes a property whose name is an empty string, this is only supported using the -AsHashTable switch.
    ///  
    /// </summary>
    internal static string EmptyKeyInJsonString {
        get {
            return ResourceManager.GetString("EmptyKeyInJsonString", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot convert the JSON string because a dictionary that was converted from the string contains the duplicated key '{0}'.
    ///  
    /// </summary>
    internal static string DuplicateKeysInJsonString {
        get {
            return ResourceManager.GetString("DuplicateKeysInJsonString", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The response content cannot be parsed because the Internet Explorer engine is not available, or Internet Explorer's first-launch configuration is not complete. Specify the UseBasicParsing parameter and try again. 
    ///  
    /// </summary>
    internal static string IEDomNotSupported {
        get {
            return ResourceManager.GetString("IEDomNotSupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot follow an insecure redirection by default. Reissue the command specifying the -AllowInsecureRedirect switch. 
    ///  
    /// </summary>
    internal static string InsecureRedirection {
        get {
            return ResourceManager.GetString("InsecureRedirection", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot convert the JSON string because it contains keys with different casing. Please use the -AsHashTable switch instead. The key that was attempted to be added to the existing key '{0}' was '{1}'.
    ///  
    /// </summary>
    internal static string KeysWithDifferentCasingInJsonString {
        get {
            return ResourceManager.GetString("KeysWithDifferentCasingInJsonString", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The maximum redirection count has been exceeded. To increase the number of redirections allowed, supply a higher value to the -MaximumRedirection parameter.
    ///  
    /// </summary>
    internal static string MaximumRedirectionCountExceeded {
        get {
            return ResourceManager.GetString("MaximumRedirectionCountExceeded", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Path '{0}' can be resolved to multiple paths.
    ///  
    /// </summary>
    internal static string MultiplePathsResolved {
        get {
            return ResourceManager.GetString("MultiplePathsResolved", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The type '{0}' is not supported for serialization or deserialization of a dictionary. Keys must be strings.
    ///  
    /// </summary>
    internal static string NonStringKeyInDictionary {
        get {
            return ResourceManager.GetString("NonStringKeyInDictionary", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Path '{0}' cannot be resolved to a file.
    ///  
    /// </summary>
    internal static string NoPathResolved {
        get {
            return ResourceManager.GetString("NoPathResolved", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Path '{0}' is not a file system path. Please specify the path to a file in the file system.
    ///  
    /// </summary>
    internal static string NotFilesystemPath {
        get {
            return ResourceManager.GetString("NotFilesystemPath", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The cmdlet cannot run because the following parameter is missing: OutFile. Provide a valid OutFile parameter value when using the {0} parameter, then retry.
    ///  
    /// </summary>
    internal static string OutFileMissing {
        get {
            return ResourceManager.GetString("OutFileMissing", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The file will not be re-downloaded because the remote file is the same size as the OutFile: {0}
    ///  
    /// </summary>
    internal static string OutFileWritingSkipped {
        get {
            return ResourceManager.GetString("OutFileWritingSkipped", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The cmdlet cannot run because the following conflicting parameters are specified: ProxyCredential and ProxyUseDefaultCredentials. Specify either ProxyCredential or ProxyUseDefaultCredentials, then retry.
    ///  
    /// </summary>
    internal static string ProxyCredentialConflict {
        get {
            return ResourceManager.GetString("ProxyCredentialConflict", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The cmdlet cannot run because the following parameter is missing: Proxy. Provide a valid proxy URI for the Proxy parameter when using the ProxyCredential or ProxyUseDefaultCredentials parameters, then retry.
    ///  
    /// </summary>
    internal static string ProxyUriNotSupplied {
        get {
            return ResourceManager.GetString("ProxyUriNotSupplied", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Reading web response stream completed. Bytes downloaded: {0}
    ///  
    /// </summary>
    internal static string ReadResponseComplete {
        get {
            return ResourceManager.GetString("ReadResponseComplete", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Reading web response stream
    ///  
    /// </summary>
    internal static string ReadResponseProgressActivity {
        get {
            return ResourceManager.GetString("ReadResponseProgressActivity", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Downloaded: {0} of {1}
    ///  
    /// </summary>
    internal static string ReadResponseProgressStatus {
        get {
            return ResourceManager.GetString("ReadResponseProgressStatus", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Resume switch can only be used if OutFile targets a file but it resolves to a directory: {0}.
    ///  
    /// </summary>
    internal static string ResumeNotFilePath {
        get {
            return ResourceManager.GetString("ResumeNotFilePath", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The cmdlet cannot run because the following conflicting parameters are specified: Session and SessionVariable. Specify either Session or SessionVariable, then retry.
    ///  
    /// </summary>
    internal static string SessionConflict {
        get {
            return ResourceManager.GetString("SessionConflict", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Unable to retrieve certificates because the thumbprint is not valid. Verify the thumbprint and retry. 
    ///  
    /// </summary>
    internal static string ThumbprintNotFound {
        get {
            return ResourceManager.GetString("ThumbprintNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Web request completed. (Number of bytes processed: {0})
    ///  
    /// </summary>
    internal static string WriteRequestComplete {
        get {
            return ResourceManager.GetString("WriteRequestComplete", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Web request cancelled. (Number of bytes processed: {0})
    ///  
    /// </summary>
    internal static string WriteRequestCancelled {
        get {
            return ResourceManager.GetString("WriteRequestCancelled", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Web request status
    ///  
    /// </summary>
    internal static string WriteRequestProgressActivity {
        get {
            return ResourceManager.GetString("WriteRequestProgressActivity", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Downloaded: {0} of {1}
    ///  
    /// </summary>
    internal static string WriteRequestProgressStatus {
        get {
            return ResourceManager.GetString("WriteRequestProgressStatus", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Conversion from JSON failed with error: {0}
    ///  
    /// </summary>
    internal static string JsonDeserializationFailed {
        get {
            return ResourceManager.GetString("JsonDeserializationFailed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Response status code does not indicate success: {0} ({1}).
    ///  
    /// </summary>
    internal static string ResponseStatusCodeFailure {
        get {
            return ResourceManager.GetString("ResponseStatusCodeFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Following rel link {0}
    ///  
    /// </summary>
    internal static string FollowingRelLinkVerboseMsg {
        get {
            return ResourceManager.GetString("FollowingRelLinkVerboseMsg", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The remote server indicated it could not resume downloading. The local file will be overwritten.
    ///  
    /// </summary>
    internal static string WebMethodResumeFailedVerboseMsg {
        get {
            return ResourceManager.GetString("WebMethodResumeFailedVerboseMsg", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Received HTTP/{0} response of content type {1} of unknown size
    ///  
    /// </summary>
    internal static string WebResponseNoSizeVerboseMsg {
        get {
            return ResourceManager.GetString("WebResponseNoSizeVerboseMsg", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Retrying after interval of {0} seconds. Status code for previous attempt: {1}
    ///  
    /// </summary>
    internal static string RetryVerboseMsg {
        get {
            return ResourceManager.GetString("RetryVerboseMsg", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Resulting JSON is truncated as serialization has exceeded the set depth of {0}.
    ///  
    /// </summary>
    internal static string JsonMaxDepthReached {
        get {
            return ResourceManager.GetString("JsonMaxDepthReached", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The WebSession properties were changed between requests forcing all HTTP connections in the session to be recreated.
    ///  
    /// </summary>
    internal static string WebSessionConnectionRecreated {
        get {
            return ResourceManager.GetString("WebSessionConnectionRecreated", resourceCulture);
        }
    }

}

