/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/microsoft.security.extensions/1.4.0/ref/netstandard2.0/getfilesiginforedistwrapper.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/humanizer.core/2.14.1/lib/net6.0/Humanizer.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/json.more.net/2.1.1/lib/net9.0/Json.More.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/jsonpointer.net/5.3.1/lib/net9.0/JsonPointer.Net.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/jsonschema.net/7.4.0/lib/net9.0/JsonSchema.Net.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/markdig.signed/0.44.0/lib/net9.0/Markdig.Signed.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/microsoft.applicationinsights/2.23.0/lib/netstandard2.0/Microsoft.ApplicationInsights.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/microsoft.bcl.asyncinterfaces/10.0.5/lib/netstandard2.1/Microsoft.Bcl.AsyncInterfaces.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/microsoft.codeanalysis.csharp/5.0.0/lib/net9.0/Microsoft.CodeAnalysis.CSharp.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/microsoft.codeanalysis.common/5.0.0/lib/net9.0/Microsoft.CodeAnalysis.dll;
;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/microsoft.extensions.objectpool/10.0.5/lib/net10.0/Microsoft.Extensions.ObjectPool.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/microsoft.powershell.markdownrender/7.2.1/lib/netstandard2.0/Microsoft.PowerShell.MarkdownRender.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/microsoft.win32.registry.accesscontrol/10.0.5/lib/net10.0/Microsoft.Win32.Registry.AccessControl.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/microsoft.win32.systemevents/10.0.5/lib/net10.0/Microsoft.Win32.SystemEvents.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/newtonsoft.json/13.0.4/lib/net6.0/Newtonsoft.Json.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.codedom/10.0.5/lib/net10.0/System.CodeDom.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.componentmodel.composition/10.0.5/lib/net10.0/System.ComponentModel.Composition.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.componentmodel.composition.registration/10.0.5/lib/net10.0/System.ComponentModel.Composition.Registration.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.configuration.configurationmanager/10.0.5/lib/net10.0/System.Configuration.ConfigurationManager.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.data.odbc/10.0.5/lib/net10.0/System.Data.Odbc.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.data.oledb/10.0.5/lib/net10.0/System.Data.OleDb.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.data.sqlclient/4.9.1/lib/net8.0/System.Data.SqlClient.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.diagnostics.eventlog/10.0.5/lib/net10.0/System.Diagnostics.EventLog.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.diagnostics.performancecounter/10.0.5/lib/net10.0/System.Diagnostics.PerformanceCounter.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.directoryservices.accountmanagement/10.0.5/lib/net10.0/System.DirectoryServices.AccountManagement.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.directoryservices/10.0.5/lib/net10.0/System.DirectoryServices.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.directoryservices.protocols/10.0.5/lib/net10.0/System.DirectoryServices.Protocols.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.drawing.common/10.0.5/lib/net10.0/System.Drawing.Common.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.io.packaging/10.0.5/lib/net10.0/System.IO.Packaging.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.io.ports/10.0.5/lib/net10.0/System.IO.Ports.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.management/10.0.5/lib/net10.0/System.Management.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.net.http.winhttphandler/10.0.5/lib/net10.0/System.Net.Http.WinHttpHandler.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.drawing.common/10.0.5/lib/net10.0/System.Private.Windows.Core.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.drawing.common/10.0.5/lib/net10.0/System.Private.Windows.GdiPlus.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.reflection.context/10.0.5/lib/net10.0/System.Reflection.Context.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.runtime.caching/10.0.5/lib/net10.0/System.Runtime.Caching.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.security.cryptography.pkcs/10.0.5/lib/net10.0/System.Security.Cryptography.Pkcs.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.security.cryptography.protecteddata/10.0.5/lib/net10.0/System.Security.Cryptography.ProtectedData.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.security.cryptography.xml/10.0.5/lib/net10.0/System.Security.Cryptography.Xml.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.security.permissions/10.0.5/lib/net10.0/System.Security.Permissions.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.servicemodel.primitives/10.0.652802/ref/net10.0/System.ServiceModel.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.servicemodel.primitives/10.0.652802/ref/net10.0/System.ServiceModel.Duplex.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.servicemodel.http/10.0.652802/ref/net10.0/System.ServiceModel.Http.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.servicemodel.netframingbase/10.0.652802/ref/net10.0/System.ServiceModel.NetFramingBase.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.servicemodel.nettcp/10.0.652802/ref/net10.0/System.ServiceModel.NetTcp.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.servicemodel.primitives/10.0.652802/ref/net10.0/System.ServiceModel.Primitives.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.servicemodel.primitives/10.0.652802/ref/net10.0/System.ServiceModel.Security.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.servicemodel.syndication/10.0.5/lib/net10.0/System.ServiceModel.Syndication.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.serviceprocess.servicecontroller/10.0.5/lib/net10.0/System.ServiceProcess.ServiceController.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.speech/10.0.5/lib/net10.0/System.Speech.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.web.services.description/8.1.2/ref/netstandard2.0/System.Web.Services.Description.dll;
/tmp/tmp.XUk65Ken.tmpsh/gdmt_genpwsh/cache/nuget_packages/system.windows.extensions/10.0.5/lib/net10.0/System.Windows.Extensions.dll;
