
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class RunspaceStrings {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal RunspaceStrings() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.RunspaceStrings", typeof(RunspaceStrings).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The runspace state is not valid for this operation.
    ///  
    /// </summary>
    internal static string InvalidRunspaceStateGeneral {
        get {
            return ResourceManager.GetString("InvalidRunspaceStateGeneral", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot open the runspace because the runspace is not in the BeforeOpen state. Current state of the runspace is '{0}'.
    ///  
    /// </summary>
    internal static string CannotOpenAgain {
        get {
            return ResourceManager.GetString("CannotOpenAgain", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot perform the operation because the runspace is not in the Opened state. Current state of the runspace is '{0}'.
    ///  
    /// </summary>
    internal static string RunspaceNotInOpenedState {
        get {
            return ResourceManager.GetString("RunspaceNotInOpenedState", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot invoke the pipeline because the runspace is not in the Opened state. Current state of the runspace is '{0}'.
    ///  
    /// </summary>
    internal static string RunspaceNotOpenForPipeline {
        get {
            return ResourceManager.GetString("RunspaceNotOpenForPipeline", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The pipeline state is not valid for this operation.
    ///  
    /// </summary>
    internal static string InvalidPipelineStateStateGeneral {
        get {
            return ResourceManager.GetString("InvalidPipelineStateStateGeneral", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot invoke pipeline because it has already been invoked.
    ///  
    /// </summary>
    internal static string PipelineReInvokeNotAllowed {
        get {
            return ResourceManager.GetString("PipelineReInvokeNotAllowed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The valid value for the parameter is PipelineResultTypes.Output.
    ///  
    /// </summary>
    internal static string InvalidValueToResultError {
        get {
            return ResourceManager.GetString("InvalidValueToResultError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The pipeline does not contain a command.
    ///  
    /// </summary>
    internal static string NoCommandInPipeline {
        get {
            return ResourceManager.GetString("NoCommandInPipeline", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The pipeline was not run because a pipeline is already running. Pipelines cannot be run concurrently.
    ///  
    /// </summary>
    internal static string ConcurrentInvokeNotAllowed {
        get {
            return ResourceManager.GetString("ConcurrentInvokeNotAllowed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A nested pipeline cannot be invoked asynchronously. Use the Invoke method.
    ///  
    /// </summary>
    internal static string NestedPipelineInvokeAsync {
        get {
            return ResourceManager.GetString("NestedPipelineInvokeAsync", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    You should only run a nested pipeline from within a running pipeline.
    ///  
    /// </summary>
    internal static string NestedPipelineNoParentPipeline {
        get {
            return ResourceManager.GetString("NestedPipelineNoParentPipeline", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Runspace cannot be closed while a SessionStateProxy method call is in progress.
    ///  
    /// </summary>
    internal static string RunspaceCloseInvalidWhileSessionStateProxy {
        get {
            return ResourceManager.GetString("RunspaceCloseInvalidWhileSessionStateProxy", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Pipeline cannot be invoked while a SessionStateProxy method call is in progress.
    ///  
    /// </summary>
    internal static string NoPipelineWhenSessionStateProxyInProgress {
        get {
            return ResourceManager.GetString("NoPipelineWhenSessionStateProxyInProgress", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A SessionStateProxy method call is in progress. Concurrent SessionStateProxy method calls are not allowed.
    ///  
    /// </summary>
    internal static string AnotherSessionStateProxyInProgress {
        get {
            return ResourceManager.GetString("AnotherSessionStateProxyInProgress", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A pipeline is already running. Concurrent SessionStateProxy method calls are not allowed.
    ///  
    /// </summary>
    internal static string NoSessionStateProxyWhenPipelineInProgress {
        get {
            return ResourceManager.GetString("NoSessionStateProxyWhenPipelineInProgress", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    This property cannot be changed after the runspace has been opened.
    ///  
    /// </summary>
    internal static string ChangePropertyAfterOpen {
        get {
            return ResourceManager.GetString("ChangePropertyAfterOpen", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    One or more errors occurred processing the module '{0}' specified in the InitialSessionState object used to create this runspace. See the ErrorRecords property for a complete list of errors. The first error was: {1}
    ///  
    /// </summary>
    internal static string ErrorLoadingModulesOnRunspaceOpen {
        get {
            return ResourceManager.GetString("ErrorLoadingModulesOnRunspaceOpen", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The thread options can only be changed if the apartment state is multithreaded apartment (MTA), the current options are UseNewThread or UseCurrentThread, and the new value is ReuseThread.
    ///  
    /// </summary>
    internal static string InvalidThreadOptionsChange {
        get {
            return ResourceManager.GetString("InvalidThreadOptionsChange", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0} cannot be false when language mode is {1} or {2}.
    ///  
    /// </summary>
    internal static string UseLocalScopeNotAllowed {
        get {
            return ResourceManager.GetString("UseLocalScopeNotAllowed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    You cannot disconnect a local-only runspace.
    ///  
    /// </summary>
    internal static string DisconnectNotSupported {
        get {
            return ResourceManager.GetString("DisconnectNotSupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Connect operation is not supported on local runspaces.
    ///  
    /// </summary>
    internal static string ConnectNotSupported {
        get {
            return ResourceManager.GetString("ConnectNotSupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The session is busy. You will be connected to the session as soon as it is available. To cancel the Enter-PSSession command, press Ctrl-C.
    ///  
    /// </summary>
    internal static string RunspaceNotReady {
        get {
            return ResourceManager.GetString("RunspaceNotReady", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The command cannot be completed. Script invocation is not supported in this session configuration. This can occur if the session configuration is in no-language mode.
    ///  
    /// </summary>
    internal static string NotSupportedOnRestrictedRunspace {
        get {
            return ResourceManager.GetString("NotSupportedOnRestrictedRunspace", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    You cannot use Disconnect and Connect operations on local runspaces.
    ///  
    /// </summary>
    internal static string DisconnectConnectNotSupported {
        get {
            return ResourceManager.GetString("DisconnectConnectNotSupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot connect the pipeline because the runspace is not in the Opened state.  Current state of runspace is '{0}'.
    ///  
    /// </summary>
    internal static string RunspaceNotOpenForPipelineConnect {
        get {
            return ResourceManager.GetString("RunspaceNotOpenForPipelineConnect", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot construct a RemoteRunspace. The provided RunspacePool object is not valid.
    ///  
    /// </summary>
    internal static string InvalidRunspacePool {
        get {
            return ResourceManager.GetString("InvalidRunspacePool", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    There is no disconnected command associated with this runspace.
    ///  
    /// </summary>
    internal static string NoDisconnectedCommand {
        get {
            return ResourceManager.GetString("NoDisconnectedCommand", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The disconnection operation is not supported on the remote computer. To support disconnecting, the remote computer must be running Windows PowerShell 3.0 or a later version of Windows PowerShell and using the WSMan transport.
    ///  
    /// </summary>
    internal static string DisconnectNotSupportedOnServer {
        get {
            return ResourceManager.GetString("DisconnectNotSupportedOnServer", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot connect the PSSession because the session is not in the Disconnected state, or is not available for connection.
    ///  
    /// </summary>
    internal static string CannotConnect {
        get {
            return ResourceManager.GetString("CannotConnect", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Value for parameter cannot be PipelineResultTypes.None or PipelineResultTypes.Output.
    ///  
    /// </summary>
    internal static string InvalidMyResultError {
        get {
            return ResourceManager.GetString("InvalidMyResultError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Valid values for the parameter are PipelineResultTypes.Output or PipelineResultTypes.Null.
    ///  
    /// </summary>
    internal static string InvalidValueToResult {
        get {
            return ResourceManager.GetString("InvalidValueToResult", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Debug stream redirection is not supported on the targeted remote computer.
    ///  
    /// </summary>
    internal static string DebugRedirectionNotSupported {
        get {
            return ResourceManager.GetString("DebugRedirectionNotSupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Verbose stream redirection is not supported on the targeted remote computer.
    ///  
    /// </summary>
    internal static string VerboseRedirectionNotSupported {
        get {
            return ResourceManager.GetString("VerboseRedirectionNotSupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Warning stream redirection is not supported on the targeted remote computer.
    ///  
    /// </summary>
    internal static string WarningRedirectionNotSupported {
        get {
            return ResourceManager.GetString("WarningRedirectionNotSupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Information stream redirection is not supported on the targeted remote computer.
    ///  
    /// </summary>
    internal static string InformationRedirectionNotSupported {
        get {
            return ResourceManager.GetString("InformationRedirectionNotSupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    You have entered a session that is busy running a command or script.  Because output is routed to job "{0}", you will not see output in the console.  You can wait for the running command to finish, or cancel the command and get an input prompt by pressing Ctrl-C.
    ///
    ///  
    /// </summary>
    internal static string RunningCmdWithJob {
        get {
            return ResourceManager.GetString("RunningCmdWithJob", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    You have entered a session that is busy running a command or script and output will be displayed in the console.  You can wait for the running command to finish or cancel it and get an input prompt by pressing Ctrl-C.
    ///
    ///  
    /// </summary>
    internal static string RunningCmdWithoutJob {
        get {
            return ResourceManager.GetString("RunningCmdWithoutJob", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    You have entered a session that is currently stopped at a debug breakpoint inside a running command or script.  Use the PowerShell command line debugger to continue debugging.
    ///
    ///  
    /// </summary>
    internal static string RunningCmdDebugStop {
        get {
            return ResourceManager.GetString("RunningCmdDebugStop", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    DefaultRunspace must be a LocalRunspace
    ///  
    /// </summary>
    internal static string RunspaceNotLocal {
        get {
            return ResourceManager.GetString("RunspaceNotLocal", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The static PrimaryRunspace property can only be set once, and has already been set.
    ///  
    /// </summary>
    internal static string PrimaryRunspaceAlreadySet {
        get {
            return ResourceManager.GetString("PrimaryRunspaceAlreadySet", resourceCulture);
        }
    }

}

