
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class FormatAndOut_MshParameter {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal FormatAndOut_MshParameter() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.FormatAndOut_MshParameter", typeof(FormatAndOut_MshParameter).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot convert {0} to one of the following types {1}.
    ///  
    /// </summary>
    internal static string UnknownParameterTypeError {
        get {
            return ResourceManager.GetString("UnknownParameterTypeError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The value of a parameter was null; one of the following types was expected: {0}.
    ///  
    /// </summary>
    internal static string NullParameterTypeError {
        get {
            return ResourceManager.GetString("NullParameterTypeError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The duplicated key "{0}" conflicts with "{1}".
    ///  
    /// </summary>
    internal static string DuplicateKeyError {
        get {
            return ResourceManager.GetString("DuplicateKeyError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The "{0}" key has a type, {1}, that is not valid; expected types are {2}.
    ///  
    /// </summary>
    internal static string IllegalTypeMultiError {
        get {
            return ResourceManager.GetString("IllegalTypeMultiError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The "{0}" key has a type, {1}, that is not valid; expected type is {2}.
    ///  
    /// </summary>
    internal static string IllegalTypeSingleError {
        get {
            return ResourceManager.GetString("IllegalTypeSingleError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The {0} key is ambiguous; {1} and {2} conflict.
    ///  
    /// </summary>
    internal static string AmbiguousKeyError {
        get {
            return ResourceManager.GetString("AmbiguousKeyError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The value of a key cannot be null.
    ///  
    /// </summary>
    internal static string DictionaryKeyNullError {
        get {
            return ResourceManager.GetString("DictionaryKeyNullError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The {0} key type is not valid. The key must be a string.
    ///  
    /// </summary>
    internal static string DictionaryKeyNonStringError {
        get {
            return ResourceManager.GetString("DictionaryKeyNonStringError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The {0} key has no value.
    ///  
    /// </summary>
    internal static string MissingKeyValueError {
        get {
            return ResourceManager.GetString("MissingKeyValueError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    A mandatory entry for {0} is missing.
    ///  
    /// </summary>
    internal static string MissingKeyMandatoryEntryError {
        get {
            return ResourceManager.GetString("MissingKeyMandatoryEntryError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The {0} key is not valid.
    ///  
    /// </summary>
    internal static string IllegalKeyError {
        get {
            return ResourceManager.GetString("IllegalKeyError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The value "{0}" for key "{1}" is not valid; valid values are {2}.
    ///  
    /// </summary>
    internal static string IllegalAlignmentValueError {
        get {
            return ResourceManager.GetString("IllegalAlignmentValueError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The value "{0}" for key "{1}" should be greater than 0.
    ///  
    /// </summary>
    internal static string OutOfRangeWidthValueError {
        get {
            return ResourceManager.GetString("OutOfRangeWidthValueError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot have an empty formatting string for key "{0}".
    ///  
    /// </summary>
    internal static string EmptyFormatStringValueError {
        get {
            return ResourceManager.GetString("EmptyFormatStringValueError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The "{0}" key cannot have an empty string value.
    ///  
    /// </summary>
    internal static string MshExEmptyStringHashError {
        get {
            return ResourceManager.GetString("MshExEmptyStringHashError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    An empty string value is not allowed.
    ///  
    /// </summary>
    internal static string MshExEmptyStringError {
        get {
            return ResourceManager.GetString("MshExEmptyStringError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The "{0}" key cannot have wildcard characters in value "{1}".
    ///  
    /// </summary>
    internal static string MshExGlobbingHashError {
        get {
            return ResourceManager.GetString("MshExGlobbingHashError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Wildcard characters are not allowed in "{0}".
    ///  
    /// </summary>
    internal static string MshExGlobbingStringError {
        get {
            return ResourceManager.GetString("MshExGlobbingStringError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The EnumerableExpansion value is not valid.
    ///  
    /// </summary>
    internal static string IllegalEnumerableExpansionValue {
        get {
            return ResourceManager.GetString("IllegalEnumerableExpansionValue", resourceCulture);
        }
    }

}

