
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a dotnet run from src\ResGen folder.
//     To add or remove a member, edit your .resx file then rerun src\ResGen.
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------


using System;
using System.Reflection;

/// <summary>
///   A strongly-typed resource class, for looking up localized strings, etc.
/// </summary>
[global::System.CodeDom.Compiler.GeneratedCodeAttribute("System.Resources.Tools.StronglyTypedResourceBuilder", "4.0.0.0")]
[global::System.Diagnostics.DebuggerNonUserCodeAttribute()]
[global::System.Runtime.CompilerServices.CompilerGeneratedAttribute()]

internal class HelpErrors {

    private static global::System.Resources.ResourceManager resourceMan;

    private static global::System.Globalization.CultureInfo resourceCulture;

    /// <summary>constructor</summary>
    [global::System.Diagnostics.CodeAnalysis.SuppressMessageAttribute("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
    internal HelpErrors() {
    }

    /// <summary>
    ///   Returns the cached ResourceManager instance used by this class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Resources.ResourceManager ResourceManager {
        get {
            if (resourceMan is null) {
                global::System.Resources.ResourceManager temp = new global::System.Resources.ResourceManager("System.Management.Automation.resources.HelpErrors", typeof(HelpErrors).Assembly);
                resourceMan = temp;
            }

            return resourceMan;
        }
    }

    /// <summary>
    ///   Overrides the current threads CurrentUICulture property for all
    ///   resource lookups using this strongly typed resource class.
    /// </summary>
    [global::System.ComponentModel.EditorBrowsableAttribute(global::System.ComponentModel.EditorBrowsableState.Advanced)]
    internal static global::System.Globalization.CultureInfo Culture {
        get {
            return resourceCulture;
        }

        set {
            resourceCulture = value;
        }
    }
    

    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Get-Help could not find {0} in a help file in this session. To download updated help topics type: "Update-Help". To get help online, search for the help topic in the TechNet library at https://go.microsoft.com/fwlink/?LinkID=107116.
    ///  
    /// </summary>
    internal static string HelpNotFound {
        get {
            return ResourceManager.GetString("HelpNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot process the Help category because "{0}" is not a valid Help category.
    ///  
    /// </summary>
    internal static string HelpCategoryInvalid {
        get {
            return ResourceManager.GetString("HelpCategoryInvalid", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot load the Help file "{0}". Details: {1}.
    ///  
    /// </summary>
    internal static string HelpFileLoadFailure {
        get {
            return ResourceManager.GetString("HelpFileLoadFailure", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot access the Help file "{0}" because the current user does not have access rights to the file. Details: {1}.
    ///  
    /// </summary>
    internal static string HelpFileNotAccessible {
        get {
            return ResourceManager.GetString("HelpFileNotAccessible", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The Help file "{0}" is not a valid xml document. Details: {1}.
    ///  
    /// </summary>
    internal static string HelpFileNotValid {
        get {
            return ResourceManager.GetString("HelpFileNotValid", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    An error occurred while loading Help content for {0} from file {1}. Details: {2}. To download updated Help topics, run the Update-Help cmdlet. To get help online, search for the Help topic in the TechNet library at https://go.microsoft.com/fwlink/?LinkID=107116.
    ///  
    /// </summary>
    internal static string LoadHelpFileForTargetFailed {
        get {
            return ResourceManager.GetString("LoadHelpFileForTargetFailed", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The provider "{0}" cannot be loaded. Details: {1}.
    ///  
    /// </summary>
    internal static string ProviderLoadError {
        get {
            return ResourceManager.GetString("ProviderLoadError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot load the Help file. The following {1} errors occurred while loading the Help file "{0}".
    ///  
    /// </summary>
    internal static string HelpLoadError {
        get {
            return ResourceManager.GetString("HelpLoadError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The node "{0}" cannot have "{1}" as a child node. Node Path: {2}.
    ///  
    /// </summary>
    internal static string MamlInvalidChildNodeError {
        get {
            return ResourceManager.GetString("MamlInvalidChildNodeError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The node "{0}" can have a maximum of {2} child nodes of type "{1}". Node Path: {3}.
    ///  
    /// </summary>
    internal static string MamlInvalidChildNodeCountError {
        get {
            return ResourceManager.GetString("MamlInvalidChildNodeCountError", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot find the registry key: "{0}{1}"; using "{2}" to load Help files.
    ///  
    /// </summary>
    internal static string RegistryPathNotFound {
        get {
            return ResourceManager.GetString("RegistryPathNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    No parameter matches criteria {0}.
    ///  
    /// </summary>
    internal static string NoParmsFound {
        get {
            return ResourceManager.GetString("NoParmsFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0} is not supported by the requested Help category.
    ///  
    /// </summary>
    internal static string ParamNotSupported {
        get {
            return ResourceManager.GetString("ParamNotSupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The online version of this Help topic cannot be displayed because the Internet address (URI) of the Help topic is not specified in the command code or in the help file for the command.
    ///  
    /// </summary>
    internal static string NoURIFound {
        get {
            return ResourceManager.GetString("NoURIFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The specified URI {0} is not valid.
    ///  
    /// </summary>
    internal static string InvalidURI {
        get {
            return ResourceManager.GetString("InvalidURI", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Starting a browser to display online Help failed. No program or browser is associated to open the URI {0}.
    ///  
    /// </summary>
    internal static string CannotLaunchURI {
        get {
            return ResourceManager.GetString("CannotLaunchURI", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    The protocol specified in the Uri "{0}" is not supported. Only "{1}" and "{2}" protocols are supported.
    ///  
    /// </summary>
    internal static string ProtocolNotSupported {
        get {
            return ResourceManager.GetString("ProtocolNotSupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Multiple Help topics were found. Use only one Help topic with the -{0} option.
    ///  
    /// </summary>
    internal static string MultipleOnlineTopicsNotSupported {
        get {
            return ResourceManager.GetString("MultipleOnlineTopicsNotSupported", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot get Help from a remote runspace, because the runspace has not been opened.  Open the runspace by running an implicit remoting command, and then try running the command to get Help again.
    ///  
    /// </summary>
    internal static string RemoteRunspaceNotAvailable {
        get {
            return ResourceManager.GetString("RemoteRunspaceNotAvailable", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Access is denied. The command could not update Help topics for the PowerShell core modules, or for any modules in the $pshome\Modules directory.
    ///To update these Help topics, start PowerShell by using the "Run as Administrator" command, and try running Update-Help again.
    ///  
    /// </summary>
    internal static string UpdatableHelpRequiresElevation {
        get {
            return ResourceManager.GetString("UpdatableHelpRequiresElevation", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    To use the {0}, install Windows PowerShell ISE by using Server Manager, and then restart this application. ({1})
    ///  
    /// </summary>
    internal static string GraphicalHostAssemblyIsNotFound {
        get {
            return ResourceManager.GetString("GraphicalHostAssemblyIsNotFound", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    {0} does not work in a remote session.
    ///  
    /// </summary>
    internal static string RemotingNotSupportedForFeature {
        get {
            return ResourceManager.GetString("RemotingNotSupportedForFeature", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    ForwardHelpTargetName cannot refer to the function itself.
    ///  
    /// </summary>
    internal static string CircularDependencyInHelpForwarding {
        get {
            return ResourceManager.GetString("CircularDependencyInHelpForwarding", resourceCulture);
        }
    }


    /// <summary>
    ///   Looks up a localized string similar to 
    ///    Cannot get help from a network location when in a restricted session.
    ///  
    /// </summary>
    internal static string NoNetworkCommands {
        get {
            return ResourceManager.GetString("NoNetworkCommands", resourceCulture);
        }
    }

}

