//
// This file is auto-generated by TypeCatalogGen.exe during build of Microsoft.PowerShell.CoreCLR.AssemblyLoadContext.dll.
// This file will be compiled into Microsoft.PowerShell.CoreCLR.AssemblyLoadContext.dll.
//
// In CoreCLR, there is no way to get all loaded TPA assemblies (.NET Framework Assemblies). In order to get type based on type
// name, powershell needs to know what .NET types are available and in which TPA assemblies. So we have to generate this type
// catalog based on the reference assemblies of .NET Core.
//
using System.Collections.Generic;

namespace System.Management.Automation
{
    internal partial class PowerShellAssemblyLoadContext
    {
        private static Dictionary<string, string> InitializeTypeCatalog()
        {
            return new Dictionary<string, string>(3704, StringComparer.OrdinalIgnoreCase) {
                {"Microsoft.Security.Extensions.SignatureState", "getfilesiginforedistwrapper, Version=10.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.Security.Extensions.SignatureStateReason", "getfilesiginforedistwrapper, Version=10.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.Security.Extensions.SignatureKind", "getfilesiginforedistwrapper, Version=10.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.Security.Extensions.FileSignatureInfo", "getfilesiginforedistwrapper, Version=10.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Humanizer.EnglishArticle", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.EnglishArticles", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.ByteSizeExtensions", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.CasingExtensions", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.ClockNotationRounding", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.CollectionHumanizeExtensions", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.DateHumanizeExtensions", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.DateToOrdinalWordsExtensions", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.EnumDehumanizeExtensions", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.EnumHumanizeExtensions", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.In", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.InDate", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.On", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.OnDate", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.PrepositionsExtensions", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.GrammaticalCase", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.GrammaticalGender", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.HeadingStyle", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.HeadingExtensions", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.InflectorExtensions", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.LetterCasing", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.MetricNumeralExtensions", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.MetricNumeralFormats", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.NoMatchFoundException", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.NumberToNumberExtensions", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.NumberToTimeSpanExtensions", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.NumberToWordsExtension", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.OnNoMatch", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.OrdinalizeExtensions", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.Plurality", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.RomanNumeralExtensions", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.StringDehumanizeExtensions", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.StringExtensions", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.StringHumanizeExtensions", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.TimeOnlyToClockNotationExtensions", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.TimeSpanHumanizeExtensions", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.TimeUnitToSymbolExtensions", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.ShowQuantityAs", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.ToQuantityExtensions", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.ICulturedStringTransformer", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.IStringTransformer", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.To", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.TruncateExtensions", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.TruncateFrom", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.ITruncator", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.Truncator", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.TupleizeExtensions", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.WordForm", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.Localisation.DataUnit", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.Localisation.ResourceKeys", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.Localisation.Resources", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.Localisation.Tense", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.Localisation.TimeUnit", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.Localisation.TimeToClockNotation.ITimeOnlyToClockNotationConverter", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.Localisation.Ordinalizers.IOrdinalizer", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.Localisation.NumberToWords.INumberToWordsConverter", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.Localisation.Formatters.DefaultFormatter", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.Localisation.Formatters.IFormatter", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.Localisation.DateToOrdinalWords.IDateOnlyToOrdinalWordConverter", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.Localisation.DateToOrdinalWords.IDateToOrdinalWordConverter", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.Localisation.CollectionFormatters.ICollectionFormatter", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.Inflections.Vocabularies", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.Inflections.Vocabulary", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.DateTimeHumanizeStrategy.DefaultDateOnlyHumanizeStrategy", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.DateTimeHumanizeStrategy.DefaultDateTimeHumanizeStrategy", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.DateTimeHumanizeStrategy.DefaultDateTimeOffsetHumanizeStrategy", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.DateTimeHumanizeStrategy.DefaultTimeOnlyHumanizeStrategy", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.DateTimeHumanizeStrategy.IDateOnlyHumanizeStrategy", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.DateTimeHumanizeStrategy.IDateTimeHumanizeStrategy", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.DateTimeHumanizeStrategy.IDateTimeOffsetHumanizeStrategy", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.DateTimeHumanizeStrategy.ITimeOnlyHumanizeStrategy", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.DateTimeHumanizeStrategy.PrecisionDateOnlyHumanizeStrategy", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.DateTimeHumanizeStrategy.PrecisionDateTimeHumanizeStrategy", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.DateTimeHumanizeStrategy.PrecisionDateTimeOffsetHumanizeStrategy", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.DateTimeHumanizeStrategy.PrecisionTimeOnlyHumanizeStrategy", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.Configuration.Configurator", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.Configuration.LocaliserRegistry`1", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.Bytes.ByteRate", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.Bytes.ByteSize", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.In+One", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.In+Two", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.In+Three", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.In+Four", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.In+Five", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.In+Six", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.In+Seven", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.In+Eight", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.In+Nine", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.In+Ten", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.InDate+One", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.InDate+Two", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.InDate+Three", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.InDate+Four", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.InDate+Five", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.InDate+Six", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.InDate+Seven", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.InDate+Eight", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.InDate+Nine", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.InDate+Ten", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.On+January", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.On+February", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.On+March", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.On+April", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.On+May", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.On+June", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.On+July", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.On+August", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.On+September", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.On+October", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.On+November", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.On+December", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.OnDate+January", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.OnDate+February", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.OnDate+March", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.OnDate+April", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.OnDate+May", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.OnDate+June", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.OnDate+July", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.OnDate+August", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.OnDate+September", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.OnDate+October", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.OnDate+November", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.OnDate+December", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.Localisation.ResourceKeys+DateHumanize", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.Localisation.ResourceKeys+TimeSpanHumanize", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Humanizer.Localisation.ResourceKeys+TimeUnitSymbol", "Humanizer, Version=2.14.0.0, Culture=neutral, PublicKeyToken=979442b78dfc278e"},
                {"Json.More.EnumStringConverter`1", "Json.More, Version=2.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.More.JsonArrayTupleConverter", "Json.More, Version=2.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.More.JsonDocumentEqualityComparer", "Json.More, Version=2.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.More.JsonDocumentExtensions", "Json.More, Version=2.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.More.JsonElementEqualityComparer", "Json.More, Version=2.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.More.JsonElementExtensions", "Json.More, Version=2.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.More.JsonElementProxy", "Json.More, Version=2.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.More.JsonNodeEqualityComparer", "Json.More, Version=2.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.More.JsonNodeExtensions", "Json.More, Version=2.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.More.JsonSerializerOptionsExtensions", "Json.More, Version=2.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.More.TypeExtensions", "Json.More, Version=2.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.More.Utf8JsonWriterExtensions", "Json.More, Version=2.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.More.IWeaklyTypedJsonConverter", "Json.More, Version=2.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.More.WeaklyTypedJsonConverter`1", "Json.More, Version=2.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Pointer.EnumerableExtensions", "JsonPointer.Net, Version=5.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Pointer.JsonPointer", "JsonPointer.Net, Version=5.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Pointer.JsonPointerJsonConverter", "JsonPointer.Net, Version=5.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Pointer.PointerCreationOptions", "JsonPointer.Net, Version=5.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Pointer.PointerParseException", "JsonPointer.Net, Version=5.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Pointer.PointerSegment", "JsonPointer.Net, Version=5.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Pointer.PropertyNameResolver", "JsonPointer.Net, Version=5.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Pointer.PropertyNameResolvers", "JsonPointer.Net, Version=5.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Pointer.RelativeJsonPointer", "JsonPointer.Net, Version=5.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Pointer.RelativeJsonPointerJsonConverter", "JsonPointer.Net, Version=5.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.AdditionalItemsKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.AdditionalItemsKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.AdditionalPropertiesKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.AdditionalPropertiesKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.AllOfKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.AllOfKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.AnchorKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.AnchorKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.AnyOfKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.AnyOfKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.CommentKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.CommentKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.CommonJsonPointers", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ConstKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ConstKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ErrorMessages", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ContainsKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ContainsKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ContentEncodingKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ContentEncodingKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ContentMediaTypeKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ContentMediaTypeKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ContentSchemaKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ContentSchemaKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.DefaultKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.DefaultKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.DefinitionsKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.DefinitionsKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.DefsKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.DefsKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.DependenciesKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.DependenciesKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.SchemaOrPropertyList", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.SchemaOrPropertyListJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.DependentRequiredKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.DependentRequiredKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.DependentSchemasKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.DependentSchemasKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.DependsOnAnnotationsFromAttribute", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.DependsOnAnnotationsFromAttribute`1", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.DeprecatedKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.DeprecatedKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.DescriptionKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.DescriptionKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.Duration", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.DynamicAnchorKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.DynamicAnchorKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.DynamicRefKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.DynamicRefKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.DynamicScope", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ElseKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ElseKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.EnumerableExtensions", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.EnumKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.EnumKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.EvaluationContext", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ConstraintExtensions", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.EvaluationOptions", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.EvaluationResults", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.EvaluationResultsJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.Pre202012EvaluationResultsJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ExamplesKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ExamplesKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ExclusiveMaximumKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ExclusiveMaximumKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ExclusiveMinimumKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ExclusiveMinimumKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.Format", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.FormatKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.FormatKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.Formats", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.IBaseDocument", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ICustomSchemaCollector", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.IdKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.IdKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.IfKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.IfKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.IIdKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.IJsonSchemaKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.IKeyedSchemaCollector", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ISchemaCollector", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ISchemaContainer", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ItemsKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ItemsKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.JsonNodeBaseDocument", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.JsonNodeExtensions", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.JsonSchema", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.SchemaJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.JsonSchemaBuilder", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.JsonSchemaBuilderExtensions", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.JsonSchemaException", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.JsonSchemaExtensions", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.JsonSerializerOptionsExtensions", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.KeywordConstraint", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.KeywordEvaluation", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.MaxContainsKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.MaxContainsKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.MaximumKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.MaximumKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.MaxItemsKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.MaxItemsKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.MaxLengthKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.MaxLengthKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.MaxPropertiesKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.MaxPropertiesKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.MetaSchemas", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.MinContainsKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.MinContainsKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.MinimumKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.MinimumKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.MinItemsKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.MinItemsKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.MinLengthKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.MinLengthKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.MinPropertiesKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.MinPropertiesKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.MultipleOfKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.MultipleOfKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.NotKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.NotKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.OneOfKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.OneOfKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.OutputFormat", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.PatternKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.PatternKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.PatternPropertiesKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.PatternPropertiesKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.PredicateFormat", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.PrefixItemsKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.PrefixItemsKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.PropertiesKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.PropertiesKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.PropertyDependenciesKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.PropertyDependenciesKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.PropertyDependency", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.PropertyDependencyJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.PropertyNamesKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.PropertyNamesKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ReadOnlyKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ReadOnlyKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.RecursiveAnchorKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.RecursiveAnchorKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.RecursiveRefKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.RecursiveRefKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.RefKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.RefKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.RefResolutionException", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.RegexFormat", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.RequiredKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.RequiredKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.SchemaConstraint", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.SchemaEvaluation", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.SchemaKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.SchemaKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.SchemaKeywordAttribute", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.SchemaKeywordRegistry", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.SchemaRegistry", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.SchemaSpecVersionAttribute", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.SchemaValueType", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.SpecVersion", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ThenKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.ThenKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.TitleKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.TitleKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.TypeKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.TypeKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.UnevaluatedItemsKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.UnevaluatedItemsKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.UnevaluatedPropertiesKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.UnevaluatedPropertiesKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.UniqueItemsKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.UniqueItemsKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.UnknownFormat", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.UnrecognizedKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.UnrecognizedKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.UriExtensions", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.Vocabularies", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.Vocabulary", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.VocabularyAttribute", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.VocabularyKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.VocabularyKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.VocabularyRegistry", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.WriteOnlyKeyword", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.WriteOnlyKeywordJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.Serialization.JsonSchemaAttribute", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.Serialization.ValidatingJsonConverter", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Json.Schema.PredicateFormat+PredicateWithErrorMessage", "JsonSchema.Net, Version=7.0.0.0, Culture=neutral, PublicKeyToken=17ed63f672b0e278"},
                {"Markdig.IMarkdownExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Markdown", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.MarkdownExtensions", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.MarkdownParserContext", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.MarkdownPipeline", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.MarkdownPipelineBuilder", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.BlankLineBlock", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.Block", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.BlockExtensions", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.CharIteratorHelper", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.CodeBlock", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.ContainerBlock", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.FencedCodeBlock", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.HeadingBlock", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.HtmlBlock", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.HtmlBlockType", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.IBlock", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.IFencedBlock", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.IMarkdownObject", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.LeafBlock", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.LinkReferenceDefinition", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.LinkReferenceDefinitionExtensions", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.LinkReferenceDefinitionGroup", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.ListBlock", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.ListItemBlock", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.MarkdownDocument", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.MarkdownObject", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.MarkdownObjectExtensions", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.EmptyBlock", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.ParagraphBlock", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.QuoteBlock", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.QuoteBlockLine", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.SourceSpan", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.ThematicBreakBlock", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.Inlines.AutolinkInline", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.Inlines.CodeInline", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.Inlines.ContainerInline", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.Inlines.DelimiterInline", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.Inlines.DelimiterType", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.Inlines.EmphasisDelimiterInline", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.Inlines.EmphasisInline", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.Inlines.HtmlEntityInline", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.Inlines.HtmlInline", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.Inlines.IInline", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.Inlines.Inline", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.Inlines.LeafInline", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.Inlines.LineBreakInline", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.Inlines.LinkDelimiterInline", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.Inlines.LocalLabel", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.Inlines.LinkInline", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.Inlines.LiteralInline", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.HtmlRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.IMarkdownObjectRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.IMarkdownRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.MarkdownObjectRenderer`2", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.ObjectRendererCollection", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.RendererBase", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.TextRendererBase", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.TextRendererBase`1", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Roundtrip.CodeBlockRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Roundtrip.EmptyBlockRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Roundtrip.HeadingRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Roundtrip.HtmlBlockRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Roundtrip.LinkReferenceDefinitionGroupRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Roundtrip.LinkReferenceDefinitionRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Roundtrip.ListRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Roundtrip.ParagraphRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Roundtrip.QuoteBlockRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Roundtrip.RoundtripObjectRenderer`1", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Roundtrip.RoundtripRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Roundtrip.ThematicBreakRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Roundtrip.Inlines.AutolinkInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Roundtrip.Inlines.CodeInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Roundtrip.Inlines.DelimiterInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Roundtrip.Inlines.EmphasisInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Roundtrip.Inlines.LineBreakInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Roundtrip.Inlines.LinkInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Roundtrip.Inlines.LiteralInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Roundtrip.Inlines.RoundtripHtmlEntityInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Roundtrip.Inlines.RoundtripHtmlInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Normalize.CodeBlockRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Normalize.HeadingRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Normalize.HtmlBlockRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Normalize.LinkReferenceDefinitionGroupRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Normalize.LinkReferenceDefinitionRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Normalize.ListRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Normalize.NormalizeObjectRenderer`1", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Normalize.NormalizeOptions", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Normalize.NormalizeRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Normalize.ParagraphRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Normalize.QuoteBlockRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Normalize.ThematicBreakRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Normalize.Inlines.AutolinkInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Normalize.Inlines.CodeInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Normalize.Inlines.DelimiterInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Normalize.Inlines.EmphasisInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Normalize.Inlines.LineBreakInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Normalize.Inlines.LinkInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Normalize.Inlines.LiteralInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Normalize.Inlines.NormalizeHtmlEntityInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Normalize.Inlines.NormalizeHtmlInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Html.CodeBlockRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Html.HeadingRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Html.HtmlAttributes", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Html.HtmlAttributesExtensions", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Html.HtmlBlockRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Html.HtmlObjectRenderer`1", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Html.ListRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Html.ParagraphRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Html.QuoteBlockRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Html.ThematicBreakRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Html.Inlines.AutolinkInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Html.Inlines.CodeInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Html.Inlines.DelimiterInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Html.Inlines.EmphasisInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Html.Inlines.HtmlEntityInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Html.Inlines.HtmlInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Html.Inlines.LineBreakInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Html.Inlines.LinkInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Html.Inlines.LiteralInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.ProcessBlockDelegate", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.BlockParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.BlockParserList", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.BlockProcessor", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.BlockState", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.BlockStateExtensions", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.FencedBlockParserBase", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.FencedBlockParserBase`1", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.FencedCodeBlockParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.HeadingBlockParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.HtmlBlockParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.TryParseAttributesDelegate", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.IAttributesParseable", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.IBlockParser`1", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.IInlineParser`1", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.IMarkdownParser`1", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.IndentedCodeBlockParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.InlineParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.InlineParserList", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.ProcessInlineDelegate", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.InlineProcessor", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.IPostInlineProcessor", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.LinkOptions", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.ListBlockParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.ListInfo", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.ListItemParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.ProcessDocumentDelegate", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.MarkdownParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.NumberedListItemParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.OrderedListItemParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.ParagraphBlockParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.ParserBase`1", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.ParserList`2", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.QuoteBlockParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.ThematicBreakParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.UnorderedListItemParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.Inlines.AutolinkInlineParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.Inlines.AutolinkOptions", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.Inlines.CodeInlineParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.Inlines.EmphasisDescriptor", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.Inlines.EmphasisInlineParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.Inlines.EscapeInlineParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.Inlines.HtmlEntityParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.Inlines.LineBreakInlineParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.Inlines.LinkInlineParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.Inlines.LiteralInlineParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Helpers.CharacterMap`1", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Helpers.CharHelper", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Helpers.CharNormalizer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Helpers.DefaultObjectCache`1", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Helpers.EntityHelper", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Helpers.HtmlHelper", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Helpers.ICharIterator", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Helpers.LineReader", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Helpers.LinkHelper", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Helpers.NewLine", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Helpers.NewLineExtensions", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Helpers.ObjectCache`1", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Helpers.OrderedList`1", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Helpers.StringBuilderCache", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Helpers.StringBuilderExtensions", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Helpers.StringLine", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Helpers.StringLineGroup", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Helpers.StringSlice", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Yaml.YamlFrontMatterBlock", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Yaml.YamlFrontMatterExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Yaml.YamlFrontMatterHtmlRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Yaml.YamlFrontMatterParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Yaml.YamlFrontMatterRoundtripRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.TextRenderer.ConfigureNewLineExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.TaskLists.HtmlTaskListRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.TaskLists.NormalizeTaskListRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.TaskLists.TaskList", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.TaskLists.TaskListExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.TaskLists.TaskListInlineParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Tables.GridTableExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Tables.GridTableParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Tables.HtmlTableRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Tables.PipeTableBlockParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Tables.PipeTableDelimiterInline", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Tables.PipeTableExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Tables.PipeTableOptions", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Tables.PipeTableParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Tables.Table", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Tables.TableCell", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Tables.TableColumnAlign", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Tables.TableColumnDefinition", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Tables.TableHelper", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Tables.TableRow", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.SmartyPants.HtmlSmartyPantRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.SmartyPants.SmartyPant", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.SmartyPants.SmartyPantOptions", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.SmartyPants.SmartyPantsExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.SmartyPants.SmartyPantsInlineParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.SmartyPants.SmartyPantType", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.SelfPipeline.SelfPipelineExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.ReferralLinks.ReferralLinksExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.PragmaLines.PragmaLineExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.NoRefLinks.NoFollowLinksExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.NonAsciiNoEscape.NonAsciiNoEscapeExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.MediaLinks.HostProviderBuilder", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.MediaLinks.IHostProvider", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.MediaLinks.MediaLinkExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.MediaLinks.MediaOptions", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Mathematics.HtmlMathBlockRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Mathematics.HtmlMathInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Mathematics.MathBlock", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Mathematics.MathBlockParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Mathematics.MathExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Mathematics.MathInline", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Mathematics.MathInlineParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.ListExtras.ListExtraExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.ListExtras.ListExtraItemParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.JiraLinks.JiraLink", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.JiraLinks.JiraLinkExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.JiraLinks.JiraLinkInlineParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.JiraLinks.JiraLinkOptions", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.JiraLinks.NormalizeJiraLinksRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Hardlines.SoftlineBreakAsHardlineExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Globalization.GlobalizationExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.GenericAttributes.GenericAttributesExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.GenericAttributes.GenericAttributesParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Footnotes.Footnote", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Footnotes.FootnoteExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Footnotes.FootnoteGroup", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Footnotes.FootnoteLink", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Footnotes.FootnoteLinkReferenceDefinition", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Footnotes.FootnoteParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Footnotes.HtmlFootnoteGroupRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Footnotes.HtmlFootnoteLinkRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Footers.FooterBlock", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Footers.FooterBlockParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Footers.FooterExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Footers.HtmlFooterBlockRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Figures.Figure", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Figures.FigureBlockParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Figures.FigureCaption", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Figures.FigureExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Figures.HtmlFigureCaptionRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Figures.HtmlFigureRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.EmphasisExtras.EmphasisExtraExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.EmphasisExtras.EmphasisExtraOptions", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Emoji.EmojiExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Emoji.EmojiInline", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Emoji.EmojiMapping", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Emoji.EmojiParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Diagrams.DiagramExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.DefinitionLists.DefinitionItem", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.DefinitionLists.DefinitionList", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.DefinitionLists.DefinitionListExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.DefinitionLists.DefinitionListParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.DefinitionLists.DefinitionTerm", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.DefinitionLists.HtmlDefinitionListRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.CustomContainers.CustomContainer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.CustomContainers.CustomContainerExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.CustomContainers.CustomContainerInline", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.CustomContainers.CustomContainerParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.CustomContainers.HtmlCustomContainerInlineRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.CustomContainers.HtmlCustomContainerRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Citations.CitationExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Bootstrap.BootstrapExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.AutoLinks.AutoLinkExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.AutoLinks.AutoLinkOptions", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.AutoLinks.AutoLinkParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.AutoIdentifiers.AutoIdentifierExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.AutoIdentifiers.AutoIdentifierOptions", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.AutoIdentifiers.HeadingLinkReferenceDefinition", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Alerts.AlertBlock", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Alerts.AlertBlockRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Alerts.AlertExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Alerts.AlertInlineParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Abbreviations.Abbreviation", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Abbreviations.AbbreviationExtension", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Abbreviations.AbbreviationHelper", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Abbreviations.AbbreviationInline", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Abbreviations.AbbreviationParser", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Abbreviations.HtmlAbbreviationRenderer", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.CodeBlock+CodeBlockLine", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.ContainerBlock+Enumerator", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.LinkReferenceDefinition+CreateLinkInlineDelegate", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.Inlines.ContainerInline+Enumerator", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Syntax.Inlines.LinkInline+GetUrlDelegate", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.MarkdownObjectRenderer`2+TryWriteDelegate", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Renderers.Html.Inlines.EmphasisInlineRenderer+GetTagDelegate", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.FencedBlockParserBase+InfoParserDelegate", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.Inlines.EmphasisInlineParser+CreateEmphasisInlineDelegate", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.Inlines.EmphasisInlineParser+TryCreateEmphasisInlineDelegate", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.Inlines.EmphasisInlineParser+DelimitersObjectCache", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Parsers.Inlines.LiteralInlineParser+PostMatchDelegate", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Helpers.CompactPrefixTree`1+Enumerator", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Helpers.HexConverter+Casing", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Helpers.StringLineGroup+Enumerator", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Helpers.StringLineGroup+Iterator", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Helpers.StringLineGroup+LineOffset", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Markdig.Extensions.Tables.GridTableState+ColumnSlice", "Markdig.Signed, Version=0.44.0.0, Culture=neutral, PublicKeyToken=870da25a133885f8"},
                {"Microsoft.ApplicationInsights.Metric", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.MetricAggregationScope", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.MetricConfigurations", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.MetricConfigurationsExtensions", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.MetricDimensionNames", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.OperationTelemetryExtensions", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.TelemetryClient", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.TelemetryClientExtensions", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Metrics.IMetricSeriesConfiguration", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Metrics.MetricAggregate", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Metrics.MetricConfiguration", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Metrics.MetricConfigurationForMeasurement", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Metrics.MetricIdentifier", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Metrics.MetricManager", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Metrics.MetricsCollection", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Metrics.MetricSeries", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Metrics.MetricSeriesConfigurationForMeasurement", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Metrics.TelemetryConfigurationExtensions", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Metrics.Extensibility.IMetricSeriesAggregator", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Metrics.Extensibility.IMetricTelemetryPipeline", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Metrics.Extensibility.IMetricValueFilter", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Metrics.Extensibility.MetricAggregationCycleKind", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Metrics.Extensibility.MetricConfigurationExtensions", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Metrics.Extensibility.MetricExtensions", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Metrics.Extensibility.MetricSeriesExtensions", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Metrics.Extensibility.TelemetryClientExtensions", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.IApplicationIdProvider", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.IExtension", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.IOperationHolder`1", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.ISerializableWithWriter", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.ISerializationWriter", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.ITelemetryInitializer", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.ITelemetryModule", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.ITelemetryProcessor", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.OperationCorrelationTelemetryInitializer", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.SdkInternalOperationsMonitor", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.SequencePropertyInitializer", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.TelemetryConfiguration", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.TelemetrySink", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.AutocollectedMetricsExtractor", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.W3C.W3CActivityExtensions", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.W3C.W3COperationCorrelationTelemetryInitializer", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.W3C.W3CUtilities", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.CloudContext", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.ComponentContext", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.DeviceContext", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.HttpWebResponseWrapper", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.InternalContext", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.JsonSerializer", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.LocationContext", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.OperationContext", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.OperationTelemetry", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.SessionContext", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.TaskTimer", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.TelemetryConfigurationExtensions", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.TelemetryContextExtensions", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.TelemetryDebugWriter", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.TelemetryExtensions", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.TelemetryProcessorChain", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.TelemetryProcessorChainBuilder", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.UserContext", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.TelemetryModules", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Tracing.DiagnosticsTelemetryModule", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Tracing.Extensions", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Tracing.FileDiagnosticsTelemetryModule", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Tracing.IHeartbeatPropertyManager", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Experimental.ExperimentalFeaturesExtension", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Endpoints.EndpointContainer", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.ApplicationId.ApplicationInsightsApplicationIdProvider", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.ApplicationId.DictionaryApplicationIdProvider", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.DataContracts.AvailabilityTelemetry", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.DataContracts.DependencyTelemetry", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.DataContracts.EventTelemetry", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.DataContracts.ExceptionDetailsInfo", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.DataContracts.ExceptionHandledAt", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.DataContracts.ExceptionTelemetry", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.DataContracts.IJsonWriter", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.DataContracts.SamplingTelemetryItemTypes", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.DataContracts.SamplingDecision", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.DataContracts.ISupportAdvancedSampling", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.DataContracts.ISupportMetrics", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.DataContracts.ISupportProperties", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.DataContracts.ISupportSampling", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.DataContracts.MetricTelemetry", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.DataContracts.PageViewPerformanceTelemetry", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.DataContracts.PageViewTelemetry", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.DataContracts.PerformanceCounterTelemetry", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.DataContracts.RequestTelemetry", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.DataContracts.SessionState", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.DataContracts.SessionStateTelemetry", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.DataContracts.SeverityLevel", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.DataContracts.StackFrame", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.DataContracts.TelemetryContext", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.DataContracts.TraceTelemetry", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Channel.IAsyncFlushable", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Channel.InMemoryChannel", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Channel.ITelemetry", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Channel.ITelemetryChannel", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Channel.SamplingScoreGenerator", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Channel.Transmission", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Channel.TransmissionStatusEventArgs", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.MetricDimensionNames+TelemetryContext", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Metrics.MetricSeriesConfigurationForMeasurement+AggregateKindConstants", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Metrics.ConcurrentDatastructures.GrowingCollection`1+Enumerator", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.RichPayloadEventSource+Keywords", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Metrics.MetricTerms+Aggregation", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Metrics.MetricTerms+Extraction", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Metrics.MetricTerms+Autocollection", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Tracing.CoreEventSource+Keywords", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.MetricDimensionNames+TelemetryContext+Cloud", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.MetricDimensionNames+TelemetryContext+Component", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.MetricDimensionNames+TelemetryContext+Device", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.MetricDimensionNames+TelemetryContext+Location", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.MetricDimensionNames+TelemetryContext+Operation", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.MetricDimensionNames+TelemetryContext+Session", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.MetricDimensionNames+TelemetryContext+User", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Metrics.MetricSeriesConfigurationForMeasurement+AggregateKindConstants+DataKeysConstants", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Metrics.MetricSeriesConfigurationForMeasurement+Constants+AggregateKindDataKeys", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Metrics.MetricTerms+Aggregation+Interval", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Metrics.MetricTerms+Extraction+ProcessedByExtractors", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Metrics.MetricTerms+Autocollection+Moniker", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Metrics.MetricTerms+Autocollection+MetricId", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Metrics.MetricTerms+Autocollection+Metric", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Metrics.MetricTerms+Autocollection+Request", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Metrics.MetricTerms+Autocollection+DependencyCall", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Metrics.MetricTerms+Autocollection+TraceCount", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Metrics.MetricTerms+Autocollection+Common", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Metrics.MetricTerms+Aggregation+Interval+Moniker", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Metrics.MetricTerms+Extraction+ProcessedByExtractors+Moniker", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Metrics.MetricTerms+Autocollection+MetricId+Moniker", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Metrics.MetricTerms+Autocollection+Metric+RequestDuration", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Metrics.MetricTerms+Autocollection+Metric+DependencyCallDuration", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Metrics.MetricTerms+Autocollection+Metric+ExceptionCount", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Metrics.MetricTerms+Autocollection+Metric+TraceCount", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Metrics.MetricTerms+Autocollection+Request+PropertyNames", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Metrics.MetricTerms+Autocollection+DependencyCall+PropertyNames", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Metrics.MetricTerms+Autocollection+TraceCount+PropertyNames", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Metrics.MetricTerms+Autocollection+Common+PropertyNames", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.ApplicationInsights.Extensibility.Implementation.Metrics.MetricTerms+Autocollection+Common+PropertyValues", "Microsoft.ApplicationInsights, Version=2.23.0.29, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.ExceptionPolyfills+<G>$E6188BA5B951F1F7AA9135E0EBB76F2B", "Microsoft.Bcl.AsyncInterfaces, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ExceptionPolyfills+<G>$3F30F31B33543D5FB8E174FB4FD780B9", "Microsoft.Bcl.AsyncInterfaces, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ExceptionPolyfills+<G>$E6188BA5B951F1F7AA9135E0EBB76F2B+<M>$96F0261AC622664B8B003966835C0332", "Microsoft.Bcl.AsyncInterfaces, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ExceptionPolyfills+<G>$3F30F31B33543D5FB8E174FB4FD780B9+<M>$1F10CFA08738E6D8AF61CBECC6763DBC", "Microsoft.Bcl.AsyncInterfaces, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"Microsoft.CodeAnalysis.CSharpExtensions", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Conversion", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.CSharpCommandLineArguments", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.CSharpCommandLineParser", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.AwaitExpressionInfo", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.CSharpCompilation", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.CSharpScriptCompilationInfo", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.DeconstructionInfo", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.ForEachStatementInfo", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.QueryClauseInfo", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.CSharpCompilationOptions", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.CSharpExtensions", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.CSharpFileSystemExtensions", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.CSharpParseOptions", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.CSharpDiagnosticFormatter", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.LanguageVersion", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.LanguageVersionFacts", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.SyntaxFacts", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.CSharpGeneratorDriver", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.SymbolDisplay", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.TypedConstantExtensions", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.SyntaxFactory", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.CSharpSyntaxNode", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.CSharpSyntaxRewriter", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.CSharpSyntaxTree", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.CSharpSyntaxVisitor`1", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.CSharpSyntaxVisitor", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.CSharpSyntaxWalker", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.SyntaxExtensions", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.SyntaxKind", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.SyntaxTokenParser", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.InterceptableLocation", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.AliasQualifiedNameSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.AnonymousFunctionExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.AnonymousMethodExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ArgumentSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ArrayRankSpecifierSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.AttributeSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.AttributeTargetSpecifierSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.BaseFieldDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.BaseMethodDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.BasePropertyDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.BlockSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.BreakStatementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.CheckedStatementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ClassDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ClassOrStructConstraintSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.CompilationUnitSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ConstructorDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ContinueStatementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ConversionOperatorDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ConversionOperatorMemberCrefSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.CrefParameterSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.LocalDeclarationStatementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.DelegateDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.DestructorDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.DirectiveTriviaSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.DoStatementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.EmptyStatementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.EnumMemberDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.EventDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ExpressionStatementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ExtensionBlockDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.FixedStatementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ForEachStatementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ForEachVariableStatementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ForStatementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.GenericNameSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.GlobalStatementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.GotoStatementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.IdentifierNameSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.IfStatementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.IndexerDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.InterfaceDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.LabeledStatementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.LambdaExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.LocalFunctionStatementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.LockStatementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.MethodDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.NameColonSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.NamespaceDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.NameSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.OperatorDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.OperatorMemberCrefSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ParameterListSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ParameterSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ParenthesizedLambdaExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.PropertyDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.AccessorDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.QualifiedNameSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.RecordDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.RefTypeSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ReturnStatementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ShebangDirectiveTriviaSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.SimpleLambdaExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.SimpleNameSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.SkippedTokensTriviaSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.StackAllocArrayCreationExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.StructDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.StructuredTriviaSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.SubpatternSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.SwitchStatementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ThrowStatementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.TryStatementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.TypeDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.TypeSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.UnsafeStatementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.UsingDirectiveSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.UsingStatementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.WhileStatementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.XmlNameAttributeElementKind", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.YieldStatementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.TypeArgumentListSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.PredefinedTypeSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ArrayTypeSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.PointerTypeSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.FunctionPointerTypeSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.FunctionPointerParameterListSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.FunctionPointerCallingConventionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.FunctionPointerUnmanagedCallingConventionListSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.FunctionPointerUnmanagedCallingConventionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.NullableTypeSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.TupleTypeSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.TupleElementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.OmittedTypeArgumentSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ScopedTypeSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ExpressionOrPatternSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ParenthesizedExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.TupleExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.PrefixUnaryExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.AwaitExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.PostfixUnaryExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.MemberAccessExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ConditionalAccessExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.MemberBindingExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ElementBindingExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.RangeExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ImplicitElementAccessSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.BinaryExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.AssignmentExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ConditionalExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.InstanceExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ThisExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.BaseExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.LiteralExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.FieldExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.MakeRefExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.RefTypeExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.RefValueExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.CheckedExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.DefaultExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.TypeOfExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.SizeOfExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.InvocationExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ElementAccessExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.BaseArgumentListSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ArgumentListSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.BracketedArgumentListSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.BaseExpressionColonSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ExpressionColonSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.DeclarationExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.CastExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.RefExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.InitializerExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.BaseObjectCreationExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ImplicitObjectCreationExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ObjectCreationExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.WithExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.AnonymousObjectMemberDeclaratorSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.AnonymousObjectCreationExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ArrayCreationExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ImplicitArrayCreationExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ImplicitStackAllocArrayCreationExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.CollectionExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.CollectionElementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ExpressionElementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.SpreadElementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.QueryClauseSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.SelectOrGroupClauseSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.QueryExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.QueryBodySyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.FromClauseSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.LetClauseSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.JoinClauseSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.JoinIntoClauseSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.WhereClauseSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.OrderByClauseSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.OrderingSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.SelectClauseSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.GroupClauseSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.QueryContinuationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.OmittedArraySizeExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.InterpolatedStringExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.IsPatternExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ThrowExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.WhenClauseSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.PatternSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.DiscardPatternSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.DeclarationPatternSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.VarPatternSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.RecursivePatternSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.PositionalPatternClauseSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.PropertyPatternClauseSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ConstantPatternSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ParenthesizedPatternSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.RelationalPatternSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.TypePatternSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.BinaryPatternSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.UnaryPatternSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ListPatternSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.SlicePatternSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.InterpolatedStringContentSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.InterpolatedStringTextSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.InterpolationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.InterpolationAlignmentClauseSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.InterpolationFormatClauseSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.StatementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.VariableDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.VariableDeclaratorSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.EqualsValueClauseSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.VariableDesignationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.SingleVariableDesignationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.DiscardDesignationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ParenthesizedVariableDesignationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.CommonForEachStatementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ElseClauseSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.SwitchSectionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.SwitchLabelSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.CasePatternSwitchLabelSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.CaseSwitchLabelSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.DefaultSwitchLabelSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.SwitchExpressionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.SwitchExpressionArmSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.CatchClauseSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.CatchDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.CatchFilterClauseSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.FinallyClauseSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ExternAliasDirectiveSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.MemberDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.BaseNamespaceDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.FileScopedNamespaceDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.AttributeListSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.AttributeArgumentListSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.AttributeArgumentSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.NameEqualsSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.TypeParameterListSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.TypeParameterSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.BaseTypeDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.EnumDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.BaseListSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.BaseTypeSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.SimpleBaseTypeSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.PrimaryConstructorBaseTypeSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.TypeParameterConstraintClauseSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.TypeParameterConstraintSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ConstructorConstraintSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.TypeConstraintSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.DefaultConstraintSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.AllowsConstraintClauseSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.AllowsConstraintSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.RefStructConstraintSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.FieldDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.EventFieldDeclarationSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ExplicitInterfaceSpecifierSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ConstructorInitializerSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ArrowExpressionClauseSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.AccessorListSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.BaseParameterListSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.BracketedParameterListSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.BaseParameterSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.FunctionPointerParameterSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.IncompleteMemberSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.DocumentationCommentTriviaSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.CrefSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.TypeCrefSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.QualifiedCrefSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.MemberCrefSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.NameMemberCrefSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ExtensionMemberCrefSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.IndexerMemberCrefSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.BaseCrefParameterListSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.CrefParameterListSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.CrefBracketedParameterListSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.XmlNodeSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.XmlElementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.XmlElementStartTagSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.XmlElementEndTagSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.XmlEmptyElementSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.XmlNameSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.XmlPrefixSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.XmlAttributeSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.XmlTextAttributeSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.XmlCrefAttributeSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.XmlNameAttributeSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.XmlTextSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.XmlCDataSectionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.XmlProcessingInstructionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.XmlCommentSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.BranchingDirectiveTriviaSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ConditionalDirectiveTriviaSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.IfDirectiveTriviaSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ElifDirectiveTriviaSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ElseDirectiveTriviaSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.EndIfDirectiveTriviaSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.RegionDirectiveTriviaSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.EndRegionDirectiveTriviaSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ErrorDirectiveTriviaSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.WarningDirectiveTriviaSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.BadDirectiveTriviaSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.DefineDirectiveTriviaSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.UndefDirectiveTriviaSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.LineOrSpanDirectiveTriviaSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.LineDirectiveTriviaSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.LineDirectivePositionSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.LineSpanDirectiveTriviaSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.PragmaWarningDirectiveTriviaSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.PragmaChecksumDirectiveTriviaSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.ReferenceDirectiveTriviaSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.LoadDirectiveTriviaSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.IgnoredDirectiveTriviaSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.NullableDirectiveTriviaSyntax", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.OverloadResolution+Options", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.BoundTreeVisitor+CancelledByStackGuardException", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.SyntheticBoundNodeFactory+MissingPredefinedMember", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.SyntaxTokenParser+Result", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.EnvironmentExtensions+<G>$57DD63EC381B40E0D8049D1E01FA2D1C", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.Syntax.InternalSyntax.SlidingTextWindow+TestAccessor", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.DecisionDagBuilder+Tests+True", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.DecisionDagBuilder+Tests+False", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.DecisionDagBuilder+Tests+One", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.DecisionDagBuilder+Tests+Not", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.DecisionDagBuilder+Tests+SequenceTests", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.DecisionDagBuilder+Tests+AndSequence", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.DecisionDagBuilder+Tests+OrSequence", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.ClosureConversion+Analysis+Scope", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.ClosureConversion+Analysis+NestedFunction", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.ClosureConversion+Analysis+ClosureEnvironment", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.LocalRewriter+PatternLocalRewriter+DagTempAllocator", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CSharp.EnvironmentExtensions+<G>$57DD63EC381B40E0D8049D1E01FA2D1C+<M>$F6325BE232BD572B64A18ADD3350A572", "Microsoft.CodeAnalysis.CSharp, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CaseInsensitiveComparison", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.AnalyzerConfig", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.AnalyzerConfigOptionsResult", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.AnalyzerConfigSet", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CommandLineAnalyzerReference", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CommandLineArguments", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CommandLineParser", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CommandLineReference", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CommandLineResource", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CommandLineSourceFile", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ErrorLogOptions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SarifVersion", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SarifVersionFacts", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Compilation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CandidateReason", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CompilationOptions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ControlFlowAnalysis", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.DataFlowAnalysis", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ModelExtensions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.GeneratedKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IImportScope", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ImportedNamespaceOrType", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ImportedXmlNamespace", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.NullableContext", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.NullableContextExtensions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.NullableContextOptions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.NullableContextOptionsExtensions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.OptimizationLevel", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ParseOptions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Platform", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.PreprocessingSymbolInfo", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ScriptCompilationInfo", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SemanticModel", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SemanticModelOptions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SourceReferenceResolver", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SpeculativeBindingOption", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SubsystemVersion", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SymbolFilter", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SymbolInfo", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxTreeOptionsProvider", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.TypeInfo", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.XmlReferenceResolver", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.DesktopAssemblyIdentityComparer", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.AdditionalText", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IAnalyzerAssemblyLoader", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostic", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.DiagnosticDescriptor", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.DiagnosticFormatter", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.DiagnosticSeverity", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.FileLinePositionSpan", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.LocalizableResourceString", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.LocalizableString", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Location", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.LocationKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ReportDiagnostic", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SuppressionDescriptor", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.WellKnownDiagnosticTags", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.DocumentationCommentId", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.DocumentationProvider", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.DocumentationMode", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.EmbeddedText", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.FileSystemExtensions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.OperationKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.MetadataImportOptions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.AssemblyIdentity", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.AssemblyIdentityComparer", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.AssemblyIdentityParts", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.AssemblyMetadata", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CompilationReference", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.MetadataId", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Metadata", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.MetadataImageKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.MetadataReference", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.MetadataReferenceProperties", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.MetadataReferenceResolver", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ModuleMetadata", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.PortableExecutableReference", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.UnresolvedMetadataReference", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Optional`1", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.OutputKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ResourceDescription", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.RuleSet", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.RuleSetInclude", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.RuntimeCapability", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SourceCodeKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SourceFileResolver", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.GeneratorAttribute", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.GeneratorExecutionContext", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.GeneratorInitializationContext", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.GeneratorSyntaxContext", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.GeneratorPostInitializationContext", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.GeneratorDriver", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.GeneratorDriverOptions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.GeneratorExtensions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IIncrementalGenerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IncrementalGeneratorInitializationContext", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IncrementalGeneratorPostInitializationContext", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SourceProductionContext", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.GeneratorFilterContext", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.HostOutputProductionContext", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IncrementalGeneratorRunStep", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IncrementalStepRunReason", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IncrementalValueProvider`1", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IncrementalValuesProvider`1", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ISourceGenerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ISyntaxReceiver", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxReceiverCreator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ISyntaxContextReceiver", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxContextReceiverCreator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IncrementalGeneratorOutputKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxValueProvider", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.GeneratorAttributeSyntaxContext", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IncrementalValueProviderExtensions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.GeneratorDriverRunResult", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.GeneratorRunResult", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.GeneratedSourceResult", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.GeneratorDriverTimingInfo", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.GeneratorTimingInfo", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.WellKnownGeneratorInputs", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.WellKnownGeneratorOutputs", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SpecialType", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.DesktopStrongNameProvider", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.StrongNameProvider", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SymbolDisplayDelegateStyle", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SymbolDisplayExtensionMethodStyle", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SymbolDisplayExtensions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SymbolDisplayFormat", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SymbolDisplayGenericsOptions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SymbolDisplayGlobalNamespaceStyle", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SymbolDisplayKindOptions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SymbolDisplayLocalOptions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SymbolDisplayMemberOptions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SymbolDisplayMiscellaneousOptions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SymbolDisplayParameterOptions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SymbolDisplayPart", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SymbolDisplayPartKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SymbolDisplayPropertyStyle", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SymbolDisplayTypeQualificationStyle", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Accessibility", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.AttributeData", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CustomModifier", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IAliasSymbol", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IArrayTypeSymbol", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IAssemblySymbol", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IDiscardSymbol", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IDynamicTypeSymbol", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IErrorTypeSymbol", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IEventSymbol", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IFieldSymbol", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IFunctionPointerTypeSymbol", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ILabelSymbol", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ILocalSymbol", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IMethodSymbol", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IModuleSymbol", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.INamedTypeSymbol", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.INamespaceOrTypeSymbol", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.INamespaceSymbol", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IParameterSymbol", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IPointerTypeSymbol", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IPreprocessingSymbol", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IPropertySymbol", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IRangeVariableSymbol", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ISourceAssemblySymbol", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ISymbol", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ISymbolExtensions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ITypeParameterSymbol", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ITypeSymbol", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.LanguageNames", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.MethodKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.NamespaceKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.NullabilityInfo", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.NullableAnnotation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.NullableFlowState", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.DllImportData", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.RefKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ScopedKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SymbolEqualityComparer", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SymbolKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SymbolVisitor", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SymbolVisitor`1", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SymbolVisitor`2", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.TypedConstant", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.TypedConstantKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.TypeKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.TypeParameterKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.VarianceKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.WellKnownMemberNames", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.AnnotationExtensions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ChildSyntaxList", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ICompilationUnitSyntax", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ISkippedTokensTriviaSyntax", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IStructuredTriviaSyntax", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.LineMapping", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.LineVisibility", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SeparatedSyntaxList", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SeparatedSyntaxList`1", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxAnnotation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxList", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxList`1", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxNode", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxNodeExtensions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxNodeOrToken", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxNodeOrTokenList", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxReference", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxRemoveOptions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxToken", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxTokenList", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxTree", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxTrivia", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxTriviaList", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxWalker", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxWalkerDepth", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.XmlFileResolver", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.FlowAnalysis.IFlowCaptureOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.FlowAnalysis.IFlowCaptureReferenceOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.FlowAnalysis.IIsNullOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.FlowAnalysis.ICaughtExceptionOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.FlowAnalysis.IStaticLocalInitializationSemaphoreOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.FlowAnalysis.IFlowAnonymousFunctionOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.FlowAnalysis.BasicBlock", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.FlowAnalysis.BasicBlockKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.FlowAnalysis.CaptureId", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.FlowAnalysis.ControlFlowBranch", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.FlowAnalysis.ControlFlowBranchSemantics", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.FlowAnalysis.ControlFlowConditionKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.FlowAnalysis.ControlFlowGraph", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.FlowAnalysis.ControlFlowGraphExtensions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.FlowAnalysis.ControlFlowRegion", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.FlowAnalysis.ControlFlowRegionKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Text.LinePosition", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Text.LinePositionSpan", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Text.SourceHashAlgorithm", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Text.SourceText", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Text.SourceTextContainer", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Text.TextChange", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Text.TextChangeEventArgs", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Text.TextChangeRange", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Text.TextLine", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Text.TextLineCollection", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Text.TextSpan", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.AdditionalTextValueProvider`1", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.AnalysisResult", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.AnalyzerConfigOptions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.AnalyzerConfigOptionsProvider", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.AnalyzerFileReference", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.AnalyzerImageReference", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.AnalyzerLoadFailureEventArgs", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.AnalyzerOptions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.AnalyzerReference", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.CompilationWithAnalyzers", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.CompilationWithAnalyzersOptions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.AnalysisContext", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.GeneratedCodeAnalysisFlags", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.CompilationStartAnalysisContext", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.CompilationAnalysisContext", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.SemanticModelAnalysisContext", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.SymbolAnalysisContext", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.SymbolStartAnalysisContext", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.CodeBlockStartAnalysisContext`1", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.CodeBlockAnalysisContext", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.OperationBlockStartAnalysisContext", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.OperationBlockAnalysisContext", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.SyntaxTreeAnalysisContext", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.AdditionalFileAnalysisContext", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.SyntaxNodeAnalysisContext", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.OperationAnalysisContext", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.SuppressionAnalysisContext", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.DiagnosticAnalyzer", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.DiagnosticAnalyzerAttribute", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.DiagnosticAnalyzerExtensions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.DiagnosticSuppressor", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.SourceTextValueProvider`1", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.Suppression", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.SyntaxTreeValueProvider`1", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.UnresolvedAnalyzerReference", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.SuppressionInfo", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.Telemetry.AnalyzerTelemetryInfo", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IInvalidOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IBlockOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IVariableDeclarationGroupOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ISwitchOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ILoopOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IForEachLoopOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IForLoopOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IForToLoopOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IWhileLoopOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ILabeledOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IBranchOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IEmptyOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IReturnOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ILockOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ITryOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IUsingOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IExpressionStatementOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ILocalFunctionOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IStopOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IEndOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IRaiseEventOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ILiteralOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IConversionOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IInvocationOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IArrayElementReferenceOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ILocalReferenceOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IParameterReferenceOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IMemberReferenceOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IFieldReferenceOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IMethodReferenceOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IPropertyReferenceOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IEventReferenceOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IUnaryOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IBinaryOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IConditionalOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ICoalesceOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IAnonymousFunctionOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IObjectCreationOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ITypeParameterObjectCreationOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IArrayCreationOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IInstanceReferenceOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IIsTypeOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IAwaitOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IAssignmentOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ISimpleAssignmentOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ICompoundAssignmentOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IParenthesizedOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IEventAssignmentOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IConditionalAccessOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IConditionalAccessInstanceOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IInterpolatedStringOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IAnonymousObjectCreationOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IObjectOrCollectionInitializerOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IMemberInitializerOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ICollectionElementInitializerOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.INameOfOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ITupleOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IDynamicObjectCreationOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IDynamicMemberReferenceOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IDynamicInvocationOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IDynamicIndexerAccessOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ITranslatedQueryOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IDelegateCreationOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IDefaultValueOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ITypeOfOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ISizeOfOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IAddressOfOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IIsPatternOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IIncrementOrDecrementOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IThrowOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IDeconstructionAssignmentOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IDeclarationExpressionOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IOmittedArgumentOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ISymbolInitializerOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IFieldInitializerOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IVariableInitializerOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IPropertyInitializerOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IParameterInitializerOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IArrayInitializerOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IVariableDeclaratorOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IVariableDeclarationOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IArgumentOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ICatchClauseOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ISwitchCaseOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ICaseClauseOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IDefaultCaseClauseOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IPatternCaseClauseOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IRangeCaseClauseOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IRelationalCaseClauseOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ISingleValueCaseClauseOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IInterpolatedStringContentOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IInterpolatedStringTextOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IInterpolationOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IPatternOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IConstantPatternOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IDeclarationPatternOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ITupleBinaryOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IMethodBodyBaseOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IMethodBodyOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IConstructorBodyOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IDiscardOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ICoalesceAssignmentOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IRangeOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IReDimOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IReDimClauseOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IRecursivePatternOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IDiscardPatternOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ISwitchExpressionOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ISwitchExpressionArmOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IPropertySubpatternOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IUsingDeclarationOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.INegatedPatternOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IBinaryPatternOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ITypePatternOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IRelationalPatternOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IWithOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IInterpolatedStringHandlerCreationOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IInterpolatedStringAdditionOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IInterpolatedStringAppendOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IInterpolatedStringHandlerArgumentPlaceholderOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IFunctionPointerInvocationOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IListPatternOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ISlicePatternOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IImplicitIndexerReferenceOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IUtf8StringOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IAttributeOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.IInlineArrayAccessOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ICollectionExpressionOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ISpreadOperation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.OperationVisitor", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.OperationVisitor`2", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.ArgumentKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.BinaryOperatorKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.BranchKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.CaseKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.CommonConversion", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.InstanceReferenceKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.InterpolatedStringArgumentPlaceholderKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.LoopKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.OperationExtensions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.OperationWalker", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.OperationWalker`1", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Operations.UnaryOperatorKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Emit.EmitDifferenceOptions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Emit.EmitResult", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Emit.DebugInformationFormat", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Emit.EditAndContinueMethodDebugInformation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Emit.EmitBaseline", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Emit.EmitDifferenceResult", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Emit.EmitOptions", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Emit.InstrumentationKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Emit.MethodInstrumentation", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Emit.RuntimeRudeEdit", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Emit.SemanticEdit", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Emit.SemanticEditKind", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.IO.Hashing.XxHashShared+State", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Roslyn.Utilities.ConcurrentSet`1+KeyEnumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Roslyn.Utilities.RoslynDebug+AssertInterpolatedStringHandler", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Roslyn.Utilities.MultiDictionary`2+ValueSet", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Roslyn.Utilities.WeakList`1+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.Cci.MetadataWriter+RawTokenEncoding", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.AbstractLookupSymbolsInfo`1+ArityEnumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.AbstractLookupSymbolsInfo`1+IArityEnumerable", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CodeAnalysisEventSource+Keywords", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CodeAnalysisEventSource+Tasks", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SmallDictionary`2+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SmallDictionary`2+EnumerableImpl", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.MetadataTypeName+Key", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.AssemblyIdentityComparer+ComparisonResult", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.MethodImplAttributeExtensions+<G>$FAAB207A3F395A6E169F99D6320A19F3", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.MethodImplOptionsExtensions+<G>$D2AE889EBC4F97C82AC8DABEE0C8CE8B", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IOperation+OperationList", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.DriverStateTable+Builder", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.GeneratorRunStateTable+Builder", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.NodeStateTable`1+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.NodeStateTable`1+Builder", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.StateTableStore+Builder", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxStore+Builder", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SwitchConstantValueHelper+SwitchLabelsComparer", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ChildSyntaxList+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ChildSyntaxList+Reversed", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.GreenNode+NodeEnumerable", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.LineDirectiveMap`1+PositionState", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SeparatedSyntaxList`1+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxList`1+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxNodeOrTokenList+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxTokenList+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxTokenList+Reversed", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxTriviaList+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxTriviaList+Reversed", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ErrorReporting.FatalError+ErrorReporterHandler", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.PooledObjects.PooledDelegates+Releaser", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Syntax.InternalSyntax.SyntaxDiagnosticInfoList+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Text.TextLineCollection+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Diagnostics.AnalyzerLoadFailureEventArgs+FailureErrorCode", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Collections.OrderPreservingMultiDictionary`2+ValueSet", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Collections.ImmutableSegmentedDictionary`2+Builder", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Collections.ImmutableSegmentedDictionary`2+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Collections.ImmutableSegmentedDictionary`2+KeyCollection", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Collections.ImmutableSegmentedDictionary`2+ValueCollection", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Collections.ImmutableSegmentedHashSet`1+Builder", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Collections.ImmutableSegmentedHashSet`1+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Collections.ImmutableSegmentedList`1+Builder", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Collections.ImmutableSegmentedList`1+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Collections.SegmentedArray`1+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Collections.SegmentedDictionary`2+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Collections.SegmentedDictionary`2+KeyCollection", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Collections.SegmentedDictionary`2+ValueCollection", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Collections.SegmentedHashSet`1+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Collections.SegmentedList`1+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Collections.TemporaryArray`1+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CodeGen.SwitchStringJumpTableEmitter+EmitStringCompareAndBranch", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.CodeGen.SwitchStringJumpTableEmitter+GetStringHashCode", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.IO.Hashing.XxHashShared+State+<Accumulators>e__FixedBuffer", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.IO.Hashing.XxHashShared+State+<Buffer>e__FixedBuffer", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.IO.Hashing.XxHashShared+State+<Secret>e__FixedBuffer", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Roslyn.Utilities.MultiDictionary`2+ValueSet+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SmallDictionary`2+KeyCollection+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SmallDictionary`2+KeyCollection+EnumerableImpl", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SmallDictionary`2+ValueCollection+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SmallDictionary`2+ValueCollection+EnumerableImpl", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.DesktopAssemblyIdentityComparer+FrameworkAssemblyDictionary+Value", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.DesktopAssemblyIdentityComparer+FrameworkRetargetingDictionary+Key", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.DesktopAssemblyIdentityComparer+FrameworkRetargetingDictionary+Value", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.MethodImplAttributeExtensions+<G>$FAAB207A3F395A6E169F99D6320A19F3+<M>$FAAB207A3F395A6E169F99D6320A19F3", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.MethodImplOptionsExtensions+<G>$D2AE889EBC4F97C82AC8DABEE0C8CE8B+<M>$D2AE889EBC4F97C82AC8DABEE0C8CE8B", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IOperation+OperationList+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IOperation+OperationList+Reversed", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.NodeStateTable`1+TableEntry+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.NodeStateTable`1+TableEntry+Builder", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.ChildSyntaxList+Reversed+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.GreenNode+NodeEnumerable+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxNode+TwoEnumeratorListStack+Which", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxNode+ThreeEnumeratorListStack+Which", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxTokenList+Reversed+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.SyntaxTriviaList+Reversed+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Emit.DeltaMetadataWriter+DefinitionIndex`1+TryGetExistingIndex", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Emit.DeltaMetadataWriter+EventOrPropertyMapIndex+TryGetExistingIndex", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Collections.OrderPreservingMultiDictionary`2+ValueSet+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Collections.ImmutableSegmentedDictionary`2+Builder+KeyCollection", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Collections.ImmutableSegmentedDictionary`2+Builder+ValueCollection", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Collections.ImmutableSegmentedDictionary`2+KeyCollection+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Collections.ImmutableSegmentedDictionary`2+ValueCollection+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Collections.SegmentedArray+UnalignedSegmentEnumerable`1+ReverseEnumerable", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Collections.SegmentedArray+UnalignedSegmentEnumerator`1+Reverse", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Collections.SegmentedArray+SegmentEnumerable`1+ReverseEnumerable", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Collections.SegmentedArray+SegmentEnumerator`1+Reverse", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Collections.SegmentedDictionary`2+KeyCollection+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.Collections.SegmentedDictionary`2+ValueCollection+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.CodeAnalysis.IOperation+OperationList+Reversed+Enumerator", "Microsoft.CodeAnalysis, Version=5.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.Extensions.ObjectPool.DefaultObjectPool`1", "Microsoft.Extensions.ObjectPool, Version=10.0.0.0, Culture=neutral, PublicKeyToken=adb9793829ddae60"},
                {"Microsoft.Extensions.ObjectPool.DefaultObjectPoolProvider", "Microsoft.Extensions.ObjectPool, Version=10.0.0.0, Culture=neutral, PublicKeyToken=adb9793829ddae60"},
                {"Microsoft.Extensions.ObjectPool.DefaultPooledObjectPolicy`1", "Microsoft.Extensions.ObjectPool, Version=10.0.0.0, Culture=neutral, PublicKeyToken=adb9793829ddae60"},
                {"Microsoft.Extensions.ObjectPool.IPooledObjectPolicy`1", "Microsoft.Extensions.ObjectPool, Version=10.0.0.0, Culture=neutral, PublicKeyToken=adb9793829ddae60"},
                {"Microsoft.Extensions.ObjectPool.IResettable", "Microsoft.Extensions.ObjectPool, Version=10.0.0.0, Culture=neutral, PublicKeyToken=adb9793829ddae60"},
                {"Microsoft.Extensions.ObjectPool.LeakTrackingObjectPool`1", "Microsoft.Extensions.ObjectPool, Version=10.0.0.0, Culture=neutral, PublicKeyToken=adb9793829ddae60"},
                {"Microsoft.Extensions.ObjectPool.LeakTrackingObjectPoolProvider", "Microsoft.Extensions.ObjectPool, Version=10.0.0.0, Culture=neutral, PublicKeyToken=adb9793829ddae60"},
                {"Microsoft.Extensions.ObjectPool.ObjectPool`1", "Microsoft.Extensions.ObjectPool, Version=10.0.0.0, Culture=neutral, PublicKeyToken=adb9793829ddae60"},
                {"Microsoft.Extensions.ObjectPool.ObjectPool", "Microsoft.Extensions.ObjectPool, Version=10.0.0.0, Culture=neutral, PublicKeyToken=adb9793829ddae60"},
                {"Microsoft.Extensions.ObjectPool.ObjectPoolProvider", "Microsoft.Extensions.ObjectPool, Version=10.0.0.0, Culture=neutral, PublicKeyToken=adb9793829ddae60"},
                {"Microsoft.Extensions.ObjectPool.ObjectPoolProviderExtensions", "Microsoft.Extensions.ObjectPool, Version=10.0.0.0, Culture=neutral, PublicKeyToken=adb9793829ddae60"},
                {"Microsoft.Extensions.ObjectPool.PooledObjectPolicy`1", "Microsoft.Extensions.ObjectPool, Version=10.0.0.0, Culture=neutral, PublicKeyToken=adb9793829ddae60"},
                {"Microsoft.Extensions.ObjectPool.StringBuilderPooledObjectPolicy", "Microsoft.Extensions.ObjectPool, Version=10.0.0.0, Culture=neutral, PublicKeyToken=adb9793829ddae60"},
                {"Microsoft.PowerShell.MarkdownRender.MarkdownConversionType", "Microsoft.PowerShell.MarkdownRender, Version=7.2.1.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.PowerShell.MarkdownRender.MarkdownInfo", "Microsoft.PowerShell.MarkdownRender, Version=7.2.1.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.PowerShell.MarkdownRender.MarkdownConverter", "Microsoft.PowerShell.MarkdownRender, Version=7.2.1.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.PowerShell.MarkdownRender.MarkdownOptionInfoProperty", "Microsoft.PowerShell.MarkdownRender, Version=7.2.1.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.PowerShell.MarkdownRender.PSMarkdownOptionInfo", "Microsoft.PowerShell.MarkdownRender, Version=7.2.1.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.PowerShell.MarkdownRender.VT100EscapeSequences", "Microsoft.PowerShell.MarkdownRender, Version=7.2.1.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.PowerShell.MarkdownRender.VT100ObjectRenderer`1", "Microsoft.PowerShell.MarkdownRender, Version=7.2.1.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.PowerShell.MarkdownRender.VT100Renderer", "Microsoft.PowerShell.MarkdownRender, Version=7.2.1.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"Microsoft.Win32.RegistryAclExtensions", "Microsoft.Win32.Registry.AccessControl, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"Microsoft.Win32.PowerModeChangedEventArgs", "Microsoft.Win32.SystemEvents, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"Microsoft.Win32.PowerModeChangedEventHandler", "Microsoft.Win32.SystemEvents, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"Microsoft.Win32.PowerModes", "Microsoft.Win32.SystemEvents, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"Microsoft.Win32.SessionEndedEventArgs", "Microsoft.Win32.SystemEvents, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"Microsoft.Win32.SessionEndedEventHandler", "Microsoft.Win32.SystemEvents, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"Microsoft.Win32.SessionEndingEventArgs", "Microsoft.Win32.SystemEvents, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"Microsoft.Win32.SessionEndingEventHandler", "Microsoft.Win32.SystemEvents, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"Microsoft.Win32.SessionEndReasons", "Microsoft.Win32.SystemEvents, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"Microsoft.Win32.SessionSwitchEventArgs", "Microsoft.Win32.SystemEvents, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"Microsoft.Win32.SessionSwitchEventHandler", "Microsoft.Win32.SystemEvents, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"Microsoft.Win32.SessionSwitchReason", "Microsoft.Win32.SystemEvents, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"Microsoft.Win32.SystemEvents", "Microsoft.Win32.SystemEvents, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"Microsoft.Win32.TimerElapsedEventArgs", "Microsoft.Win32.SystemEvents, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"Microsoft.Win32.TimerElapsedEventHandler", "Microsoft.Win32.SystemEvents, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"Microsoft.Win32.UserPreferenceCategory", "Microsoft.Win32.SystemEvents, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"Microsoft.Win32.UserPreferenceChangedEventArgs", "Microsoft.Win32.SystemEvents, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"Microsoft.Win32.UserPreferenceChangedEventHandler", "Microsoft.Win32.SystemEvents, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"Microsoft.Win32.UserPreferenceChangingEventArgs", "Microsoft.Win32.SystemEvents, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"Microsoft.Win32.UserPreferenceChangingEventHandler", "Microsoft.Win32.SystemEvents, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"Newtonsoft.Json.ConstructorHandling", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.DateFormatHandling", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.DateParseHandling", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.DateTimeZoneHandling", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.DefaultJsonNameTable", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.DefaultValueHandling", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.FloatFormatHandling", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.FloatParseHandling", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Formatting", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.IArrayPool`1", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.IJsonLineInfo", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.JsonArrayAttribute", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.JsonConstructorAttribute", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.JsonContainerAttribute", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.JsonConvert", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.JsonConverter", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.JsonConverter`1", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.JsonConverterAttribute", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.JsonConverterCollection", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.JsonDictionaryAttribute", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.JsonException", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.JsonExtensionDataAttribute", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.JsonIgnoreAttribute", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.JsonNameTable", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.JsonObjectAttribute", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.JsonPropertyAttribute", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.JsonReader", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.JsonReaderException", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.JsonRequiredAttribute", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.JsonSerializationException", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.JsonSerializer", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.JsonSerializerSettings", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.JsonTextReader", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.JsonTextWriter", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.JsonToken", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.JsonValidatingReader", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.JsonWriter", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.JsonWriterException", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.MemberSerialization", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.MetadataPropertyHandling", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.MissingMemberHandling", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.NullValueHandling", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.ObjectCreationHandling", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.PreserveReferencesHandling", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.ReferenceLoopHandling", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Required", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.StringEscapeHandling", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.TypeNameAssemblyFormatHandling", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.TypeNameHandling", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.WriteState", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.CamelCaseNamingStrategy", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.CamelCasePropertyNamesContractResolver", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.DefaultContractResolver", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.DefaultNamingStrategy", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.DefaultSerializationBinder", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.DiagnosticsTraceWriter", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.DynamicValueProvider", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.ErrorContext", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.ErrorEventArgs", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.ExpressionValueProvider", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.IAttributeProvider", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.IContractResolver", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.IReferenceResolver", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.ISerializationBinder", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.ITraceWriter", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.IValueProvider", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.JsonArrayContract", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.JsonContainerContract", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.SerializationCallback", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.SerializationErrorCallback", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.ExtensionDataSetter", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.ExtensionDataGetter", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.JsonContract", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.JsonDictionaryContract", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.JsonDynamicContract", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.JsonISerializableContract", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.JsonLinqContract", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.JsonObjectContract", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.JsonPrimitiveContract", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.JsonProperty", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.JsonPropertyCollection", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.JsonStringContract", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.KebabCaseNamingStrategy", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.MemoryTraceWriter", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.NamingStrategy", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.ObjectConstructor`1", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.OnErrorAttribute", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.ReflectionAttributeProvider", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.ReflectionValueProvider", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Serialization.SnakeCaseNamingStrategy", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Schema.Extensions", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Schema.JsonSchema", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Schema.JsonSchemaException", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Schema.JsonSchemaGenerator", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Schema.JsonSchemaResolver", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Schema.JsonSchemaType", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Schema.UndefinedSchemaIdHandling", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Schema.ValidationEventArgs", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Schema.ValidationEventHandler", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Linq.CommentHandling", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Linq.DuplicatePropertyNameHandling", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Linq.Extensions", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Linq.IJEnumerable`1", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Linq.JArray", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Linq.JConstructor", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Linq.JContainer", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Linq.JEnumerable`1", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Linq.JObject", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Linq.JProperty", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Linq.JPropertyDescriptor", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Linq.JRaw", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Linq.JsonCloneSettings", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Linq.JsonLoadSettings", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Linq.JsonMergeSettings", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Linq.JsonSelectSettings", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Linq.JToken", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Linq.JTokenEqualityComparer", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Linq.JTokenReader", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Linq.JTokenType", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Linq.JTokenWriter", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Linq.JValue", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Linq.LineInfoHandling", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Linq.MergeArrayHandling", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Linq.MergeNullValueHandling", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Converters.BinaryConverter", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Converters.BsonObjectIdConverter", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Converters.CustomCreationConverter`1", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Converters.DataSetConverter", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Converters.DataTableConverter", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Converters.DateTimeConverterBase", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Converters.DiscriminatedUnionConverter", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Converters.EntityKeyMemberConverter", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Converters.ExpandoObjectConverter", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Converters.IsoDateTimeConverter", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Converters.JavaScriptDateTimeConverter", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Converters.KeyValuePairConverter", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Converters.RegexConverter", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Converters.StringEnumConverter", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Converters.UnixDateTimeConverter", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Converters.VersionConverter", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Converters.XmlNodeConverter", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Bson.BsonObjectId", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Bson.BsonReader", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"Newtonsoft.Json.Bson.BsonWriter", "Newtonsoft.Json, Version=13.0.0.0, Culture=neutral, PublicKeyToken=30ad4fe6b2a6aeed"},
                {"System.CodeDom.CodeArgumentReferenceExpression", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeArrayCreateExpression", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeArrayIndexerExpression", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeAssignStatement", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeAttachEventStatement", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeAttributeArgument", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeAttributeArgumentCollection", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeAttributeDeclaration", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeAttributeDeclarationCollection", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeBaseReferenceExpression", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeBinaryOperatorExpression", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeBinaryOperatorType", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeCastExpression", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeCatchClause", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeCatchClauseCollection", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeChecksumPragma", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeComment", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeCommentStatement", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeCommentStatementCollection", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeCompileUnit", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeConditionStatement", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeConstructor", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeDefaultValueExpression", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeDelegateCreateExpression", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeDelegateInvokeExpression", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeDirectionExpression", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeDirective", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeDirectiveCollection", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeEntryPointMethod", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeEventReferenceExpression", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeExpression", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeExpressionCollection", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeExpressionStatement", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeFieldReferenceExpression", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeGotoStatement", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeIndexerExpression", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeIterationStatement", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeLabeledStatement", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeLinePragma", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeMemberEvent", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeMemberField", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeMemberMethod", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeMemberProperty", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeMethodInvokeExpression", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeMethodReferenceExpression", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeMethodReturnStatement", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeNamespace", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeNamespaceCollection", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeNamespaceImport", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeNamespaceImportCollection", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeObject", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeObjectCreateExpression", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeParameterDeclarationExpression", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeParameterDeclarationExpressionCollection", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodePrimitiveExpression", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodePropertyReferenceExpression", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodePropertySetValueReferenceExpression", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeRegionDirective", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeRegionMode", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeRemoveEventStatement", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeSnippetCompileUnit", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeSnippetExpression", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeSnippetStatement", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeSnippetTypeMember", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeStatement", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeStatementCollection", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeThisReferenceExpression", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeThrowExceptionStatement", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeTryCatchFinallyStatement", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeTypeConstructor", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeTypeDeclaration", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeTypeDeclarationCollection", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeTypeDelegate", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeTypeMember", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeTypeMemberCollection", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeTypeOfExpression", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeTypeParameter", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeTypeParameterCollection", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeTypeReferenceOptions", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeTypeReference", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeTypeReferenceCollection", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeTypeReferenceExpression", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeVariableDeclarationStatement", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.CodeVariableReferenceExpression", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.FieldDirection", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.MemberAttributes", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.Compiler.CodeCompiler", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.Compiler.CodeDomProvider", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.Compiler.CodeGenerator", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.Compiler.CodeGeneratorOptions", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.Compiler.CodeParser", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.Compiler.CompilerError", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.Compiler.CompilerErrorCollection", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.Compiler.CompilerInfo", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.Compiler.CompilerParameters", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.Compiler.CompilerResults", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.Compiler.Executor", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.Compiler.GeneratorSupport", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.Compiler.ICodeCompiler", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.Compiler.ICodeGenerator", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.Compiler.ICodeParser", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.Compiler.LanguageOptions", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.CodeDom.Compiler.TempFileCollection", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"Microsoft.VisualBasic.VBCodeProvider", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"Microsoft.CSharp.CSharpCodeProvider", "System.CodeDom, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ComponentModel.Composition.AttributedModelServices", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.CatalogReflectionContextAttribute", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.ChangeRejectedException", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.CompositionContractMismatchException", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.CompositionError", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.CompositionException", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.CreationPolicy", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.ExportAttribute", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.ExportFactory`1", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.ExportFactory`2", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.ExportLifetimeContext`1", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.ExportMetadataAttribute", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.ICompositionService", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.ImportAttribute", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.ImportCardinalityMismatchException", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.ImportingConstructorAttribute", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.ImportManyAttribute", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.ImportSource", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.InheritedExportAttribute", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.IPartImportsSatisfiedNotification", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.MetadataAttributeAttribute", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.MetadataViewImplementationAttribute", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.PartCreationPolicyAttribute", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.PartMetadataAttribute", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.PartNotDiscoverableAttribute", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.AdaptationConstants", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.ReflectionModel.LazyMemberInfo", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.ReflectionModel.ReflectionModelServices", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Primitives.ComposablePart", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Primitives.ComposablePartCatalog", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Primitives.ComposablePartDefinition", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Primitives.ComposablePartException", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Primitives.ContractBasedImportDefinition", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Primitives.Export", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Primitives.ExportDefinition", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Primitives.ExportedDelegate", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Primitives.ICompositionElement", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Primitives.ImportCardinality", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Primitives.ImportDefinition", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Hosting.AggregateCatalog", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Hosting.AggregateExportProvider", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Hosting.ApplicationCatalog", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Hosting.AssemblyCatalog", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Hosting.AtomicComposition", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Hosting.CatalogExportProvider", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Hosting.CatalogExtensions", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Hosting.ComposablePartCatalogChangeEventArgs", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Hosting.ComposablePartExportProvider", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Hosting.CompositionBatch", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Hosting.CompositionConstants", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Hosting.CompositionContainer", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Hosting.CompositionOptions", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Hosting.CompositionScopeDefinition", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Hosting.CompositionService", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Hosting.DirectoryCatalog", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Hosting.ExportProvider", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Hosting.ExportsChangeEventArgs", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Hosting.FilteredCatalog", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Hosting.ImportEngine", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Hosting.INotifyComposablePartCatalogChanged", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Hosting.ScopingExtensions", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Hosting.TypeCatalog", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.Composition.Diagnostics.DebuggerTraceWriter+TraceEventType", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.MetadataViewGenerator+MetadataViewFactory", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Hosting.CompositionLock+CompositionLockHolder", "System.ComponentModel.Composition, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Registration.ExportBuilder", "System.ComponentModel.Composition.Registration, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Registration.ImportBuilder", "System.ComponentModel.Composition.Registration, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Registration.ParameterImportBuilder", "System.ComponentModel.Composition.Registration, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Registration.PartBuilder", "System.ComponentModel.Composition.Registration, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Registration.PartBuilder`1", "System.ComponentModel.Composition.Registration, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.ComponentModel.Composition.Registration.RegistrationBuilder", "System.ComponentModel.Composition.Registration, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.UriIdnScope", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.TraceConfiguration", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Configuration.SystemDrawingSection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.DictionarySectionHandler", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.DpapiProtectedConfigurationProvider", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.IConfigurationSystem", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.IdnElement", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.IPersistComponentSettings", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ApplicationSettingsBase", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ApplicationScopedSettingAttribute", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ApplicationSettingsGroup", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.AppSettingsReader", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.AppSettingsSection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.CallbackValidator", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.CallbackValidatorAttribute", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ClientSettingsSection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.CommaDelimitedStringCollectionConverter", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.Configuration", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationAllowDefinition", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationAllowExeDefinition", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationCollectionAttribute", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationConverterBase", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationElement", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationElementCollection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationElementCollectionType", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationElementProperty", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationErrorsException", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationException", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationFileMap", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationLocation", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationLocationCollection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationLockCollection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationManager", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationProperty", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationPropertyAttribute", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationPropertyCollection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationPropertyOptions", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationSaveMode", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationSection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationSectionCollection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationSectionGroup", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationSectionGroupCollection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationSettings", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationUserLevel", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationValidatorAttribute", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationValidatorBase", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigXmlDocument", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConnectionStringSettings", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConnectionStringSettingsCollection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConnectionStringsSection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ContextInformation", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.DefaultSection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.DefaultSettingValueAttribute", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.DefaultValidator", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ElementInformation", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ExeConfigurationFileMap", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ExeContext", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.GenericEnumConverter", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.IApplicationSettingsProvider", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.IConfigurationSectionHandler", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.IgnoreSection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.IgnoreSectionHandler", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.InfiniteIntConverter", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.InfiniteTimeSpanConverter", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.IntegerValidator", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.IntegerValidatorAttribute", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.IriParsingElement", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ISettingsProviderService", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.KeyValueConfigurationCollection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.KeyValueConfigurationElement", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.LocalFileSettingsProvider", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.LongValidator", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.LongValidatorAttribute", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.NameValueConfigurationCollection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.NameValueConfigurationElement", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.NameValueFileSectionHandler", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.NameValueSectionHandler", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.OverrideMode", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.PositiveTimeSpanValidator", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.PositiveTimeSpanValidatorAttribute", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.PropertyInformation", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.PropertyInformationCollection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.PropertyValueOrigin", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ProtectedConfiguration", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ProtectedConfigurationProvider", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ProtectedConfigurationProviderCollection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ProtectedConfigurationSection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ProtectedProviderSettings", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ProviderSettings", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ProviderSettingsCollection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.RegexStringValidator", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.RegexStringValidatorAttribute", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.RsaProtectedConfigurationProvider", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SchemeSettingElement", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SchemeSettingElementCollection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SectionInformation", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingAttribute", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingChangingEventArgs", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingChangingEventHandler", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingElement", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingElementCollection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingsAttributeDictionary", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingsBase", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingsContext", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingsDescriptionAttribute", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingsGroupDescriptionAttribute", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingsGroupNameAttribute", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingsLoadedEventArgs", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingsLoadedEventHandler", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingsManageability", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingsManageabilityAttribute", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingsPropertyCollection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingsPropertyIsReadOnlyException", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingsPropertyNotFoundException", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingsPropertyValueCollection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingsPropertyWrongTypeException", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingsProvider", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingsProviderAttribute", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingsProviderCollection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingsSavingEventHandler", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingsSerializeAsAttribute", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingValueElement", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SingleTagSectionHandler", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SpecialSetting", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SpecialSettingAttribute", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.CommaDelimitedStringCollection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.StringValidator", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.StringValidatorAttribute", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SubclassTypeValidator", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SubclassTypeValidatorAttribute", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.TimeSpanMinutesConverter", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.TimeSpanMinutesOrInfiniteConverter", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.TimeSpanSecondsConverter", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.TimeSpanSecondsOrInfiniteConverter", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.TimeSpanValidator", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.TimeSpanValidatorAttribute", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.TypeNameConverter", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.UriSection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.UserScopedSettingAttribute", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.UserSettingsGroup", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ValidatorCallback", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.WhiteSpaceTrimStringConverter", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingsProperty", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingsPropertyValue", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.SettingsSerializeAs", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.NoSettingsVersionUpgradeAttribute", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.Provider.ProviderBase", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.Provider.ProviderCollection", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.Provider.ProviderException", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.Internal.DelegatingConfigHost", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.Internal.IConfigErrorInfo", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.Internal.IConfigSystem", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.Internal.IConfigurationManagerHelper", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.Internal.IConfigurationManagerInternal", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.Internal.IInternalConfigClientHost", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.Internal.IInternalConfigConfigurationFactory", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.Internal.IInternalConfigHost", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.Internal.IInternalConfigRecord", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.Internal.IInternalConfigRoot", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.Internal.IInternalConfigSettingsFactory", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.Internal.IInternalConfigSystem", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.Internal.InternalConfigEventArgs", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.Internal.InternalConfigEventHandler", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.Internal.StreamChangeCallback", "System.Configuration.ConfigurationManager, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Odbc.ODBC32", "System.Data.Odbc, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Odbc.OdbcCommand", "System.Data.Odbc, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Odbc.OdbcCommandBuilder", "System.Data.Odbc, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Odbc.OdbcConnection", "System.Data.Odbc, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Odbc.OdbcConnectionStringBuilder", "System.Data.Odbc, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Odbc.OdbcDataAdapter", "System.Data.Odbc, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Odbc.OdbcDataReader", "System.Data.Odbc, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Odbc.OdbcError", "System.Data.Odbc, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Odbc.OdbcErrorCollection", "System.Data.Odbc, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Odbc.OdbcException", "System.Data.Odbc, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Odbc.OdbcFactory", "System.Data.Odbc, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Odbc.OdbcInfoMessageEventArgs", "System.Data.Odbc, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Odbc.OdbcInfoMessageEventHandler", "System.Data.Odbc, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Odbc.OdbcMetaDataCollectionNames", "System.Data.Odbc, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Odbc.OdbcMetaDataColumnNames", "System.Data.Odbc, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Odbc.OdbcParameter", "System.Data.Odbc, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Odbc.OdbcParameterCollection", "System.Data.Odbc, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Odbc.OdbcRowUpdatedEventArgs", "System.Data.Odbc, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Odbc.OdbcRowUpdatedEventHandler", "System.Data.Odbc, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Odbc.OdbcRowUpdatingEventArgs", "System.Data.Odbc, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Odbc.OdbcRowUpdatingEventHandler", "System.Data.Odbc, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Odbc.OdbcTransaction", "System.Data.Odbc, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Odbc.OdbcType", "System.Data.Odbc, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Odbc.ODBC32+RETCODE", "System.Data.Odbc, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OleDb.OleDbCommand", "System.Data.OleDb, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OleDb.OleDbCommandBuilder", "System.Data.OleDb, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OleDb.OleDbConnection", "System.Data.OleDb, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OleDb.OleDbConnectionStringBuilder", "System.Data.OleDb, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OleDb.OleDbDataAdapter", "System.Data.OleDb, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OleDb.OleDbDataReader", "System.Data.OleDb, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OleDb.OleDbEnumerator", "System.Data.OleDb, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OleDb.OleDbError", "System.Data.OleDb, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OleDb.OleDbErrorCollection", "System.Data.OleDb, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OleDb.OleDbException", "System.Data.OleDb, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OleDb.OleDbFactory", "System.Data.OleDb, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OleDb.OleDbInfoMessageEventArgs", "System.Data.OleDb, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OleDb.OleDbInfoMessageEventHandler", "System.Data.OleDb, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OleDb.OleDbLiteral", "System.Data.OleDb, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OleDb.OleDbMetaDataCollectionNames", "System.Data.OleDb, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OleDb.OleDbMetaDataColumnNames", "System.Data.OleDb, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OleDb.OleDbParameter", "System.Data.OleDb, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OleDb.OleDbParameterCollection", "System.Data.OleDb, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OleDb.OleDbRowUpdatedEventArgs", "System.Data.OleDb, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OleDb.OleDbRowUpdatedEventHandler", "System.Data.OleDb, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OleDb.OleDbRowUpdatingEventArgs", "System.Data.OleDb, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OleDb.OleDbRowUpdatingEventHandler", "System.Data.OleDb, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OleDb.OleDbSchemaGuid", "System.Data.OleDb, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OleDb.OleDbTransaction", "System.Data.OleDb, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OleDb.OleDbType", "System.Data.OleDb, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OperationAbortedException", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlTypes.SqlFileStream", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.ApplicationIntent", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.OnChangeEventHandler", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.PoolBlockingPeriod", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SortOrder", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlBulkCopy", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlBulkCopyColumnMapping", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlBulkCopyColumnMappingCollection", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlBulkCopyOptions", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlClientFactory", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlClientMetaDataCollectionNames", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlCommand", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlCommandBuilder", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlConnection", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlConnectionStringBuilder", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlCredential", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlDataAdapter", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlDataReader", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlDependency", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlError", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlErrorCollection", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlException", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlInfoMessageEventArgs", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlInfoMessageEventHandler", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlNotificationEventArgs", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlNotificationInfo", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlNotificationSource", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlNotificationType", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlParameter", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlParameterCollection", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlRowsCopiedEventArgs", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlRowsCopiedEventHandler", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlRowUpdatedEventArgs", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlRowUpdatedEventHandler", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlRowUpdatingEventArgs", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlRowUpdatingEventHandler", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.SqlClient.SqlTransaction", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Data.Sql.SqlNotificationRequest", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"Microsoft.SqlServer.Server.DataAccessKind", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"Microsoft.SqlServer.Server.Format", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"Microsoft.SqlServer.Server.IBinarySerialize", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"Microsoft.SqlServer.Server.InvalidUdtException", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"Microsoft.SqlServer.Server.SqlDataRecord", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"Microsoft.SqlServer.Server.SqlFacetAttribute", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"Microsoft.SqlServer.Server.SqlFunctionAttribute", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"Microsoft.SqlServer.Server.SqlMetaData", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"Microsoft.SqlServer.Server.SqlMethodAttribute", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"Microsoft.SqlServer.Server.SqlUserDefinedAggregateAttribute", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"Microsoft.SqlServer.Server.SqlUserDefinedTypeAttribute", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"Microsoft.SqlServer.Server.SystemDataAccessKind", "System.Data.SqlClient, Version=4.6.1.6, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Diagnostics.EntryWrittenEventArgs", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.EntryWrittenEventHandler", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.EventInstance", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.EventLog", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.EventLogEntry", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.EventLogEntryCollection", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.EventLogEntryType", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.EventLogTraceListener", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.EventSourceCreationData", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.OverflowAction", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.EventBookmark", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.EventKeyword", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.EventLevel", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.EventLogConfiguration", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.EventLogException", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.EventLogInformation", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.EventLogInvalidDataException", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.EventLogIsolation", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.EventLogLink", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.EventLogMode", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.EventLogNotFoundException", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.EventLogPropertySelector", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.EventLogProviderDisabledException", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.EventLogQuery", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.EventLogReader", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.EventLogReadingException", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.EventLogRecord", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.EventLogSession", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.EventLogStatus", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.EventLogType", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.EventLogWatcher", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.EventMetadata", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.EventOpcode", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.EventProperty", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.EventRecord", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.EventRecordWrittenEventArgs", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.EventTask", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.PathType", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.ProviderMetadata", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.SessionAuthentication", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.StandardEventKeywords", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.StandardEventLevel", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.StandardEventOpcode", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.Eventing.Reader.StandardEventTask", "System.Diagnostics.EventLog, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.CounterCreationData", "System.Diagnostics.PerformanceCounter, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.CounterCreationDataCollection", "System.Diagnostics.PerformanceCounter, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.CounterSample", "System.Diagnostics.PerformanceCounter, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.CounterSampleCalculator", "System.Diagnostics.PerformanceCounter, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.ICollectData", "System.Diagnostics.PerformanceCounter, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.InstanceData", "System.Diagnostics.PerformanceCounter, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.InstanceDataCollection", "System.Diagnostics.PerformanceCounter, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.InstanceDataCollectionCollection", "System.Diagnostics.PerformanceCounter, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.PerformanceCounter", "System.Diagnostics.PerformanceCounter, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.PerformanceCounterCategory", "System.Diagnostics.PerformanceCounter, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.PerformanceCounterCategoryType", "System.Diagnostics.PerformanceCounter, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.PerformanceCounterInstanceLifetime", "System.Diagnostics.PerformanceCounter, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.PerformanceCounterManager", "System.Diagnostics.PerformanceCounter, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.PerformanceCounterType", "System.Diagnostics.PerformanceCounter, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.PerformanceData.CounterData", "System.Diagnostics.PerformanceCounter, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.PerformanceData.CounterSet", "System.Diagnostics.PerformanceCounter, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.PerformanceData.CounterSetInstance", "System.Diagnostics.PerformanceCounter, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.PerformanceData.CounterSetInstanceCounterDataSet", "System.Diagnostics.PerformanceCounter, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.PerformanceData.CounterSetInstanceType", "System.Diagnostics.PerformanceCounter, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.PerformanceData.CounterType", "System.Diagnostics.PerformanceCounter, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.DirectoryServices.AccountManagement.AdvancedFilters", "System.DirectoryServices.AccountManagement, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.DirectoryServices.AccountManagement.AuthenticablePrincipal", "System.DirectoryServices.AccountManagement, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.DirectoryServices.AccountManagement.ComputerPrincipal", "System.DirectoryServices.AccountManagement, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.DirectoryServices.AccountManagement.ContextOptions", "System.DirectoryServices.AccountManagement, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.DirectoryServices.AccountManagement.ContextType", "System.DirectoryServices.AccountManagement, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.DirectoryServices.AccountManagement.DirectoryObjectClassAttribute", "System.DirectoryServices.AccountManagement, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.DirectoryServices.AccountManagement.DirectoryPropertyAttribute", "System.DirectoryServices.AccountManagement, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.DirectoryServices.AccountManagement.DirectoryRdnPrefixAttribute", "System.DirectoryServices.AccountManagement, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.DirectoryServices.AccountManagement.GroupPrincipal", "System.DirectoryServices.AccountManagement, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.DirectoryServices.AccountManagement.GroupScope", "System.DirectoryServices.AccountManagement, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.DirectoryServices.AccountManagement.IdentityType", "System.DirectoryServices.AccountManagement, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.DirectoryServices.AccountManagement.MatchType", "System.DirectoryServices.AccountManagement, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.DirectoryServices.AccountManagement.MultipleMatchesException", "System.DirectoryServices.AccountManagement, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.DirectoryServices.AccountManagement.NoMatchingPrincipalException", "System.DirectoryServices.AccountManagement, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.DirectoryServices.AccountManagement.PasswordException", "System.DirectoryServices.AccountManagement, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.DirectoryServices.AccountManagement.Principal", "System.DirectoryServices.AccountManagement, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.DirectoryServices.AccountManagement.PrincipalCollection", "System.DirectoryServices.AccountManagement, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.DirectoryServices.AccountManagement.PrincipalContext", "System.DirectoryServices.AccountManagement, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.DirectoryServices.AccountManagement.PrincipalException", "System.DirectoryServices.AccountManagement, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.DirectoryServices.AccountManagement.PrincipalExistsException", "System.DirectoryServices.AccountManagement, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.DirectoryServices.AccountManagement.PrincipalOperationException", "System.DirectoryServices.AccountManagement, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.DirectoryServices.AccountManagement.PrincipalSearcher", "System.DirectoryServices.AccountManagement, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.DirectoryServices.AccountManagement.PrincipalSearchResult`1", "System.DirectoryServices.AccountManagement, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.DirectoryServices.AccountManagement.PrincipalServerDownException", "System.DirectoryServices.AccountManagement, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.DirectoryServices.AccountManagement.PrincipalValueCollection`1", "System.DirectoryServices.AccountManagement, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.DirectoryServices.AccountManagement.UserPrincipal", "System.DirectoryServices.AccountManagement, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.DirectoryServices.ActiveDirectoryAccessRule", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectoryAuditRule", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectoryRights", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectorySecurity", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectorySecurityInheritance", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.AuthenticationTypes", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.CreateChildAccessRule", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.DeleteChildAccessRule", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.DeleteTreeAccessRule", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.DereferenceAlias", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.DirectoryEntries", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.DirectoryEntry", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.DirectoryEntryConfiguration", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.DirectorySearcher", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.DirectoryServicesCOMException", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.DirectorySynchronization", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.DirectorySynchronizationOptions", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.DirectoryVirtualListView", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.DirectoryVirtualListViewContext", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ExtendedDN", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ExtendedRightAccessRule", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ListChildrenAccessRule", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.PasswordEncodingMethod", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.PropertyAccess", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.PropertyAccessRule", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.PropertyCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.PropertySetAccessRule", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.PropertyValueCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ReferralChasingOption", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ResultPropertyCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ResultPropertyValueCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.SchemaNameCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.SearchResult", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.SearchResultCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.SearchScope", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.SecurityMasks", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.SortDirection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.SortOption", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ActiveDirectoryInterSiteTransport", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ActiveDirectoryObjectExistsException", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ActiveDirectoryObjectNotFoundException", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ActiveDirectoryOperationException", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ActiveDirectoryPartition", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ActiveDirectoryReplicationMetadata", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ActiveDirectoryRole", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ActiveDirectoryRoleCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ActiveDirectorySchedule", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ActiveDirectorySchema", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ActiveDirectorySchemaClass", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ActiveDirectorySchemaClassCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ActiveDirectorySchemaProperty", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ActiveDirectorySchemaPropertyCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ActiveDirectoryServerDownException", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ActiveDirectorySite", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ActiveDirectorySiteCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ActiveDirectorySiteLink", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ActiveDirectorySiteLinkBridge", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ActiveDirectorySiteLinkCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ActiveDirectorySiteOptions", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ActiveDirectorySubnet", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ActiveDirectorySubnetCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ActiveDirectorySyntax", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ActiveDirectoryTransportType", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.AdamInstance", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.AdamInstanceCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.AdamRole", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.AdamRoleCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ApplicationPartition", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ApplicationPartitionCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.AttributeMetadata", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.AttributeMetadataCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ConfigurationSet", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.DirectoryContext", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.DirectoryContextType", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.DirectoryServer", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.DirectoryServerCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.Domain", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.DomainCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.DomainCollisionOptions", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.DomainController", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.DomainControllerCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.DomainMode", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.Forest", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ForestMode", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ForestTrustCollisionException", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ForestTrustCollisionType", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ForestTrustDomainInfoCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ForestTrustDomainInformation", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ForestTrustDomainStatus", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ForestTrustRelationshipCollision", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ForestTrustRelationshipCollisionCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ForestTrustRelationshipInformation", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.GlobalCatalog", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.GlobalCatalogCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.HourOfDay", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.LocatorOptions", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.MinuteOfHour", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.NotificationStatus", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.PropertyTypes", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ReadOnlyActiveDirectorySchemaClassCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ReadOnlyActiveDirectorySchemaPropertyCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ReadOnlyDirectoryServerCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ReadOnlySiteCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ReadOnlySiteLinkBridgeCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ReadOnlySiteLinkCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ReadOnlyStringCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ReplicationConnection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ReplicationConnectionCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ReplicationCursor", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ReplicationCursorCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ReplicationFailure", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ReplicationFailureCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ReplicationNeighbor", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ReplicationNeighborCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ReplicationOperation", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ReplicationOperationCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ReplicationOperationInformation", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ReplicationOperationType", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ReplicationSecurityLevel", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ReplicationSpan", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.SchemaClassType", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.SyncFromAllServersErrorCategory", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.SyncFromAllServersErrorInformation", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.SyncFromAllServersEvent", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.SyncFromAllServersOperationException", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.SyncFromAllServersOptions", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.SyncUpdateCallback", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.TopLevelName", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.TopLevelNameCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.TopLevelNameCollisionOptions", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.TopLevelNameStatus", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.TrustDirection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.TrustRelationshipInformation", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.TrustRelationshipInformationCollection", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.TrustType", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.ActiveDirectory.ReplicationNeighbor+ReplicationNeighborOptions", "System.DirectoryServices, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.AddRequest", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.AddResponse", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.AsqRequestControl", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.AsqResponseControl", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.AuthType", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.BerConversionException", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.BerConverter", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.CompareRequest", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.CompareResponse", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.CrossDomainMoveControl", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.DeleteRequest", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.DeleteResponse", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.DereferenceAlias", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.DereferenceConnectionCallback", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.DirectoryAttribute", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.DirectoryAttributeCollection", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.DirectoryAttributeModification", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.DirectoryAttributeModificationCollection", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.DirectoryAttributeOperation", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.DirectoryConnection", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.DirectoryControl", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.DirectoryControlCollection", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.DirectoryException", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.DirectoryIdentifier", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.DirectoryNotificationControl", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.DirectoryOperation", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.DirectoryOperationException", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.DirectoryRequest", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.DirectoryResponse", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.DirectorySynchronizationOptions", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.DirSyncRequestControl", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.DirSyncResponseControl", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.DomainScopeControl", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.DsmlAuthRequest", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.ExtendedDNControl", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.ExtendedDNFlag", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.ExtendedRequest", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.ExtendedResponse", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.LazyCommitControl", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.LdapConnection", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.LdapDirectoryIdentifier", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.LdapException", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.LdapSessionOptions", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.LocatorFlags", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.ModifyDNRequest", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.ModifyDNResponse", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.ModifyRequest", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.ModifyResponse", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.NotifyOfNewConnectionCallback", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.PageResultRequestControl", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.PageResultResponseControl", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.PartialResultProcessing", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.PartialResultsCollection", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.PermissiveModifyControl", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.QueryClientCertificateCallback", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.QueryForConnectionCallback", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.QuotaControl", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.ReferralCallback", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.ReferralChasingOptions", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.ResultCode", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.SearchOption", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.SearchOptionsControl", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.SearchRequest", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.SearchResponse", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.SearchResultAttributeCollection", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.SearchResultEntry", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.SearchResultEntryCollection", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.SearchResultReference", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.SearchResultReferenceCollection", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.SearchScope", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.SecurityDescriptorFlagControl", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.SecurityMasks", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.SecurityPackageContextConnectionInformation", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.SecurityProtocol", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.ShowDeletedControl", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.SortKey", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.SortRequestControl", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.SortResponseControl", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.TlsOperationException", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.TreeDeleteControl", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.VerifyNameControl", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.VerifyServerCertificateCallback", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.VlvRequestControl", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.DirectoryServices.Protocols.VlvResponseControl", "System.DirectoryServices.Protocols, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Drawing.Bitmap", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.BitmapSuffixInSameAssemblyAttribute", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.BitmapSuffixInSatelliteAssemblyAttribute", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Brush", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Brushes", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.BufferedGraphics", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.BufferedGraphicsContext", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.BufferedGraphicsManager", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.CharacterRange", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.ContentAlignment", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.CopyPixelOperation", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Font", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.FontConverter", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.FontFamily", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.FontStyle", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Graphics", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.GraphicsUnit", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Icon", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.IconConverter", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.IDeviceContext", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Image", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.ImageAnimator", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.ImageConverter", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.ImageFormatConverter", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Pen", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Pens", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Region", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.RotateFlipType", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.SolidBrush", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.StockIconId", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.StockIconOptions", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.StringAlignment", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.StringDigitSubstitute", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.StringFormat", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.StringFormatFlags", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.StringTrimming", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.StringUnit", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.SystemBrushes", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.SystemFonts", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.SystemIcons", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.SystemPens", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.TextureBrush", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.ToolboxBitmapAttribute", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Text.FontCollection", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Text.GenericFontFamilies", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Text.HotkeyPrefix", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Text.InstalledFontCollection", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Text.PrivateFontCollection", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Text.TextRenderingHint", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.StandardPrintController", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.Duplex", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.InvalidPrinterException", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.Margins", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.MarginsConverter", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.PageSettings", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.PaperKind", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.PaperSize", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.PaperSource", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.PaperSourceKind", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.PreviewPageInfo", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.PreviewPrintController", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.PrintAction", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.PrintController", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.PrintDocument", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.PrinterResolution", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.PrinterResolutionKind", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.PrinterSettings", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.PrinterUnit", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.PrinterUnitConvert", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.PrintEventArgs", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.PrintEventHandler", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.PrintPageEventArgs", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.PrintPageEventHandler", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.PrintRange", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.QueryPageSettingsEventArgs", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.QueryPageSettingsEventHandler", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Interop.LOGFONT", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.BitmapData", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.CachedBitmap", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.ColorAdjustType", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.ColorChannelFlag", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.ColorMap", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.ColorMapType", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.ColorMatrix", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.ColorMatrixFlag", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.ColorMode", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.ColorPalette", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.DitherType", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.EmfPlusRecordType", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.EmfType", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.Encoder", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.EncoderParameter", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.EncoderParameters", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.EncoderParameterValueType", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.EncoderValue", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.FrameDimension", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.ImageAttributes", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.ImageCodecFlags", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.ImageCodecInfo", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.ImageFlags", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.ImageFormat", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.ImageLockMode", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.Metafile", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.MetafileFrameUnit", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.MetafileHeader", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.MetafileType", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.MetaHeader", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.PaletteFlags", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.PaletteType", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.PixelFormat", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.PlayRecordCallback", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.PropertyItem", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.WmfPlaceableFileHeader", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.Effects.BlackSaturationCurveEffect", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.Effects.BlurEffect", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.Effects.BrightnessContrastEffect", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.Effects.ColorBalanceEffect", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.Effects.ColorCurveEffect", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.Effects.ColorLookupTableEffect", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.Effects.ColorMatrixEffect", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.Effects.ContrastCurveEffect", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.Effects.CurveChannel", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.Effects.DensityCurveEffect", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.Effects.Effect", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.Effects.ExposureCurveEffect", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.Effects.GrayScaleEffect", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.Effects.HighlightCurveEffect", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.Effects.InvertEffect", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.Effects.LevelsEffect", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.Effects.MidtoneCurveEffect", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.Effects.SepiaEffect", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.Effects.ShadowCurveEffect", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.Effects.SharpenEffect", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.Effects.TintEffect", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.Effects.VividEffect", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Imaging.Effects.WhiteSaturationCurveEffect", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Design.CategoryNameCollection", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.DashCap", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.AdjustableArrowCap", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.Blend", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.ColorBlend", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.CombineMode", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.CompositingMode", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.CompositingQuality", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.CoordinateSpace", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.CustomLineCap", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.DashStyle", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.FillMode", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.FlushIntention", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.GraphicsContainer", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.GraphicsPath", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.GraphicsPathIterator", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.GraphicsState", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.HatchBrush", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.HatchStyle", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.InterpolationMode", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.LinearGradientBrush", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.LinearGradientMode", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.LineCap", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.LineJoin", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.Matrix", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.MatrixOrder", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.PathData", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.PathGradientBrush", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.PathPointType", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.PenAlignment", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.PenType", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.PixelOffsetMode", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.QualityMode", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.RegionData", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.SmoothingMode", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.WarpMode", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Drawing2D.WrapMode", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"Windows.Win32.__byte_256+<Value>e__FixedBuffer", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"Windows.Win32.UI.Shell.SHSTOCKICONINFO+<szPath>e__FixedBuffer", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.FontConverter+FontNameConverter", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.FontConverter+FontUnitConverter", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Graphics+DrawImageAbort", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Graphics+EnumerateMetafileProc", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Image+GetThumbnailImageAbort", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.PrinterSettings+PaperSizeCollection", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.PrinterSettings+PaperSourceCollection", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.PrinterSettings+PrinterResolutionCollection", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.PrinterSettings+StringCollection", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Interop.LOGFONT+<_lfFaceName>e__FixedBuffer", "System.Drawing.Common, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.IO.FileFormatException", "System.IO.Packaging, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IO.Packaging.CompressionOption", "System.IO.Packaging, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IO.Packaging.EncryptionOption", "System.IO.Packaging, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IO.Packaging.Package", "System.IO.Packaging, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IO.Packaging.PackagePart", "System.IO.Packaging, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IO.Packaging.PackagePartCollection", "System.IO.Packaging, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IO.Packaging.PackageProperties", "System.IO.Packaging, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IO.Packaging.PackageRelationship", "System.IO.Packaging, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IO.Packaging.PackageRelationshipCollection", "System.IO.Packaging, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IO.Packaging.PackageRelationshipSelector", "System.IO.Packaging, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IO.Packaging.PackageRelationshipSelectorType", "System.IO.Packaging, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IO.Packaging.PackUriHelper", "System.IO.Packaging, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IO.Packaging.TargetMode", "System.IO.Packaging, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IO.Packaging.ZipPackage", "System.IO.Packaging, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IO.Packaging.ZipPackagePart", "System.IO.Packaging, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IO.Ports.Handshake", "System.IO.Ports, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.IO.Ports.Parity", "System.IO.Ports, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.IO.Ports.SerialData", "System.IO.Ports, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.IO.Ports.SerialDataReceivedEventArgs", "System.IO.Ports, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.IO.Ports.SerialDataReceivedEventHandler", "System.IO.Ports, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.IO.Ports.SerialError", "System.IO.Ports, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.IO.Ports.SerialErrorReceivedEventArgs", "System.IO.Ports, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.IO.Ports.SerialErrorReceivedEventHandler", "System.IO.Ports, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.IO.Ports.SerialPinChange", "System.IO.Ports, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.IO.Ports.SerialPinChangedEventArgs", "System.IO.Ports, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.IO.Ports.SerialPinChangedEventHandler", "System.IO.Ports, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.IO.Ports.SerialPort", "System.IO.Ports, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.IO.Ports.StopBits", "System.IO.Ports, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Management.AuthenticationLevel", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.CimType", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.CodeLanguage", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ComparisonSettings", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.CompletedEventArgs", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.CompletedEventHandler", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ConnectionOptions", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.DeleteOptions", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.EnumerationOptions", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.EventArrivedEventArgs", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.EventArrivedEventHandler", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.EventQuery", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.EventWatcherOptions", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ImpersonationLevel", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.InvokeMethodOptions", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ManagementBaseObject", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ManagementClass", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ManagementDateTimeConverter", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ManagementEventArgs", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ManagementEventWatcher", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ManagementException", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ManagementNamedValueCollection", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ManagementObject", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ManagementObjectCollection", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ManagementObjectSearcher", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ManagementOperationObserver", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ManagementOptions", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ManagementPath", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ManagementQuery", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ManagementScope", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ManagementStatus", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.MethodData", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.MethodDataCollection", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ObjectGetOptions", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ObjectPutEventArgs", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ObjectPutEventHandler", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ObjectQuery", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ObjectReadyEventArgs", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ObjectReadyEventHandler", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ProgressEventArgs", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ProgressEventHandler", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.PropertyData", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.PropertyDataCollection", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.PutOptions", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.PutType", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.QualifierData", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.QualifierDataCollection", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.RelatedObjectQuery", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.RelationshipQuery", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.SelectQuery", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.StoppedEventArgs", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.StoppedEventHandler", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.TextFormat", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.WqlEventQuery", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.WqlObjectQuery", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.ManagementObjectCollection+ManagementObjectEnumerator", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.MethodDataCollection+MethodDataEnumerator", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.PropertyDataCollection+PropertyDataEnumerator", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Management.QualifierDataCollection+QualifierDataEnumerator", "System.Management, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Net.Http.CookieUsePolicy", "System.Net.Http.WinHttpHandler, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Net.Http.WindowsProxyUsePolicy", "System.Net.Http.WinHttpHandler, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Net.Http.WinHttpHandler", "System.Net.Http.WinHttpHandler, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"Windows.Win32.PInvokeCore+EnumChildWindowsCallback", "System.Private.Windows.Core, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"Windows.Win32.PInvokeCore+EnumDisplayMonitorsCallback", "System.Private.Windows.Core, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"Windows.Win32.PInvokeCore+EnumThreadWindowsCallback", "System.Private.Windows.Core, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"Windows.Win32.PInvokeCore+EnumWindowsCallback", "System.Private.Windows.Core, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"Windows.Win32.__byte_32+<Value>e__FixedBuffer", "System.Private.Windows.Core, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"Windows.Win32.__char_32+<Value>e__FixedBuffer", "System.Private.Windows.Core, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"Windows.Win32.__char_260+<Value>e__FixedBuffer", "System.Private.Windows.Core, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"Windows.Win32.System.Com.IComCallableWrapper+Vtbl", "System.Private.Windows.Core, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"Windows.Win32.System.Com.IComCallableWrapper+Interface", "System.Private.Windows.Core, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"Windows.Win32.System.Com.IDispatch+Interface", "System.Private.Windows.Core, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"Windows.Win32.Graphics.Gdi.ArgbBuffer+<_stackBuffer>e__FixedBuffer", "System.Private.Windows.Core, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.DisposableBase+Finalizable", "System.Private.Windows.Core, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.Drawing.IIconExtensions", "System.Private.Windows.GdiPlus, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"Windows.Win32.__float_25+<Value>e__FixedBuffer", "System.Private.Windows.GdiPlus, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.Reflection.Context.CustomReflectionContext", "System.Reflection.Context, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b77a5c561934e089"},
                {"System.Runtime.Caching.CacheEntryChangeMonitor", "System.Runtime.Caching, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Runtime.Caching.CacheEntryRemovedArguments", "System.Runtime.Caching, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Runtime.Caching.CacheEntryRemovedCallback", "System.Runtime.Caching, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Runtime.Caching.CacheEntryRemovedReason", "System.Runtime.Caching, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Runtime.Caching.CacheEntryUpdateArguments", "System.Runtime.Caching, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Runtime.Caching.CacheEntryUpdateCallback", "System.Runtime.Caching, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Runtime.Caching.CacheItem", "System.Runtime.Caching, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Runtime.Caching.CacheItemPolicy", "System.Runtime.Caching, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Runtime.Caching.CacheItemPriority", "System.Runtime.Caching, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Runtime.Caching.ChangeMonitor", "System.Runtime.Caching, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Runtime.Caching.DefaultCacheCapabilities", "System.Runtime.Caching, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Runtime.Caching.FileChangeMonitor", "System.Runtime.Caching, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Runtime.Caching.HostFileChangeMonitor", "System.Runtime.Caching, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Runtime.Caching.MemoryCache", "System.Runtime.Caching, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Runtime.Caching.ObjectCache", "System.Runtime.Caching, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Runtime.Caching.OnChangedCallback", "System.Runtime.Caching, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Runtime.Caching.Hosting.IApplicationIdentifier", "System.Runtime.Caching, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Runtime.Caching.Hosting.IFileChangeNotificationSystem", "System.Runtime.Caching, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Runtime.Caching.Hosting.IMemoryCacheManager", "System.Runtime.Caching, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.CryptographicAttributeObject", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.CryptographicAttributeObjectCollection", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.CryptographicAttributeObjectEnumerator", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Xml.X509IssuerSerial", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.Pkcs9AttributeObject", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.Pkcs9ContentType", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.Pkcs9DocumentDescription", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.Pkcs9DocumentName", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.Pkcs9MessageDigest", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.Pkcs9SigningTime", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.AlgorithmIdentifier", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.CmsRecipient", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.CmsRecipientCollection", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.CmsRecipientEnumerator", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.ContentInfo", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.EnvelopedCms", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.KeyAgreeRecipientInfo", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.KeyTransRecipientInfo", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.PublicKeyInfo", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.RecipientInfo", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.RecipientInfoCollection", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.RecipientInfoEnumerator", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.RecipientInfoType", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.SubjectIdentifier", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.SubjectIdentifierOrKey", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.SubjectIdentifierOrKeyType", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.SubjectIdentifierType", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.CmsSigner", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.SignedCms", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.SignerInfo", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.SignerInfoCollection", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.SignerInfoEnumerator", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.Pkcs12Builder", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.Pkcs12CertBag", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.Pkcs12ConfidentialityMode", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.Pkcs12Info", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.Pkcs12IntegrityMode", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.Pkcs12KeyBag", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.Pkcs12SafeBag", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.Pkcs12SafeContents", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.Pkcs12SafeContentsBag", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.Pkcs12SecretBag", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.Pkcs12ShroudedKeyBag", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.Pkcs8PrivateKeyInfo", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.Pkcs9LocalKeyId", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.Rfc3161TimestampRequest", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.Rfc3161TimestampToken", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Pkcs.Rfc3161TimestampTokenInfo", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.HexConverter+Casing", "System.Security.Cryptography.Pkcs, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.DataProtectionScope", "System.Security.Cryptography.ProtectedData, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.ProtectedData", "System.Security.Cryptography.ProtectedData, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Security.Cryptography.Xml.CipherData", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.CipherReference", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.CryptoSignedXmlRecursionException", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.DataObject", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.DataReference", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.DSAKeyValue", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.EncryptedData", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.EncryptedKey", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.EncryptedReference", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.EncryptedType", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.EncryptedXml", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.EncryptionMethod", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.EncryptionProperty", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.EncryptionPropertyCollection", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.IRelDecryptor", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.KeyInfo", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.KeyInfoClause", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.KeyInfoEncryptedKey", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.KeyInfoName", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.KeyInfoNode", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.KeyInfoRetrievalMethod", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.KeyInfoX509Data", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.KeyReference", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.Reference", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.ReferenceList", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.RSAKeyValue", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.Signature", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.SignedInfo", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.SignedXml", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.Transform", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.TransformChain", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.XmlDecryptionTransform", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.XmlDsigBase64Transform", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.XmlDsigC14NTransform", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.XmlDsigC14NWithCommentsTransform", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.XmlDsigEnvelopedSignatureTransform", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.XmlDsigExcC14NTransform", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.XmlDsigExcC14NWithCommentsTransform", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.XmlDsigXPathTransform", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.XmlDsigXsltTransform", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.Xml.XmlLicenseTransform", "System.Security.Cryptography.Xml, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ApplicationIdentity", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Xaml.Permissions.XamlLoadPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Web.AspNetHostingPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Web.AspNetHostingPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Web.AspNetHostingPermissionLevel", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Transactions.DistributedTransactionPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Transactions.DistributedTransactionPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceProcess.ServiceControllerPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceProcess.ServiceControllerPermissionAccess", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceProcess.ServiceControllerPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceProcess.ServiceControllerPermissionEntry", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceProcess.ServiceControllerPermissionEntryCollection", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.CodeAccessPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.HostProtectionException", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.HostSecurityManager", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.HostSecurityManagerOptions", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.IEvidenceFactory", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.ISecurityPolicyEncodable", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.NamedPermissionSet", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.PolicyLevelType", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.SecurityContext", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.SecurityContextSource", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.SecurityManager", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.SecurityState", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.SecurityZone", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.XmlSyntaxException", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.AllMembershipCondition", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.ApplicationDirectory", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.ApplicationDirectoryMembershipCondition", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.ApplicationTrust", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.ApplicationTrustCollection", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.ApplicationTrustEnumerator", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.ApplicationVersionMatch", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.CodeConnectAccess", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.CodeGroup", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.FileCodeGroup", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.FirstMatchCodeGroup", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.GacInstalled", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.GacMembershipCondition", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.Hash", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.HashMembershipCondition", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.IIdentityPermissionFactory", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.IMembershipCondition", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.NetCodeGroup", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.PermissionRequestEvidence", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.PolicyException", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.PolicyLevel", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.PolicyStatement", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.PolicyStatementAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.Publisher", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.PublisherMembershipCondition", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.Site", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.SiteMembershipCondition", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.StrongName", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.StrongNameMembershipCondition", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.TrustManagerContext", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.TrustManagerUIContext", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.UnionCodeGroup", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.Url", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.UrlMembershipCondition", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.Zone", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Policy.ZoneMembershipCondition", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.DataProtectionPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.DataProtectionPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.DataProtectionPermissionFlags", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.EnvironmentPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.EnvironmentPermissionAccess", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.EnvironmentPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.FileDialogPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.FileDialogPermissionAccess", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.FileDialogPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.FileIOPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.FileIOPermissionAccess", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.FileIOPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.GacIdentityPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.GacIdentityPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.HostProtectionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.HostProtectionResource", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.IsolatedStorageContainment", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.IsolatedStorageFilePermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.IsolatedStorageFilePermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.IsolatedStoragePermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.IsolatedStoragePermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.IUnrestrictedPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.KeyContainerPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.KeyContainerPermissionAccessEntry", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.KeyContainerPermissionAccessEntryCollection", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.KeyContainerPermissionAccessEntryEnumerator", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.KeyContainerPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.KeyContainerPermissionFlags", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.MediaPermissionAudio", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.MediaPermissionVideo", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.MediaPermissionImage", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.MediaPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.MediaPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.PermissionSetAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.PrincipalPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.PrincipalPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.PublisherIdentityPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.PublisherIdentityPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.ReflectionPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.ReflectionPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.ReflectionPermissionFlag", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.RegistryPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.RegistryPermissionAccess", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.RegistryPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.ResourcePermissionBase", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.ResourcePermissionBaseEntry", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.SecurityPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.SiteIdentityPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.SiteIdentityPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.StorePermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.StorePermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.StorePermissionFlags", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.StrongNameIdentityPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.StrongNameIdentityPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.StrongNamePublicKeyBlob", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.TypeDescriptorPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.TypeDescriptorPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.TypeDescriptorPermissionFlags", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.UIPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.UIPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.UIPermissionClipboard", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.UIPermissionWindow", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.UrlIdentityPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.UrlIdentityPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.WebBrowserPermissionLevel", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.WebBrowserPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.WebBrowserPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.ZoneIdentityPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Permissions.ZoneIdentityPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Net.DnsPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Net.DnsPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Net.EndpointPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Net.NetworkAccess", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Net.SocketPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Net.SocketPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Net.TransportType", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Net.WebPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Net.WebPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Net.PeerToPeer.PnrpPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Net.PeerToPeer.PnrpPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Net.PeerToPeer.PnrpScope", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Net.PeerToPeer.Collaboration.PeerCollaborationPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Net.PeerToPeer.Collaboration.PeerCollaborationPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Net.NetworkInformation.NetworkInformationAccess", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Net.NetworkInformation.NetworkInformationPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Net.NetworkInformation.NetworkInformationPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Net.Mail.SmtpAccess", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Net.Mail.SmtpPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Net.Mail.SmtpPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.PrintingPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.PrintingPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Drawing.Printing.PrintingPermissionLevel", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.DirectoryServices.DirectoryServicesPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.DirectoryServices.DirectoryServicesPermissionAccess", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.DirectoryServices.DirectoryServicesPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.DirectoryServices.DirectoryServicesPermissionEntry", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.DirectoryServices.DirectoryServicesPermissionEntryCollection", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.EventLogPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.EventLogPermissionAccess", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.EventLogPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.EventLogPermissionEntry", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.EventLogPermissionEntryCollection", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.PerformanceCounterPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.PerformanceCounterPermissionAccess", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.PerformanceCounterPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.PerformanceCounterPermissionEntry", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Diagnostics.PerformanceCounterPermissionEntryCollection", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.SqlClient.SqlClientPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.SqlClient.SqlClientPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OracleClient.OraclePermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OracleClient.OraclePermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OleDb.OleDbPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.OleDb.OleDbPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Odbc.OdbcPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Odbc.OdbcPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Common.DBDataPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Data.Common.DBDataPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationPermission", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Configuration.ConfigurationPermissionAttribute", "System.Security.Permissions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.BasicHttpBinding", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.BasicHttpMessageCredentialType", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.BasicHttpMessageSecurity", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.BasicHttpsBinding", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.BasicHttpSecurity", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.BasicHttpSecurityMode", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.BasicHttpsSecurity", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.BasicHttpsSecurityMode", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.HttpBindingBase", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.HttpClientCredentialType", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.HttpProxyCredentialType", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.HttpTransportSecurity", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.MessageSecurityOverHttp", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.NetHttpBinding", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.NetHttpMessageEncoding", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.NetHttpsBinding", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.NonDualMessageSecurityOverHttp", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.WS2007HttpBinding", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.WSHttpBinding", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.WSHttpBindingBase", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.WSHttpSecurity", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.WSMessageEncoding", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.HttpRequestMessageProperty", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.HttpResponseMessageProperty", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.HttpsTransportBindingElement", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.HttpTransportBindingElement", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.IHttpCookieContainerManager", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.WebSocketTransportSettings", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.WebSocketTransportUsage", "System.ServiceModel.Http, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.ConnectionOrientedTransportBindingElement", "System.ServiceModel.NetFramingBase, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.IConnection", "System.ServiceModel.NetFramingBase, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.IConnectionInitiator", "System.ServiceModel.NetFramingBase, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.IConnectionPoolSettings", "System.ServiceModel.NetFramingBase, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.NetFramingTransportChannelFactory`1", "System.ServiceModel.NetFramingBase, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.SslStreamSecurityBindingElement", "System.ServiceModel.NetFramingBase, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.StreamSecurityUpgradeInitiator", "System.ServiceModel.NetFramingBase, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.StreamSecurityUpgradeProvider", "System.ServiceModel.NetFramingBase, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.StreamUpgradeBindingElement", "System.ServiceModel.NetFramingBase, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.StreamUpgradeInitiator", "System.ServiceModel.NetFramingBase, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.StreamUpgradeProvider", "System.ServiceModel.NetFramingBase, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.WindowsStreamSecurityBindingElement", "System.ServiceModel.NetFramingBase, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.MessageSecurityOverTcp", "System.ServiceModel.NetTcp, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.NetTcpBinding", "System.ServiceModel.NetTcp, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.NetTcpSecurity", "System.ServiceModel.NetTcp, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.TcpClientCredentialType", "System.ServiceModel.NetTcp, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.TcpTransportSecurity", "System.ServiceModel.NetTcp, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.TcpConnectionPoolSettings", "System.ServiceModel.NetTcp, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.TcpTransportBindingElement", "System.ServiceModel.NetTcp, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.ActionNotSupportedException", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.CacheSetting", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.ChannelFactory", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.ChannelFactory`1", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.ChannelTerminatedException", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.ClientBase`1", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.ClientCredentialsSecurityTokenManager", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.CommunicationException", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.CommunicationObjectAbortedException", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.CommunicationObjectFaultedException", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.CommunicationState", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.DataContractFormatAttribute", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.DeliveryRequirementsAttribute", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.EndpointAddress", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.EndpointAddressBuilder", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.EndpointIdentity", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.EndpointNotFoundException", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.EnvelopeVersion", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.ExceptionDetail", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.ExtensionCollection`1", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.FaultCode", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.FaultContractAttribute", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.FaultException", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.FaultException`1", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.FaultReason", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.FaultReasonText", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.HostNameComparisonMode", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.IClientChannel", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.ICommunicationObject", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.IContextChannel", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.IDefaultCommunicationTimeouts", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.IExtensibleObject`1", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.IExtension`1", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.IExtensionCollection`1", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.InvalidMessageContractException", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.MessageHeaderAttribute", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.MessageHeaderArrayAttribute", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.MessageBodyMemberAttribute", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.MessageContractAttribute", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.MessageContractMemberAttribute", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.MessagePropertyAttribute", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.MessageCredentialType", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.MessageHeader`1", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.MessageHeaderException", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.MessageParameterAttribute", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.OperationContext", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.OperationContextScope", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.OperationContractAttribute", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.OperationFormatStyle", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.OperationFormatUse", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.OptionalReliableSession", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.ProtocolException", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.QueuedDeliveryRequirementsMode", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.QuotaExceededException", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.ReliableMessagingVersion", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.ReliableSession", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.SecurityMode", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.ServerTooBusyException", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.ServiceActivationException", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.ServiceContractAttribute", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.SessionMode", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.ServiceKnownTypeAttribute", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.TransferMode", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.UnknownMessageReceivedEventArgs", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.XmlSerializerFormatAttribute", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.InstanceContext", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.DnsEndpointIdentity", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.MessageSecurityVersion", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.ServiceSecurityContext", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.SpnEndpointIdentity", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.UpnEndpointIdentity", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.X509CertificateEndpointIdentity", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.CallbackBehaviorAttribute", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.ConcurrencyMode", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.DuplexChannelFactory`1", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.DuplexClientBase`1", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.HttpDigestClientCredential", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.ISecuritySession", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.MessageSecurityException", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.SecurityAccessDeniedException", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.SecurityCredentialsManager", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.SecurityNegotiationException", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.UserNamePasswordClientCredential", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.WindowsClientCredential", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.X509ServiceCertificateAuthentication", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.X509CertificateInitiatorClientCredential", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.X509CertificateRecipientClientCredential", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.X509CertificateValidationMode", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.SecurityKeyEntropyMode", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.BasicSecurityProfileVersion", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.IdentityVerifier", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.SecurityMessageProperty", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.SecureConversationVersion", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.SecurityAlgorithmSuite", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.SecurityTokenAttachmentMode", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.SecurityTokenSpecification", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.SecurityPolicyVersion", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.SecurityVersion", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.SupportingTokenSpecification", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.TrustVersion", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.Tokens.BinarySecretSecurityToken", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.Tokens.InitiatorServiceModelSecurityTokenRequirement", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.Tokens.IssuedSecurityTokenParameters", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.Tokens.SecureConversationSecurityTokenParameters", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.Tokens.SecurityTokenParameters", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.Tokens.ServiceModelSecurityTokenRequirement", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.Tokens.SspiSecurityToken", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.Tokens.SupportingTokenParameters", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Security.Tokens.UserNameSecurityTokenParameters", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Dispatcher.ClientOperation", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Dispatcher.ClientRuntime", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Dispatcher.DispatchOperation", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Dispatcher.DispatchRuntime", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Dispatcher.ChannelDispatcher", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Dispatcher.EndpointDispatcher", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Dispatcher.FaultContractInfo", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Dispatcher.IChannelInitializer", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Dispatcher.IClientMessageFormatter", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Dispatcher.IClientMessageInspector", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Dispatcher.IDispatchMessageInspector", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Dispatcher.IClientOperationSelector", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Dispatcher.IInteractiveChannelInitializer", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Dispatcher.IParameterInspector", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Dispatcher.IErrorHandler", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Description.CallbackDebugBehavior", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Description.ClientCredentials", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Description.ContractDescription", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Description.DataContractSerializerOperationBehavior", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Description.FaultDescription", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Description.FaultDescriptionCollection", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Description.IContractBehavior", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Description.IEndpointBehavior", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Description.IOperationBehavior", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Description.MessageBodyDescription", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Description.MessageDescription", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Description.MessageDescriptionCollection", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Description.MessageDirection", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Description.MessageHeaderDescription", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Description.MessageHeaderDescriptionCollection", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Description.MessagePartDescription", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Description.MessagePartDescriptionCollection", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Description.MessagePropertyDescription", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Description.MessagePropertyDescriptionCollection", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Description.OperationDescription", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Description.OperationDescriptionCollection", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Description.ServiceEndpoint", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Description.XmlSerializerOperationBehavior", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Description.TypedMessageConverter", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.AddressHeader", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.AddressHeaderCollection", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.AddressingVersion", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.BinaryMessageEncodingBindingElement", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.Binding", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.BindingContext", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.BindingElement", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.BindingElementCollection", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.BindingParameterCollection", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.BodyWriter", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.BufferManager", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.ChannelBase", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.ChannelFactoryBase", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.ChannelFactoryBase`1", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.ChannelManagerBase", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.ChannelParameterCollection", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.CommunicationObject", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.CompressionFormat", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.CustomBinding", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.FaultConverter", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.IChannel", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.IChannelFactory", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.IChannelFactory`1", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.IBindingDeliveryCapabilities", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.IDuplexChannel", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.IDuplexSession", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.IDuplexSessionChannel", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.IInputChannel", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.IInputSession", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.IInputSessionChannel", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.IMessageProperty", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.IOutputChannel", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.IOutputSession", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.IOutputSessionChannel", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.IRequestChannel", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.IRequestSessionChannel", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.ISession", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.ISessionChannel`1", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.Message", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.MessageBuffer", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.MessageEncoder", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.MessageEncoderFactory", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.MessageEncodingBindingElement", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.MessageFault", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.MessageHeader", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.MessageHeaderInfo", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.MessageHeaders", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.MessageProperties", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.MessageState", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.MessageVersion", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.MtomMessageEncodingBindingElement", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.ReliableSessionBindingElement", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.RequestContext", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.TextMessageEncodingBindingElement", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.TransportBindingElement", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.UnderstoodHeaders", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.ISecurityCapabilities", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.LocalClientSecuritySettings", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.SecurityBindingElement", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.SecurityHeaderLayout", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.TransportSecurityBindingElement", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Channels.ITransportCompressionSupport", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Policy.IAuthorizationPolicy", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Policy.IAuthorizationComponent", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Policy.AuthorizationContext", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Policy.EvaluationContext", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Claims.ClaimSet", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Claims.Claim", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Claims.ClaimTypes", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Claims.DefaultClaimSet", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Claims.X509CertificateClaimSet", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Tokens.SecurityKey", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Tokens.SecurityKeyIdentifier", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Tokens.SecurityKeyIdentifierClause", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Tokens.SecurityToken", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Tokens.SymmetricSecurityKey", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Tokens.GenericXmlSecurityToken", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Tokens.GenericXmlSecurityKeyIdentifierClause", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Tokens.SecurityKeyType", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Tokens.SecurityKeyUsage", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Tokens.SecurityTokenException", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Tokens.SecurityTokenValidationException", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Tokens.UserNameSecurityToken", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Tokens.X509SecurityToken", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Selectors.SecurityTokenAuthenticator", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Selectors.SecurityTokenManager", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Selectors.SecurityTokenProvider", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Selectors.SecurityTokenRequirement", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Selectors.SecurityTokenResolver", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Selectors.SecurityTokenSerializer", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Selectors.SecurityTokenVersion", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Selectors.X509CertificateValidator", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.IdentityModel.Selectors.X509SecurityTokenAuthenticator", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Collections.Generic.SynchronizedCollection`1", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Collections.Generic.KeyedByTypeCollection`1", "System.ServiceModel.Primitives, Version=10.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceModel.Syndication.Atom10FeedFormatter", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.Atom10FeedFormatter`1", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.Atom10ItemFormatter", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.Atom10ItemFormatter`1", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.AtomPub10CategoriesDocumentFormatter", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.AtomPub10ServiceDocumentFormatter", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.AtomPub10ServiceDocumentFormatter`1", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.CategoriesDocument", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.CategoriesDocumentFormatter", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.InlineCategoriesDocument", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.ReferencedCategoriesDocument", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.ResourceCollectionInfo", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.Rss20FeedFormatter", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.Rss20FeedFormatter`1", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.Rss20ItemFormatter", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.Rss20ItemFormatter`1", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.ServiceDocument", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.ServiceDocumentFormatter", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.SyndicationCategory", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.SyndicationContent", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.SyndicationElementExtension", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.SyndicationElementExtensionCollection", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.SyndicationFeed", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.TryParseDateTimeCallback", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.TryParseUriCallback", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.SyndicationFeedFormatter", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.SyndicationItem", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.SyndicationItemFormatter", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.SyndicationLink", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.SyndicationPerson", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.SyndicationTextInput", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.SyndicationVersions", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.TextSyndicationContent", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.TextSyndicationContentKind", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.UrlSyndicationContent", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.Workspace", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.XmlDateTimeData", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.XmlSyndicationContent", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceModel.Syndication.XmlUriData", "System.ServiceModel.Syndication, Version=10.0.0.5, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.ServiceProcess.PowerBroadcastStatus", "System.ServiceProcess.ServiceController, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceProcess.ServiceBase", "System.ServiceProcess.ServiceController, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceProcess.ServiceController", "System.ServiceProcess.ServiceController, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceProcess.ServiceControllerStatus", "System.ServiceProcess.ServiceController, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceProcess.ServiceProcessDescriptionAttribute", "System.ServiceProcess.ServiceController, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceProcess.ServiceStartMode", "System.ServiceProcess.ServiceController, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceProcess.ServiceType", "System.ServiceProcess.ServiceController, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceProcess.SessionChangeDescription", "System.ServiceProcess.ServiceController, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceProcess.SessionChangeReason", "System.ServiceProcess.ServiceController, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.ServiceProcess.TimeoutException", "System.ServiceProcess.ServiceController, Version=10.0.0.5, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Speech.Synthesis.BookmarkReachedEventArgs", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.FilePrompt", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.InstalledVoice", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.PhonemeReachedEventArgs", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.Prompt", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.PromptBreak", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.PromptBuilder", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.PromptEmphasis", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.PromptEventArgs", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.PromptRate", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.PromptStyle", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.PromptVolume", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.SayAs", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.SpeakCompletedEventArgs", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.SpeakProgressEventArgs", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.SpeakStartedEventArgs", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.SpeechSynthesizer", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.StateChangedEventArgs", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.SynthesisMediaType", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.SynthesisTextFormat", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.SynthesizerEmphasis", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.SynthesizerState", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.VisemeReachedEventArgs", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.VoiceAge", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.VoiceChangeEventArgs", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.VoiceGender", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.VoiceInfo", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.TtsEngine.ContourPoint", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.TtsEngine.ContourPointChangeType", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.TtsEngine.EmphasisBreak", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.TtsEngine.EmphasisWord", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.TtsEngine.EventParameterType", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.TtsEngine.FragmentState", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.TtsEngine.ITtsEngineSite", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.TtsEngine.Prosody", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.TtsEngine.ProsodyNumber", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.TtsEngine.ProsodyPitch", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.TtsEngine.ProsodyRange", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.TtsEngine.ProsodyRate", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.TtsEngine.ProsodyUnit", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.TtsEngine.ProsodyVolume", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.TtsEngine.SayAs", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.TtsEngine.SkipInfo", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.TtsEngine.SpeakOutputFormat", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.TtsEngine.SpeechEventInfo", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.TtsEngine.TextFragment", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.TtsEngine.TtsEngineAction", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.TtsEngine.TtsEngineSsml", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Synthesis.TtsEngine.TtsEventId", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.AudioLevelUpdatedEventArgs", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.AudioSignalProblem", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.AudioSignalProblemOccurredEventArgs", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.AudioState", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.AudioStateChangedEventArgs", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.Choices", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.DictationGrammar", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.DisplayAttributes", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.EmulateRecognizeCompletedEventArgs", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.Grammar", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.GrammarBuilder", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.LoadGrammarCompletedEventArgs", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.RecognitionEventArgs", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.RecognitionResult", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.RecognizeCompletedEventArgs", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.RecognizedAudio", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.RecognizedPhrase", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.RecognizedWordUnit", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.RecognizeMode", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.RecognizerInfo", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.RecognizerState", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.RecognizerUpdateReachedEventArgs", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.ReplacementText", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.SemanticResultKey", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.SemanticResultValue", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.SemanticValue", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.SpeechDetectedEventArgs", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.SpeechHypothesizedEventArgs", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.SpeechRecognitionEngine", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.SpeechRecognitionRejectedEventArgs", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.SpeechRecognizedEventArgs", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.SpeechRecognizer", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.SpeechUI", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.StateChangedEventArgs", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.SubsetMatchingMode", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.SrgsGrammar.SrgsDocument", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.SrgsGrammar.SrgsElement", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.SrgsGrammar.SrgsGrammarCompiler", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.SrgsGrammar.SrgsGrammarMode", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.SrgsGrammar.SrgsItem", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.SrgsGrammar.SrgsNameValueTag", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.SrgsGrammar.SrgsOneOf", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.SrgsGrammar.SrgsPhoneticAlphabet", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.SrgsGrammar.SrgsRule", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.SrgsGrammar.SrgsRuleRef", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.SrgsGrammar.SrgsRulesCollection", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.SrgsGrammar.SrgsRuleScope", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.SrgsGrammar.SrgsSemanticInterpretationTag", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.SrgsGrammar.SrgsSubset", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.SrgsGrammar.SrgsText", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.Recognition.SrgsGrammar.SrgsToken", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.AudioFormat.AudioBitsPerSample", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.AudioFormat.AudioChannel", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.AudioFormat.EncodingFormat", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Speech.AudioFormat.SpeechAudioFormatInfo", "System.Speech, Version=10.0.0.5, Culture=neutral, PublicKeyToken=31bf3856ad364e35"},
                {"System.Web.Services.WsiProfiles", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.Binding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.BindingCollection", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.DocumentableItem", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.FaultBinding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.FaultBindingCollection", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.HttpAddressBinding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.HttpBinding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.HttpOperationBinding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.HttpUrlEncodedBinding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.HttpUrlReplacementBinding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.Import", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.ImportCollection", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.InputBinding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.Message", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.MessageBinding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.MessageCollection", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.MessagePart", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.MessagePartCollection", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.MimeContentBinding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.MimeMultipartRelatedBinding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.MimePart", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.MimePartCollection", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.MimeTextBinding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.MimeTextMatch", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.MimeTextMatchCollection", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.MimeXmlBinding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.NamedItem", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.Operation", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.OperationBinding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.OperationBindingCollection", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.OperationCollection", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.OperationFault", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.OperationFaultCollection", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.OperationFlow", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.OperationInput", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.OperationMessage", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.OperationMessageCollection", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.OperationOutput", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.OutputBinding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.Port", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.PortCollection", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.PortType", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.PortTypeCollection", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.Service", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.ServiceCollection", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.ServiceDescription", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.ServiceDescriptionBaseCollection", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.ServiceDescriptionCollection", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.ServiceDescriptionFormatExtension", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.ServiceDescriptionFormatExtensionCollection", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.Soap12AddressBinding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.Soap12Binding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.Soap12BodyBinding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.Soap12FaultBinding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.Soap12HeaderBinding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.Soap12OperationBinding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.SoapAddressBinding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.SoapBinding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.SoapBindingStyle", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.SoapBindingUse", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.SoapBodyBinding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.SoapFaultBinding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.SoapHeaderBinding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.SoapHeaderFaultBinding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.SoapOperationBinding", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Description.Types", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Configuration.XmlFormatExtensionAttribute", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Configuration.XmlFormatExtensionPointAttribute", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Web.Services.Configuration.XmlFormatExtensionPrefixAttribute", "System.Web.Services.Description, Version=8.1.2.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"},
                {"System.Xaml.Permissions.XamlAccessLevel", "System.Windows.Extensions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.X509Certificates.X509Certificate2UI", "System.Windows.Extensions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Security.Cryptography.X509Certificates.X509SelectionFlag", "System.Windows.Extensions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Media.SoundPlayer", "System.Windows.Extensions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Media.SystemSound", "System.Windows.Extensions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
                {"System.Media.SystemSounds", "System.Windows.Extensions, Version=10.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51"},
            };
        }
    }
}
